#include "Cgcns.h"
#include "CompositeGridOperators.h"
#include "ParallelUtility.h"
#include "App.h"
#include "GenericGraphicsInterface.h"
#include "CnsParameters.h"
// int 
// getAmrErrorFunction(realCompositeGridFunction & u, 
//                     real t,
//                     Parameters & parameters,
//                     realCompositeGridFunction & error,
//                     bool computeOnFinestLevel =false );


// ===================================================================================================================
/// \brief Return a grid function for plotting and for output that may contain extra variables, such
/// as errors or such as the pressure for the compressible NS.
///
/// \param gf0 (input) : input grid function.
/// \param v (input) : grid function to hold the result, IF extra variables area added.
/// \return a realCompositeGridFunction holding the possibly augmented solution.
/// 
// ==================================================================================================================
realCompositeGridFunction & Cgcns::
getAugmentedSolution( GridFunction & gf0, realCompositeGridFunction & v )
{
  checkArrayIDs(sPrintF("getAugmentedSolution: start") ); 

  // No need to compute the augmented solution if graphics plotting is off
  if( parameters.dbase.get<GenericGraphicsInterface* >("ps")!=NULL && !parameters.dbase.get<GenericGraphicsInterface* >("ps")->graphicsIsOn() ) return gf0.u;

  CompositeGrid & cg = gf0.cg;
  realCompositeGridFunction & u = gf0.u;

  const int numberOfComponents = parameters.dbase.get<int >("numberOfComponents");
  
  bool plotMoreComponents = false;
  int extra=0;
  int dc;   // save divergence here

  if( parameters.isAdaptiveGridProblem() && parameters.dbase.get<int >("showAmrErrorFunction") )
    extra+=1;

  if( parameters.dbase.get<int >("showResiduals") || 
      parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")==Parameters::steadyStateRungeKutta )
  {
    // plot residuals too
    extra+=numberOfComponents;
    dc=2*numberOfComponents;
  }
  
  

  if( parameters.dbase.get<bool >("twilightZoneFlow") )
  {
    // For twilight zone flow we also plot the errors
    plotMoreComponents=true;
    dc+=numberOfComponents;
    
    Range all;
    Index I1,I2,I3;
    v.updateToMatchGrid(cg,all,all,all,numberOfComponents*2+extra);  // extra space for errors
    Range N(0,numberOfComponents-1);
    Range N2(numberOfComponents,2*numberOfComponents-1);

    for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    {
      MappedGrid & mg = cg[grid];
      getIndex(mg.dimension(),I1,I2,I3);

      const realArray & x= mg.center();
      bool ok=true;
      #ifdef USE_PPP
        realSerialArray xLocal; getLocalArrayWithGhostBoundaries(x,xLocal);
        realSerialArray uLocal; getLocalArrayWithGhostBoundaries(u[grid],uLocal);
        realSerialArray vLocal; getLocalArrayWithGhostBoundaries(v[grid],vLocal);
        const int includeGhost=1;
        ok = ParallelUtility::getLocalArrayBounds(u[grid],uLocal,I1,I2,I3,includeGhost); 
      #else
        const realSerialArray & xLocal = x;
        const realSerialArray & uLocal = u[grid];
        const realSerialArray & vLocal = v[grid];
      #endif

      if( ok )
      {
        vLocal(I1,I2,I3,N)=uLocal(I1,I2,I3,N);

        bool isRectangular=false;
        realSerialArray u0(I1,I2,I3);
	for( int n=N.getBase(), n2=N2.getBase(); n<=N.getBound(); n++, n2++ )
	  {
	    u0.reshape(I1,I2,I3,Range(n,n)); //kkc 070920 some TW functions (like Pulse) need the last index to match n in the call below
	    parameters.dbase.get<OGFunction* >("exactSolution")->gd( u0,xLocal,mg.numberOfDimensions(),isRectangular,0,0,0,0,I1,I2,I3,n,gf0.t);
	    vLocal(I1,I2,I3,n2)=vLocal(I1,I2,I3,n)-u0;
	  }
      }

//       v[grid](I1,I2,I3,N)=u[grid](I1,I2,I3,N);
//       v[grid](I1,I2,I3,N2)=v[grid](I1,I2,I3,N)-(*parameters.dbase.get<OGFunction* >("exactSolution"))(cg[grid],I1,I2,I3,N,gf0.t);

    }
    for( int n=0; n<numberOfComponents; n++ )
    {
      v.setName(u.getName(n),n);
      v.setName(u.getName(n)+" (error)",n+numberOfComponents);
    }
  }
  else 
  {
    // for compressible NS we also plot p
    plotMoreComponents=true;

    if( parameters.dbase.get<CnsParameters::KnownSolutionsEnum >("knownSolution")!=CnsParameters::noKnownSolution )
    {
      // plot errors and true solution
      extra+=numberOfComponents*2;
    }


    int numberOfAugmentedComponents=0;
    if( parameters.dbase.get<CnsParameters::PDE >("pde")==CnsParameters::compressibleNavierStokes )
    {
      numberOfAugmentedComponents=numberOfComponents+1+extra;
    }
    else if( parameters.dbase.get<CnsParameters::PDE >("pde")==CnsParameters::compressibleMultiphase )
    {
      numberOfAugmentedComponents=numberOfComponents+2+extra; 
    }
    else
    {
      Overture::abort("error");
    }
    

    const int rc=parameters.dbase.get<int >("rc");
    const int tc=parameters.dbase.get<int >("tc");
    const int sc=parameters.dbase.get<int >("sc");
    
    Range all;
    Index I1,I2,I3;
    v.updateToMatchGrid(cg,all,all,all,numberOfAugmentedComponents);

    Range N(0,numberOfComponents-1);
    int pn=numberOfComponents;
    
    for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    {
      getIndex(cg[grid].dimension(),I1,I2,I3);
      // v[grid](I1,I2,I3,N)=u[grid](I1,I2,I3,N);
      assign( v[grid], u[grid], all,all,all,N ); 

      // evaluate the pressure:
      parameters.getDerivedFunction("pressure",u[grid],v[grid],grid,pn,gf0.t,parameters);
    }
    
    for( int n=0; n<numberOfComponents; n++ )
      v.setName(u.getName(n),n);

    if( parameters.dbase.get<CnsParameters::PDE >("pde")==CnsParameters::compressibleNavierStokes )
    {
      v.setName("p",pn);
    }
    else if( parameters.dbase.get<CnsParameters::PDE >("pde")==CnsParameters::compressibleMultiphase )
    { // there are two pressures added in this case
      v.setName("ps",pn);
      pn++;  
      v.setName("pg",pn);
    }

    if( parameters.dbase.get<CnsParameters::KnownSolutionsEnum >("knownSolution")!=CnsParameters::noKnownSolution )
    {
      realCompositeGridFunction & uKnown = parameters.getKnownSolution( cg, gf0.t );

      const int rc=parameters.dbase.get<int >("rc");
      const int uc=parameters.dbase.get<int >("uc");
      const int vc=parameters.dbase.get<int >("vc");
      const int wc=parameters.dbase.get<int >("wc"); // for swirl equation if present
      const int tc=parameters.dbase.get<int >("tc");
      const int pc=parameters.dbase.get<int >("pc");

      for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
      {
	getIndex(cg[grid].dimension(),I1,I2,I3);
	// save the true solution
        int c0=pn+1;
	int cc=c0;
        v[grid](I1,I2,I3,cc)=uKnown[grid](I1,I2,I3,rc);
	cc++;
        v[grid](I1,I2,I3,cc)=uKnown[grid](I1,I2,I3,uc);
	cc++;
        v[grid](I1,I2,I3,cc)=uKnown[grid](I1,I2,I3,vc);
	cc++;
	if ( parameters.dbase.get<bool >("axisymmetricWithSwirl") )
	  {
	    v[grid](I1,I2,I3,cc)=uKnown[grid](I1,I2,I3,wc);
	    cc++;
	  }
        v[grid](I1,I2,I3,cc)=uKnown[grid](I1,I2,I3,tc);
	cc++;
       
        // here are the errors
        c0+=4; 
        v[grid](I1,I2,I3,cc)=u[grid](I1,I2,I3,rc)-uKnown[grid](I1,I2,I3,rc);
	cc++;
        v[grid](I1,I2,I3,cc)=u[grid](I1,I2,I3,uc)-uKnown[grid](I1,I2,I3,uc);
	cc++;
        v[grid](I1,I2,I3,cc)=u[grid](I1,I2,I3,vc)-uKnown[grid](I1,I2,I3,vc);
	cc++;
	if ( parameters.dbase.get<bool >("axisymmetricWithSwirl") )
	  {
	    v[grid](I1,I2,I3,cc)=u[grid](I1,I2,I3,wc)-uKnown[grid](I1,I2,I3,wc);
	    cc++;
	  }
        v[grid](I1,I2,I3,cc)=u[grid](I1,I2,I3,tc)-uKnown[grid](I1,I2,I3,tc);
	cc++;

      }
      int c0=pn+1;
      int cc=c0;
      v.setName("rhoTrue",cc);cc++;
      v.setName("uTrue"  ,cc);cc++;
      v.setName("vTrue"  ,cc);cc++;
      if ( parameters.dbase.get<bool >("axisymmetricWithSwirl") )
	{
	  v.setName("wTrue"  ,cc);cc++;
	}
      v.setName("TTrue"  ,cc);cc++;
      c0+=4;
      
      v.setName("rhoErr",cc);cc++;
      v.setName("uErr"  ,cc);cc++;
      v.setName("vErr"  ,cc);cc++;
      if ( parameters.dbase.get<bool >("axisymmetricWithSwirl") )
	{
	  v.setName("wErr"  ,cc);cc++;
	}
      v.setName("TErr"  ,cc);cc++;
    }
    
  } // if !twilight zone
  
  if( parameters.isAdaptiveGridProblem() && parameters.dbase.get<int >("showAmrErrorFunction") )
  {
    // add on the AMR error estimate
    real timeAmr=getCPU();
    Range all;
    if( !plotMoreComponents )
    {
      v.updateToMatchGrid(cg,all,all,all,numberOfComponents+extra); 
      plotMoreComponents=true;
    }
    
    int ec = v.getComponentBound(0);
    #ifndef USE_PPP
      realCompositeGridFunction error;
      error.link(v,Range(ec,ec));
      bool computeOnFinestLevel=true;
      getAmrErrorFunction(u,gf0.t,error,computeOnFinestLevel);
    #else
      realCompositeGridFunction error(cg);
      bool computeOnFinestLevel=true;
      getAmrErrorFunction(u,gf0.t,error,computeOnFinestLevel);
      for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
      {
        assign(v[grid],all,all,all,ec, u[grid],all,all,all, 0 );
      }
      
    #endif

    // do NOT count time spent here as part of AMR, but rather part of plotting
    parameters.dbase.get<RealArray>("timing")(Parameters::timeForAmrErrorFunction)-=getCPU()-timeAmr; 
    v.setName("error estimate",ec);
  }
  
  if( parameters.dbase.get<int >("showResiduals") || parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")==Parameters::steadyStateRungeKutta )
  {
    if( !plotMoreComponents )
    {
      Range all;
      v.updateToMatchGrid(cg,all,all,all,numberOfComponents+extra); 
      plotMoreComponents=true;

      Range N=numberOfComponents;
      for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
        v[grid](all,all,all,N)=u[grid](all,all,all,N);    // copy existing components
      for( int n=0; n<numberOfComponents; n++ )
        v.setName(u.getName(n),n);
    }

    Range N(0,numberOfComponents-1);
    int offset=numberOfComponents;
    if( parameters.dbase.get<bool >("twilightZoneFlow") )
      offset+=numberOfComponents;
    if ( !parameters.dbase.get<bool >("twilightZoneFlow")  )
      { // kkc 080721 pressure is not computed when twilight zone is turned on
	if (parameters.dbase.get<CnsParameters::PDE >("pde")==CnsParameters::compressibleNavierStokes ||
	    parameters.dbase.get<CnsParameters::PDE >("pde")==CnsParameters::compressibleMultiphase )
	  { // kkc 080721 increment the offset for the pressure slot(s)
	    offset++;
	    if ( parameters.dbase.get<CnsParameters::PDE >("pde")==CnsParameters::compressibleMultiphase ) offset++;
	  }
      }

    realCompositeGridFunction & residual = fn[0];   // pointer to du/dt

    // printf(" getAug: *** residual.numberOfComponentGrids=%i **\n",residual.numberOfComponentGrids());
    // printf(" getAug: *** v.numberOfComponentGrids=%i **\n",v.numberOfComponentGrids());
    
    Index I1,I2,I3;
    for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    {
      getIndex(cg[grid].dimension(),I1,I2,I3);
      v[grid](I1,I2,I3,N+offset)=residual[grid](I1,I2,I3,N);
      where( cg[grid].mask()(I1,I2,I3)<=0 )
      {
        for( int n=N.getBase(); n<=N.getBound(); n++ )
          v[grid](I1,I2,I3,n+offset)=0.;   // set residual to zero at unused and interp points
      }
      
    }
    for( int n=0; n<numberOfComponents; n++ )
    {
      v.setName(u.getName(n)+" (residual)",n+offset);
    }
  }
  
  checkArrayIDs(sPrintF("getAugmentedSolution: near end") ); 

  realCompositeGridFunction & ur = plotMoreComponents ? v : gf[current].u;
//   if( true ) // *wdh* 060228
//   {
//     gf0.u.display("\n\n getAugmentedSolution: gf0.u","%5.2f ");
//     ur.display("\n\n getAugmentedSolution: ur","%5.2f ");
//   }
  return ur;
}
