// ---------------------------------------------------------------------------------------
//  Main class for handling deforming bodies.
//
//
// Original version by Petri Fast. for elastic-filamanents  c. 2000? 
// Changes to support deforming grids by wdh. c. 2006
// Major updates by wdh. 2008 
// ---------------------------------------------------------------------------------------

#include "DeformingBodyMotion.h"
#include "DeformingGrid.h"
#include "DeformingGridGenerationInformation.h"
#include "ElasticFilament.h"
#include "SplineMapping.h"
#include "FilamentMapping.h"
#include "HyperbolicMapping.h"
#include "MappingInformation.h"
#include "NurbsMapping.h"
#include "DomainSolver.h"
#include "Parameters.h"
#include "ParallelUtility.h"

namespace
{

enum UserDefinedDeformingBodyMotionEnum
{
  iceDeform,
  ellipseDeform,
  sphereDeform,
  advectBody,
  elasticShell,
  interfaceDeform,
  userDeformingSurface  // deforming surface is defined by the function: userDefinedDeformingSurface
}; 
 
}

#define  FOR_3(i1,i2,i3,I1,I2,I3)\
  I1Base=I1.getBase(); I2Base=I2.getBase(); I3Base=I3.getBase();\
  I1Bound=I1.getBound(); I2Bound=I2.getBound(); I3Bound=I3.getBound();\
  for( i3=I3Base; i3<=I3Bound; i3++ )  \
  for( i2=I2Base; i2<=I2Bound; i2++ )  \
  for( i1=I1Base; i1<=I1Bound; i1++ )

#define  FOR_3D(i1,i2,i3,I1,I2,I3)\
  int I1Base,I2Base,I3Base;\
  int I1Bound,I2Bound,I3Bound;\
  I1Base=I1.getBase(); I2Base=I2.getBase(); I3Base=I3.getBase();\
  I1Bound=I1.getBound(); I2Bound=I2.getBound(); I3Bound=I3.getBound();\
  for( i3=I3Base; i3<=I3Bound; i3++ )  \
  for( i2=I2Base; i2<=I2Bound; i2++ )  \
  for( i1=I1Base; i1<=I1Bound; i1++ )

//==========================================================================================
/// \brief Constructor:  Set up grids, but do not initialize the mapping in the composite grid. 
///    And don't instantiate the `physics object' yet
/// \param params (input) : the DomainSolver Parameters object.
// ===================================================================================
DeformingBodyMotion::
DeformingBodyMotion( Parameters & params,
                     int numberOfTimeLevels,             /* =3  */
		     GenericGraphicsInterface *pGIDebug00, /* =NULL */ 
		     int debug00                           /* =0 */)
  : parameters(params)
{
  debug = debug00;
  if( debug & 2 ) 
    printF("\nDeformingBodyMotion -- Constructor called\n");

  // new way: add parameters to the data-base:
  deformingBodyDataBase.put<int>("numberOfDeformingGrids",0);

  deformingBodyDataBase.put<int>("numberOfFaces",0);
  deformingBodyDataBase.put<IntegerArray>("boundaryFaces");

  deformingBodyDataBase.put<int>("numberOfTimeLevels",numberOfTimeLevels);

  deformingBodyDataBase.put<DeformingBodyType>("deformingBodyType",elasticFilament);


  // These Mapping's will replace the Mapping's in the CompositeGrid's and allow for deformation. 
  // These will often be a HyperbolicMapping
  deformingBodyDataBase.put<vector<Mapping*> >("transform");
  // deformingBodyDataBase.put<int>("transformToUse",0);

  // The Mapping's will define the "surface" from which the volume mapping will be created.
  // This will often be a NurbsMapping. In 2D the surface will be a curve.
  deformingBodyDataBase.put<vector<Mapping*> >("surface");

  // These arrays hold the grid points that define the deforming "surface" 
  // These values may be interpolated to form a Mapping "surface" 
  deformingBodyDataBase.put<vector<RealArray*> >("surfaceArray",NULL);
  deformingBodyDataBase.put<vector<real*> >("surfaceArrayTime",NULL); // corresponding times

  // gridEvolution holds a time history of grids from which we can compute velocity and acceleration
  deformingBodyDataBase.put<vector<GridEvolution*> >("gridEvolution"); 
  GridEvolution::debug=debug;

  // We parameterize the boundary curve by chord-length or index (flor fluid structure problems
  //  we should parameterize by index so the grid points match)
  deformingBodyDataBase.put<int>("boundaryParameterization");
  deformingBodyDataBase.get<int>("boundaryParameterization")=NurbsMapping::parameterizeByChordLength; 
  
  // order of accuracy for the acceleration computation: 
  deformingBodyDataBase.put<int>("accelerationOrderOfAccuracy",1);
  // order of accurcy for the velocity computation
  deformingBodyDataBase.put<int>("velocityOrderOfAccuracy",1);

  pElasticFilament = NULL;  //.. zero the physics objects

  debug=debug00;
  pGIDebug = pGIDebug00;
  if( pGIDebug != NULL )
  {
    pMapInfoDebug = new MappingInformation;
    pMapInfoDebug->graphXInterface = pGIDebug;
  } 
  else 
  {
    pMapInfoDebug = NULL;
  }

  //..DeformingGrid list -- just one componentGrid for now
  //..SHOULD USE a 'DeformingGridCollection'

  pDeformingGrid=NULL;

}


DeformingBodyMotion::
~DeformingBodyMotion()
{
  delete(pElasticFilament);
  delete(pDeformingGrid);

  delete pMapInfoDebug;

  // for userDefined

  vector<Mapping*> & transform =deformingBodyDataBase.get<vector<Mapping*> >("transform");
  for( int i=0; i<transform.size(); i++ )
    delete transform[i];

  vector<Mapping*> & surface = deformingBodyDataBase.get<vector<Mapping*> >("surface");
  for( int i=0; i<surface.size(); i++ )
    delete surface[i];

  vector<RealArray*> & surfaceArray = deformingBodyDataBase.get<vector<RealArray*> >("surfaceArray");
  vector<real*> & surfaceArrayTime = deformingBodyDataBase.get<vector<real*> >("surfaceArrayTime"); 
  for( int i=0; i<surfaceArray.size(); i++ )
  {
    delete [] surfaceArray[i];
    delete [] surfaceArrayTime[i];
  }
  


  vector<GridEvolution*> & gridEvolution = deformingBodyDataBase.get<vector<GridEvolution*> >("gridEvolution");
  for( int i=0; i<gridEvolution.size(); i++ )
    delete gridEvolution[i];

}

//
// -------------- Services
//

// ============================================================================================
/// \brief return the order of accuracy used to compute the acceleration.
// ============================================================================================
int DeformingBodyMotion::
getAccelerationOrderOfAccuracy() const
{
  return deformingBodyDataBase.get<int>("accelerationOrderOfAccuracy");
}

// ============================================================================================
/// \brief return the order of accuracy used to compute the velocity.
// ============================================================================================
int DeformingBodyMotion::
getVelocityOrderOfAccuracy() const
{
  return deformingBodyDataBase.get<int>("velocityOrderOfAccuracy");
}


// ============================================================================================
/// \brief set the order of accuracy used to compute the acceleration.
/// \param order (input) : a positive integer
// ============================================================================================
int DeformingBodyMotion::
setAccelerationOrderOfAccuracy( int order )
{
  deformingBodyDataBase.get<int>("accelerationOrderOfAccuracy")=order;
  return 0;
}

// ============================================================================================
/// \brief set the order of accuracy used to compute the velocity.
/// \param order (input) : a positive integer
// ============================================================================================
int DeformingBodyMotion::
setVelocityOrderOfAccuracy( int order )
{
  deformingBodyDataBase.get<int>("velocityOrderOfAccuracy")=order;
  return 0;
}

// ================================================================================================
/// \brief Define the deforming body in terms of faces on grids.
/// \param numberOfFaces (input) : 
/// \param boundaryFaces (input) : boundaryFaces(0:2,f) = (side,axis,grid) for face f 
// ================================================================================================
int DeformingBodyMotion::
defineBody( int numberOfFaces_, IntegerArray & boundaryFaces_ )
{
  int & numberOfDeformingGrids = deformingBodyDataBase.get<int>("numberOfDeformingGrids");
  int & numberOfFaces = deformingBodyDataBase.get<int>("numberOfFaces");
  IntegerArray & boundaryFaces = deformingBodyDataBase.get<IntegerArray>("boundaryFaces");

  numberOfFaces=numberOfFaces_;
  boundaryFaces.redim(3,numberOfFaces);
  boundaryFaces(Range(0,2),Range(numberOfFaces))=boundaryFaces_(Range(0,2),Range(numberOfFaces));

  
  numberOfDeformingGrids=numberOfFaces;  // assume this is true for now 
  

  DeformingBodyType & deformingBodyType = 
                  deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");

  if( !deformingBodyDataBase.has_key("userDefinedDeformingBodyMotionOption") )
    deformingBodyDataBase.put<UserDefinedDeformingBodyMotionEnum>("userDefinedDeformingBodyMotionOption",iceDeform);
  UserDefinedDeformingBodyMotionEnum & userDefinedDeformingBodyMotionOption = 
    deformingBodyDataBase.get<UserDefinedDeformingBodyMotionEnum>("userDefinedDeformingBodyMotionOption");

  if( numberOfFaces>1 &&
      ( deformingBodyType==elasticFilament ||
        (deformingBodyType==userDefinedDeformingBody && userDefinedDeformingBodyMotionOption==interfaceDeform) ) )
  {
    int sideToMove=boundaryFaces(0,0);
    int axisToMove=boundaryFaces(1,0);
    int gridToMove=boundaryFaces(2,0); 

    printF("\n ***DeformingBodyMotion:defineBody: Grid=%i has more than one face that deforms. This is not allowed.\n",gridToMove);
    printF("      ... specify fewer faces when defining the grid motion.\n");
    Overture::abort("error");
  }


  return 0;
}

// ===============================================================================================
/// \brief return the initial state (e.g. position, velocity, acceleration)
/// \details return the initial state of the deforming grid.
// ===============================================================================================

// return the initial state (position, velocity, acceleration)
int DeformingBodyMotion::
getInitialState( InitialStateOptionEnum stateOption, 
		 const real time, const int grid, 
		 const Index &I1, const Index &I2, const Index &I3, 
		 realArray & state )
{
  DeformingBodyType & deformingBodyType = 
                  deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");

  if( true || debug & 2 ) 
    printF("DeformingBodyMotion::getInitialState: stateOption=%i, grid=%i, t=%9.3e\n",(int)stateOption,grid,time);

  const int numberOfDimensions = parameters.dbase.get<int>("numberOfDimensions");
  Range Rx=numberOfDimensions;
  if( deformingBodyType==elasticFilament )
  {
    if( stateOption==initialVelocity || stateOption==initialAcceleration )
    {
      state(I1,I2,I3,Rx)=0.;
    }
    else
    {
      OV_ABORT("getInitialState:ERROR: finish me");
    }
  }
  else if( deformingBodyType==userDefinedDeformingBody )
  {
    // for now we only support constant initial grid velocity and grid acceleration
    if( stateOption==initialVelocity )
    {
      if( !deformingBodyDataBase.has_key("initialVelocity") )
      {
	deformingBodyDataBase.put<real [3]>("initialVelocity");
        real *v0 = deformingBodyDataBase.get<real [3]>("initialVelocity");
        v0[0]=v0[1]=v0[2]=0.;
      }
      real *v0 = deformingBodyDataBase.get<real [3]>("initialVelocity");
      if( true || debug & 2 )
	printF("DeformingBodyMotion::getInitialState: Setting grid velocity to (%9.3e,%9.3e,%9.3e)\n",v0[0],v0[1],v0[2]);

      for( int axis=0; axis<numberOfDimensions; axis++ )
        state(I1,I2,I3,axis)=v0[axis];
    }
    else if( stateOption==initialAcceleration )
    {
      if( !deformingBodyDataBase.has_key("initialAcceleration") )
      {
	deformingBodyDataBase.put<real [3]>("initialAcceleration");
	real *a0 = deformingBodyDataBase.get<real [3]>("initialAcceleration");
	a0[0]=a0[1]=a0[2]=0.;
      }
      real *a0 = deformingBodyDataBase.get<real [3]>("initialAcceleration");
      for( int axis=0; axis<numberOfDimensions; axis++ )
        state(I1,I2,I3,axis)=a0[axis];
    }
    else
    {
      OV_ABORT("getInitialState:ERROR: finish me");
    }
  }
  else
  {
    OV_ABORT("getInitialState:ERROR: unknown deformingBodyType");
  }
  return 0;

}


// ==================================================================================
/// \brief Return the face number that corresponds to a given grid.
///        This is a protected routine. 
// ==================================================================================
int DeformingBodyMotion::
getFace( int grid ) const 
{
  int & numberOfFaces = deformingBodyDataBase.get<int>("numberOfFaces");
  IntegerArray & boundaryFaces = deformingBodyDataBase.get<IntegerArray>("boundaryFaces");
  int face=-1;
  for( int f=0; f<numberOfFaces; f++ )
  {
    if( boundaryFaces(2,f)==grid )
    {
      face=f;
    }
  }
  return face;
}


// ===============================================================================================
/// \brief Return the grid velocity 
/// \details This routine uses the GridEvolution class to compute the grid velocity. The
/// GridEvolution class keeps a sequence of past grids and computes the time derivative 
/// of the grid motion using these grids. 
// ===============================================================================================
int DeformingBodyMotion::
getVelocity( const real time0, 
	     const int grid, 
	     CompositeGrid & cg,
	     realArray & gridVelocity)
{

  DeformingBodyType & deformingBodyType = 
                  deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");

  if( deformingBodyType==elasticFilament )
  {
    if(debug&2) 
      cout << "DeformingBodyMotion::getVelocity, grid=" <<  grid << " called\n";

    assert( pDeformingGrid != NULL );
    pDeformingGrid->getVelocity( time0, grid, cg, gridVelocity);
  }
  else if( deformingBodyType==userDefinedDeformingBody )
  {
    if(debug&2)
       printF("DeformingBodyMotion:getVelocity called for userDefinedDeformingBody for t=%9.3e\n",time0);

    Index Iv[3], & I1 = Iv[0], &I2=Iv[1], &I3=Iv[2];

    int extra=1; // evaluate the grid velocity at one ghost point too (is this needed?)
    getIndex(cg[grid].gridIndexRange(),I1,I2,I3,extra);
    
    if( time0<=0 )
    {
      // for initial times we use the following function: 
      getInitialState( initialVelocity,time0,grid,I1,I2,I3,gridVelocity);
    }
    else
    {
      vector<GridEvolution*> & gridEvolution = deformingBodyDataBase.get<vector<GridEvolution*> >("gridEvolution");
      int face=getFace( grid );
      if( face>=0 && face<gridEvolution.size() )
      {
	gridEvolution[face]->getVelocity(time0,gridVelocity,I1,I2,I3);
      }
      else
      {
	if( true || debug&2 )
	{
	  printF("DeformingBodyMotion:getVelocity:WARNING: There is no velocity available for t=%9.3e\n",time0);
	  printF("DeformingBodyMotion: grid=%i, face=%i,  gridEvolution.size()=%i\n",grid,face,gridEvolution.size());

	  int & numberOfFaces = deformingBodyDataBase.get<int>("numberOfFaces");
	  IntegerArray & boundaryFaces = deformingBodyDataBase.get<IntegerArray>("boundaryFaces");
	  printF("DeformingBodyMotion: numberOfFaces=%i\n",numberOfFaces);
	  ::display(boundaryFaces,"boundaryFaces");
	}
      
	gridVelocity=0.;
      }
    }
  }
  else
  {
    Overture::abort("ERROR: unknown deformingBodyType");
  }
  return 0;
}


// ===============================================================================================
/// \brief Return the grid velocity at specified points 
// ===============================================================================================
int DeformingBodyMotion::
getVelocityBC( const real time0, const int grid, const Index &I1, const Index &I2, const Index &I3, 
	       realArray & bcVelocity)
{
  DeformingBodyType & deformingBodyType = 
                  deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");

  if( deformingBodyType==elasticFilament )
  {
    assert( pElasticFilament!=NULL );
    bcVelocity = 0.;
    pElasticFilament->getVelocityBC( time0, I1,I2,I3, bcVelocity );
  }
  else if( deformingBodyType==userDefinedDeformingBody )
  {
    if( debug & 2 )
      printF("DeformingBodyMotion:getVelocityBC called for userDefinedDeformingBody for t=%9.3e\n",time0);

    // bcVelocity = 0.;
    if( time0<=0 ) 
    {
      // for initial times we use the following function: 
      getInitialState( initialVelocity,time0,grid,I1,I2,I3,bcVelocity);
    }
    else
    {
      vector<GridEvolution*> & gridEvolution = deformingBodyDataBase.get<vector<GridEvolution*> >("gridEvolution");
      int face=getFace( grid );
      if( face>=0 && face<gridEvolution.size() )
      {
	gridEvolution[face]->getVelocity(time0,bcVelocity,I1,I2,I3);
      }
      else
      {
	if( true || debug&2 )
	  printF("DeformingBodyMotion:getVelocityBC:WARNING: There is no velocity available for t=%9.3e\n",time0);
      }
    }
  }
  else
  {
    Overture::abort("ERROR: unknown deformingBodyType");
  } 
  return 0;
}


// ===============================================================================================
/// \brief Return the grid acceleration on boundary points.
// ===============================================================================================
int DeformingBodyMotion::
getAccelerationBC( const real time0, const int grid, const Index &I1, const Index &I2, const Index &I3, 
		   realArray & bcAcceleration)
{
  DeformingBodyType & deformingBodyType = 
                  deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");

  if( deformingBodyType==elasticFilament )
  {
    assert( pElasticFilament!=NULL );
    bcAcceleration = 0.;
    pElasticFilament->getAccelerationBC( time0, I1,I2,I3, bcAcceleration );
  }
  else if( deformingBodyType==userDefinedDeformingBody )
  {
    // bcAcceleration = 0.;
    if( time0<=0 ) 
    {
      // for initial times we use the following function: 
      getInitialState( initialAcceleration,time0,grid,I1,I2,I3,bcAcceleration );
    }
    else
    {
      vector<GridEvolution*> & gridEvolution = deformingBodyDataBase.get<vector<GridEvolution*> >("gridEvolution");
      int face=getFace( grid );
      if( face>=0 && face<gridEvolution.size() )
      {
	gridEvolution[face]->getAcceleration(time0,bcAcceleration,I1,I2,I3);
      }
      else
      {
	if( true || debug&2 )
	  printF("DeformingBodyMotion:getAccelerationBC:WARNING: There is no data available for t=%9.3e\n",time0);
      }
    }
  }
  else
  {
    Overture::abort("ERROR: unknown deformingBodyType");
  } 
  return 0;
}

// int DeformingBodyMotion::
// getAccelerationBC( const real time0, const int grid, 
// 		   CompositeGrid & cg, realArray & bcAcceleration)
// {
//   cout<<"DeformingBodyMotion::getAccelBC(time,grid,cg,bcAccel) -- not used!!\n";
//   //  bcAcceleration = 0.;
//   //pElasticFilament->getAccelerationBC( time0, grid, cg, bcAcceleration );
//   return 0;
// }

//========================================================================================
/// \brief: return the number of grids that form the deforming body
//========================================================================================
int DeformingBodyMotion::
getNumberOfGrids()
{
  int & numberOfDeformingGrids = deformingBodyDataBase.get<int>("numberOfDeformingGrids");

  return numberOfDeformingGrids; 
}



// ========================================================================================
//  \brief Initialize the component grid that is deforming
//
// ========================================================================================
int DeformingBodyMotion::
initialize( CompositeGrid & cg, real t /* = 0. */ )
{
  int ierr=0;
  const int numberOfDimensions = cg.numberOfDimensions();
  int & numberOfTimeLevels = deformingBodyDataBase.get<int>("numberOfTimeLevels");
  int & numberOfDeformingGrids = deformingBodyDataBase.get<int>("numberOfDeformingGrids");
  int & numberOfFaces = deformingBodyDataBase.get<int>("numberOfFaces");
  IntegerArray & boundaryFaces = deformingBodyDataBase.get<IntegerArray>("boundaryFaces");
  DeformingBodyType & deformingBodyType = 
    deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");


  for( int face=0; face<numberOfFaces; face++ )
  {
    int sideToMove=boundaryFaces(0,face);
    int axisToMove=boundaryFaces(1,face);
    int gridToMove=boundaryFaces(2,face); 

    if( deformingBodyType==elasticFilament )
    {

      // do this for now -- treat more grids later 
      assert( numberOfDeformingGrids==1 );
      assert( numberOfFaces==1 );


      ierr=1;
    
      // *wdh* put this here:
      pDeformingGrid = new DeformingGrid( numberOfTimeLevels, pGIDebug, debug);


      //.... We assume moving grid is an "ElasticFilament'
      //.... Use mapping in the grid
      //.... >> Create new ElasticFilament & init. with FilamMap from the grid
      pElasticFilament = new ElasticFilament();

      Mapping *pMap = (FilamentMapping*) cg[gridToMove].mapping().mapPointer;
      assert( pMap != NULL);
      if ( pMap->getClassName() == "FilamentMapping" )
      {
	FilamentMapping *pTempFilament = (FilamentMapping*) pMap;
	assert( pTempFilament != NULL );
	if(debug&4) {
	  cout << endl << "-----------------------------------------\n";
	  cout << "DeformingBodyMotion::init -- calling printHyperbDims:\n";
	}
	printFilamentHyperbolicDimensions(cg, gridToMove);
    
	pElasticFilament->initializeFromFilamentMapping( pTempFilament );
      }
      else 
      {
	ierr=0;
	cout <<endl
	     <<"DeformingBodyMotion::initializeDeformingBody:\n"
	     <<"   error: trying to move FilamentMapping "
	     << "grid #"<<gridToMove<<", of type "
	     << pMap->getClassName() 
	     <<endl<<endl;
	pElasticFilament=NULL;
      }
    }
    else if( deformingBodyType==userDefinedDeformingBody )
    {
      printF("*** DeformingBodyMotion::initialize userDefinedDeformingBody ***\n");
    
      if( !deformingBodyDataBase.has_key("userDefinedDeformingBodyMotionOption") )
	deformingBodyDataBase.put<UserDefinedDeformingBodyMotionEnum>("userDefinedDeformingBodyMotionOption",iceDeform);
      UserDefinedDeformingBodyMotionEnum & userDefinedDeformingBodyMotionOption = 
	deformingBodyDataBase.get<UserDefinedDeformingBodyMotionEnum>("userDefinedDeformingBodyMotionOption");


      vector<Mapping*> & transform =deformingBodyDataBase.get<vector<Mapping*> >("transform");
    

      // The "surface" Mapping holds the start curve in 2D
      vector<Mapping*> & surface = deformingBodyDataBase.get<vector<Mapping*> >("surface");
      if( face>=surface.size() )
	surface.push_back(new NurbsMapping);

      const int numberOfSurfaceArrays=3;
      vector<RealArray*> & surfaceArray = deformingBodyDataBase.get<vector<RealArray*> >("surfaceArray");
      vector<real*> & surfaceArrayTime = deformingBodyDataBase.get<vector<real*> >("surfaceArrayTime"); 

      RealArray *ppx = new RealArray [numberOfSurfaceArrays];
      real *psat = new real [numberOfSurfaceArrays];
      if( face>=surfaceArray.size() )
      {
	surfaceArray.push_back(ppx);
	surfaceArrayTime.push_back(psat);
      }
      

      RealArray *px = surfaceArray[face];
      RealArray &x0 = px[0], &x1 = px[1], &x2=px[2];
      real & tx0 = surfaceArrayTime[face][0];
      tx0=t;  // set time associated with the x0 points
      
      real rad=.5, theta; 

      if( userDefinedDeformingBodyMotionOption==iceDeform )
      {
	// ************************************************      
	// ********** Specified Ice Motion ****************      
	// ************************************************      
	if( cg.numberOfDimensions()==2 )
	{
	  // Two_dimensions
	  const int n1=13;
	  x0.redim(n1,2); x1.redim(n1,2); x2.redim(n1,2);

	  // x0: points on a arc
	  x0(0,0)=0.; x0(0,1)=1.*rad;
	  theta=Pi* 5./180.; x0(1,0)=-rad*sin(theta); x0(1,1)=rad*cos(theta);
	  theta=Pi*10./180.; x0(2,0)=-rad*sin(theta); x0(2,1)=rad*cos(theta);
	  theta=Pi*15./180.; x0(3,0)=-rad*sin(theta); x0(3,1)=rad*cos(theta);
	  theta=Pi*20./180.; x0(4,0)=-rad*sin(theta); x0(4,1)=rad*cos(theta);

	  theta=Pi*50./180.; x0(5,0)=-rad*sin(theta); x0(5,1)=rad*cos(theta);
	  theta=Pi*90./180.; x0(6,0)=-rad*sin(theta); x0(6,1)=rad*cos(theta);
	  theta=Pi*130./180.;x0(7,0)=-rad*sin(theta); x0(7,1)=rad*cos(theta);

	  x0( 8,0)=x0(4,0); x0( 8,1)=-x0(4,1);
	  x0( 9,0)=x0(3,0); x0( 9,1)=-x0(3,1);
	  x0(10,0)=x0(2,0); x0(10,1)=-x0(2,1);
	  x0(11,0)=x0(1,0); x0(11,1)=-x0(1,1);
	  x0(12,0)=x0(0,0); x0(12,1)=-x0(0,1);


	  // a second start curve (deformed version)
	  x1=x0;
	  x1(5,0)=-.7; x1(5,1)=  .5;
	  x1(6,0)=-.6; x1(6,1)=  0.;
	  x1(7,0)=-.7; x1(7,1)= -.5;
	}
	else if( cg.numberOfDimensions()==3 )
	{
	  // Three dimensional case (starting from cylDeform.hdf)

	  // get width in z-direction from the backGround grid (assumed to be grdi 0)
	  assert( cg[0].isRectangular() );
	  real dx[3],xab[2][3];
	  cg[0].getRectangularGridParameters( dx, xab );


//      const int n1=13, n2=4;
	  const int n1=13, n2=11;
	  x0.redim(n1,n2,3); x1.redim(n1,n2,3); x2.redim(n1,n2,3);
	  Range all;

	  real za=xab[0][2], zb=xab[1][2];  // get these from backGround grid
	  // x0: points on a arc
	  for( int i2=0; i2<n2; i2++ )
	  {
	    real z0=za+(zb-za)*i2/(n2-1);
	
	    x0(0,i2,0)=0.; x0(0,i2,1)=1.*rad;
	    theta=Pi* 5./180.; x0(1,i2,0)=-rad*sin(theta); x0(1,i2,1)=rad*cos(theta);
	    theta=Pi*10./180.; x0(2,i2,0)=-rad*sin(theta); x0(2,i2,1)=rad*cos(theta);
	    theta=Pi*15./180.; x0(3,i2,0)=-rad*sin(theta); x0(3,i2,1)=rad*cos(theta);
	    theta=Pi*20./180.; x0(4,i2,0)=-rad*sin(theta); x0(4,i2,1)=rad*cos(theta);

	    theta=Pi*50./180.; x0(5,i2,0)=-rad*sin(theta); x0(5,i2,1)=rad*cos(theta);
	    theta=Pi*90./180.; x0(6,i2,0)=-rad*sin(theta); x0(6,i2,1)=rad*cos(theta);
	    theta=Pi*130./180.;x0(7,i2,0)=-rad*sin(theta); x0(7,i2,1)=rad*cos(theta);

	    x0( 8,i2,0)=x0(4,i2,0); x0( 8,i2,1)=-x0(4,i2,1);
	    x0( 9,i2,0)=x0(3,i2,0); x0( 9,i2,1)=-x0(3,i2,1);
	    x0(10,i2,0)=x0(2,i2,0); x0(10,i2,1)=-x0(2,i2,1);
	    x0(11,i2,0)=x0(1,i2,0); x0(11,i2,1)=-x0(1,i2,1);
	    x0(12,i2,0)=x0(0,i2,0); x0(12,i2,1)=-x0(0,i2,1);

	    x0(all,i2,2)=z0;
	  }

	  // a second start curve (deformed version)
	  x1=x0;
	  for( int i2=0; i2<n2; i2++ )
	  {
	    x1(5,i2,0)=-.7; x1(5,i2,1)=  .5;
	    x1(6,i2,0)=-.6; x1(6,i2,1)=  0.;
	    x1(7,i2,0)=-.7; x1(7,i2,1)= -.5;
	  }
      
	}
	else
	{
	  Overture::abort("ERROR: unexpected value for numberOfDimensions!");
	}
      }
      else if( userDefinedDeformingBodyMotionOption==ellipseDeform )
      {
	rad=1.;  // initial radius 
	const int n1=51;
	x0.redim(n1,2);

	for( int i=0; i<n1; i++ )
	{
	  theta = i*2.*Pi/(n1-1);
	  x0(i,0) = rad*cos(theta);
	  x0(i,1) = rad*sin(theta);
	}

	NurbsMapping & startCurve = *((NurbsMapping*) surface[face]);
	startCurve.setIsPeriodic(axis1,Mapping::functionPeriodic);

      }
      else if( userDefinedDeformingBodyMotionOption==sphereDeform )
      {

	Mapping & map = cg[gridToMove].mapping().getMapping();
	assert( map.getClassName()=="HyperbolicMapping" );
	HyperbolicMapping & hype = (HyperbolicMapping&)map;

        Mapping *pStartSurface = (Mapping*)hype.getSurface();
        assert( pStartSurface!=NULL );
        Mapping & startSurface = *pStartSurface;

	GenericGraphicsInterface & gi = *Overture::getGraphicsInterface();

	if( false )
	{
	  gi.stopReadingCommandFile();
	  printF("DeformingBodyMotion:initialize: surface Mapping for hyperbolic grid=%i\n",
		 gridToMove);
	
	  gi.erase();
	  PlotIt::plot(gi,startSurface);
	}
	
        NurbsMapping & surf = *((NurbsMapping*)surface[face]);

        int degree=3;
        surf.interpolate(startSurface,degree,NurbsMapping::parameterizeByIndex);

        realArray surfGrid; surfGrid = surf.getGrid();
        #ifdef USE_PPP
	  getLocalArrayWithGhostBoundaries( surfGrid, x0 );
        #else
	  x0.reference(surfGrid);
        #endif 
        Index I1=x0.dimension(0), I2=x0.dimension(1);
        x0.reshape(I1,I2,numberOfDimensions);

      }
      else if( userDefinedDeformingBodyMotionOption==advectBody ||
               userDefinedDeformingBodyMotionOption==interfaceDeform )
      {
        // ** do this for now **

	Mapping & map = cg[gridToMove].mapping().getMapping();
	assert( map.getClassName()=="HyperbolicMapping" );
	HyperbolicMapping & hype = (HyperbolicMapping&)map;

        Mapping *pStartSurface = (Mapping*)hype.getSurface();
        assert( pStartSurface!=NULL );
        Mapping & startSurface = *pStartSurface;

	GenericGraphicsInterface & gi = *Overture::getGraphicsInterface();

	if( false )
	{
	  gi.stopReadingCommandFile();
	  printF("DeformingBodyMotion:initialize: surface Mapping for hyperbolic grid=%i\n",
		 gridToMove);
	
	  gi.erase();
	  PlotIt::plot(gi,startSurface);
	}
	
        NurbsMapping & surf = *((NurbsMapping*)surface[face]);

        int degree=3;
        surf.interpolate(startSurface,degree,NurbsMapping::parameterizeByIndex);

        realArray surfGrid; surfGrid = surf.getGrid();
        #ifdef USE_PPP
	  getLocalArrayWithGhostBoundaries( surfGrid, x0 );
        #else
	  x0.reference(surfGrid);
        #endif 
        Index I1=x0.dimension(0), I2=x0.dimension(1);
	if( numberOfDimensions==2 )
	  x0.reshape(I1,numberOfDimensions);
        else 
	  x0.reshape(I1,I2,numberOfDimensions);

      }
      else if( userDefinedDeformingBodyMotionOption==elasticShell ||
               userDefinedDeformingBodyMotionOption==userDeformingSurface )
      {

	Index Ib1,Ib2,Ib3;
        const int numGhost=1;  // include ghost points 
	getBoundaryIndex(cg[gridToMove].gridIndexRange(),sideToMove,axisToMove,Ib1,Ib2,Ib3,numGhost);

        Range Rx=numberOfDimensions;
	x0.redim(Ib1,Ib2,Ib3,Rx);
	cg[gridToMove].update(MappedGrid::THEvertex | MappedGrid::THEcenter); // do this for now 
	
        #ifdef USE_PPP
          RealArray vertex; getLocalArrayWithGhostBoundaries(cg[gridToMove].vertex(),vertex);
        #else
          RealArray & vertex = cg[gridToMove].vertex();
        #endif

	// Here is the undeformed state
        x0=vertex(Ib1,Ib2,Ib3,Rx);

      }
      else
      {
	printF("DeformingBodyMotion:ERROR: unexpected value for userDefinedDeformingBodyMotionOption=%i\n",
	       (int)userDefinedDeformingBodyMotionOption);
	Overture::abort("DeformingBodyMotion:ERROR: unexpected value for userDefinedDeformingBodyMotionOption!");
      }
    
    
//     const IntegerArray & gid = cg[gridToMove].gridIndexRange();
//     for( int axis=0; axis<cg.numberOfDimensions(); axis++ )
//       deformingGridPoints[axis]=gid(1,axis)-gid(0,axis)+1;

//     printF("DeformingBodyMotion::initialize: deformingGridPoints=%i, %i, %i\n",
// 	   deformingGridPoints[0],deformingGridPoints[1],deformingGridPoints[2]);
    
    }
    else
    {
      Overture::abort("ERROR: unknown deformingBodyType");
    }
    
  } // end for face 
  
  initializeGrid( cg,t );  // *wdh* 080221

  return( ierr );
}

// ==================================================================================
///  \brief This initialization routine will adjust the initial CompositeGrid so that
///  it can be deformed. This usually means that we change the Mapping for the grid that deforms.
/// \param cg (input) : CompositeGrid to use
/// \t (input) : time 
// ==================================================================================
int DeformingBodyMotion::
initializeGrid(CompositeGrid & cg, real t /* =0. */ ) 
{
  int & numberOfDeformingGrids = deformingBodyDataBase.get<int>("numberOfDeformingGrids");
  int & numberOfFaces = deformingBodyDataBase.get<int>("numberOfFaces");
  IntegerArray & boundaryFaces = deformingBodyDataBase.get<IntegerArray>("boundaryFaces");
  DeformingBodyType & deformingBodyType = 
                  deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");

  for( int face=0; face<numberOfFaces; face++ )
  {
    int gridToMove= boundaryFaces(2,face);  


    if( deformingBodyType==userDefinedDeformingBody )
    {

      if( debug & 2 ) 
        printF(">>>DeformingBodyMotion::initializeGrid: grid=%i (face=%i)\n",gridToMove,face);
    
      // allocate a new GridEvolution for this grid 
      vector<GridEvolution*> & gridEvolution = deformingBodyDataBase.get<vector<GridEvolution*> >("gridEvolution");
      if( face >=gridEvolution.size() )
        gridEvolution.push_back(new GridEvolution);

      // set the order of accuracy for computing the velocity and acceleration
      gridEvolution[face]->setVelocityOrderOfAccuracy(deformingBodyDataBase.get<int>("velocityOrderOfAccuracy"));
      gridEvolution[face]->setAccelerationOrderOfAccuracy(deformingBodyDataBase.get<int>("accelerationOrderOfAccuracy"));

      Mapping & map = cg[gridToMove].mapping().getMapping();

      //  ---------------------------------------------------------------------------------
      //  --- if the grid is already a hyperbolic mapping we don't need to replace it  ----
      //  ---------------------------------------------------------------------------------

      HyperbolicMapping *pHype=NULL;
      if( map.getClassName()!="HyperbolicMapping" )
      {  
        // ************ The Mapping is NOT a HyperbolicMapping : create one ******************


	// The "surface" Mapping holds the start curve in 2D
	vector<Mapping*> & surface = deformingBodyDataBase.get<vector<Mapping*> >("surface");
	assert( face<surface.size() );
	NurbsMapping & startCurve = *((NurbsMapping*)surface[face]);

	vector<RealArray*> & surfaceArray = deformingBodyDataBase.get<vector<RealArray*> >("surfaceArray");
	assert( face<surfaceArray.size() );
	RealArray *px = surfaceArray[face];

	assert( px!=NULL );
	RealArray &x0 = px[0];
        #ifdef USE_PPP
	  Overture::abort("fix me");
        #else	
	  // *wdh* 081107 startCurve.interpolate(x0);
          int option=0, degree=3;
          const int boundaryParameterization = deformingBodyDataBase.get<int>("boundaryParameterization");
          startCurve.interpolate(x0,option,Overture::nullRealDistributedArray(),degree,
                                 (NurbsMapping::ParameterizationTypeEnum)boundaryParameterization);
        #endif
	vector<Mapping*> & transform =deformingBodyDataBase.get<vector<Mapping*> >("transform");

	if( face >=transform.size() )
	  transform.push_back(new HyperbolicMapping);
      

	assert( transform[face]->getClassName()=="HyperbolicMapping" );
        pHype = (HyperbolicMapping *)transform[face];
	HyperbolicMapping & hyp = *pHype;
    

	// hyp.setSurface(startCurve,false);
	const bool isSurfaceGrid=false; // this is a volume grid (in 2d or 3d) 
	const bool init=false; // this means keep existing hype parameters such as distanceToMarch, linesToMarch etc.
	hyp.setSurface(startCurve,isSurfaceGrid,init);

        // **** we need to set the marching parameters ****
        printF(">>>DeformingBodyMotion::initializeGrid: You must build the initial hyperbolic grid now...\n");
      
        GenericGraphicsInterface & gi = *Overture::getGraphicsInterface();
	
	hyp.interactiveUpdate(gi);
	// hyp.generate();

	cg[gridToMove].reference(hyp);  

	cg[gridToMove].update(MappedGrid::THEcenter | MappedGrid::THEvertex);    


      }
      else
      {
        pHype = &( (HyperbolicMapping &)map);
        
      }
      
      


      assert( pHype!=NULL );
      HyperbolicMapping & hyp = *pHype;

      if( debug & 4 )
      {
        printF("DeformingBodyMotion::initializeGrid: spacingOption=%i\n",hyp.spacingOption);
	printF("DeformingBodyMotion::initializeGrid: HyperbolicMapping Parameters:");
        hyp.display();
      }


#ifndef USE_PPP
      // turn this off temporarily for PPP until we update Overture
      const DataPointMapping *pdpm = hyp.getDataPointMapping();
#else
      DataPointMapping *pdpm = NULL; 
#endif
      assert( pdpm!=NULL );
      DataPointMapping & dpm = (DataPointMapping&)(*pdpm);
      // dpm.getDataPoints() includes ghost points

      if( debug & 2 ) 
	printF(">>>DeformingBodyMotion::add initial grid to GridEvolution, gridToMove=%i, t=%8.2e\n",gridToMove,t);

      gridEvolution[face]->addGrid(dpm.getDataPoints(),t);
      

    } // end if deformingBodyType==userDefinedDeformingBody
  
  } // end for face
  
  return 0;
}

//
// .. CALLED in the beginning of a computation to initialize
//    the mapping history list with reasonable values so the
//    grid velocity gets computed OK from the very first tstep
//
// .. This will have to change once the surface gets it's own physics
// .. --> use x_t, x_tt at t=0 to find boundary curve at t= -dt, t=-2*dt
//
int DeformingBodyMotion::
initializePast( real time00, real dt00, CompositeGrid & cg)
{
  int & numberOfDeformingGrids = deformingBodyDataBase.get<int>("numberOfDeformingGrids");
  int & numberOfFaces = deformingBodyDataBase.get<int>("numberOfFaces");
  IntegerArray & boundaryFaces = deformingBodyDataBase.get<IntegerArray>("boundaryFaces");
  DeformingBodyType & deformingBodyType = 
                  deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");

  // do this for now -- treat more grids later 
  for( int face=0; face<numberOfFaces; face++ )
  {
    int gridToMove= boundaryFaces(2,face);  


    if( deformingBodyType==elasticFilament )
    {
      assert( numberOfDeformingGrids==1 );
      assert( numberOfFaces==1 );

      assert(pElasticFilament!=NULL);
      assert(pDeformingGrid!=NULL);

      if(debug&2) cout << "DeformingBodyMotion::initializePast called\n";
      real time2=time00-2*dt00;  //past time levels
      real time1=time00-dt00;

      HyperbolicMapping *pHyper;
      aString newMappingName;   // for debugging: change name of hyperb!!

      //..Initialize past time-2*dt
      pDeformingGrid->getNewMapping( time2, pHyper); // t=-2*dt
      assert( pHyper != NULL);
      pElasticFilament->evaluateSurfaceAtTime(time2); 
      if (debug &4) {
	pDeformingGrid->createMappingName( time2, newMappingName );
	pElasticFilament->copyBodyFittedMapping( *pHyper, &newMappingName );
	cout << "initPast: time level 2 = time-2dt\n";
	if(pMapInfoDebug!=NULL) pHyper->update( *pMapInfoDebug ); 
      } else {
	pElasticFilament->copyBodyFittedMapping( *pHyper ); //non debug code
      }
      pElasticFilament->referenceMap( gridToMove,  cg); //TRYING TO FORCE FILAM TO SAVE

      //..Initialize past time1-dt
      pDeformingGrid->getNewMapping( time1, pHyper); // t=  -dt
      assert( pHyper != NULL);
      pElasticFilament->evaluateSurfaceAtTime(time1);
      if (debug &4) {
	pDeformingGrid->createMappingName( time1, newMappingName );
	pElasticFilament->copyBodyFittedMapping( *pHyper, &newMappingName );
	cout << "initPast: time level 1 = time-dt\n";
	if(pMapInfoDebug!=NULL) pHyper->update( *pMapInfoDebug ); 
      } else {
	pElasticFilament->copyBodyFittedMapping( *pHyper ); //non debug code
    
      }
      //TEMP FIX **pf--not needed, feb26,2001?---still needed, closed filam;-(
      if( pMapInfoDebug!=NULL) pHyper->update( *pMapInfoDebug );

      pElasticFilament->referenceMap( gridToMove,  cg); //TRYING TO FORCE FILAM TO SAVE
  
      //..Initialize current TIME!!
      pDeformingGrid->getNewMapping( time00, pHyper); // t=  -dt
      assert( pHyper != NULL);
      //pElasticFilament->evaluateSurfaceAtTime( time00 );
      pElasticFilament-> initializeSurfaceData( time00 );

      pElasticFilament->copyBodyFittedMapping( *pHyper, &newMappingName );
      if (debug &4) {
	pDeformingGrid->createMappingName( time00, newMappingName );
	pElasticFilament->copyBodyFittedMapping( *pHyper, &newMappingName );
	cout << "initPast: time level 0 = current\n";
	//if(pMapInfoDebug!=NULL) pHyper->update( *pMapInfoDebug ); 
      } else {
	pElasticFilament->copyBodyFittedMapping( *pHyper ); //non debug code
      }
      pElasticFilament->referenceMap( gridToMove,  cg); //TRYING TO FORCE FILAM TO SAVE

    }
    else if( deformingBodyType==userDefinedDeformingBody )
    {
      printF("DeformingBodyMotion:initializePast called for userDefinedDeformingBody for t=%9.3e, dt=%9.3e\n",
	     time00,dt00);
    }
    else
    {
      Overture::abort("ERROR: unknown deformingBodyType");
    }

  } // end for face

  return 0;
}

// ===========================================================================================
/// \brief Advance the elastic shell in time.
//
/// \param option : option=0 : predictor-step, option=1 corrector step. 
//
// Here is a simple approximation to an elastic shell
// 
// The equation of motion of the interface x(s,t) is:
//
//  (rhoe*ds) dv/dt = p ds n - ke*( x-x0 ) + te*(x-x0).ss - be*v 
//    dx/dt = v 
//
//  where:
//    x0(s) : undeformed state.
//    rhoe : "density" of the shell. 
//    ds : arc length 
//    ke : restoring "spring" coefficient
//    te : tension term 
//    be : damping coefficient
// 
// ===========================================================================================
int DeformingBodyMotion::
advanceElasticShell(real t1, real t2, real t3, 
		    GridFunction & cgf1,
		    GridFunction & cgf2,
		    GridFunction & cgf3,
		    realCompositeGridFunction & stress,
                    int option )
{
  int ierr=0;

  const int numberOfDimensions = cgf1.cg.numberOfDimensions();
  int & numberOfDeformingGrids = deformingBodyDataBase.get<int>("numberOfDeformingGrids");
  int & numberOfFaces = deformingBodyDataBase.get<int>("numberOfFaces");
  IntegerArray & boundaryFaces = deformingBodyDataBase.get<IntegerArray>("boundaryFaces");
  DeformingBodyType & deformingBodyType = 
                  deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");
  UserDefinedDeformingBodyMotionEnum & userDefinedDeformingBodyMotionOption = 
    deformingBodyDataBase.get<UserDefinedDeformingBodyMotionEnum>("userDefinedDeformingBodyMotionOption");

  assert( deformingBodyType==userDefinedDeformingBody );
  assert( userDefinedDeformingBodyMotionOption==elasticShell );

  CompositeGrid & cg = cgf3.cg;

  real tForce=t2;
  real tNew=t3;
  
  if( debug&2 )
    printF("--DeformingBodyMotion::advanceElasticShell called for tNew=%f, forces at t=%f\n",tNew, tForce);

  for( int face=0; face<numberOfFaces; face++ )
  {
    int sideToMove=boundaryFaces(0,face);
    int axisToMove=boundaryFaces(1,face);
    int gridToMove=boundaryFaces(2,face); 


    const int uc = parameters.dbase.get<int >("uc");
    const int vc = parameters.dbase.get<int >("vc");
    const int wc = parameters.dbase.get<int >("wc");
    const int pc = parameters.dbase.get<int >("pc");

    // The "surface" Mapping holds the start curve in 2D
    vector<Mapping*> & surface = deformingBodyDataBase.get<vector<Mapping*> >("surface");
    assert( face<surface.size() );
    NurbsMapping & startCurve = *((NurbsMapping*)surface[face]);


    // We need to grab the surface points from cgf1
    const real dt = t3-t2;   // do this for now -- assume the solution x0 is at time t2

    real *par = deformingBodyDataBase.get<real [10]>("elasticShellParameters");
    real & rhoe = par[0], &te=par[1], &ke=par[2], &be=par[3], ad2=par[4];
    assert( rhoe>0. );

    // The undeformed surface, x0,  is stored here: 
    vector<RealArray*> & surfaceArray = deformingBodyDataBase.get<vector<RealArray*> >("surfaceArray");
    vector<real*> & surfaceArrayTime = deformingBodyDataBase.get<vector<real*> >("surfaceArrayTime"); 
    assert( face<surfaceArray.size() );
    RealArray *px = surfaceArray[face];
    RealArray &x0 = px[0];
    assert( face<surfaceArrayTime.size() );
    real & tx0= surfaceArrayTime[face][0];

    const int numGhost=1;  // include ghost points 
    Range Rx=numberOfDimensions;
    Index Ib1,Ib2,Ib3;

    // --- The current surface position and velocity are stored in the GridFunction data-base ---
    for( int m=0; m<3; m++ )
    {
      GridFunction & gf = m==0 ? cgf1 : m==1 ? cgf2 : cgf3;
      if( !gf.dbase.has_key("xShell") )
      {
	// note: include ghost points:
	getBoundaryIndex(cgf1.cg[gridToMove].gridIndexRange(),sideToMove,axisToMove,Ib1,Ib2,Ib3,numGhost);
	gf.dbase.put<RealArray>("xShell");
	gf.dbase.put<RealArray>("vShell");

	RealArray & x = gf.dbase.get<RealArray>("xShell");
	RealArray & v = gf.dbase.get<RealArray>("vShell");
 
	gf.cg[gridToMove].update(MappedGrid::THEvertex | MappedGrid::THEcenter); // do this for now 
#ifdef USE_PPP
	RealArray vertex; getLocalArrayWithGhostBoundaries(gf.cg[gridToMove].vertex(),vertex);
#else
	RealArray & vertex = gf.cg[gridToMove].vertex();
#endif
	x= vertex(Ib1,Ib2,Ib3,Rx);
	v.redim(Ib1,Ib2,Ib3,Rx);

	v=0.;                       // do this for now   ** fix me **
      }
    }
	
    getBoundaryIndex(cgf1.cg[gridToMove].gridIndexRange(),sideToMove,axisToMove,Ib1,Ib2,Ib3);

    RealArray & x1 = cgf1.dbase.get<RealArray>("xShell");
    RealArray & v1 = cgf1.dbase.get<RealArray>("vShell");
	
    RealArray & x2 = cgf2.dbase.get<RealArray>("xShell");
    RealArray & v2 = cgf2.dbase.get<RealArray>("vShell");
	
    RealArray & x3 = cgf3.dbase.get<RealArray>("xShell");
    RealArray & v3 = cgf3.dbase.get<RealArray>("vShell");
	

    // RealArray x1,x2,x3, v1,v2,v3;

#ifdef USE_PPP
    RealArray u2; getLocalArrayWithGhostBoundaries(cgf2.u[gridToMove],u2);
    const RealArray & normal2 = cgf2.cg[gridToMove].vertexBoundaryNormalArray(sideToMove,axisToMove);
    RealArray vertex2; getLocalArrayWithGhostBoundaries(cgf2.cg[gridToMove].vertex(),vertex2);
#else
    RealArray & u2 = cgf2.u[gridToMove];
    const RealArray & normal2 = cgf2.cg[gridToMove].vertexBoundaryNormal(sideToMove,axisToMove);
    const RealArray & vertex2 = cgf2.cg[gridToMove].vertex();
#endif
	
    if( numberOfDimensions==2 )
    {
      assert( uc>=0 && vc>=0 && pc>=0 );
      int axisp = (axisToMove + 1) % numberOfDimensions;  // axis in the tangential direction 
      // Index I1=x0.dimension(0), I2=x0.dimension(1);
      int iv[3], &i1=iv[0], &i2=iv[1], &i3=iv[2];
      int ivp[3], &i1p=ivp[0], &i2p=ivp[1], &i3p=ivp[2];
      int ivm[3], &i1m=ivm[0], &i2m=ivm[1], &i3m=ivm[2];
      FOR_3D(i1,i2,i3,Ib1,Ib2,Ib3)
      {
	// int j1 = iv[axisp];  // j1 = i1 or i2 or i3 
	i1p=i1, i2p=i2, i3p=i3;
	i1m=i1, i2m=i2, i3m=i3;
	ivp[axisp]+=1;
	ivm[axisp]-=1;
	real p2=u2(i1,i2,i3,pc);
	    
	// compute the local arc-length (use vertex2 since it has ghost points)
	real ds = .5*( sqrt( SQR(vertex2(i1p,i2p,i3p,0)-vertex2(i1m,i2m,i3m,0))+
			     SQR(vertex2(i1p,i2p,i3p,1)-vertex2(i1m,i2m,i3m,1)) ) );

	real ac[3]; // holds acceleration 
	for( int dir=0; dir<numberOfDimensions; dir++ )
	{
	  // Curvature term: dxss = (x-x0).ss 
	  // ** we need ghost point values for this ***
	  real dxss =( (vertex2(i1p,i2p,i3p,dir)-2.*vertex2(i1,i2,i3,dir)+vertex2(i1m,i2m,i3m,dir)) - 
		       (x0(i1p,i2p,i3p,dir)-2.*x0(i1,i2,i3,dir)+x0(i1m,i2m,i3m,dir)) )/(ds*ds);

	  // smoothing terms
	  real vss = v2(i1p,i2p,i3p,dir)-2.*v2(i1,i2,i3,dir)+v2(i1m,i2m,i3m,dir);
	  real xss = x2(i1p,i2p,i3p,dir)-2.*x2(i1,i2,i3,dir)+x2(i1m,i2m,i3m,dir);

	  ac[dir]=( p2*normal2(i1,i2,i3,dir) -(ke/ds)*( x2(i1,i2,i3,dir)-x0(i1,i2,i3,dir) ) 
		    + te*dxss -be*v2(i1,i2,i3,dir) )/(rhoe) + ad2*vss;
	  v3(i1,i2,i3,dir) = v1(i1,i2,i3,dir) + dt*ac[dir];
	  x3(i1,i2,i3,dir) = x1(i1,i2,i3,dir) + dt*( v2(i1,i2,i3,dir) +ad2*dxss*ds*ds ) ;
	}
	if( false )
	{
	  printF("elasticShell: x=(%5.2f,%5.2f) x0=(%5.2f,%5.2f) ds=%8.2e v=(%5.2f,%5.2f) a=(%8.2e,%8.2e) "
		 "(u,v,p)=(%8.2e,%8.2e,%8.2e)\n",
		 x3(i1,i2,i3,0),x3(i1,i2,i3,1),x0(i1,i2,i3,0),x0(i1,i2,i3,1),ds, 
		 v3(i1,i2,i3,0),v3(i1,i2,i3,1),ac[0],ac[1],u2(i1,i2,i3,uc),u2(i1,i2,i3,vc),p2);
	      
	}
      }

      // *** do this for now ***
      startCurve.setIsPeriodic(axis1,Mapping::functionPeriodic);

      x3.reshape(x3.dimension(axisp),Rx);

      #ifdef USE_PPP
	Overture::abort("fix me");
      #else
      int option=0, degree=3;
      const int boundaryParameterization = deformingBodyDataBase.get<int>("boundaryParameterization");
      startCurve.interpolate(x3,option,Overture::nullRealDistributedArray(),degree,
                             (NurbsMapping::ParameterizationTypeEnum)boundaryParameterization,numGhost);
      // 081107 startCurve.interpolate(x3,option,Overture::nullRealDistributedArray(),degree,
      //                               NurbsMapping::parameterizeByChordLength,numGhost);
      #endif
      getBoundaryIndex(cgf1.cg[gridToMove].gridIndexRange(),sideToMove,axisToMove,Ib1,Ib2,Ib3,numGhost);
      x3.reshape(Ib1,Ib2,Ib3,Rx);

    }
    else if( numberOfDimensions==3)
    {
      Overture::abort("error");
    }
	
    tx0=t3; // x0 points now live at this time

    
  } // end for face
  
  return ierr;
}



//\begin{>>DeformingBodyMotionInclude.tex}{\subsection{integrate}} 
int DeformingBodyMotion::
integrate( real t1, real t2, real t3, 
           GridFunction & cgf1,
           GridFunction & cgf2,
           GridFunction & cgf3,
           realCompositeGridFunction & stress )
//==================================================================================
// /Description:
//    Integrate the Deforming body equations from time t1 to time t3 using the solution at time t2
//   to compute the forces etc. 
// \begin{verbatim}
//          u3(t3) <- u1(t1) + (t3-t1)*d(u2(t2))/dt
// \end{verbatim}
//
//  cgf1 and cgf3 must be DIFFERENT
//
//  /t1,cgf1 (input): grid and solution at time t1
//  /t2,cgf2 (input): grid velocity is taken from this time
//
//  /cgf3 (output): cgf3 can be used to hold the new solution at time t3 for variables that are
//         computed with by Deforming body class. 
//
//
//\end{DeformingBodyMotionInclude.tex}  
//=========================================================================================
{

  const int numberOfDimensions = cgf1.cg.numberOfDimensions();
  int & numberOfDeformingGrids = deformingBodyDataBase.get<int>("numberOfDeformingGrids");
  int & numberOfFaces = deformingBodyDataBase.get<int>("numberOfFaces");
  
  IntegerArray & boundaryFaces = deformingBodyDataBase.get<IntegerArray>("boundaryFaces");
  DeformingBodyType & deformingBodyType = 
    deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");
  UserDefinedDeformingBodyMotionEnum & userDefinedDeformingBodyMotionOption = 
    deformingBodyDataBase.get<UserDefinedDeformingBodyMotionEnum>("userDefinedDeformingBodyMotionOption");

  CompositeGrid & cg = cgf3.cg;

  const int numberOfComponentGrids = cg.numberOfComponentGrids();

  real tForce=t2;
  real tNew=t3;
  
  if( debug&2 )
    printF("--DeformingBodyMotion::Integrate called for tNew=%f, forces at t=%f\n",tNew, tForce);
  int ierr=0;

  if( deformingBodyType==userDefinedDeformingBody && 
      userDefinedDeformingBodyMotionOption==elasticShell )
  {
    int option=0;  // predictor 
    ierr=advanceElasticShell(t1,t2,t3,cgf1,cgf2,cgf3,stress,option);
    return ierr;
   
  }
  else if( deformingBodyType==userDefinedDeformingBody && 
	   userDefinedDeformingBodyMotionOption==userDeformingSurface )
  {
    int option=0;  // predictor 
    ierr=userDefinedDeformingSurface(t1,t2,t3,cgf1,cgf2,cgf3,option);
    return ierr;
   
  }

  for( int face=0; face<numberOfFaces; face++ )
  {
    int sideToMove=boundaryFaces(0,face);
    int axisToMove=boundaryFaces(1,face);
    int gridToMove=boundaryFaces(2,face); 

    assert( gridToMove>=0 && gridToMove<numberOfComponentGrids );

    if( deformingBodyType==elasticFilament )
    {
      //call DeformingBodySolver --> now just special case=ElasticFilament
      assert( pElasticFilament!=NULL );
      ierr=pElasticFilament->integrate(tForce, stress, tNew);
      //int pElasticFilament->evaluateSurfaceAtTime( tNew );
    }
    else if( deformingBodyType==userDefinedDeformingBody )
    {

      if(  userDefinedDeformingBodyMotionOption==advectBody )
      {
        // advect the body from time t1 to t3 using velocity from time t2 

	if( debug & 2 )
	  printF("DeformingBodyMotion:integrate called for userDefinedDeformingBody for t=%9.3e\n",tNew);

	const int uc = parameters.dbase.get<int >("uc");
	const int vc = parameters.dbase.get<int >("vc");
	const int wc = parameters.dbase.get<int >("wc");

	vector<RealArray*> & surfaceArray = deformingBodyDataBase.get<vector<RealArray*> >("surfaceArray");
        vector<real*> & surfaceArrayTime = deformingBodyDataBase.get<vector<real*> >("surfaceArrayTime"); 
	assert( face<surfaceArray.size() );
	RealArray *px = surfaceArray[face];
	RealArray &x0 = px[0], &x1 = px[1], &x2=px[2];
        assert( face<surfaceArrayTime.size() );
	real & tx0= surfaceArrayTime[face][0];
	
	// Mapping & map = cg[gridToMove].mapping().getMapping();
	// assert( map.getClassName()=="HyperbolicMapping" );
	// HyperbolicMapping & hyp = (HyperbolicMapping&)map;
    
        // We need to grab the surface points from cgf1
        // const real dt = t3-t2;   // do this for now -- assume the solution x0 is at time t2
        // *wdh* 081211 -- use the proper time (needed for advancePC which takes an initial back step to t=-dt)
        const real dt = t3-tx0;   
        tx0=t3; // x0 will now live at this time
        realArray & u = cgf2.u[gridToMove];
	
	Index Ib1,Ib2,Ib3;
	getBoundaryIndex(cgf1.cg[gridToMove].gridIndexRange(),sideToMove,axisToMove,Ib1,Ib2,Ib3);

	if( numberOfDimensions==2 )
	{
          assert( uc>=0 && vc>=0 );
          int axisp = (axisToMove + 1) % numberOfDimensions;
	  Index I1=x0.dimension(0), I2=x0.dimension(1);
	  int iv[3], &i1=iv[0], &i2=iv[1], &i3=iv[2];
          FOR_3D(i1,i2,i3,Ib1,Ib2,Ib3)
	  {
            int j1 = iv[axisp];  // j1 = i1 or i2 or i3 
	    real x = x0(j1,0), y=x0(j1,1);
	    real u0=u(i1,i2,i3,uc), v0=u(i1,i2,i3,vc);
	      
            u0=1.;  v0=0.; // do this for testing 
	     
	    x0(j1,0) = x + dt*u0;
	    x0(j1,1) = y + dt*v0;

	    // printF("advectBody: x=(%5.2f,%5.2f) u=(%5.2f,%5.2f) x0=(%5.2f,%5.2f)\n",x,y,u0,v0,x0(j1,0),x0(j1,1));
	    
	  }


	}
	else if( numberOfDimensions==3)
	{
          assert( uc>=0 && vc>=0 && wc>= 0 );

	  Index I1=x0.dimension(0), I2=x0.dimension(1);
	  int i1,i2,i3;
	  for( int i2=I2.getBase(); i2<=I2.getBound(); i2++ )
	    for( int i1=I1.getBase(); i1<=I1.getBound(); i1++ )
	    {
	      real x = x0(i1,i2,0), y=x0(i1,i2,1), z=x0(i1,i2,2);
	      real u0=u(i1,i2,i3,uc), v0=u(i1,i2,i3,vc), w0=u(i1,i2,i3,wc);
	      
	      x0(i1,i2,0) = x + dt*u0;
	      x0(i1,i2,1) = y + dt*v0;
	      x0(i1,i2,2) = z + dt*w0;

	    }
	}
	
	// The "surface" Mapping holds the start curve in 2D
	vector<Mapping*> & surface = deformingBodyDataBase.get<vector<Mapping*> >("surface");
	assert( face<surface.size() );
	NurbsMapping & startCurve = *((NurbsMapping*)surface[face]);

	// *** do this for now *** FIX ME ******************************************************
	// startCurve.setIsPeriodic(axis1,Mapping::functionPeriodic);

	// startCurve.setIsPeriodic(axis1,Mapping::derivativePeriodic);

        #ifdef USE_PPP
  	  Overture::abort("fix me");
        #else
	  // *wdh* 081107 startCurve.interpolate(x0);
          int option=0, degree=3;
          const int boundaryParameterization = deformingBodyDataBase.get<int>("boundaryParameterization");
          startCurve.interpolate(x0,option,Overture::nullRealDistributedArray(),degree,
                                 (NurbsMapping::ParameterizationTypeEnum)boundaryParameterization);
        #endif
        // NOTE: We should evaluate the GRID-POINTS at interpolation pts by interpolating
        //       the grid-points from other grids -- this should make sure that overlapping grids
        //       stay connected. 
	
      }
      else if( userDefinedDeformingBodyMotionOption==interfaceDeform )
      {
        // ----------------------------------------------------------------------------------------------------
        // --- The interface is defined by the "boundaryData" array used for the RHS to boundary conditions ---
        // ----------------------------------------------------------------------------------------------------

	if( debug & 4 )
	  printF("DeformingBodyMotion:integrate called for interfaceDeform at t=%9.3e\n",tNew);

	

	// Here is the array that defines the domain interfaces, interfaceType(side,axis,grid) 
	const IntegerArray & interfaceType = parameters.dbase.get<IntegerArray >("interfaceType");


        RealArray & bd = parameters.getBoundaryData(sideToMove,axisToMove,gridToMove,cg[gridToMove]);
 	if( debug & 8  )
 	{
 	  ::display(bd,"DeformingBodyMotion::integrate:interfaceDeform: Here is the boundary data","%6.3f ");
 	}
	
	const int uc = parameters.dbase.get<int >("uc");
	const int vc = parameters.dbase.get<int >("vc");
	const int wc = parameters.dbase.get<int >("wc");

	vector<RealArray*> & surfaceArray = deformingBodyDataBase.get<vector<RealArray*> >("surfaceArray");
        vector<real*> & surfaceArrayTime = deformingBodyDataBase.get<vector<real*> >("surfaceArrayTime"); 
	assert( face<surfaceArray.size() );
	RealArray *px = surfaceArray[face];
	RealArray &x0 = px[0], &x1 = px[1], &x2=px[2];
        assert( face<surfaceArrayTime.size() );
	real & tx0= surfaceArrayTime[face][0];

	// Mapping & map = cg[gridToMove].mapping().getMapping();
	// assert( map.getClassName()=="HyperbolicMapping" );
	// HyperbolicMapping & hyp = (HyperbolicMapping&)map;
    
        // We need to grab the surface points from cgf1
        // *wdh* const real dt = t3-t2;   // do this for now -- assume the solution x0 is at time t2
        // *wdh* 081211 -- use the proper time (needed for advancePC which takes an initial back step to t=-dt)
        const real dt = t3-tx0;   
        tx0=t3; // x0 will not live at this time

        realArray & u = cgf2.u[gridToMove];
	
	Index Ib1,Ib2,Ib3;
	getBoundaryIndex(cgf1.cg[gridToMove].gridIndexRange(),sideToMove,axisToMove,Ib1,Ib2,Ib3);
	int iv[3], &i1=iv[0], &i2=iv[1], &i3=iv[2];
	
        if( true )
	{

	  assert( numberOfDimensions==2 );
          // *** fix me ***

          if( Ib1.getLength()>1 ) // *** NOTE: we know which (side,axis) the boundaryData lives on -- fix this --
	  { // the boundary follows "i1"
            if( x0.getBase()!=Ib1.getBase() || x0.getBound(0)!=Ib1.getBound() )
	    {
	      printF("DeformingBodyMotion::integrate:interfaceDeform:ERROR: boundary data array does not "
		     "start curve dimensions!\n"
		     " gridToMove=%i sideToMove=%i axisToMove=%i uc=%i\n"
		     " Ib1=[%i,%i], Ib2=[%i,%i], Ib2=[%i,%i],   x0 : [%i,%i]\n",
		     gridToMove,sideToMove,axisToMove,uc,
		     Ib1.getBase(),Ib1.getBound(),
		     Ib2.getBase(),Ib2.getBound(),
		     Ib3.getBase(),Ib3.getBound(),
		     x0.getBase(0),x0.getBound(0));
              Overture::abort("error");
	    }
            
	    FOR_3D(i1,i2,i3,Ib1,Ib2,Ib3)
	    {
              for( int m=0; m<numberOfDimensions; m++ )
		x0(i1,m)=bd(i1,i2,i3,m+uc);
	    }
	  }
	  else
	  { // the boundary follows "i2"
            if( x0.getBase()!=Ib2.getBase() || x0.getBound(0)!=Ib2.getBound() )
	    {
	      printF("DeformingBodyMotion::integrate:interfaceDeform:ERROR: boundary data array does not "
		     "start curve dimensions!\n"
		     " gridToMove=%i sideToMove=%i axisToMove=%i uc=%i\n"
		     " Ib1=[%i,%i], Ib2=[%i,%i], Ib2=[%i,%i],   x0 : [%i,%i]\n",
		     gridToMove,sideToMove,axisToMove,uc,
		     Ib1.getBase(),Ib1.getBound(),
		     Ib2.getBase(),Ib2.getBound(),
		     Ib3.getBase(),Ib3.getBound(),
		     x0.getBase(0),x0.getBound(0));
              Overture::abort("error");
	    }
	    

	    FOR_3D(i1,i2,i3,Ib1,Ib2,Ib3)
	    {
              for( int m=0; m<numberOfDimensions; m++ )
		x0(i2,m)=bd(i1,i2,i3,m+uc);
	    }
	  }
	  
	  if( debug & 8  )
	  {
	    printF("*** DeformingBodyMotion::integrate: gridToMove=%i sideToMove=%i axisToMove=%i uc=%i\n",
		   gridToMove,sideToMove,axisToMove,uc);
	    ::display(x0,"DeformingBodyMotion::integrate:interfaceDeform: Here are the start curve pts","%7.4f ");
	  }

	}
	else if( numberOfDimensions==2 )
	{
          // ***** finish me *****

          assert( uc>=0 && vc>=0 );

          int axisp = (axisToMove + 1) % numberOfDimensions;
	  Index I1=x0.dimension(0), I2=x0.dimension(1);
          FOR_3D(i1,i2,i3,Ib1,Ib2,Ib3)
	  {
            int j1 = iv[axisp];  // j1 = i1 or i2 or i3 
	    real x = x0(j1,0), y=x0(j1,1);
	    // real u0=u(i1,i2,i3,uc), v0=u(i1,i2,i3,vc);
	      
            real u0=1.,  v0=0.; 
	     
	    x0(j1,0) = x + dt*u0;
	    x0(j1,1) = y + dt*v0;

	    printF("interfaceDeform: x=(%5.2f,%5.2f) u=(%5.2f,%5.2f) x0=(%5.2f,%5.2f)\n",x,y,u0,v0,x0(j1,0),x0(j1,1));
	    
	  }

	}
	else if( numberOfDimensions==3)
	{
          // ***** finish me *****

          assert( uc>=0 && vc>=0 && wc>= 0 );

	  Index I1=x0.dimension(0), I2=x0.dimension(1);
	  for( int i2=I2.getBase(); i2<=I2.getBound(); i2++ )
	    for( int i1=I1.getBase(); i1<=I1.getBound(); i1++ )
	    {
	      real x = x0(i1,i2,0), y=x0(i1,i2,1), z=x0(i1,i2,2);
	      // real u0=u(i1,i2,i3,uc), v0=u(i1,i2,i3,vc), w0=u(i1,i2,i3,wc);
	      
              real u0=1., v0=0., w0=0.; 

	      x0(i1,i2,0) = x + dt*u0;
	      x0(i1,i2,1) = y + dt*v0;
	      x0(i1,i2,2) = z + dt*w0;

	    }
	}
	
	// The "surface" Mapping holds the start curve in 2D
	vector<Mapping*> & surface = deformingBodyDataBase.get<vector<Mapping*> >("surface");
	assert( face<surface.size() );
	NurbsMapping & startCurve = *((NurbsMapping*)surface[face]);

	// *** do this for now *** FIX ME ******************************************************
	// startCurve.setIsPeriodic(axis1,Mapping::functionPeriodic);

	// startCurve.setIsPeriodic(axis1,Mapping::derivativePeriodic);

        #ifdef USE_PPP
  	  Overture::abort("fix me");
        #else
	  // *wdh* 081107 startCurve.interpolate(x0);
          int option=0, degree=3;
          const int boundaryParameterization = deformingBodyDataBase.get<int>("boundaryParameterization");
          startCurve.interpolate(x0,option,Overture::nullRealDistributedArray(),degree,
                                (NurbsMapping::ParameterizationTypeEnum)boundaryParameterization);
        #endif
        // NOTE: We should evaluate the GRID-POINTS at interpolation pts by interpolating
        //       the grid-points from other grids -- this should make sure that overlapping grids
        //       stay connected. 
	
      }
      else if( userDefinedDeformingBodyMotionOption==elasticShell ||
               userDefinedDeformingBodyMotionOption==userDeformingSurface )
      {
        // this is done above 


      }

    }
    else
    {
      Overture::abort("ERROR: unknown deformingBodyType");
    } 
  }
  
  return ierr;
}


//==================================================================================
/// \brief Corrector step for deforming grids.
/// \details This function is called at the corrector step to update the moving grids. 
///  the acceleration on the boundary ).
/// \param t1,cgf1 (input) : solution at the old time
/// \param t2,cgf2 (input) : solution at the new time (these are valid values)
//==================================================================================
int DeformingBodyMotion::
correct( real t1, real t2, 
	 GridFunction & cgf1,GridFunction & cgf2 )
{
  if(debug&2)
    printF("--DeformingBodyMotion::correct called at t=%f (Not Implemented!!)\n", t2);

  DeformingBodyType & deformingBodyType = 
                  deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");

  if( deformingBodyType==elasticFilament )
  {
  }
  else if( deformingBodyType==userDefinedDeformingBody )
  {
    // *wdh* new 081206  -- add something like this ---
   // regenerateComponentGrids( t2,cgf2.cg,correctGrid );
    
  }
  else
  {
    Overture::abort("ERROR: unknown deformingBodyType");
  }
  return 0;
}


//........Access routines 

// ===================================================================================
/// \brief: Set the deforming body type. 
///    One of elasticFilament or userDefinedDeformingBody. 
// ===================================================================================
int DeformingBodyMotion::
setType( const DeformingBodyType bodyType )
{
  DeformingBodyType & deformingBodyType = 
                  deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");

  deformingBodyType=bodyType;
  return 0;
}


// ===================================================================================
/// \brief: regenerate all the component grids associated with this deforming body.
// ===================================================================================
int DeformingBodyMotion::
regenerateComponentGrids( const real newT, CompositeGrid & cg)
{
  int & numberOfDeformingGrids = deformingBodyDataBase.get<int>("numberOfDeformingGrids");
  int & numberOfFaces = deformingBodyDataBase.get<int>("numberOfFaces");
  IntegerArray & boundaryFaces = deformingBodyDataBase.get<IntegerArray>("boundaryFaces");
  DeformingBodyType & deformingBodyType = 
                  deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");

  // do this for now -- treat more grids later 
  for( int face=0; face<numberOfFaces; face++ )
  {
    int gridToMove= boundaryFaces(2,face);  

    if( deformingBodyType==elasticFilament )
    {
      assert( numberOfFaces==1 );

      if( debug&4 )
	printF("DeformingBodyMotion::regenCompGrid called, gridToMove=%i\n",gridToMove);

      assert( pElasticFilament!=NULL );
      assert(pDeformingGrid   !=NULL );

      HyperbolicMapping *pHyper;
      aString newMappingName;

      pDeformingGrid->getNewMapping( newT, pHyper );
      assert( pHyper != NULL );
      pElasticFilament->evaluateSurfaceAtTime(newT);
      if (debug &4) 
      {
	pDeformingGrid->createMappingName( newT, newMappingName );
	pElasticFilament->copyBodyFittedMapping( *pHyper, &newMappingName );
	cout << "DeformingBodyMotion::regenCompGrid -- "
	     << "look at stored time level\n";
	if(pMapInfoDebug!=NULL) pHyper->update( *pMapInfoDebug ); 
      } 
      else 
      {
	//cout << "DeformingBodyMotion::regenCompGrid -- "
	//	 << "look at stored time level @@ NON DEBUG\n";
	pElasticFilament->copyBodyFittedMapping( *pHyper ); //non debug code
      }
      pElasticFilament->referenceMap( gridToMove,  cg); //TRYING TO FORCE FILAM TO SAVE?? Is this needed feb26,2001**pf??
    }
    else if( deformingBodyType==userDefinedDeformingBody )
    {

      if( !deformingBodyDataBase.has_key("userDefinedDeformingBodyMotionOption") )
	deformingBodyDataBase.put<UserDefinedDeformingBodyMotionEnum>("userDefinedDeformingBodyMotionOption",iceDeform);
      UserDefinedDeformingBodyMotionEnum & userDefinedDeformingBodyMotionOption = 
	deformingBodyDataBase.get<UserDefinedDeformingBodyMotionEnum>("userDefinedDeformingBodyMotionOption");

      if( debug & 2 )
	printF("DeformingBodyMotion:regenerateComponentGrid: userDefinedDeformingBody for gridToMove=%i, t=%9.3e\n",
	       gridToMove,newT);
    
      vector<RealArray*> & surfaceArray = deformingBodyDataBase.get<vector<RealArray*> >("surfaceArray");
      vector<real*> & surfaceArrayTime = deformingBodyDataBase.get<vector<real*> >("surfaceArrayTime"); 
      assert( face<surfaceArray.size() );
      RealArray *px = surfaceArray[face];
      RealArray &x0 = px[0], &x1 = px[1], &x2=px[2];
      assert( face<surfaceArrayTime.size() );
      real & tx0= surfaceArrayTime[face][0];

      Mapping & map = cg[gridToMove].mapping().getMapping();
      assert( map.getClassName()=="HyperbolicMapping" );
      HyperbolicMapping & hyp = (HyperbolicMapping&)map;
    
      // The "surface" Mapping holds the start curve in 2D
      vector<Mapping*> & surface = deformingBodyDataBase.get<vector<Mapping*> >("surface");
      assert( face<surface.size() );
      NurbsMapping & startCurve = *((NurbsMapping*)surface[face]);

      if( !deformingBodyDataBase.has_key("deformationFrequency") )
	deformingBodyDataBase.put<real>("deformationFrequency",1.);
      real & deformationFrequency = deformingBodyDataBase.get<real>("deformationFrequency");
      if( !deformingBodyDataBase.has_key("deformationAmplitude") )
	deformingBodyDataBase.put<real>("deformationAmplitude",1.);
      real & deformationAmplitude = deformingBodyDataBase.get<real>("deformationAmplitude");


      real omega=pow(sin(deformationFrequency*Pi*newT),2.);  // omega varies in the interval [0,1]
    
      if( userDefinedDeformingBodyMotionOption==iceDeform )
      {
	if( true )
	{
	  // perturbation varies along the axis of of the cylinder:
	  Range I1=x2.dimension(0), I2=x2.dimension(1);
	  int ns= max(2,I2.getLength());
	  for( int i2=I2.getBase(); i2<=I2.getBound(); i2++ )
	  {
	    real s = real(i2)/(ns-1);  // axial variable

	    real varOmega=deformationAmplitude*omega*SQR(cos(2.*Pi*s));

	    x2(I1,i2) = (1.-varOmega)*x0(I1,i2) + varOmega*x1(I1,i2);
	  }
      
	}
	else
	{
	  x2 = (1.-omega)*x0 + omega*x1;
	}

        #ifdef USE_PPP
  	  Overture::abort("fix me");
        #else
	  // *wdh* 081107 startCurve.interpolate(x2);
          int option=0, degree=3;
          const int boundaryParameterization = deformingBodyDataBase.get<int>("boundaryParameterization");
          startCurve.interpolate(x2,option,Overture::nullRealDistributedArray(),degree,
                                 (NurbsMapping::ParameterizationTypeEnum)boundaryParameterization);
        #endif
      }
      else  if( userDefinedDeformingBodyMotionOption==ellipseDeform )
      {
	real rad=1.;  // initial radius 
	real da=rad*.1*deformationAmplitude;
	real db=rad*.1*deformationAmplitude;
	real a = rad + da*sin(omega);
	real b = rad - db*sin(omega);
      
	const int n1=51;
	x0.redim(n1,2);

	for( int i=0; i<n1; i++ )
	{
	  real theta = i*2.*Pi/(n1-1);
	  x0(i,0) = a*cos(theta);
	  x0(i,1) = b*sin(theta);
	}

	startCurve.setIsPeriodic(axis1,Mapping::functionPeriodic);
        #ifdef USE_PPP
  	  Overture::abort("fix me");
        #else
	  // *wdh* 081107 startCurve.interpolate(x0);
          int option=0, degree=3;
          const int boundaryParameterization = deformingBodyDataBase.get<int>("boundaryParameterization");
          startCurve.interpolate(x0,option,Overture::nullRealDistributedArray(),degree,
                                 (NurbsMapping::ParameterizationTypeEnum)boundaryParameterization);
        #endif
      }
      else  if( userDefinedDeformingBodyMotionOption==sphereDeform )
      {
        // Here is a sphere that expands to an ellipse

        // -- The array x0(i1,i2,0:2) holds the grid points on the original surface --

        real omega=sin(deformationFrequency*Pi*newT);  // omega varies in the interval [-1,1]

        real a0= .8*deformationAmplitude, 
             b0=1.2*deformationAmplitude, 
             c0=1.2*deformationAmplitude;  // final ellipse shape

        real a  = (1. + (a0-1.)*omega); 
        real b  = (1. + (b0-1.)*omega); 
        real c  = (1. + (c0-1.)*omega); 

        // realArray x0;
	// x0= startCurve.getGrid();
        Index I1=x0.dimension(0), I2=x0.dimension(1);
        // x0.reshape(I1,I2,3);
        RealArray x1(I1,I2,cg.numberOfDimensions());
        int i1,i2,i3;
	for( int i2=I2.getBase(); i2<=I2.getBound(); i2++ )
	for( int i1=I1.getBase(); i1<=I1.getBound(); i1++ )
	{
	  real x = x0(i1,i2,0), y=x0(i1,i2,1), z=x0(i1,i2,2);
	  real r = sqrt(x*x + y*y + z*z);
	  
          x1(i1,i2,0) = x*a;
          x1(i1,i2,1) = y*b;
          x1(i1,i2,2) = z*c;
	  

	}
        #ifdef USE_PPP
  	  Overture::abort("fix me");
        #else
	  // *wdh* 081107 startCurve.interpolate(x1);
          int option=0, degree=3;
          const int boundaryParameterization = deformingBodyDataBase.get<int>("boundaryParameterization");
          startCurve.interpolate(x1,option,Overture::nullRealDistributedArray(),degree,
                                 (NurbsMapping::ParameterizationTypeEnum)boundaryParameterization);
        #endif
      }
      else if( userDefinedDeformingBodyMotionOption==advectBody ||
               userDefinedDeformingBodyMotionOption==interfaceDeform ||
               userDefinedDeformingBodyMotionOption==elasticShell ||
               userDefinedDeformingBodyMotionOption==userDeformingSurface)
      {
      }
      else
      {
	printF("DeformingBodyMotion:regenerateComponentGrid:ERROR: unexpected value for "
               "userDefinedDeformingBodyMotionOption=%i\n",
	       (int)userDefinedDeformingBodyMotionOption);
	Overture::abort("ERROR: unexpected value for userDefinedDeformingBodyMotionOption!");
      }

      if( debug & 2 )
	printF("DeformingBodyMotion:regenerateComponentGrid: call hype to generate the grid t=%9.3e\n",
	       newT);

      // printF("DeformingBodyMotion::regenerateComponentGrid: BEFORE hyp.setSurface spacingOption=%i\n",hyp.spacingOption);

      const bool isSurfaceGrid=false; // this is a volume grid (in 2d or 3d) 
      const bool init=false; // this means keep existing hype parameters such as distanceToMarch, linesToMarch etc.
      hyp.setSurface(startCurve,isSurfaceGrid,init);

      // printF("DeformingBodyMotion::regenerateComponentGrid: AFTER hyp.setSurface spacingOption=%i\n",hyp.spacingOption);
      if( debug & 4 )
      {
	printF("DeformingBodyMotion::regenerateComponentGrid: HyperbolicMapping Parameters:");
	hyp.display();
      }

      if( false  ) // turn on for debugging 
      {
	GenericGraphicsInterface & gi = *Overture::getGraphicsInterface();
	
        gi.stopReadingCommandFile();
	printF("DeformingBodyMotion:regenerateComponentGrid: gridToMove=%i, t=%9.3e\n",gridToMove,newT);
	hyp.interactiveUpdate(gi);
      }
      else
      {
	int returnCode = hyp.generate();


        if( returnCode!=0 )
	{
	  printF("DeformingBodyMotion:regenerateComponentGrid:ERROR return from HyperbolicMapping::generate\n"
                 "  ... I am going to enter interactive mode so you can generate the hyperbolic grid interactively\n");
	  printF("  ... gridToMove=%i, t=%9.3e\n",gridToMove,newT);

	  if( true )
	    display(startCurve.getGrid(),"Here are points on the start curve","%6.2f");

	  GenericGraphicsInterface & gi = *Overture::getGraphicsInterface();
	
	  gi.stopReadingCommandFile();
	  hyp.interactiveUpdate(gi);
	}
	
      }
      

#ifndef USE_PPP
      // turn this off temporarily for PPP until we update Overture
      const DataPointMapping *pdpm = hyp.getDataPointMapping();
#else
      DataPointMapping *pdpm = NULL; 
#endif
      assert( pdpm!=NULL );
      DataPointMapping & dpm = (DataPointMapping&)(*pdpm);
      // dpm.getDataPoints() includes ghost points

      vector<GridEvolution*> & gridEvolution = deformingBodyDataBase.get<vector<GridEvolution*> >("gridEvolution"); 

      assert( face>=0 && face<gridEvolution.size() );
      
      gridEvolution[face]->addGrid(dpm.getDataPoints(),newT);
    

    }
    else
    {
      Overture::abort("ERROR: unknown deformingBodyType");
    }  
  }  // end for face
  
  return 0;
}


int DeformingBodyMotion::
update(  GenericGraphicsInterface & gi )
// ***** check this -- called by MovingGrids ---
{
  // GUI --> ADD
  printF("DeformingBodyMotion::update called\n");
  int ierr=0;

  DeformingBodyType & deformingBodyType = 
                  deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");

  if( deformingBodyType==elasticFilament )
  {
    assert( pElasticFilament!= NULL );
    ierr=pElasticFilament->update( gi );
  }
  else if( deformingBodyType==userDefinedDeformingBody )
  {
  }
  else
  {
    Overture::abort("ERROR: unknown deformingBodyType");
  }  
  return ierr;
}

//..DEBUG interface
void DeformingBodyMotion::
printFilamentHyperbolicDimensions(CompositeGrid & cg, int gridToMove)
{
  DeformingBodyType & deformingBodyType = 
                  deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");

  if( deformingBodyType==elasticFilament )
  {
    Mapping *pMap = (FilamentMapping*) cg[gridToMove].mapping().mapPointer;
    assert( pMap != NULL);
    if ( pMap->getClassName() == "FilamentMapping" )
    {
      FilamentMapping *pTempFilament = (FilamentMapping*) pMap;
      assert( pTempFilament != NULL );
  
      cout << "DBM.init Filam->pHyper->gridDimension(axis1) = " // **pf DEBUG
	   << pTempFilament->pHyper->getGridDimensions(axis1) << endl;
      cout << "DBM.init Filam->pHyper->gridDimension(axis2) = " // **pf DEBUG
	   << pTempFilament->pHyper->getGridDimensions(axis2) << endl;
    
      if ( (debug&2) && (pMapInfoDebug != NULL) ) 
      {
	cout << "DeformingBodyMotion::printHyperbDimensions "
	     << "LOOK AT THE *HYPERB* IN THE GRID!\n";
	pTempFilament->pHyper->update( *pMapInfoDebug );
	cout << "DeformingBodyMotion::printHyperbDimensions "
	     << "LOOK AT THE FILAM IN THE GRID!\n";
	pTempFilament->update( *pMapInfoDebug );
      }
    }
  }
  
}

int DeformingBodyMotion::
getPastLevelGrid( const int level, 
		  const int grid, 
		  CompositeGrid & cg,
		  realArray & gridVelocity)
{
  if(debug&4)
    cout << "DeformingBodyMotion::getPastLevelGrid, grid="
	 <<  grid << ", level=" << level << " called\n";

  DeformingBodyType & deformingBodyType = 
                  deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");

  if( deformingBodyType==elasticFilament )
  {
    assert( pDeformingGrid != NULL );
    pDeformingGrid->getPastLevelGrid( grid, cg, level, gridVelocity);
  }
  else if( deformingBodyType==userDefinedDeformingBody )
  {
    printF("DeformingBodyMotion:getPastLevelGrid called userDefinedDeformingBody for grid=%i, level=%i\n",
	   grid,level);
    
    // update cg[grid] here 

  }
  else
  {
    Overture::abort("ERROR: unknown deformingBodyType");
  } 

  return 0;
}

void DeformingBodyMotion::
simpleGetVelocity( const real vTime, 
		   const int grid00, 
		   CompositeGrid & cg,
		   realArray & gridVelocity)
		   //   realArray & xpoints1, realArray & xpoints2)
{
  DeformingBodyType & deformingBodyType = 
                  deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");

  if( deformingBodyType==elasticFilament )
  {
    assert( pDeformingGrid != NULL );
    pDeformingGrid->simpleGetVelocity( vTime,  grid00,  cg, gridVelocity);
    // xpoints1, xpoints2);
  }
  else if( deformingBodyType==userDefinedDeformingBody )
  {
    printF("DeformingBodyMotion:simpleGetVelocity called for userDefinedDeformingBody time=%9.3e\n",
	   vTime);
    gridVelocity=0.;
  }
  else
  {
    Overture::abort("ERROR: unknown deformingBodyType");
  }
  
}

//================================PRIVATE:

//..Handle component grids
//.......Should we have a 'deformingGridMapping'???
//.........Would know which Body it belongs to
//.........Could encapsulate Hyperb., DPM, or more general
void DeformingBodyMotion::
registerDeformingComponentGrid( const int grid, CompositeGrid & cg)
{
  // not needed?
}


//\begin{>>DeformingBodyMotionInclude.tex}{\subsection{update}} 
  // interactive update
int DeformingBodyMotion::update(CompositeGrid & cg, GenericGraphicsInterface & gi )
// =================================================================================================
// /Description:
//    Define deforming grid properties interactively.
// /Author:
//    WDH
//\end{DeformingBodyMotionInclude.tex}  
// =================================================================================================
{
  DeformingBodyType & deformingBodyType = 
                  deformingBodyDataBase.get<DeformingBodyType>("deformingBodyType");

  if( !deformingBodyDataBase.has_key("userDefinedDeformingBodyMotionOption") )
    deformingBodyDataBase.put<UserDefinedDeformingBodyMotionEnum>("userDefinedDeformingBodyMotionOption",iceDeform);
  UserDefinedDeformingBodyMotionEnum & userDefinedDeformingBodyMotionOption = 
                 deformingBodyDataBase.get<UserDefinedDeformingBodyMotionEnum>("userDefinedDeformingBodyMotionOption");
  

  aString answer,buff;

  const aString deformMenu[]=
    {
      "elastic filament",
      "user defined deforming body",
      ">types",
        "ice deform",
        "ellipse deform",
        "sphere deform",
        "advect body",
        "elastic shell",
        "interface deform",
        "user defined deforming surface",
      "<>parameters",
        "debug",
        "deformation frequency",
        "deformation amplitude",
        "elastic shell parameters",
        "boundary parameterization",
        "velocity order of accuracy",
        "acceleration order of accuracy",
        "initial velocity",
        "initial acceleration",
      "<done",
      ""
    }; 

  gi.appendToTheDefaultPrompt("deform>");
  for( ;; ) 
  {
	    
    int response=gi.getMenuItem(deformMenu,answer,"Choose the deformation type");
    

    if( answer=="done" || answer=="exit" )
    {
      break;
    }
    else if( answer=="elastic filament" )
    {
      deformingBodyType=elasticFilament;
    }
    else if( answer=="user defined deforming body" )
    {
      deformingBodyType=userDefinedDeformingBody; 
    }
    else if( answer=="ice deform" )
    {
      userDefinedDeformingBodyMotionOption=iceDeform;
    }
    else if( answer=="ellipse deform" )
    {
      userDefinedDeformingBodyMotionOption=ellipseDeform;
    }
    else if( answer=="sphere deform" )
    {
      userDefinedDeformingBodyMotionOption=sphereDeform;
    }
    else if( answer=="advect body" )
    {
      userDefinedDeformingBodyMotionOption=advectBody;
    }
    else if( answer=="interface deform" )
    {
      userDefinedDeformingBodyMotionOption=interfaceDeform;
      printF(" INFO interface deform is used for multi-domain deforming interfaces\n");
    }
    else if( answer=="elastic shell" )
    {
      userDefinedDeformingBodyMotionOption=elasticShell;
    }
    else if( answer=="user defined deforming surface" )
    {
      userDefinedDeformingBodyMotionOption=userDeformingSurface;
    }
    else if( answer=="deformation frequency" )
    {
      if( !deformingBodyDataBase.has_key("deformationFrequency") )
	deformingBodyDataBase.put<real>("deformationFrequency",1.);
      real & deformationFrequency = deformingBodyDataBase.get<real>("deformationFrequency");

      gi.inputString(answer,sPrintF("Enter the deformation frequency (for user defined motion) (current=%6.2f)",
				     deformationFrequency));
      sScanF(answer,"%e",&deformationFrequency);
      printF("Setting the deformation frequency to %9.3e for the user defined motion\n",deformationFrequency);
    }
    else if( answer=="deformation amplitude" )
    {
      if( !deformingBodyDataBase.has_key("deformationAmplitude") )
	deformingBodyDataBase.put<real>("deformationAmplitude",1.);
      real & deformationAmplitude = deformingBodyDataBase.get<real>("deformationAmplitude");

      gi.inputString(answer,sPrintF("Enter the deformation amplitude (for user defined motion) (current=%6.2f)",
				     deformationAmplitude));
      sScanF(answer,"%e",&deformationAmplitude);
      printF("Setting the deformation amplitude to %9.3e for the user defined motion\n",deformationAmplitude);
    }
    else if( answer=="elastic shell parameters" )
    {
      if( !deformingBodyDataBase.has_key("elasticShellParameters") )
	deformingBodyDataBase.put<real [10]>("elasticShellParameters");
      real *par = deformingBodyDataBase.get<real [10]>("elasticShellParameters");
      par[0]=10.; // rhoe
      par[1]=1.;  // te 
      par[2]=5.;  // ke 
      par[3]=.1;  // be 
      par[4]=1.;  // ad2 : coefficient of artificial diffusion

      gi.inputString(answer,sPrintF("Enter rhoe,te,ke,be,ad2 (default=(%g,%g,%g,%g,%g)",par[0],par[1],par[2],par[3],par[4]));
      sScanF(answer,"%e %e %e %e %e",&par[0],&par[1],&par[2],&par[3],&par[4]);
      printF("Setting rhoe=%g, te=%e, ke=%e, be=%e ad2=%e for the elastic shell\n",par[0],par[1],par[2],par[3],par[4]);
    }
    else if( answer=="boundary parameterization" )
    {
     int & boundaryParameterization = deformingBodyDataBase.get<int>("boundaryParameterization");
      
     printF("INFO: you should set the boundary parameterization to `index' for fluid structure problems\n"
            "      so that the grid lines match across the interface\n");
     gi.inputString(answer,sPrintF(buff,"Enter the boundary parameterization: %i=chord, %i=index (current=%i)",
		(int)NurbsMapping::parameterizeByChordLength,(int)NurbsMapping::parameterizeByIndex,boundaryParameterization));
      sScanF(answer,"%i",&boundaryParameterization);
      if( boundaryParameterization!=NurbsMapping::parameterizeByChordLength &&
          boundaryParameterization!=NurbsMapping::parameterizeByIndex )
      {
        printF("DeformingBodyMotion::update:ERROR: invalid value for boundary parameterization = %i\n",boundaryParameterization);
	gi.stopReadingCommandFile();

	boundaryParameterization=NurbsMapping::parameterizeByChordLength;
      }
      printF("Setting the boundary parameterization to %i\n",boundaryParameterization);

    }
    else if( answer=="velocity order of accuracy" )
    {
      int & order=deformingBodyDataBase.get<int>("velocityOrderOfAccuracy");
      gi.inputString(answer,sPrintF("Enter the velocity order of accuracy (default=%i)",order));
      sScanF(answer,"%i",&order);
      printF("DeformingBodyMotion::update: set the velocity order of accuracy =%i\n",order);
    }
    else if( answer=="acceleration order of accuracy" )
    {
      int & order=deformingBodyDataBase.get<int>("accelerationOrderOfAccuracy");
      gi.inputString(answer,sPrintF("Enter the acceleration order of accuracy (default=%i)",order));
      sScanF(answer,"%i",&order);
      printF("DeformingBodyMotion::update: set the acceleration order of accuracy =%i\n",order);
    }
    else if( answer=="initial velocity" )
    {
      if( !deformingBodyDataBase.has_key("initialVelocity") )
      {
	deformingBodyDataBase.put<real [3]>("initialVelocity");
        real *v0 = deformingBodyDataBase.get<real [3]>("initialVelocity");
        v0[0]=v0[1]=v0[2]=0.;
      }
      real *v0 = deformingBodyDataBase.get<real [3]>("initialVelocity");
      gi.inputString(answer,sPrintF("Enter the initial velocity: v0,v1,v2 (default=(%g,%g,%g)",v0[0],v0[1],v0[2]));
      sScanF(answer,"%e %e %e",&v0[0],&v0[1],&v0[2]);
      printF("Setting the inital velocity to (%9.3e,%9.3e,%9.3e)\n",v0[0],v0[1],v0[2]);
    }
    else if( answer=="initial acceleration" )
    {
      if( !deformingBodyDataBase.has_key("initialAcceleration") )
      {
	deformingBodyDataBase.put<real [3]>("initialAcceleration");
	real *a0 = deformingBodyDataBase.get<real [3]>("initialAcceleration");
	a0[0]=a0[1]=a0[2]=0.;
      }
      real *a0 = deformingBodyDataBase.get<real [3]>("initialAcceleration");
      gi.inputString(answer,sPrintF("Enter the initial acceleration: a0,a1,a2 (default=(%g,%g,%g)",a0[0],a0[1],a0[2]));
      sScanF(answer,"%e %e %e",&a0[0],&a0[1],&a0[2]);
      printF("Setting the inital acceleration to (%9.3e,%9.3e,%9.3e)\n",a0[0],a0[1],a0[2]);
    }
    else if( answer=="debug" )
    {
      gi.inputString(answer,sPrintF("Enter debug (default=%i)",debug));
      sScanF(answer,"%i",&debug);
      printF("DeformingBodyMotion::update: set debug=%i\n",debug);
    }
    else
    {
      printF("DeformingBodyMotion::update:ERRROR:unknown response=[%s]\n",(const char*)answer);
      gi.stopReadingCommandFile();
    }
    
  }
    
//   delete [] gridMenu;
  gi.unAppendTheDefaultPrompt();

  return 0;

}



