#include "GridEvolution.h"

int GridEvolution::debug=0;

// ============================================================================================
/// \brief GridEvolution contructor.
// ============================================================================================
GridEvolution::GridEvolution()
{
  maximumNumberOfTimeLevels=4; // keep at most this many time levels
  numberOfTimeLevels=0;
  current=-1;

  accelerationOrderOfAccuracy=1;  // order of accuracy for the acceleration computation
  velocityOrderOfAccuracy=1;      // order of accurcy for the velocity computation

  time.redim(maximumNumberOfTimeLevels);
}

// ============================================================================================
/// \brief GridEvolution destructor
// ============================================================================================
GridEvolution::~GridEvolution()
{
}

// ============================================================================================
/// \brief return the order of accuracy used to compute the acceleration.
// ============================================================================================
int GridEvolution::
getAccelerationOrderOfAccuracy() const
{
  return accelerationOrderOfAccuracy;
}

// ============================================================================================
/// \brief return the order of accuracy used to compute the velocity.
// ============================================================================================
int GridEvolution::
getVelocityOrderOfAccuracy() const
{
  return velocityOrderOfAccuracy;
}


// ============================================================================================
/// \brief set the order of accuracy used to compute the acceleration.
/// \param order (input) : a positive integer
// ============================================================================================
int GridEvolution::
setAccelerationOrderOfAccuracy( int order )
{
  accelerationOrderOfAccuracy=order;
  return 0;
}

// ============================================================================================
/// \brief set the order of accuracy used to compute the velocity.
/// \param order (input) : a positive integer
// ============================================================================================
int GridEvolution::
setVelocityOrderOfAccuracy( int order )
{
  velocityOrderOfAccuracy=order;
  return 0;
}




// ===================================================================================
/// \brief Add a grid to the time history
///
/// \param x (input) : normally x is the vertex array from a MappedGrid
/// \param t (input) : time to associate with this grid.
// ===================================================================================
// add a new grid at time t
int GridEvolution::
addGrid( const realArray & x, real t )
{
  // debug=3;  // ****
  
  if( numberOfTimeLevels!=0 )
  {
    if( t <= time(current) )
    {
      // -- This could be a grid with t<0 for starting a simulation. --
      //    Add it to the start of the list

      // NOTE: We should generalize this to  adding any number of previous grids
      // NOTE: What happens if we want to replace a grid with an improved one ??
         

      if( t <time(current) && time(current)==0.  && numberOfTimeLevels==1 && current==0 )
      {
        gridList.addElement(x,0); // insert into the start of the list

	current = (current +1) % maximumNumberOfTimeLevels;

        time(current)=time(0);
        time(0)=t;
	
        numberOfTimeLevels=min(numberOfTimeLevels+1,maximumNumberOfTimeLevels);

	if( debug & 4 )
	  printF("GridEvolution::addGrid:INFO: adding a grid for t<0, t=%9.3e, (current=%i, t=%9.3e)\n",t,
		 current,time(current));
	
        return 1;
      }

      printF("GridEvolution::addGrid:ERROR: t=%10.4e is less than or equal to time(current)=%10.4e \n",
            t,time(current));
      printF("GridEvolution::addGrid:... skipping this grid...\n");
      return 0;
    }
  }

  current = (current +1) % maximumNumberOfTimeLevels;
  
  
  if( numberOfTimeLevels<maximumNumberOfTimeLevels )
  {
    gridList.addElement(x,current);
  }
  else
  { // replace existing element:
   gridList.deleteElement(current);
   gridList.addElement(x,current);
  }
  
  time(current)=t;

  numberOfTimeLevels=min(numberOfTimeLevels+1,maximumNumberOfTimeLevels);
  
  if( debug & 4 )
    printF(" *** GridEvolution::addGrid t=%9.3e, current=%i, numberOfTimeLevels=%i\n",t,current,numberOfTimeLevels);
  
  return 0;
}

// ===========================================================================================
/// \brief Compute the grid velocity from a set of grids over time.
/// \details The GridEvolution class keeps a sequence of past grids and computes the time derivative 
/// of the grid motion using these grids. 
// ==================================================================================================
int GridEvolution::
getVelocity( real t, realArray & gridVelocity, 
	     const Index &I1, const Index &I2, const Index &I3 ) const
{
  if( numberOfTimeLevels>=2 )
  {
    assert( velocityOrderOfAccuracy>0 );

    int i=(current-1 + maximumNumberOfTimeLevels) % maximumNumberOfTimeLevels;
    for( int j=0; j<numberOfTimeLevels-1; j++ )
    {
      if( time(i)<=t )
      {
	break;
      }
      i = (i-1 + maximumNumberOfTimeLevels) % maximumNumberOfTimeLevels;
    }
    
    // we should have time(i) < t < time(i+1) 


    Range Rx = gridVelocity.dimension(3);
    if( velocityOrderOfAccuracy==1 || numberOfTimeLevels==2 )
    {
      if( velocityOrderOfAccuracy != 1 )
	printF("GridEvolution::getVelocity:WARNING: t=%9.3e : only computing the velocity to 1st order accuracy\n"
	       " since there are only 2 time levels available, requesed accuracy=%i\n",t,velocityOrderOfAccuracy);

      int ip1 = (i+1) % maximumNumberOfTimeLevels;
      realArray & x0 = gridList[i];
      realArray & x1 = gridList[ip1];
      real dt0= time(ip1)-time(i);
      assert( dt0>0. );
      gridVelocity(I1,I2,I3,Rx) = (x1(I1,I2,I3,Rx) - x0(I1,I2,I3,Rx))/dt0;

    }
    else 
    {
      assert( numberOfTimeLevels>=3 );
      if( velocityOrderOfAccuracy != 2 )
	printF("GridEvolution::getVelocity:WARNING: only computing the velocity to 2nd order accuracy"
	       " requested accuracy=%i\n",velocityOrderOfAccuracy);

      int im1 = (i-1 + maximumNumberOfTimeLevels) % maximumNumberOfTimeLevels;
      int ip1 = (i+1) % maximumNumberOfTimeLevels;
      realArray & x0 = gridList[im1];
      realArray & x1 = gridList[i];
      realArray & x2 = gridList[ip1];
      const real t0=time(im1), t1=time(i), t2=time(ip1);
      real dt0= t1-t0;
      real dt1= t2-t1;
      assert( dt0>0. && dt1>0. );

      // Compute the time derivative of the Lagrange polynomial: 
      //    x(t) = l0(t)*x0 + l1(t)*x1 + l2(t)*x2 
      // where 
      //   l0 = (t-t1)*(t-t2)/( (t0-t1)*(t0-t2) );
      //   l1 = (t-t2)*(t-t0)/( (t1-t2)*(t1-t0) );
      //   l2 = (t-t0)*(t-t1)/( (t2-t0)*(t2-t1) );
      
      real l0t = (2.*t-(t1+t2))/( (t0-t1)*(t0-t2) );
      real l1t = (2.*t-(t2+t0))/( (t1-t2)*(t1-t0) );
      real l2t = (2.*t-(t0+t1))/( (t2-t0)*(t2-t1) );
      
      gridVelocity(I1,I2,I3,Rx) = l0t*x0(I1,I2,I3,Rx) + l1t*x1(I1,I2,I3,Rx) + l2t*x2(I1,I2,I3,Rx);
      
    }
    
//     printf("GridEvolution::getVelocity: t=%9.3e, current=%i i,ip1=%i,%i, "
// 	   "max(vel)=%8.2e min(vel)=%8.2e\n",t,current,i,ip1,max(gridVelocity(I1,I2,I3,Rx)),
//            min(gridVelocity(I1,I2,I3,Rx)));

  }
  else
  {
    printF("GridEvolution::getVelocity:WARNING: there are only %i time levels (t=%9.3e), setting gridVelocity=0.\n",
	   numberOfTimeLevels,t);

    Range Rx=gridVelocity.dimension(3);
    gridVelocity(I1,I2,I3,Rx)=0.;
  }
  


  return 0;
}

// ========================================================================================
///  \brief Compute the grid acceleration.
// ========================================================================================
int GridEvolution::
getAcceleration( real t, realArray & gridAcceleration, 
		 const Index &I1, const Index &I2, const Index &I3 ) const
{
  if( numberOfTimeLevels>=3 )
  {
    assert( accelerationOrderOfAccuracy>0 );

    int i=(current-1 + maximumNumberOfTimeLevels) % maximumNumberOfTimeLevels;
    for( int j=0; j<numberOfTimeLevels-1; j++ )
    {
      if( time(i)<=t )
      {
	break;
      }
      i = (i-1 + maximumNumberOfTimeLevels) % maximumNumberOfTimeLevels;
    }
    
    // we should have time(i) < t < time(i+1) 

    int im1 = (i-1 + maximumNumberOfTimeLevels) % maximumNumberOfTimeLevels;
    int ip1 = (i+1) % maximumNumberOfTimeLevels;
    realArray & x0 = gridList[im1];
    realArray & x1 = gridList[i];
    realArray & x2 = gridList[ip1];
    

    real dt0= time(i)-time(im1);
    real dt1= time(ip1)-time(i);
    assert( dt0>0. && dt1>0. );
    Range Rx = gridAcceleration.dimension(3);
    
    if( accelerationOrderOfAccuracy != 1 )
    {
      printF("GridEvolution::getAcceleration:WARNING: only computing the acceleration to 1st order accuracy\n"
             "     requested order =%i\n",accelerationOrderOfAccuracy);
    }

    gridAcceleration(I1,I2,I3,Rx) = ( (x2(I1,I2,I3,Rx) - x1(I1,I2,I3,Rx))/dt1
				     -(x1(I1,I2,I3,Rx) - x0(I1,I2,I3,Rx))/dt0 )/(.5*(dt0+dt1));
    
    if( false )
      printf("GridEvolution::getAcceleration: t=%9.3e, current=%i, im1,i,ip1=%i,%i,%i min(acel)=%8.2e "
	     " max(accel)=%9.2e\n",
	     t, current,im1,i,ip1,min(gridAcceleration(I1,I2,I3,Rx)),max(gridAcceleration(I1,I2,I3,Rx)));
    
  }
  else
  {
    printf("GridEvolution::getAcceleration:WARNING: there are only %i time levels (t=%9.3e)\n",
	   numberOfTimeLevels,t);

    Range Rx=gridAcceleration.dimension(3);
    gridAcceleration(I1,I2,I3,Rx)=0.;
  }
  


  return 0;
}

