//                              -*- c++ -*-
#ifndef MOVING_GRIDS_H
#define MOVING_GRIDS_H

#include "Overture.h"

//..Forward definitions
class GridFunction;
class GenericGraphicsInterface;
class RigidBodyMotion;
class Parameters;
class Integrate;
class Ogshow;

class DeformingBodyMotion;


// ========================================================================
//    This class coordinates the movement of grids for CG
// ========================================================================

class MovingGrids
{
 public:

  // This enum defines how the grid for a body moves
  enum MovingGridOption
  {
    notMoving,
    rotate,
    shift,
    oscillate,
    scale,
    rigidBody,
    deformingBody,
    userDefinedMovingGrid,
    numberOfMovingGridOptions
  };

  

  MovingGrids(Parameters & parameters);
  MovingGrids(const MovingGrids & mg);
  virtual ~MovingGrids();

  // interactive update
  virtual int update(CompositeGrid & cg, GenericGraphicsInterface & gi );

  bool isMovingGridProblem() const;

  int setIsMovingGridProblem( bool trueOrFalse=TRUE );

  bool gridIsMoving(int grid) const;

  MovingGridOption movingGridOption(int grid) const;

  aString movingGridOptionName(MovingGridOption option) const;
  
  const RealArray & getMoveParameters() const;

  real getTimeStepForRigidBodies() const; 

//  virtual int move( CompositeGrid & cg, real time );

//  virtual int getGridVelocity(CompositeGrid & cg, real time, realCompositeGridFunction & gridVelocity );

//  virtual int getGridAcceleration(CompositeGrid & cg, real time, realCompositeGridFunction & gridAcceleration );
  
  virtual int moveGrids(const real & t1, 
			const real & t2, 
			const real & t3,
			const real & dt0,
			GridFunction & cgf1,  
			GridFunction & cgf2,
			GridFunction & cgf3 );
  
  // correction step: 
  virtual int correctGrids(const real t1,
			   const real t2, 
			   GridFunction & cgf1,
			   GridFunction & cgf2 );

  virtual int getGridVelocity( GridFunction & gf0, const real & tGV );

  virtual int gridAccelerationBC(const int & grid,
				 const real & t0,
				 MappedGrid & c,
				 realMappedGridFunction & u ,
				 realMappedGridFunction & f ,
				 realMappedGridFunction & gridVelocity ,
				 realArray & normal,
				 const Index & I1,
				 const Index & I2,
				 const Index & I3,
				 const Index & I1g,
				 const Index & I2g,
				 const Index & I3g    );
  
  virtual int getBoundaryAcceleration( MappedGrid & c, realArray & gtt, int grid, real t0, int option );

  virtual int rigidBodyMotion(const real & t1, 
			      const real & t2, 
			      const real & t3,
			      const real & dt0,
			      GridFunction & cgf1,  
			      GridFunction & cgf2,
			      GridFunction & cgf3 );
 
  virtual int moveDeformingBodies(const real & t1, 
				  const real & t2, 
				  const real & t3,
				  const real & dt0,
				  GridFunction & cgf1,  
				  GridFunction & cgf2,
				  GridFunction & cgf3 );
  
  int updateToMatchGrid( CompositeGrid & cg );

  // general motion
  int userDefinedMotion(const real & t1, 
			const real & t2, 
			const real & t3,
			const real & dt0,
			GridFunction & cgf1,  
			GridFunction & cgf2,
			GridFunction & cgf3 );
  
  // motions that use a MatrixTransform -- shift, rotate, scale
  int userDefinedTransformMotion(const real & t1, 
				 const real & t2, 
				 const real & t3,
				 const real & dt0,
				 GridFunction & cgf1,  
				 GridFunction & cgf2,
				 GridFunction & cgf3,
				 const int grid );
  
  int getUserDefinedGridVelocity( GridFunction & gf0, const real & t0, const int grid );

  int userDefinedGridAccelerationBC(const int & grid,
				    const real & t0,
				    MappedGrid & c,
				    realMappedGridFunction & u ,
				    realMappedGridFunction & f ,
				    realMappedGridFunction & gridVelocity ,
				    realArray & normal,
				    const Index & I1,
				    const Index & I2,
				    const Index & I3,
				    const Index & I1g,
				    const Index & I2g,
				    const Index & I3g );

  int getUserDefinedBoundaryAcceleration( MappedGrid & mg, realArray & gtt, int grid, real t0, int option,
					  const int side, const int axis );

  int updateUserDefinedMotion(CompositeGrid & cg, GenericGraphicsInterface & gi);


  virtual int saveToShowFile() const;
  
  int detectCollisions(GridFunction & cgf1);
  
  int getNumberOfRigidBodies() const;
  RigidBodyMotion & getRigidBody(int bodyNumber) const;

  static int debug0;
  int debug(){return debug0;}

  // get from a data base file
  int get( const GenericDataBase & dir, const aString & name);

  // put to a data base file
  int put( GenericDataBase & dir, const aString & name) const;


  protected:

  int initialize();
  int getRamp(real t, real rampInterval, real & ramp, real & rampSpeed, real & rampAcceleration );
  

  Parameters & parameters;

  IntegerArray gridsToMove;

  bool isInitialized;

  bool movingGridProblem;
  IntegerArray moveOption;
  IntegerArray movingGrid;         // gridIsMoving(grid) is true if a grid is moving
  RealArray moveParameters;          // moveParameters(.,grid) : parameters for the type of movement

  //..Rigid body information
  int numberOfRigidBodies;
  RigidBodyMotion **body;     // array of pointers one or more rigid bodies.
  Integrate *integrate;       // used to integrate stresses on bodies.
  bool useHybridGridsForSurfaceIntegrals;

  int rigidBodyInfoCount;
  RealArray rigidBodyInfo;    // save time sequence of some rigid body info.
  RealArray rigidBodyInfoTime;
  
  int numberOfRigidBodyInfoNames;
  aString *rigidBodyInfoName;

  bool limitForces;
  real maximumAllowableForce, maximumAllowableTorque;

  //..Deforming body information
  int numberOfDeformingBodies;
  DeformingBodyMotion **deformingBodyList;
  
};

#endif
