#ifndef RIGID_BODY_MOTION_H
#define RIGID_BODY_MOTION_H

// Class for keeping track of the position and orientation of a rigid body moving
// under the influence of forces and torques.

#include "Overture.h"

class GenericGraphicsInterface;

class RigidBodyMotion
{
  
 public:

  enum PositionConstraintEnum
  {
    positionHasNoConstraint=0,
    positionConstrainedToAPlane,
    positionConstrainedToALine,
    positionIsFixed
  };

  enum RotationConstraintEnum
  {
    rotationHasNoConstraint=0,
    rotationConstrainedToAPlane,
    rotationConstrainedToALine,
    rotationIsFixed
  };

  enum TimeSteppingMethoEnum
  {
    leapFrogTrapezoidal=0,
    improvedEuler
  } timeSteppingMethod;

  RigidBodyMotion(int numberOfDimensions = 3);
  ~RigidBodyMotion();
  
  bool centerOfMassHasBeenInitialized() const;
  bool massHasBeenInitialized() const;

  real getMass() const;
  void setMass( const real totalMass );

  RealArray getMomentsOfInertia() const;

  real getDensity() const; // return the density, if known. If not known return a negative value.

  int getAxesOfInertia( real t, RealArray & axesOfInertia  ) const;
  int getAcceleration( real t, RealArray & vCM  ) const;
  int getPosition( real t, RealArray & xCM ) const;
  int getVelocity( real t, RealArray & vCM  ) const;
  int getAngularVelocities( real t, RealArray & omega  ) const;
  int getAngularAcceleration( real t, RealArray & omegaDot  ) const;
  int getRotationMatrix(real t, 
                        RealArray & r,                     
                        RealArray & rDot = Overture::nullRealArray(),
                        RealArray & rDotDot = Overture::nullRealArray() ) const;
  
  // determine the velocity or acceleration of a point p(t) belonging to the body.
  int getPointVelocity( real t, const RealArray & p, RealArray & vp) const;
  int getPointAcceleration( real t, const RealArray & p, RealArray & ap) const;
  // use this next routine to compute the point velocity or acceleration for many points.
  int getPointTransformationMatrix( real t, 
				    RealArray & rDotRt  = Overture::nullRealArray(),
				    RealArray & rDotDotRt = Overture::nullRealArray() ) const ;

  // general purpose routine:
  int getCoordinates( real t, 
		      RealArray & xCM      = Overture::nullRealArray(), 
		      RealArray & vCM      = Overture::nullRealArray(),
		      RealArray & rotation = Overture::nullRealArray(), 
		      RealArray & omega    = Overture::nullRealArray(), 
		      RealArray & omegaDot = Overture::nullRealArray(), 
		      RealArray & aCM      = Overture::nullRealArray(),
                      RealArray & axesOfInertia = Overture::nullRealArray()  ) const;

  
  int getInitialConditions(RealArray & x0 = Overture::nullRealArray() , 
			   RealArray & v0 = Overture::nullRealArray() , 
			   RealArray & w0 = Overture::nullRealArray() ,
			   RealArray & axesOfInertia = Overture::nullRealArray()  ) const ;

  real getTimeStepEstimate() const;

  int setInitialCentreOfMass( const RealArray & x0 );

  int setInitialConditions( real t0=0., 
			    const RealArray & x0 = Overture::nullRealArray() , 
			    const RealArray & v0 = Overture::nullRealArray() , 
			    const RealArray & w0 = Overture::nullRealArray() ,
			    const RealArray & axesOfInertia = Overture::nullRealArray()  );

  int setProperties(real totalMass, 
		    const RealArray & momentsOfInertia, 
		    const int numberOfDimensions = 3 );

  int setPositionConstraint( PositionConstraintEnum constraint, const RealArray & values );

  int setRotationConstraint( RotationConstraintEnum constraint, const RealArray & values );

  // integrate to a new time
  int integrate( real tf, 
		 const RealArray & force, 
		 const RealArray & torque,
                 real t, 
                 RealArray & xCM, 
		 RealArray & rotation );

  // correct solution at time t using new values of the forces at time t.
  int correct( real t, 
	       const RealArray & force, 
	       const RealArray & torque,
	       RealArray & xCM, 
	       RealArray & rotation );

  int momentumTransfer( real t, RealArray & v );

  int update( GenericGraphicsInterface & gi );


  // get from a data base file
  int get( const GenericDataBase & dir, const aString & name);

  // put to a data base file
  int put( GenericDataBase & dir, const aString & name) const;

 protected:

    

  enum ConstraintApplicationEnum
  {
    applyConstraintsToPosition=0,
    applyConstraintsToVelocity,
    applyConstraintsToForces
  };


  int applyConstraints( const int step, const ConstraintApplicationEnum option );

  real mass;               // total mass
  int numberOfDimensions;  // 2 or 3 dimensional problem
  int current;             // position of current solution in arrays.
  int numberOfSteps;       // number of time steps taken
  int numberSaved;         // number of different times we have saved force and moment data.
  int maximumNumberToSave; // save at most this many previous time values of various quantities
  int initialConditionsGiven; // keeps track of which initial conditions have been given
  real density;            // In cases where the volume can be determined we can use the density to get the mass.

  real damping;  // for damping oscillations.

  RealArray mI;     // 3 moment of inertia

  RealArray e, e0;    // e(0:2,0:2) : 3 axes of inertia, e0=e at t=0

  RealArray x, v;  // centre of mass: position, velocity
  RealArray x0, v0, w0;  // initial centre of mass,  velocity and angular velocity
  
  RealArray f, g;  // force and torque

  RealArray w;      // angular velocities about axes of inertia

  RealArray time;   // arrays of times for which we know data.

  RealArray bodyForce;

  const Range R;          // R=Range(0,2) 

  PositionConstraintEnum positionConstraint;
  RotationConstraintEnum rotationConstraint;
  RealArray constraintValues;
  
};


#endif
