#include "CgSolver.h"
#include "NameList.h"
#include "GenericGraphicsInterface.h"
#include "MaterialProperties.h"
#include "Chemkin.h"
#include "Ogshow.h"
#ifndef OV_USE_OLD_STL_HEADERS
#include <vector>
#include <algorithm>	   // STL algorithms class library
OV_USINGNAMESPACE(std);
#else
#include <vector.h>
#include <algorithm.h>	   // STL algorithms class library
#endif

//\begin{>>CgSolverInclude.tex}{\subsection{setDefaultDataForABoundaryCondition}} 
static int 
setDefaultDataForABoundaryCondition(const int & side,
				    const int & axis,
				    const int & grid,
                                    CompositeGrid & cg, Parameters & parameters)
// ============================================================================================
// /Description:
//    Assign the default values for the data required by a boundary condition.
// /grid,side,axis (input) : defines a boundary
//  
//\end{CgSolverInclude.tex}  
// ============================================================================================
{
  Range all;
  switch (cg[grid].boundaryCondition()(side,axis)) 
  {
  case Parameters::inflowWithVelocityGiven:
  case Parameters::subSonicInflow:
    // data is set n.u = ...
    parameters.bcData(all,side,axis,grid)=0.;
    break;
  case Parameters::outflow:
  case Parameters::subSonicOutflow:
  case Parameters::convectiveOutflow:
  case Parameters::tractionFree:
   //  data is a*p+b*p.n = c
    parameters.bcData(0,side,axis,grid)=1.;
    parameters.bcData(1,side,axis,grid)=1.;
    parameters.bcData(2,side,axis,grid)=0.;
    
  }
  return 0;
  
}

//\begin{>>CgSolverInclude.tex}{\subsection{setDefaultDataForBoundaryConditions}} 
int CgSolver::
setDefaultDataForBoundaryConditions()
// ============================================================================================
// /Description:
//    Assign the default values for the data required by the boundary conditions.
//\end{CgSolverInclude.tex}  
// ============================================================================================
{
  for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
  {
    for( int axis=0; axis<cg.numberOfDimensions(); axis++ )
    {
      for( int side=Start; side<=End; side++ )
      {
	setDefaultDataForABoundaryCondition(side,axis,grid,cg,parameters);
      }
    }
  }
  return 0;
}

//\begin{>>CgSolverInclude.tex}{\subsection{checkForValidBoundaryCondition}} 
int 
checkForValidBoundaryCondition( const int & bc, Parameters & parameters, bool reportErrors=true )
// =================================================================================
// /Description:
//   Check that a given boundary condition is valid for a given PDE.
//\end{CgSolverInclude.tex}  
// =================================================================================
{
  // here are BC's that are valid for all equations:
  if( bc<=0 ||
      bc==(int)Parameters::dirichletBoundaryCondition ||
      bc==(int)Parameters::neumannBoundaryCondition ||
      bc==(int)Parameters::symmetry ||
      bc==(int)Parameters::interfaceBoundaryCondition )
    return 0;


  if( parameters.dbase.get<Parameters::PDE >("pde")==Parameters::incompressibleNavierStokes )
  {
    if( bc==(int)Parameters::noSlipWall ||
	bc==(int)Parameters::slipWall ||
        bc==(int)Parameters::inflowWithVelocityGiven ||
        bc==(int)Parameters::inflowWithPressureAndTangentialVelocityGiven ||
        bc==(int)Parameters::outflow ||
        bc==(int)Parameters::axisymmetric ||
        bc==(int)Parameters::tractionFree ||
        bc==(int)Parameters::dirichletBoundaryCondition  )
    {
      return 0;
    }
  }
  else if( parameters.dbase.get<Parameters::PDE >("pde")==Parameters::compressibleNavierStokes ||
           parameters.dbase.get<Parameters::PDE >("pde")==Parameters::compressibleMultiphase )
  {
    if( bc==(int)Parameters::noSlipWall ||
	bc==(int)Parameters::slipWall ||
        bc==(int)Parameters::superSonicInflow ||
        bc==(int)Parameters::superSonicOutflow ||
        bc==(int)Parameters::subSonicInflow ||
        bc==(int)Parameters::subSonicOutflow ||
        bc==(int)Parameters::outflow ||
        bc==(int)Parameters::farField ||
        bc==(int)Parameters::axisymmetric )
    {
      return 0;
    }
  }
  else if( parameters.dbase.get<Parameters::PDE >("pde")==Parameters::allSpeedNavierStokes )
  {
    if( bc==(int)Parameters::noSlipWall ||
	bc==(int)Parameters::slipWall ||
        bc==(int)Parameters::superSonicInflow ||
        bc==(int)Parameters::superSonicOutflow ||
        bc==(int)Parameters::subSonicInflow ||
        bc==(int)Parameters::subSonicOutflow ||
        bc==(int)Parameters::convectiveOutflow  ||
        bc==(int)Parameters::tractionFree )
    {
      return 0;
    }
  }
  else
    return 0;  // fix this
  

  if( bc>=0 && bc<Parameters::numberOfBCNames )
  {
    if( reportErrors )
    {
      printF("checkForValidBoundaryConditionERROR::ERROR: The boundary condition %s is not valid for this method\n",
	     (const char*)parameters.dbase.get< >("bcName")[bc]);
    }
  }
  else
  {
    printF("checkForValidBoundaryConditionERROR: Unknown boundary condition number = %i.\n"
           "Something is wrong here Bill! \n",bc);
  }
  return 1;
}

//  static int
//  chooseUserDefinedBoundaryValues(int side,int axis,int grid)
//  {
//    printF("chooseUserDefinedBoundaryValues: **** fix me ****\n");
//    return 0;
//  }

//\begin{>>CgSolverInclude.tex}{\subsection{setBoundaryConditionValues}}   
static int 
setBoundaryConditionValues(const aString & answer,
			   const IntegerArray & originalBoundaryCondition,
			   CompositeGrid & cg, Parameters & parameters )
// ============================================================================================
// /Description:
//     Parse the string that defines the boundary conditions and options.
//  Change the boundary conditions and set the data for the boundary conditions.
//\end{CgSolverInclude.tex}  
// ============================================================================================
{
  int grid,side,axis,n;
//    aString answer2, answer3;
//    char buff[200];

  aString gridName, bcName;
  Range G,S,A;
  int bc;
  RealArray value;	
  int len=0;

  int length=answer.length();
  int i,mark=-1;
  for( i=0; i<length; i++ )
  {
    if( answer[i]=='(' || answer[i]=='=' )
    {
      mark=i-1;
      break;
    }
  }
  if( mark<0 )
  {
    printF("unknown form of answer=[%s]. Try again or type `help' for examples.\n",(const char *)answer);
    // gi.stopReadingCommandFile();
    return 1; 
  }

  gridName=answer(0,mark);  // this is the name of the grid or `all'
  S=Range(-1,-1);
  A=Range(-1,-1);
  if( answer[mark+1]=='(' )
  { // determine which side and axis to assign
    int side=-1,axis=-1;
    int numRead=sscanf(answer(mark+1,length-1),"(%i,%i)",&side,&axis);
    if( numRead==2 )
    {
      if( side>=0 && side<=1 && axis>=0 && axis<=cg.numberOfDimensions()-1 )
      {
	S=Range(side,side);
	A=Range(axis,axis);
	for( i=mark+1; i<length; i++ )
	{
	  if( answer[i]=='=' )
	  {
	    mark=i-1;
	    break;
	  }
	}
      }
      else
      {
	printf("invalid values for side=%i or axis=%i, 0<=side<=1, 0<=axis<=%i \n",side,axis,
	       cg.numberOfDimensions()-1);
	return 1;
      }
    }
  }
  else
  { // assign all sides:
    S=Range(0,1);
    A=Range(0,cg.numberOfDimensions()-1);
  }
  if( S.getBase()==-1 || A.getBase()==-1 || mark+2>length-1 )
  {
    printF("unknown form of answer=[%s]. Try again or type `help' for examples.\n",(const char *)answer);
    // gi.stopReadingCommandFile();
    return 1; 
  }
  // search for a blank separating the bc name from any options
  int endOfName=length-1;
  for( i=mark+3; i<length; i++ )
  {
    if( answer[i]==' ' || answer[i]==',' )
    {
      endOfName=i-1;
      break;
    }
  }
  bcName=answer(mark+2,endOfName);
  mark=endOfName+1;

  G=Range(-1,-1);
  int changeBoundaryConditionNumber=-1;  // for bcNumber#=bcName

  if( gridName=="all" )
  {
    G=Range(0,cg.numberOfComponentGrids()-1);
  }
  else if( len=gridName.matches("bcNumber") )
  {
    // BC of the form
    //    bcNumber3=noSlipWall
    G=Range(0,cg.numberOfComponentGrids()-1); // check all grids

    sScanF(gridName(len,gridName.length()-1),"%i",&changeBoundaryConditionNumber);
    printF("setting BC number %i to be %s\n",changeBoundaryConditionNumber,(const char*)bcName);

  }
  else
  { // search for the name of the grid
    for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    {
      if( gridName==cg[grid].mapping().getName(Mapping::mappingName) )
      {
	G=Range(grid,grid);
	break;
      }
    }
  }
  if( G.getBase()==-1  )
  {
    printF("Unknown grid name = <%s> \n",(const char *)gridName);
    // gi.stopReadingCommandFile();
    return 1;
  }
  // search for the name of the boundary condition
  bc=-1;
  for( int i=1; i<Parameters::numberOfBCNames; i++ )
  {
    if( bcName==parameters.dbase.get< >("bcName")[i] )
    {
      bc=i;
      break;
    }
  }
  if( bc<=0 || checkForValidBoundaryCondition( bc,parameters )!=0 )
  {
    printF("Unknown boundary condition = <%s> \n",(const char *)bcName);
    // gi.stopReadingCommandFile();
    return 1; 
  }

  // *********************************************************
  // search for options that define the data for the BC's:
  //    uniform(c0=v0,c1=v1,...)
  //    parabolic(par0=val0,par1=val1,...)
  //    oscillate(par0=val0,...)
  //    ramp(par0=val0,...)
  // *********************************************************
  mark=endOfName+1;
  const char *line = answer;

  char *uniform, *parabolic, *blasius, *oscillate, *pressure, *mixedDerivative, *ramp, *jet, *userDefinedBoundaryData;
  uniform=strstr(line,"uniform");
  oscillate=strstr(line,"oscillate");
  ramp=strstr(line,"ramp");
  parabolic=strstr(line,"parabolic");
  blasius=strstr(line,"blasius");
  pressure=strstr(line,"pressure");
  mixedDerivative=strstr(line,"mixedDerivative");
  userDefinedBoundaryData=strstr(line,"userDefinedBoundaryData");
  jet=strstr(line,"jet");
	  
  const int maxNumberOfParameters=20;
//  char c[maxNumberOfParameters][10];
  aString c[maxNumberOfParameters];
  real val[maxNumberOfParameters];
	  
  int numRead=0;
  if( uniform!=NULL || parabolic!=NULL || blasius!=NULL || jet!=NULL )
  {
    // *** extract values for the components rho,u,v,... that are specfied by a line of the form:
    //     prefix(name1=val1,name2=val2,...)
    // where prefix=uniform, or prefix=parabolic, ...

    aString prefix = uniform!=NULL ? "uniform" : parabolic!=NULL ? "parabolic" : blasius!=NULL ? "blasius" : "jet";
    int lenPrefix=prefix.length();
    // aString line=uniform;
    aString line=uniform!=NULL   ? uniform : 
                 parabolic!=NULL ? parabolic : 
                 blasius!=NULL   ? blasius : 
                 jet!=NULL       ? jet : 
                                   " ";
    line=line(lenPrefix+1,line.length()); // strip off "prefix("
    int len=line.length();
    int ie=len-1;
    for( int i=0; i<len; i++ )  // look for ending ')'
    {
      if( line[i]==')' )
      {
	ie=i-1;
	break;
      }
    }
    line=line(0,ie);
    // look for components that are assigned values, c0=v1,
    if( parameters.dbase.get<int >("debug") & 4 )
      printF("setBC: scanning for %s: \n",(const char*)prefix);
    numRead=parameters.parseValues(line,c,val,maxNumberOfParameters);

    // **** Fix this to look for longer component names *****

//       parameters.dbase.get<RealArray >("initialConditions")=(real)Parameters::defaultValue;
//       parameters.inputParameterValues(answer,"initial conditions", parameters.dbase.get<RealArray >("initialConditions") );

    // parameters.parseValues( const aString & answer, aString *name, real *value, int maxNumber )

    // look for components that are assigned values, c0=v1,
//     numRead=sScanF(uniform,"uniform(%1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e",
// 		   c[0],&val[0],c[1],&val[1],c[2],&val[2],c[3],&val[3],c[4],&val[4],c[5],&val[5],
// 		   c[6],&val[6],c[7],&val[7],c[8],&val[8],c[9],&val[9]);
//     numRead/=2;
    // printF("scanning for uniform flow: [%s=%e], [%s=%e], [%s=%e]\n",c[0],val[0],c[1],val[1],c[2],val[2]);
  }
//   if( parabolic!=NULL )
//   {
//     // look for components that are assigned values, c0=v1,
//     numRead=sScanF(parabolic,"parabolic(%1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e",
// 		   c[0],&val[0],c[1],&val[1],c[2],&val[2],c[3],&val[3],c[4],&val[4],c[5],&val[5],
// 		   c[6],&val[6],c[7],&val[7],c[8],&val[8],c[9],&val[9]);
//     numRead/=2;
//     // printF("scanning for parabolic flow: [%s=%e], [%s=%e], %s=%e\n",c[0],val[0],c[1],val[1],c[2],val[2]);
//   }
//   if ( blasius!=NULL )
//   {
//     numRead=sScanF(blasius,"blasius(%1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e",
// 		   c[0],&val[0],c[1],&val[1],c[2],&val[2],c[3],&val[3],c[4],&val[4],c[5],&val[5],
// 		   c[6],&val[6],c[7],&val[7],c[8],&val[8],c[9],&val[9]);
//     numRead/=2;
//     printF("scanning for blasius flow: [%s=%e], [%s=%e], %s=%e\n",c[0],val[0],c[1],val[1],c[2],val[2]);
//   }
 
  real jetRadius=.5, jetBoundaryLayerWidth=.1, jetX0=0., jetY0=0., jetZ0=0.;
  if( jet!=NULL )
  {
//     numRead=sScanF(jet,"jet(%1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e",
// 		   c[0],&val[0],c[1],&val[1],c[2],&val[2],c[3],&val[3],c[4],&val[4],c[5],&val[5],
// 		   c[6],&val[6],c[7],&val[7],c[8],&val[8],c[9],&val[9]);
//     numRead/=2;

    // look for some extra parameters
    real d=.1,r0=.5,x0=0.,y0=0.,z0=0.;
	    
    char *pos;
    pos=strstr(jet,"d=");
    if( pos!=NULL ) sScanF(pos+2,"%e",&jetBoundaryLayerWidth);
    pos=strstr(jet,"r=");
    if( pos!=NULL ) sScanF(pos+2,"%e",&jetRadius);
    pos=strstr(jet,"x=");
    if( pos!=NULL ) sScanF(pos+2,"%e",&jetX0);
    pos=strstr(jet,"y=");
    if( pos!=NULL ) sScanF(pos+2,"%e",&jetY0);
    pos=strstr(jet,"z=");
    if( pos!=NULL ) sScanF(pos+2,"%e",&jetZ0);

    printF("Jet: parameters: jetBoundaryLayerWidth=%e, r=%e, x=%e, y=%e, z=%e\n",jetBoundaryLayerWidth,
	   jetRadius,jetX0,jetY0,jetZ0);
	    
  }
	  

  real oscillateT0=0., oscillateOmega=1., oscillateAmplitude0=0., oscillateAmplitude1=1., 
    oscillateOffsetU=0., oscillateOffsetV=0., oscillateOffsetW=0.;
  if( oscillate!=NULL )
  {
    // get parameters for oscillating inflow
    char *pos;
    pos=strstr(oscillate,"t0=");
    if( pos!=NULL )
      sScanF(pos+3,"%e",&oscillateT0);
    pos=strstr(oscillate,"omega=");
    if( pos!=NULL )                           
      sScanF(pos+6,"%e",&oscillateOmega);
    pos=strstr(oscillate,"a0=");
    if( pos!=NULL )                           
      sScanF(pos+3,"%e",&oscillateAmplitude0);
    pos=strstr(oscillate,"a1=");
    if( pos!=NULL )                           
      sScanF(pos+3,"%e",&oscillateAmplitude1);
    pos=strstr(oscillate,"u0=");
    if( pos!=NULL )                           
      sScanF(pos+3,"%e",&oscillateOffsetU);
    pos=strstr(oscillate,"v0=");
    if( pos!=NULL )                           
      sScanF(pos+3,"%e",&oscillateOffsetV);
    pos=strstr(oscillate,"w0=");
    if( pos!=NULL )                           
      sScanF(pos+3,"%e",&oscillateOffsetW);
    printF(" Reading oscillation parameters: omega=%e, t0=%e, a0=%e, a1=%e, (u0,v0,w0)=(%e,%e,%e) \n",
	   oscillateOmega,oscillateT0,oscillateAmplitude0,oscillateAmplitude1,
	   oscillateOffsetU,oscillateOffsetV,oscillateOffsetW);
  }

  real ta=0.,tb=1.,ua=0.,ub=1.,va=0.,vb=0.,wa=0.,wb=0.,Ta=1.,Tb=1.;;
  if( ramp!=NULL )
  {
    // get parameters for ramped inflow
    char *pos;
    pos=strstr(ramp,"ta=");
    if( pos!=NULL )
      sScanF(pos+3,"%e",&ta);
    pos=strstr(ramp,"tb=");
    if( pos!=NULL )
      sScanF(pos+3,"%e",&tb);
    pos=strstr(ramp,"ua=");
    if( pos!=NULL )
      sScanF(pos+3,"%e",&ua);
    pos=strstr(ramp,"ub=");
    if( pos!=NULL )
      sScanF(pos+3,"%e",&ub);
    pos=strstr(ramp,"va=");
    if( pos!=NULL )
      sScanF(pos+3,"%e",&va);
    pos=strstr(ramp,"vb=");
    if( pos!=NULL )
      sScanF(pos+3,"%e",&vb);
    pos=strstr(ramp,"wa=");
    if( pos!=NULL )
      sScanF(pos+3,"%e",&wa);
    pos=strstr(ramp,"wb=");
    if( pos!=NULL )
      sScanF(pos+3,"%e",&wb);
    pos=strstr(ramp,"Ta=");
    if( pos!=NULL )
      sScanF(pos+3,"%e",&Ta);
    pos=strstr(ramp,"Tb=");
    if( pos!=NULL )
      sScanF(pos+3,"%e",&Tb);

    printF(" Reading ramp parameters: ta=%6.2e, tb=%6.2e, ua=%6.2e, ub=%6.2e, va=%6.2e, vb=%6.2e,"
	   " wa=%6.2e, wb=%6.2e, Ta=%6.2e, Tb=%6.2e\n",ta,tb,ua,ub,va,vb,wa,wb,Ta,Tb);
  }
	  

  int component=-1;
  real mixedDerivValue[3];  // *wdh* 051201 -- to support adiabatic noSlipWall
  if( pressure!=NULL )
  {
    numRead=sScanF(pressure,"pressure(%e*p+%e*p.n=%e",&mixedDerivValue[0],&mixedDerivValue[1],&mixedDerivValue[2]);
    if( numRead==3 )
    {
      printF("pressure BC: %e*p+%e*p.n=%e \n",mixedDerivValue[0],mixedDerivValue[1],mixedDerivValue[2]);
    }
    else
    {
      Overture::abort("error");
    }
    component=parameters.dbase.get<int >("pc");
    // numRead=0;
  }
  if( mixedDerivative!=NULL )
  {
    numRead=sScanF(mixedDerivative,"mixedDerivative(%e*t+%e*t.n=%e",&mixedDerivValue[0],&mixedDerivValue[1],&mixedDerivValue[2]);
    if( numRead==3 )
    {
      printF("mixed derivative BC: %e*t+%e*t.n=%e \n",mixedDerivValue[0],mixedDerivValue[1],mixedDerivValue[2]);
    }
    else
    {
      Overture::abort("error");
    }
    component=parameters.dbase.get<int >("tc");
    // numRead=0;
  }
  // int code=0;
  if( userDefinedBoundaryData!=NULL )
  {
    numRead=0;
//             char *pos;
//             pos=strstr(userDefinedBoundaryData,"code=");
//             if( pos!=NULL )
//     	      sScanF(pos+5,"%i",&code);
  }

  value.redim(parameters.dbase.get<int >("numberOfComponents")+3); value=0;
  real parabolicBoundaryLayerWidth=.1;
  real reynoldsNumberX = 1000.;
	  

  // ****************************************************************************
  // fill in the value() array from c[] and val[]: 
  //             value(parameters.dbase.get<int >("uc"))=val[i] if c[i]='u'
  // ****************************************************************************
  if ( parabolic!=NULL )
  {
    // fill in the value array: value(parameters.dbase.get<int >("uc"))=val[i] if c[i]='u'
    //  "d" is an extra name to add to the list of known names.
    const int dc = parameters.dbase.get<int >("numberOfComponents")+1;
    value(dc)=parabolicBoundaryLayerWidth;  // give a default value
    parameters.assignParameterValues("boundary conditions",value,numRead,c,val, "d",dc ); 
    parabolicBoundaryLayerWidth=value(dc);
  }
  else if ( blasius!=NULL )
  {
    // fill in the value array: value(parameters.dbase.get<int >("uc"))=val[i] if c[i]='u'
    //  "R" is an extra name to add to the list of known names.
    const int rxc= parameters.dbase.get<int >("numberOfComponents")+1;
    value(rxc) = reynoldsNumberX;
    parameters.assignParameterValues("boundary conditions",value,numRead,c,val, "R",rxc );
    reynoldsNumberX = value(rxc);
  }
  else
  {

    // default case:
    parameters.assignParameterValues("boundary conditions",value,numRead,c,val ); 

    int nc=parameters.dbase.get<int >("numberOfComponents");
    if( mixedDerivative!=NULL )
    {
      value(nc  )=mixedDerivValue[0];
      value(nc+1)=mixedDerivValue[1];
      value(nc+2)=mixedDerivValue[2];
    }

  }
	  

  RealArray timeDependenceParameters;
  if( bc==Parameters::inflowWithVelocityGiven || bc==Parameters::subSonicInflow ||
      bc==Parameters::noSlipWall )
  {
    if( parabolic!=NULL )
    {
      if( parabolicBoundaryLayerWidth <=0. )
      {
	parabolicBoundaryLayerWidth=.1;
	printf("ERROR: width of the parabolic boundary layer is specified to be <= 0.\n"
	       "Setting to the default value of %e \n",parabolicBoundaryLayerWidth);
      }
      printF("width of the parabolic boundary layer = %e \n",parabolicBoundaryLayerWidth);
    }
    if( oscillate!=NULL )
    {
      timeDependenceParameters.redim(7);
      timeDependenceParameters(0)=oscillateOmega;
      timeDependenceParameters(1)=oscillateT0;
      timeDependenceParameters(2)=oscillateAmplitude0;
      timeDependenceParameters(3)=oscillateAmplitude1;
      timeDependenceParameters(4)=oscillateOffsetU;
      timeDependenceParameters(5)=oscillateOffsetV;
      timeDependenceParameters(6)=oscillateOffsetW;
	      
      printF("the oscillating inflow has the form: a0 + a1*cos((t-t0)*(2*pi*omega))*( uniform/parabolic)");
      printF(" omega=%e, t0=%e, a0=%e, a1=%e \n",oscillateOmega,oscillateT0,oscillateAmplitude0,
	     oscillateAmplitude1);
    }
    if( ramp!=NULL )
    {
      int nDim=max(int(parameters.dbase.get<int >("uc")),max(int(parameters.dbase.get<int >("vc")),max(int(parameters.dbase.get<int >("wc")),int(parameters.dbase.get<int >("tc")))))+1;

      #define TIMEPAR(n,c) timeDependenceParameters(2+(n)+2*(c))

      timeDependenceParameters.redim(2+2*nDim);
      timeDependenceParameters=0.;

      timeDependenceParameters(0)=ta;
      timeDependenceParameters(1)=tb;

      TIMEPAR(0,parameters.dbase.get<int >("uc"))=ua;  TIMEPAR(1,parameters.dbase.get<int >("uc"))=ub;
      TIMEPAR(0,parameters.dbase.get<int >("vc"))=va;  TIMEPAR(1,parameters.dbase.get<int >("vc"))=vb;
      if( parameters.dbase.get<int >("wc")>=0 )
      { 
        TIMEPAR(0,parameters.dbase.get<int >("wc"))=wa;  TIMEPAR(1,parameters.dbase.get<int >("wc"))=wb;
      }
      if( parameters.dbase.get<int >("tc")>=0 )
      {
	TIMEPAR(0,parameters.dbase.get<int >("tc"))=Ta;  TIMEPAR(1,parameters.dbase.get<int >("tc"))=Tb;
      }
      
//       timeDependenceParameters(2)=ua;
//       timeDependenceParameters(3)=ub;
//       timeDependenceParameters(4)=va;
//       timeDependenceParameters(5)=vb;
//       timeDependenceParameters(6)=wa;
//       timeDependenceParameters(7)=wb;
//       timeDependenceParameters(8)=Ta;
//       timeDependenceParameters(9)=Tb;

      #undef TIMEPAR
    }
	    
  }
  else if( bc==Parameters::outflow || 
	   bc==Parameters::subSonicOutflow ||
	   bc==Parameters::convectiveOutflow ||
	   bc==Parameters::tractionFree )
  {
    // Save info for mixed derivative, alpha*p + beta*p.n = gamma
    //   value(c) = "gamma" for component c
    //   value(c+NC*1) = "alpha"   (NC=numberOfComponents)
    //   value(c+NC*2) = "beta"
    value.redim(parameters.dbase.get<int >("numberOfComponents")*3);
    value=1.;
    if( (pressure!=NULL || mixedDerivative!=NULL ) && component>=0 && component<parameters.dbase.get<int >("numberOfComponents") )
    {
      value(component)=mixedDerivValue[2];  // note: rhs value goes first
      value(component+parameters.dbase.get<int >("numberOfComponents")*1)=mixedDerivValue[0];
      value(component+parameters.dbase.get<int >("numberOfComponents")*2)=mixedDerivValue[1];
    }
    else
      value(Range(0,parameters.dbase.get<int >("numberOfComponents")-1))=0.;
  }
	  

// 	  if( userDefinedBoundaryData!=NULL )
// 	  {
// 	    value.redim(1);
//             value(0)=code;
// 	  }
	  

	  // assign boundary conditions and data
  if( gridName=="all")
  {
    printF("Setting boundary condition to %s (bc=%i) on all grids.\n",(const char*)parameters.dbase.get< >("bcName")[bc],bc);
  }
  bool bcWasAssigned=false;
  for( grid=G.getBase(); grid<=G.getBound(); grid++ )
  {
    for( int axis=A.getBase(); axis<=A.getBound(); axis++ )
    {
      for( int side=S.getBase(); side<=S.getBound(); side++ )
      {
	// printF("grid,side,axis,original,change=%i,%i,%i,  %i,%i\n",grid,side,axis,
	//        originalBoundaryCondition(side,axis,grid),changeBoundaryConditionNumber);
		
	if( cg[grid].boundaryCondition(side,axis) > 0 && 
	    (changeBoundaryConditionNumber==-1 || 
	     originalBoundaryCondition(side,axis,grid)==changeBoundaryConditionNumber) )
	{
	  bcWasAssigned=true;

	  cg[grid].setBoundaryCondition(side,axis,bc);
	  // set underlying mapping too (for moving grids)
	  cg[grid].mapping().getMapping().setBoundaryCondition(side,axis,bc);

	  setDefaultDataForABoundaryCondition(side,axis,grid,cg,parameters);

	  // the array parameters.bcData(.,side,axis,grid) holds the data that is
	  // passed directly to the boundary condition routines (if the BC doesn't vary in space)

	  if( gridName!="all" )
	    printF("Setting (side,axis)=(%i,%i) of grid %s to bc=%s, (number=%i)\n",side,axis,
		   (const char*)cg[grid].mapping().getName(Mapping::mappingName),
		   (const char*)parameters.dbase.get< >("bcName")[bc],bc);
	  if( value.getBound(0) > parameters.dbase.get<RealArray >("bcData").getBound(0) )
	  {
	    Range all,R0(parameters.dbase.get<RealArray >("bcData").getBound(0)+1,value.getBound(0));
		    
	    parameters.dbase.get<RealArray >("bcData").resize(value.getLength(0),parameters.dbase.get<RealArray >("bcData").getLength(1),
				     parameters.dbase.get<RealArray >("bcData").getLength(2),parameters.dbase.get<RealArray >("bcData").getLength(3));
                    
	    parameters.bcData(R0,all,all,all)=0.; // give initial values
	  }

          // *** assign parameter values for this face:
	  for( n=0; n<=value.getBound(0); n++ )
	    parameters.bcData(n,side,axis,grid)=value(n);

	  if( bc==Parameters::inflowWithVelocityGiven || bc==Parameters::subSonicInflow )
	  {
	    if ( blasius!=NULL )
	    {
	      parameters.setBcType(side,axis,grid,Parameters::blasiusProfile);
	      parameters.setTimeDependenceBoundaryConditionParameters(side,axis,grid,timeDependenceParameters);		    
	      parameters.setBcVariesInSpace(side,axis,grid,true);
	      parameters.bcParameters(1,side,axis,grid)=reynoldsNumberX;

	    }
	    else if( parabolic!=NULL )
	    {
	      if( ramp!=NULL )
	      {
		parameters.setBcType(side,axis,grid,Parameters::parabolicInflowRamped);
		parameters.setBcIsTimeDependent(side,axis,grid,true);
		parameters.setTimeDependenceBoundaryConditionParameters(side,axis,grid,timeDependenceParameters);
			
//                         int rpStart=1;
// 			assert( parameters.dbase.get<RealArray >("bcParameters").getLength(0)>value.getLength(0) );
// 			for( n=0; n<=value.getBound(0); n++ )
// 			  parameters.bcParameters(rpStart+n,side,axis,grid)=value(n);
	      }
	      else if( oscillate!=NULL )
	      {
		parameters.setBcType(side,axis,grid,Parameters::parabolicInflowOscillating );
		parameters.setBcIsTimeDependent(side,axis,grid,true);
		parameters.setTimeDependenceBoundaryConditionParameters(side,axis,grid,timeDependenceParameters);
	      }
	      else if( userDefinedBoundaryData!=NULL )
	      {
		parameters.setBcType(side,axis,grid,Parameters::parabolicInflowOscillating );
		parameters.setBcIsTimeDependent(side,axis,grid,true);

		parameters.setBcVariesInSpace(side,axis,grid,true);  // always assume this for user BC's
		parameters.chooseUserDefinedBoundaryValues(side,axis,grid,cg);
	      }
	      else
		parameters.setBcType(side,axis,grid,Parameters::parabolicInflow);

	      parameters.setBcVariesInSpace(side,axis,grid,true);

	      parameters.bcParameters(0,side,axis,grid)=parabolicBoundaryLayerWidth;



	      printF(" +++ setting parameters.bcParameters(0,%i,%i,%i)=%e\n",
		     side,axis,grid,parabolicBoundaryLayerWidth);
	    }
	    else if( uniform!=NULL )
	    {
	      parameters.setBcType(side,axis,grid,Parameters::uniformInflow);
	      if( ramp!=NULL )
	      {
		parameters.setTimeDependenceBoundaryConditionParameters(side,axis,grid,timeDependenceParameters);
	      }
	      else if( oscillate!=NULL )
	      {
		parameters.setTimeDependenceBoundaryConditionParameters(side,axis,grid,timeDependenceParameters);
	      }
	    }
	    else if( ramp!=NULL )
	    {
	      parameters.setBcIsTimeDependent(side,axis,grid,true);
	      parameters.setBcType(side,axis,grid,Parameters::rampInflow);
	      parameters.setTimeDependenceBoundaryConditionParameters(side,axis,grid,timeDependenceParameters);

	    }
	    else if( oscillate!=NULL )
	    {
	    }
	    else if( jet!=NULL )
	    {
	      // the jet values are actually assigned in jetInflow (bcForcing.C)
	      parameters.setBcType(side,axis,grid,Parameters::jetInflow);
	      parameters.setBcVariesInSpace(side,axis,grid,true);
		      
	      parameters.bcParameters(0,side,axis,grid)=jetRadius;
	      parameters.bcParameters(1,side,axis,grid)=jetBoundaryLayerWidth;
	      parameters.bcParameters(2,side,axis,grid)=jetX0;
	      parameters.bcParameters(3,side,axis,grid)=jetY0;
	      parameters.bcParameters(4,side,axis,grid)=jetZ0;
		      
	    }
	    else if( userDefinedBoundaryData!=NULL )
	    {
	      parameters.setBcVariesInSpace(side,axis,grid,true);  // always assume this for user BC's
	      parameters.chooseUserDefinedBoundaryValues(side,axis,grid,cg);
/* ------------		      
   parameters.setBcIsTimeDependent(side,axis,grid,true);  // query for this?


   parameters.setBcType(side,axis,grid,Parameters::userDefinedBoundaryData);
   parameters.variableBoundaryData(grid)=TRUE; 
   ----- */

	    }
		    
		    
	  }
	  else if( bc==Parameters::noSlipWall || bc==Parameters::slipWall )
	  {
	    if( userDefinedBoundaryData!=NULL )
	    {
	      // **** new ****

	      parameters.setBcVariesInSpace(side,axis,grid,true);  // always assume this for user BC's
	      parameters.chooseUserDefinedBoundaryValues(side,axis,grid,cg);
	    }
		    
	    // ***finish other cases ***
	    if( ramp!=NULL )
	    {
              printF(" ***NEW: setting bcType=ramped for (grid,side,axis)=(%i,%i,%i)\n",grid,side,axis);
	      
	      parameters.setBcIsTimeDependent(side,axis,grid,true);
	      parameters.setBcType(side,axis,grid,Parameters::ramped); // ** new generic way ***
	      parameters.setTimeDependenceBoundaryConditionParameters(side,axis,grid,timeDependenceParameters);

	    }

	  }
		  
		  
	}
      }
    }
  } // end for grid
  if( !bcWasAssigned && changeBoundaryConditionNumber>0 )
  {
    printF("\n setBoundaryConditions:INFO: There were no boundaries assigned for bcNumber%i=%s\n\n",
	   changeBoundaryConditionNumber,(const char*)bcName );
  }
	  

  return 0;
}



//\begin{>>CgSolverInclude.tex}{\subsection{setBoundaryConditionsInteractively}}   
int CgSolver::
setBoundaryConditionsInteractively(const aString & answer,
				   const IntegerArray & originalBoundaryCondition )
// ============================================================================================
// /Description:
//   Old way: parse an answer holding a boundary condition specification.
//\end{CgSolverInclude.tex}  
// ============================================================================================
{
  assert( parameters.dbase.get<GenericGraphicsInterface* >("ps") !=NULL );
  GenericGraphicsInterface & gi = *parameters.dbase.get<GenericGraphicsInterface* >("ps");
  // PlotStuffParameters & psp = parameters.dbase.get<GraphicsParameters >("psp");
  
  int grid,side,axis,n;
  aString answer2, answer3;
  char buff[200];

  if( answer=="boundary conditions" )
  {
    gi.appendToTheDefaultPrompt("bc>");
	
    aString gridName, bcName;
    Range G,S,A;
    int bc;
    RealArray value;	
    int len=0;

    for(;;)
    {
      gi.inputString(answer2,"Enter a new boundary condition (or type `help' or `done')");
      if( answer2=="done" || answer2=="exit" )
	break;
      else if( answer2=="help" )
      {
	printf("BC: Enter a change to the boundary conditions. Type a string of the form     \n"
               "                                                                             \n"
	       "       <grid name>(side,axis)=<boundary condition name> [,option] [,option] ...\n"
               "                                                                             \n"
	       " to change the boundary condition on a given side of a given grid, or        \n"
               "                                                                             \n"
	       "       <grid name>=<boundary condition name>  [,option] [,option] ...          \n"
               "                                                                             \n"
               " to change all boundaries on a given grid, or                                \n"
               "                                                                             \n"
	       "       bcNumber<num>=<boundary condition name>  [,option] [,option] ... \n"
               "                                                                             \n"
               " to change all boundaries that *originally* had a boundary condition value of `num'\n"
               " Here \n"
	       "   <grid name> is the name of the grid, side=0,1 and axis=0,1,2.  <grid name> can also be `all'.\n"
               "                                                                             \n"
               " The optional arguments specify data for the boundary conditions:            \n"
               "    option = `uniform(p=1.,u=1.,...)'       : to specify a uniform inflow profile   \n"
               "    option = `parabolic(d=2,p=1.,...)'      : to specify a parabolic inflow profile \n"
	       "    option = `blasius(p=1.,u=1.,R=1000)'  : to specify a laminar boundary layer profile \n"
               "    option = `pressure(.1*p+1.*p.n=0.)'     : pressure boundary condition at outflow\n"
               "    option = `oscillate(t0=.5,omega=1.,a0=0.,a1=1.,u0=0,v0=0.,w0=0.)'  : oscillating values    \n"
               "    option = `ramp(ta=0.,tb=1.,ua=0.,ub=1.,va=0.,..)  : ramped inflow parameters    \n" 
               "    option = `jet(d=.1,r=1.,x=0.,y=0,z=0.,p=1.,...)': to specify a parabolic jet inflow profile \n" 
               "    option = `userDefinedBoundaryData : choose user defined data  \n"
	       " The pressure outflow boundary condition is a*p + b*p.n = c \n"
	       "  where p=pressure, p.n=normal derivative of p.\n"
	       "  Note that a and b should have the same sign or else the condition is unstable\n"
               "                                                                               \n"
               " The oscillate option is used with `uniform' or `parabolic' and takes the form:\n" 
               "         [u0,v0,w0] + { a0+a1*cos((t-t0)*(2*pi*omega)) }*[ uniform or parabolic profile ]  \n"
               "                                                                             \n"
               "The userDefinedBoundaryData option calls the function userDefinedBoundaryValues that\n"
               "may be changed to suit your purposes. \n"
	       " Examples: \n"
	       "     square(0,0)=inflowWithVelocityGiven , uniform(p=1.,u=1.)     \n"
	       "     annulus=noSlipWall                         \n"
	       "     all=slipWall                               \n"
	       "     bcNumber2=slipWall                         \n"
	       "     square(0,1)=outflow , pressure(.1*p+1.*p.n=0.) \n"
                );
      
          
	for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
	  printF(" grid %i : name=%s \n",grid,(const char*)cg[grid].mapping().getName(Mapping::mappingName));
	printf("Available boundary conditions are: \n");
	for( int i=1; i<Parameters::numberOfBCNames; i++ )
	  printF(" %s \n",(const char*)parameters.dbase.get< >("bcName")[i]);
      }
      else 
      {
	int length=answer2.length();
	int i,mark=-1;
	for( i=0; i<length; i++ )
	{
	  if( answer2[i]=='(' || answer2[i]=='=' )
	  {
	    mark=i-1;
	    break;
	  }
	}
	if( mark<0 )
	{
	  printF("unknown form of answer=[%s]. Try again or type `help' for examples.\n",(const char *)answer2);
	  gi.stopReadingCommandFile();
	  continue;
	}
	else
	{

	  setBoundaryConditionValues(answer2,originalBoundaryCondition,cg,parameters);

// --- start
 
//  	  gridName=answer2(0,mark);  // this is the name of the grid or `all'
//  	  S=Range(-1,-1);
//  	  A=Range(-1,-1);
//  	  if( answer2[mark+1]=='(' )
//  	  { // determine which side and axis to assign
//  	    int side=-1,axis=-1;
//  	    int numRead=sscanf(answer2(mark+1,length-1),"(%i,%i)",&side,&axis);
//  	    if( numRead==2 )
//  	    {
//  	      if( side>=0 && side<=1 && axis>=0 && axis<=cg.numberOfDimensions()-1 )
//  	      {
//  		S=Range(side,side);
//  		A=Range(axis,axis);
//  		for( i=mark+1; i<length; i++ )
//  		{
//  		  if( answer2[i]=='=' )
//  		  {
//  		    mark=i-1;
//  		    break;
//  		  }
//  		}
//  	      }
//                else
//  	      {
//  		printf("invalid values for side=%i or axis=%i, 0<=side<=1, 0<=axis<=%i \n",side,axis,
//                         cg.numberOfDimensions()-1);
//                  continue;
//  	      }
//  	    }
//  	  }
//  	  else
//  	  { // assign all sides:
//  	    S=Range(0,1);
//  	    A=Range(0,cg.numberOfDimensions()-1);
//  	  }
//  	  if( S.getBase()==-1 || A.getBase()==-1 || mark+2>length-1 )
//  	  {
//  	    printF("unknown form of answer=[%s]. Try again or type `help' for examples.\n",(const char *)answer2);
//  	    gi.stopReadingCommandFile();
//  	    continue;
//  	  }
//  	  // search for a blank separating the bc name from any options
//            int endOfName=length-1;
//            for( i=mark+3; i<length; i++ )
//  	  {
//  	    if( answer2[i]==' ' || answer2[i]==',' )
//  	    {
//  	      endOfName=i-1;
//  	      break;
//  	    }
//  	  }
//  	  bcName=answer2(mark+2,endOfName);
//            mark=endOfName+1;

//  	  G=Range(-1,-1);
//            int changeBoundaryConditionNumber=-1;  // for bcNumber#=bcName

//  	  if( gridName=="all" )
//  	  {
//  	    G=Range(0,cg.numberOfComponentGrids()-1);
//  	  }
//            else if( len=gridName.matches("bcNumber") )
//  	  {
//  	    // BC of the form
//              //    bcNumber3=noSlipWall
//              G=Range(0,cg.numberOfComponentGrids()-1); // check all grids

//              sScanF(gridName(len,gridName.length()-1),"%i",&changeBoundaryConditionNumber);
//              printF("setting BC number %i to be %s\n",changeBoundaryConditionNumber,(const char*)bcName);

//  	  }
//  	  else
//  	  { // search for the name of the grid
//  	    for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
//  	    {
//  	      if( gridName==cg[grid].mapping().getName(Mapping::mappingName) )
//  	      {
//  		G=Range(grid,grid);
//  		break;
//  	      }
//  	    }
//  	  }
//  	  if( G.getBase()==-1  )
//  	  {
//  	    printF("Unknown grid name = <%s> \n",(const char *)gridName);
//  	    gi.stopReadingCommandFile();
//  	    continue;
//  	  }
//  	  // search for the name of the boundary condition
//  	  bc=-1;
//  	  for( int i=1; i<Parameters::numberOfBCNames; i++ )
//  	  {
//  	    if( bcName==parameters.dbase.get< >("bcName")[i] )
//  	    {
//  	      bc=i;
//  	      break;
//  	    }
//  	  }
//  	  if( bc<=0 || checkForValidBoundaryCondition( bc,parameters )!=0 )
//  	  {
//  	    printF("Unknown boundary condition = <%s> \n",(const char *)bcName);
//  	    gi.stopReadingCommandFile();
//  	    continue;
//  	  }

//            // *********************************************************
//            // search for options that define the data for the BC's:
//            //    uniform(c0=v0,c1=v1,...)
//            //    parabolic(par0=val0,par1=val1,...)
//            //    oscillate(par0=val0,...)
//            //    ramp(par0=val0,...)
//            // *********************************************************
//            mark=endOfName+1;
//            const char *line = answer2;

//            char *uniform, *parabolic, *blasius, *oscillate, *pressure, *mixedDerivative, *ramp, *jet, *userDefinedBoundaryData;
//  	  uniform=strstr(line,"uniform");
//  	  oscillate=strstr(line,"oscillate");
//  	  ramp=strstr(line,"ramp");
//  	  parabolic=strstr(line,"parabolic");
//  	  blasius=strstr(line,"blasius");
//  	  pressure=strstr(line,"pressure");
//  	  mixedDerivative=strstr(line,"mixedDerivative");
//  	  userDefinedBoundaryData=strstr(line,"userDefinedBoundaryData");
//  	  jet=strstr(line,"jet");
	  
//            const int maxNumberOfParameters=10;
//            char c[maxNumberOfParameters][10];
//  	  real val[maxNumberOfParameters];
	  
//            int numRead=0;
//  	  if( uniform!=NULL )
//  	  {
//              // look for components that are assigned values, c0=v1,
//              numRead=sScanF(uniform,"uniform(%1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e",
//                             c[0],&val[0],c[1],&val[1],c[2],&val[2],c[3],&val[3],c[4],&val[4],c[5],&val[5],
//                             c[6],&val[6],c[7],&val[7],c[8],&val[8],c[9],&val[9]);
//              numRead/=2;
//              // printF("scanning for uniform flow: [%s=%e], [%s=%e], [%s=%e]\n",c[0],val[0],c[1],val[1],c[2],val[2]);
//  	  }
//  	  if( parabolic!=NULL )
//  	  {
//              // look for components that are assigned values, c0=v1,
//              numRead=sScanF(parabolic,"parabolic(%1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e",
//                             c[0],&val[0],c[1],&val[1],c[2],&val[2],c[3],&val[3],c[4],&val[4],c[5],&val[5],
//                             c[6],&val[6],c[7],&val[7],c[8],&val[8],c[9],&val[9]);
//              numRead/=2;
//              // printF("scanning for parabolic flow: [%s=%e], [%s=%e], %s=%e\n",c[0],val[0],c[1],val[1],c[2],val[2]);
//  	  }
//  	  if ( blasius!=NULL )
//  	  {
//  	    numRead=sScanF(blasius,"blasius(%1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e",
//  			   c[0],&val[0],c[1],&val[1],c[2],&val[2],c[3],&val[3],c[4],&val[4],c[5],&val[5],
//                             c[6],&val[6],c[7],&val[7],c[8],&val[8],c[9],&val[9]);
//  	    numRead/=2;
//              printF("scanning for blasius flow: [%s=%e], [%s=%e], %s=%e\n",c[0],val[0],c[1],val[1],c[2],val[2]);
//  	  }
 
//  	  real jetRadius=.5, jetBoundaryLayerWidth=.1, jetX0=0., jetY0=0., jetZ0=0.;
//            if( jet!=NULL )
//  	  {
//               numRead=sScanF(jet,"jet(%1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e %1s=%e",
//                             c[0],&val[0],c[1],&val[1],c[2],&val[2],c[3],&val[3],c[4],&val[4],c[5],&val[5],
//                             c[6],&val[6],c[7],&val[7],c[8],&val[8],c[9],&val[9]);
//              numRead/=2;

//              real d=.1,r0=.5,x0=0.,y0=0.,z0=0.;
	    
//              char *pos;
//              pos=strstr(jet,"d=");
//              if( pos!=NULL ) sScanF(pos+2,"%e",&jetBoundaryLayerWidth);
//              pos=strstr(jet,"r=");
//              if( pos!=NULL ) sScanF(pos+2,"%e",&jetRadius);
//              pos=strstr(jet,"x=");
//              if( pos!=NULL ) sScanF(pos+2,"%e",&jetX0);
//              pos=strstr(jet,"y=");
//              if( pos!=NULL ) sScanF(pos+2,"%e",&jetY0);
//              pos=strstr(jet,"z=");
//              if( pos!=NULL ) sScanF(pos+2,"%e",&jetZ0);

//              printF("Jet: parameters: jetBoundaryLayerWidth=%e, r=%e, x=%e, y=%e, z=%e\n",jetBoundaryLayerWidth,
//  		   jetRadius,jetX0,jetY0,jetZ0);
	    
//  	  }
	  

//            real oscillateT0=0., oscillateOmega=1., oscillateAmplitude0=0., oscillateAmplitude1=1., 
//                 oscillateOffsetU=0., oscillateOffsetV=0., oscillateOffsetW=0.;
//            if( oscillate!=NULL )
//  	  {
//              // get parameters for oscillating inflow
//              char *pos;
//              pos=strstr(oscillate,"t0=");
//              if( pos!=NULL )
//      	      sScanF(pos+3,"%e",&oscillateT0);
//              pos=strstr(oscillate,"omega=");
//              if( pos!=NULL )                           
//      	      sScanF(pos+6,"%e",&oscillateOmega);
//              pos=strstr(oscillate,"a0=");
//              if( pos!=NULL )                           
//      	      sScanF(pos+3,"%e",&oscillateAmplitude0);
//              pos=strstr(oscillate,"a1=");
//              if( pos!=NULL )                           
//      	      sScanF(pos+3,"%e",&oscillateAmplitude1);
//              pos=strstr(oscillate,"u0=");
//              if( pos!=NULL )                           
//      	      sScanF(pos+3,"%e",&oscillateOffsetU);
//              pos=strstr(oscillate,"v0=");
//              if( pos!=NULL )                           
//      	      sScanF(pos+3,"%e",&oscillateOffsetV);
//              pos=strstr(oscillate,"w0=");
//              if( pos!=NULL )                           
//      	      sScanF(pos+3,"%e",&oscillateOffsetW);
//               printF(" Reading oscillation parameters: omega=%e, t0=%e, a0=%e, a1=%e, (u0,v0,w0)=(%e,%e,%e) \n",
//                 oscillateOmega,oscillateT0,oscillateAmplitude0,oscillateAmplitude1,
//                 oscillateOffsetU,oscillateOffsetV,oscillateOffsetW);
//  	  }

//            real ta=0.,tb=1.,ua=0.,ub=1.,va=0.,vb=0.,wa=0.,wb=0.;
//            if( ramp!=NULL )
//  	  {
//              // get parameters for ramped inflow
//              char *pos;
//              pos=strstr(ramp,"ta=");
//              if( pos!=NULL )
//      	      sScanF(pos+3,"%e",&ta);
//              pos=strstr(ramp,"tb=");
//              if( pos!=NULL )
//      	      sScanF(pos+3,"%e",&tb);
//              pos=strstr(ramp,"ua=");
//              if( pos!=NULL )
//      	      sScanF(pos+3,"%e",&ua);
//              pos=strstr(ramp,"ub=");
//              if( pos!=NULL )
//      	      sScanF(pos+3,"%e",&ub);
//              pos=strstr(ramp,"va=");
//              if( pos!=NULL )
//      	      sScanF(pos+3,"%e",&va);
//              pos=strstr(ramp,"vb=");
//              if( pos!=NULL )
//      	      sScanF(pos+3,"%e",&vb);
//              pos=strstr(ramp,"wa=");
//              if( pos!=NULL )
//      	      sScanF(pos+3,"%e",&wa);
//              pos=strstr(ramp,"wb=");
//              if( pos!=NULL )
//      	      sScanF(pos+3,"%e",&wb);

//               printF(" Reading ramp parameters: ta=%6.2e, tb=%6.2e, ua=%6.2e, ub=%6.2e, va=%6.2e, vb=%6.2e,"
//                      " wa=%6.2e, wb=%6.2e\n",ta,tb,ua,ub,va,vb,wa,wb);
//  	  }
	  

//            int component=-1;
//  	  if( pressure!=NULL )
//  	  {
//              numRead=sScanF(pressure,"pressure(%e*p+%e*p.n=%e",&val[0],&val[1],&val[2]);
//              if( numRead==3 )
//  	    {
//                printF("pressure BC: %e*p+%e*p.n=%e \n",val[0],val[1],val[2]);
//  	    }
//  	    else
//  	    {
//                Overture::abort("error");
//  	    }
//              component=parameters.dbase.get<int >("pc");
//              numRead=0;
//  	  }
//  	  if( mixedDerivative!=NULL )
//  	  {
//              numRead=sScanF(mixedDerivative,"mixedDerivative(%e*t+%e*t.n=%e",&val[0],&val[1],&val[2]);
//              if( numRead==3 )
//  	    {
//                printF("mixed derivative BC: %e*t+%e*t.n=%e \n",val[0],val[1],val[2]);
//  	    }
//  	    else
//  	    {
//                Overture::abort("error");
//  	    }
//              component=parameters.dbase.get<int >("tc");
//              numRead=0;
//  	  }
//            // int code=0;
//            if( userDefinedBoundaryData!=NULL )
//  	  {
//              numRead=0;
//  //             char *pos;
//  //             pos=strstr(userDefinedBoundaryData,"code=");
//  //             if( pos!=NULL )
//  //     	      sScanF(pos+5,"%i",&code);
//  	  }

//            value.redim(parameters.dbase.get<int >("numberOfComponents")+3); value=0;
//  	  real parabolicBoundaryLayerWidth=.1;
//  	  real reynoldsNumberX = 1000.;
	  

//  	  if ( parabolic!=NULL )
//  	  {
//              // fill in the value array: value(parameters.dbase.get<int >("uc"))=val[i] if c[i]='u'
//              //  "d" is an extra name to add to the list of known names.
//  	    const int dc = parameters.dbase.get<int >("numberOfComponents")+1;
//  	    value(dc)=parabolicBoundaryLayerWidth;  // give a default value
//  	    parameters.assignParameterValues("boundary conditions",value,numRead,c,val, "d",dc ); 
//  	    parabolicBoundaryLayerWidth=value(dc);
//  	  }
//  	  else if ( blasius!=NULL )
//  	  {
//              // fill in the value array: value(parameters.dbase.get<int >("uc"))=val[i] if c[i]='u'
//              //  "R" is an extra name to add to the list of known names.
//              const int rxc= parameters.dbase.get<int >("numberOfComponents")+1;
//  	    value(rxc) = reynoldsNumberX;
//  	    parameters.assignParameterValues("boundary conditions",value,numRead,c,val, "R",rxc );
//  	    reynoldsNumberX = value(rxc);
//  	  }
//            else
//  	  {
//              // fill in the value array: value(parameters.dbase.get<int >("uc"))=val[i] if c[i]='u'
//              parameters.assignParameterValues("boundary conditions",value,numRead,c,val ); 
//  	  }
	  

//  	  RealArray timeDependenceParameters;
//  	  if( bc==Parameters::inflowWithVelocityGiven || bc==Parameters::subSonicInflow )
//  	  {
//              if( parabolic!=NULL )
//  	    {
//                if( parabolicBoundaryLayerWidth <=0. )
//  	      {
//                  parabolicBoundaryLayerWidth=.1;
//  		printf("ERROR: width of the parabolic boundary layer is specified to be <= 0.\n"
//                         "Setting to the default value of %e \n",parabolicBoundaryLayerWidth);
//  	      }
//  	      printF("width of the parabolic boundary layer = %e \n",parabolicBoundaryLayerWidth);
//  	    }
//              if( oscillate!=NULL )
//  	    {
//  	      timeDependenceParameters.redim(4);
//  	      timeDependenceParameters(0)=oscillateOmega;
//  	      timeDependenceParameters(1)=oscillateT0;
//                timeDependenceParameters(2)=oscillateAmplitude0;
//                timeDependenceParameters(3)=oscillateAmplitude1;
//                timeDependenceParameters(4)=oscillateOffsetU;
//                timeDependenceParameters(5)=oscillateOffsetV;
//                timeDependenceParameters(6)=oscillateOffsetW;
	      
//  	      printF("the oscillating inflow has the form: a0 + a1*cos((t-t0)*(2*pi*omega))*( uniform/parabolic)");
//  	      printF(" omega=%e, t0=%e, a0=%e, a1=%e \n",oscillateOmega,oscillateT0,oscillateAmplitude0,
//                       oscillateAmplitude1);
//  	    }
//              if( ramp!=NULL )
//  	    {
//  	      timeDependenceParameters.redim(8);
//                timeDependenceParameters(0)=ta;
//                timeDependenceParameters(1)=tb;
//                timeDependenceParameters(2)=ua;
//                timeDependenceParameters(3)=ub;
//                timeDependenceParameters(4)=va;
//                timeDependenceParameters(5)=vb;
//                timeDependenceParameters(6)=wa;
//                timeDependenceParameters(7)=wb;
//  	    }
	    
//  	  }
//  	  else if( bc==Parameters::outflow || 
//                     bc==Parameters::subSonicOutflow ||
//                     bc==Parameters::convectiveOutflow ||
//                     bc==Parameters::tractionFree )
//  	  {
//              // Save info for mixed derivative, alpha*p + beta*p.n = gamma
//              //   value(c) = "gamma" for component c
//              //   value(c+NC*1) = "alpha"   (NC=numberOfComponents)
//              //   value(c+NC*2) = "beta"
//  	    value.redim(parameters.dbase.get<int >("numberOfComponents")*3);
//              value=1.;
//              if( (pressure!=NULL || mixedDerivative!=NULL ) && component>=0 && component<parameters.dbase.get<int >("numberOfComponents") )
//  	    {
//  	      value(component)=val[2];  // note: rhs value goes first
//  	      value(component+parameters.dbase.get<int >("numberOfComponents")*1)=val[0];
//  	      value(component+parameters.dbase.get<int >("numberOfComponents")*2)=val[1];
//  	    }
//  	    else
//   	      value(Range(0,parameters.dbase.get<int >("numberOfComponents")-1))=0.;
//  	  }
	  

//  // 	  if( userDefinedBoundaryData!=NULL )
//  // 	  {
//  // 	    value.redim(1);
//  //             value(0)=code;
//  // 	  }
	  

//  	  // assign boundary conditions and data
//  	  if( gridName=="all")
//  	  {
//  	    printF("Setting boundary condition to %s on all grids.\n",(const char*)parameters.dbase.get< >("bcName")[bc]);
//  	  }
//            bool bcWasAssigned=false;
//  	  for( grid=G.getBase(); grid<=G.getBound(); grid++ )
//  	  {
//  	    for( int axis=A.getBase(); axis<=A.getBound(); axis++ )
//  	    {
//  	      for( int side=S.getBase(); side<=S.getBound(); side++ )
//  	      {
//                  // printF("grid,side,axis,original,change=%i,%i,%i,  %i,%i\n",grid,side,axis,
//  		//        originalBoundaryCondition(side,axis,grid),changeBoundaryConditionNumber);
		
//  		if( cg[grid].boundaryCondition(side,axis) > 0 && 
//                          (changeBoundaryConditionNumber==-1 || 
//  			 originalBoundaryCondition(side,axis,grid)==changeBoundaryConditionNumber) )
//  		{
//                    bcWasAssigned=true;

//  		  cg[grid].setBoundaryCondition(side,axis,bc);
//                    // set underlying mapping too (for moving grids)
//                    cg[grid].mapping().getMapping().setBoundaryCondition(side,axis,bc);

//  		  setDefaultDataForABoundaryCondition(side,axis,grid,cg,parameters);

//                    // the array parameters.bcData(.,side,axis,grid) holds the data that is
//                    // passed directly to the boundary condition routines (if the BC doesn't vary in space)

//  		  if( gridName!="all" )
//  		    printF("Setting (side,axis)=(%i,%i) of grid %s to bc=%s, (number=%i)\n",side,axis,
//  			   (const char*)cg[grid].mapping().getName(Mapping::mappingName),
//                             (const char*)parameters.dbase.get< >("bcName")[bc],bc);
//  		  if( value.getBound(0) > parameters.dbase.get<RealArray >("bcData").getBound(0) )
//  		  {
//                      Range all,R0(parameters.dbase.get<RealArray >("bcData").getBound(0)+1,value.getBound(0));
		    
//  		    parameters.dbase.get<RealArray >("bcData").resize(value.getLength(0),parameters.dbase.get<RealArray >("bcData").getLength(1),
//  					     parameters.dbase.get<RealArray >("bcData").getLength(2),parameters.dbase.get<RealArray >("bcData").getLength(3));
                    
//                      parameters.bcData(R0,all,all,all)=0.; // give initial values
//  		  }
//                    for( n=0; n<=value.getBound(0); n++ )
//  		    parameters.bcData(n,side,axis,grid)=value(n);

//  		  if( bc==Parameters::inflowWithVelocityGiven || bc==Parameters::subSonicInflow )
//  		  {
//  		    if ( blasius!=NULL )
//  		    {
//  		      parameters.setBcType(side,axis,grid,Parameters::blasiusProfile);
//  		      parameters.setTimeDependenceBoundaryConditionParameters(side,axis,grid,timeDependenceParameters);		    
//                        parameters.setBcVariesInSpace(side,axis,grid,true);
//  		      parameters.bcParameters(1,side,axis,grid)=reynoldsNumberX;

//  		    }
//  		    else if( parabolic!=NULL )
//  		    {
//                        if( ramp!=NULL )
//  		      {
//                          parameters.setBcType(side,axis,grid,Parameters::parabolicInflowRamped);
//                          parameters.setBcIsTimeDependent(side,axis,grid,true);
//  			parameters.setTimeDependenceBoundaryConditionParameters(side,axis,grid,timeDependenceParameters);
			
//  //                         int rpStart=1;
//  // 			assert( parameters.dbase.get<RealArray >("bcParameters").getLength(0)>value.getLength(0) );
//  // 			for( n=0; n<=value.getBound(0); n++ )
//  // 			  parameters.bcParameters(rpStart+n,side,axis,grid)=value(n);
//  		      }
//                        else if( oscillate!=NULL )
//  		      {
//                          parameters.setBcType(side,axis,grid,Parameters::parabolicInflowOscillating );
//                          parameters.setBcIsTimeDependent(side,axis,grid,true);
//  			parameters.setTimeDependenceBoundaryConditionParameters(side,axis,grid,timeDependenceParameters);
//  		      }
//                        else if( userDefinedBoundaryData!=NULL )
//  		      {
//                          parameters.setBcType(side,axis,grid,Parameters::parabolicInflowOscillating );
//                          parameters.setBcIsTimeDependent(side,axis,grid,true);

//  			parameters.setBcVariesInSpace(side,axis,grid,true);  // always assume this for user BC's
//  			parameters.chooseUserDefinedBoundaryValues(side,axis,grid,cg);
//  		      }
//  		      else
//                          parameters.setBcType(side,axis,grid,Parameters::parabolicInflow);

//                        parameters.setBcVariesInSpace(side,axis,grid,true);

//  		      parameters.bcParameters(0,side,axis,grid)=parabolicBoundaryLayerWidth;



//                        printF(" +++ setting parameters.bcParameters(0,%i,%i,%i)=%e\n",
//  			     side,axis,grid,parabolicBoundaryLayerWidth);
//  		    }
//  		    else if( uniform!=NULL )
//  		    {
//  		      parameters.setBcType(side,axis,grid,Parameters::uniformInflow);
//                        if( ramp!=NULL )
//  		      {
//  			parameters.setTimeDependenceBoundaryConditionParameters(side,axis,grid,timeDependenceParameters);
//  		      }
//  		      else if( oscillate!=NULL )
//  		      {
//  			parameters.setTimeDependenceBoundaryConditionParameters(side,axis,grid,timeDependenceParameters);
//  		      }
//  		    }
//                      else if( ramp!=NULL )
//  		    {
//                        parameters.setBcIsTimeDependent(side,axis,grid,true);
//  		      parameters.setBcType(side,axis,grid,Parameters::rampInflow);
//  		      parameters.setTimeDependenceBoundaryConditionParameters(side,axis,grid,timeDependenceParameters);

//  		    }
//                      else if( oscillate!=NULL )
//  		    {
//  		    }
//                      else if( jet!=NULL )
//  		    {
//  		      // the jet values are actually assigned in jetInflow (bcForcing.C)
//                        parameters.setBcType(side,axis,grid,Parameters::jetInflow);
//  		      parameters.setBcVariesInSpace(side,axis,grid,true);
		      
//  		      parameters.bcParameters(0,side,axis,grid)=jetRadius;
//  		      parameters.bcParameters(1,side,axis,grid)=jetBoundaryLayerWidth;
//  		      parameters.bcParameters(2,side,axis,grid)=jetX0;
//  		      parameters.bcParameters(3,side,axis,grid)=jetY0;
//  		      parameters.bcParameters(4,side,axis,grid)=jetZ0;
		      
//  		    }
//  		    else if( userDefinedBoundaryData!=NULL )
//  		    {
//                        parameters.setBcVariesInSpace(side,axis,grid,true);  // always assume this for user BC's
//                        parameters.chooseUserDefinedBoundaryValues(side,axis,grid,cg);
//  /* ------------		      
//                        parameters.setBcIsTimeDependent(side,axis,grid,true);  // query for this?


//                        parameters.setBcType(side,axis,grid,Parameters::userDefinedBoundaryData);
//                        parameters.variableBoundaryData(grid)=TRUE; 
//  ----- */

//  		    }
		    
		    
//  		  }
//  		  else if( bc==Parameters::noSlipWall || bc==Parameters::slipWall )
//  		  {
//  		    if( userDefinedBoundaryData!=NULL )
//  		    {
//  		      // **** new ****

//  		      parameters.setBcVariesInSpace(side,axis,grid,true);  // always assume this for user BC's
//  		      parameters.chooseUserDefinedBoundaryValues(side,axis,grid,cg);
//  		    }
		    
//  		  }
		  
		  
//  		}
//  	      }
//  	    }
//  	  } // end for grid
//  	  if( !bcWasAssigned && changeBoundaryConditionNumber>0 )
//  	  {
//  	    printF("\n setBoundaryConditions:INFO: There were no boundaries assigned for bcNumber%i=%s\n\n",
//                        changeBoundaryConditionNumber,(const char*)bcName );
//  	  }
	  
  	}
      }
    }
    
  }

  else if( answer=="data for boundary conditions" )
  {
    // fill in the parameters.bcData(.,side,axis,grid) array with data for boundary conditions
    // for example an inflow BC will need some values

    gi.appendToTheDefaultPrompt("bc data>");



    RealArray value;
    // first make a list of the distinct BC's
    intArray distinctBC(cg.numberOfComponentGrids()*cg.numberOfDimensions()*2);
    int numberOfDistinctBC=0, bc;
    for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    {
      for( axis=0; axis<cg.numberOfDimensions(); axis++ )
      {
	for( side=Start; side<=End; side++ )
	{
	  bc=cg[grid].boundaryCondition()(side,axis);
	  if( bc > 0 )
	  {
	    for( int i=0; i<numberOfDistinctBC; i++ )
	    {
	      if( bc==distinctBC(i) )
	      {
		bc=-1;  // this bc is already in the list
		break;
	      }
	    }
	    if( bc>0 && bc<Parameters::numberOfBCNames )
	      distinctBC(numberOfDistinctBC++)=bc;
	    else if( bc > Parameters::numberOfBCNames )
	    {
	      printF("ERROR: There is an unknown BC=%i on side (%i,%i,%i) = (%s,side,axis) \n",
		     bc,grid,side,axis,(const char *)cg[grid].mapping().getName(Mapping::mappingName));
	    }
	  }
	}
      }
    }
    if( numberOfDistinctBC==0 )
    {
      printF("ERROR : There are no physical boundaries to give data for!\n");
    }
    else
    {
      aString *bcMenu = new aString[numberOfDistinctBC+2];
      for( int i=0; i<numberOfDistinctBC; i++ )
	bcMenu[i]=parameters.dbase.get< >("bcName")[distinctBC(i)]; //  + sPrintF(buff," (number %i)",distinctBC(i));
      bcMenu[numberOfDistinctBC]="done";
      bcMenu[numberOfDistinctBC+1]="";
      
      gi.appendToTheDefaultPrompt("bc>");
      for(;;)
      {
	bc=gi.getMenuItem(bcMenu,answer2,"Set data for which boundary condition?");
	if( answer2=="done" )
	  break;
	else if( bc<0 )
	{
	  cout << "unknown response=[" << answer2 << "]\n";
	  gi.stopReadingCommandFile();
	  continue;
	}
	// make a list of all sides with this boundary condition
	bc=distinctBC(bc);
	aString *sideMenu = new aString[cg.numberOfComponentGrids()*cg.numberOfDimensions()*2+3];
	int i=0;
	for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
	{
	  for( axis=0; axis<cg.numberOfDimensions(); axis++ )
	  {
	    for( side=Start; side<=End; side++ )
	    {
	      if( bc==cg[grid].boundaryCondition()(side,axis) )
	      {
		sideMenu[i++]=sPrintF(buff,"(%i,%i,%i) = (%s,side,axis)",grid,side,axis,
				      (const char*)cg[grid].mapping().getName(Mapping::mappingName));
	      }
	    }
	  }
	}
	sideMenu[i++]="all";
	sideMenu[i++]="done";
	sideMenu[i]="";
	gi.appendToTheDefaultPrompt("side>");
	for( ;; )
	{
	  int response=gi.getMenuItem(sideMenu,answer2,sPrintF(buff,"Set %s data for which side(s)?",
							       (const char*)parameters.dbase.get< >("bcName")[bc]));
	  if( response<0 )
	  {
	    printF("Unknown response = %s, try again..\n",(const char *) answer2);
	    gi.stopReadingCommandFile();
	    continue;
	  }
	  else if( answer2=="done" )
	    break;
	  else 
	  {
	    int inflowProfile=-1;
	    if( bc==Parameters::inflowWithVelocityGiven || bc==Parameters::subSonicInflow )
	    {
	      aString bcTypeMenu[]=
	      {
		"uniform",
		"parabolic",
		"uniform with oscillation",
		"parabolic with oscillation",
		""
	      };
	      gi.getMenuItem(bcTypeMenu,answer3,"Choose the inflow profile");
	      if( answer3=="uniform" || answer3=="uniform with oscillation" )
	      {
		inflowProfile=Parameters::uniformInflow;
	      }
	      else if( answer3=="parabolic" || answer3=="parabolic with oscillation" )
	      {
		inflowProfile=Parameters::parabolicInflow;
	      }
	      else
	      {
		cout << "Unknown response=" << answer3 << endl;
		gi.stopReadingCommandFile();
	      }

	      value.redim(parameters.dbase.get<int >("numberOfComponents")+2); value=0;
	      for( n=0; n<parameters.dbase.get<int >("numberOfComponents"); n++ )
	      {
		gi.inputString(answer3,sPrintF(buff,"Enter bc value for %s",
					       (const char *)parameters.dbase.get<aString* >("componentName")[n]));
		if( answer3!="" )
		  sScanF(answer3,"%e",&value(n));
	      }
	      printF(" %s boundary condition data: (",(const char*)parameters.dbase.get< >("bcName")[bc]);
	      for( n=0; n<parameters.dbase.get<int >("numberOfComponents"); n++ )
		printf("%s=%f,",(const char *)parameters.dbase.get<aString* >("componentName")[n],value(n));
	      printF(")\n");
	      if( answer3=="uniform with oscillation" || answer3=="parabolic with oscillation" )
	      {
		n=parameters.dbase.get<int >("numberOfComponents");
		value(n)=1.;
		value(n+1)=0.;
		printf("the oscillating inflow has time dependence of cos[ (t-t0)*(2*pi*omega) ] ");
		gi.inputString(answer3,"Enter omega and t0");
		if( answer3!="" )
		  sScanF(answer3,"%e %e",&value(n),&value(n+1));
	      }
	    }
	    else if( bc==Parameters::outflow || 
                     bc==Parameters::subSonicOutflow ||
                     bc==Parameters::convectiveOutflow ||
                     bc==Parameters::tractionFree )
	    {
	      value.redim(3);
	      value(0)=1., value(1)=1., value(2)=0.;
	      printF("The outflow/convectiveOutflow/tractionFree boundary condition for p is a*p + b*p.n = c \n"
		     " where p=pressure, p.n=normal derivative of p.\n"
		     " Note that a and b should have the same sign or else the condition is unstable\n"
		     " Default values are a=%e, b=%e, c=%e \n",value(0),value(1),value(2));
	      gi.inputString(answer3,"Enter a,b,c for bc: a*p + b*p.n = c");
	      if( answer3!="" )
		sScanF(answer3,"%e %e %e",&value(0),&value(1),&value(1));
	      printF(" new values: a=%e, b=%e, c=%e \n",value(0),value(1),value(2));
	    }
	    else if( bc==Parameters::noSlipWall )
	    {
	      value.redim(parameters.dbase.get<int >("numberOfComponents")); value=0;
	      for( n=0; n<parameters.dbase.get<int >("numberOfComponents"); n++ )
	      {
		gi.inputString(answer3,sPrintF(buff,"Enter bc value for %s",
					       (const char *)parameters.dbase.get<aString* >("componentName")[n]));
		if( answer3!="" )
		  sScanF(answer3,"%e",&value(n));
	      }
	      printF(" %s boundary condition data: (",(const char*)parameters.dbase.get< >("bcName")[bc]);
	      for( n=0; n<parameters.dbase.get<int >("numberOfComponents"); n++ )
		printf("%s=%f,",(const char *)parameters.dbase.get<aString* >("componentName")[n],value(n));
	      printF(")\n");
	    }
	    else
	    {
	      printF("Sorry: I do not know what data to give this boundary condition\n");
	      break;
	    }
	      
	    // now fill in the bcData array
	    if( answer2=="all" )
	    {
	      for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
		for( axis=0; axis<cg.numberOfDimensions(); axis++ )
		  for( side=Start; side<=End; side++ )
		    if( bc==cg[grid].boundaryCondition()(side,axis) )
		    {
		      assert( value.getBound(0)<=parameters.dbase.get<RealArray >("bcData").getBound(0) );
		      for( n=0; n<=value.getBound(0); n++ )
			parameters.bcData(n,side,axis,grid)=value(n);

		      if( inflowProfile==Parameters::parabolicInflow )
		      {
			parameters.setBcType(side,axis,grid,Parameters::parabolicInflow);
			gi.inputString(answer3,sPrintF(buff,
						       "(grid=%s,side=%i,axis=%i) Enter the width of the parabolic boundary layer",
						       (const char*)cg[grid].mapping().getName(Mapping::mappingName),side,axis));
			if( answer3!="" )
			{
			  sScanF(answer3,"%e",&parameters.bcParameters(0,side,axis,grid));
			  printF("width of the parabolic boundary layer = %e \n",
                                 parameters.bcParameters(0,side,axis,grid));
			}
		      }
		    }
	    }
	    else
	    {
	      sScanF(answer2,"(%i %i %i)",&grid,&side,&axis);
	      assert( grid>=0 && grid<cg.numberOfComponentGrids() && side>=0 && side<=1 
		      && axis>=0 && axis<cg.numberOfDimensions() );
	      assert( value.getBound(0)<=parameters.dbase.get<RealArray >("bcData").getBound(0) );

	      for( n=0; n<=value.getBound(0); n++ )
		parameters.bcData(n,side,axis,grid)=value(n);

	      if( inflowProfile==Parameters::parabolicInflow )
	      {
		parameters.setBcType(side,axis,grid,Parameters::parabolicInflow);
		gi.inputString(answer3,"Enter the width of the parabolic boundary layer");
		if( answer3!="" )
		{
		  sScanF(answer3,"%e",&parameters.bcParameters(0,side,axis,grid));
		  printF("width of the parabolic boundary layer = %e \n",parameters.bcParameters(0,side,axis,grid));
		}
	      }
	    }
	  }
	} // for(;;)  this bc
	gi.unAppendTheDefaultPrompt();
	delete [] sideMenu;
      }
      gi.unAppendTheDefaultPrompt();
      delete [] bcMenu;
    }
    gi.unAppendTheDefaultPrompt();
  }
  return 0;
  
}

//  int Parameters::
//  updateBoundaryConditionDialog( int side, int axis, int grid,
//                                 CompositeGrid & cg )
//  // =======================================================================================
//  // /Description:
//  //    Prompt for changes to boundary condition parameters for a particular face of a grid.
//  //
//  //     This function knows how to assign parameters and rhs data for all boundary conditions.
//  // =======================================================================================
//  {
//    assert( ps !=NULL );
//    GenericGraphicsInterface & gi = *ps;

//    GUIState dialog;
//    // DialogData & dialog = dialog.getDialogSibling();

//    dialog.setExitCommand("close","close");

//    assert( grid>=0 && grid<cg.numberOfComponentGrids() );
  
//    int bc0=cg[grid].boundaryCondition(side,axis);
//    assert( bc0>0 && bc0<numberOfBCNames );

//    BoundaryCondition bc = BoundaryCondition(bc0);
  
//    const aString sideLabel[]={ "left","right","bottom","top","back","front"};  //
//    aString label= sideLabel[side+2*axis];

//    const int numberOfTextStrings=20;
//    aString textLabels[numberOfTextStrings];
//    aString textStrings[numberOfTextStrings];    
//    int nt=0;

//    if( bc==noSlipWall )
//    {
//      aString name="noSlipWall";
//      dialog.setWindowTitle(label+" bc : "+name);
   
//      textLabels[nt] = "wall values:";  sPrintF(textStrings[nt], "u=%g, v=%g, w=%g",0.,0.,0.);  nt++; 

//  	// null strings terminal list
//      textLabels[nt]="";   textStrings[nt]="";  assert( nt<numberOfTextStrings );
//      dialog.setTextBoxes(textLabels, textLabels, textStrings);

//    }
//    else if( bc==inflowWithVelocityGiven )
//    {
//      aString name="inflowWithVelocityGiven";
//      dialog.setWindowTitle(label+" bc : "+name);
   
//      aString tbCommands[] = {"uniform", 
//  			    "parabolic",
//  			    "oscillate",
//  			    ""};
//      int tbState[4];
//      tbState[0] = 0;
//      tbState[1] = 0;
//      tbState[2] = 0;
//      tbState[3] = 0;
//      int numColumns=1;
//      dialog.setToggleButtons(tbCommands, tbCommands, tbState, numColumns); 

//      textLabels[nt] = "uniform";    sPrintF(textStrings[nt], "u=%g",0.);  nt++; 
//      textLabels[nt] = "parabolic";  sPrintF(textStrings[nt], "width=%g",.25);  nt++; 

//  	// null strings terminal list
//      textLabels[nt]="";   textStrings[nt]="";  assert( nt<numberOfTextStrings );
//      dialog.setTextBoxes(textLabels, textLabels, textStrings);

//    }
//    else if( bc==tractionFree )
//    {
//      gi.outputString("The traction free BC is:  P n_i - alpha nu n_i( D_j u_i + D_i u_j ) where alpha=1 (default)");

//      aString name="tractionFree";
//      dialog.setWindowTitle(label+" bc : "+name);
   
//      textLabels[nt] = "alpha :";  sPrintF(textStrings[nt], "alpha=%g",1.);  nt++; 

//  	// null strings terminal list
//      textLabels[nt]="";   textStrings[nt]="";  assert( nt<numberOfTextStrings );
//      dialog.setTextBoxes(textLabels, textLabels, textStrings);
//    }
//    else
//    {
//      printF("updateBoundaryConditionDialog:ERROR: unknown value for bc=%i\n",bc);
//      return 1;
//    }
  

//    aString answer,line;

//    gi.pushGUI(dialog);
//    for( ;; )
//    {
//      gi.getAnswer(answer,"");
//      if( answer=="close" )
//      {
//        break;
//      }
//      else
//      {
//        cout << "Unknown command: " << answer << endl;
//      }
//    }

//    gi.popGUI();

//    return 0;
  
//  }

static aString bcBoundaryChoice,faceChoice,bcTypeName,bcOption1,bcOption2;

// return the BC command
static aString 
getCommand()
{
  aString bcCommand;
  char buff[500];
  bcCommand = sPrintF(buff,"%s%s=%s %s %s",(const char*)bcBoundaryChoice,(const char*)faceChoice,
                      (const char*)bcTypeName,(const char*)bcOption1,(const char*)bcOption2);
  return bcCommand;
}


int Parameters::
defineBoundaryConditions(CompositeGrid & cg, 
                         const IntegerArray & originalBoundaryCondition,
                         const aString & command /* = nullString */,
			 DialogData *interface /* =NULL */ )
// =====================================================================================
// /Description:
//     Define the boundary conditions. This is the *new* dialog based way to specify
//  boundary conditions.
// =====================================================================================
{
  int returnValue=0;

  assert( ps !=NULL );
  GenericGraphicsInterface & gi = *ps;

  // For now turn off the prefix -- not really needed anymore
  // aString prefix = "OBBC:"; // prefix for commands to make them unique.
//    const bool executeCommand = command!=nullString;
//    if( executeCommand && command(0,prefix.length()-1)!=prefix && command!="build dialog" )
//      return 1;

  const bool executeCommand = false;
  aString prefix = ""; // prefix for commands to make them unique.


  aString answer,line;
  char buff[100];
//  const int numberOfDimensions = cg.numberOfDimensions();
  

  GUIState gui;
  gui.setWindowTitle("Boundary conditions");
  gui.setExitCommand("done", "continue");
  DialogData & dialog = interface!=NULL ? *interface : (DialogData &)gui;


  const int maxCommands=max(cg.numberOfGrids()+2,numberOfBCNames+1);
  aString *cmd = new aString[maxCommands];


  RealArray defaultState(numberOfComponents);
  defaultState=0.;

  int activeGrid=0;
  
  bcBoundaryChoice="all";
  faceChoice="";
  bcTypeName="noSlipWall";
  bcOption1=""; // ", uniform(p=0.,u=0.,v=0.)";
  bcOption2="";

  // Make a list of names of possible boundary conditions that are valid for this pde
  aString bcNameList[numberOfBCNames+1];
  aString bcNameLabel[numberOfBCNames+1];
  int numberOfValidBoundaryConditions=0;
  int i;
  for( i=0; i<numberOfBCNames; i++ )
  {
    if( checkForValidBoundaryCondition(i,*this,false)==0 && i!=Parameters::interpolation )
    {
      bcNameList[numberOfValidBoundaryConditions]=bcName[i];
      bcNameLabel[numberOfValidBoundaryConditions]=sPrintF(buff,"bc: %s",(const char*)bcName[i]);
      if( numberOfValidBoundaryConditions==0 )
        bcTypeName=bcName[i];  // set default
      
      numberOfValidBoundaryConditions++;
    }
  }
  bcNameList[numberOfValidBoundaryConditions]="";
  bcNameLabel[numberOfValidBoundaryConditions]="";


  if( interface==NULL || command=="build dialog" )
  {

    dialog.setOptionMenuColumns(1);

    aString *gridName = new aString [cg.numberOfComponentGrids()+2];
    aString *gridLabel = new aString [cg.numberOfComponentGrids()+2];
    gridName[0]="all";
    gridLabel[0]="boundary: all";
    int grid;
    for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    {
      gridName[grid+1]=sPrintF(buff,"%s",(const char*)cg[grid].getName());
      gridLabel[grid+1]=sPrintF(buff,"boundary: %s",(const char*)cg[grid].getName());
    }
    gridName[cg.numberOfComponentGrids()+1]="";
    gridLabel[cg.numberOfComponentGrids()+1]="";
    
    dialog.addOptionMenu("boundary:", gridLabel, gridName,  0);


    // Find all the distinct boundary condition numbers *** should use old values of the bc ****
    std::vector<int> bcNumbers;
    for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    {
      for( int side=0; side<=1; side++ )
      {
	for( int axis=0; axis<cg.numberOfDimensions(); axis++ )
	{
          int bc = originalBoundaryCondition(side,axis,grid);
	  if( bc>0 && find(bcNumbers.begin(),bcNumbers.end(),bc)==bcNumbers.end() )
	  {
	    bcNumbers.push_back(bc);
	  }
	}
      }
    }
    // *** sort the vector ***
    sort(bcNumbers.begin(),bcNumbers.end());
    
    const int numberOfBoundaryConditions=bcNumbers.size();
    // build a list of valid boundary condition numbers
    aString *bcNumberCommand = new aString [numberOfBoundaryConditions+1];   
    aString *bcNumberLabel   = new aString [numberOfBoundaryConditions+1];   
    for( i=0; i<numberOfBoundaryConditions; i++ )
    {
      bcNumberCommand[i]=sPrintF("bc=%i",bcNumbers[i]);
      bcNumberLabel[i]=sPrintF("bcNumber: %i",bcNumbers[i]);
    }
    bcNumberCommand[numberOfBoundaryConditions]="";
    bcNumberLabel[numberOfBoundaryConditions]="";
    
    dialog.addOptionMenu("bc number:", bcNumberLabel, bcNumberCommand, 0);
    delete [] bcNumberCommand;
    delete [] bcNumberLabel;

    aString sideName[]= {"all", "left","right","bottom","top","back","front",""  };//
    aString sideLabel[]= {"set face all", "set face left","set face right","set face bottom","set face top",
                              "set face back","set face front",""  };

    if( cg.numberOfDimensions()==2 )
    {
      sideName[5]="";
      sideLabel[5]="";
    }
      
    dialog.addOptionMenu("set face:", sideLabel, sideName, 0 );
   
    dialog.addOptionMenu("bc:", bcNameLabel, bcNameList,0);

    // Main options

//          "    option = `uniform(p=1.,u=1.,...)'       : to specify a uniform inflow profile   \n"
//          "    option = `parabolic(d=2,p=1.,...)'      : to specify a parabolic inflow profile \n"
//          "    option = `blasius(p=1.,u=1.,R=1000)'  : to specify a laminar boundary layer profile \n"
//          "    option = `pressure(.1*p+1.*p.n=0.)'     : pressure boundary condition at outflow\n"
//          "    option = `oscillate(t0=.5,omega=1.,a0=0.,a1=1.,u0=0,v0=0.,w0=0.)'  : oscillating values    \n"
//          "    option = `ramp(ta=0.,tb=1.,ua=0.,ub=1.,va=0.,..)  : ramped inflow parameters    \n" 
//          "    option = `jet(d=.1,r=1.,x=0.,y=0,z=0.,p=1.,...)': to specify a parabolic jet inflow profile \n" 
//          "    option = `userDefinedBoundaryData : choose user defined data  \n"

    aString optionCommand[] = { "no option",
                                "uniform",
                                "parabolic",
                                "blasius",
                                "pressure",
                                "jet",
                                "user defined",
                                "" };  // 
    dialog.addOptionMenu("option:", optionCommand, optionCommand, 0);

    aString option2Command[] = { "no option2",
                                 "oscillate",
				 "ramp",
				 "" };  // 
    dialog.addOptionMenu("option2:", option2Command, option2Command, 0);


    aString extrapOptionCommand[] = { "polynomial extrapolation",
                                      "limited extrapolation",
				      "" };  // 
    dialog.addOptionMenu("extrap option:", extrapOptionCommand, extrapOptionCommand, (int)extrapolationOption);

/* ---
    aString sideName[3][2]= {"left:    ","right:  ","bottom:","top:     ","back:    ","front:   "  };//
    aString sideLabel[3][2]= {"set bc left ","set bc right","set bc bottom","set bc top",
                              "set bc back","set bc front"  };//
    aString periodicName[] ={ "periodic",""};  //
    aString interpolationName[] ={ "interpolation",""};  //
      
    int axis,side;
    for( axis=0; axis<cg.numberOfDimensions(); axis++ )
    {
      for( side=0; side<=1; side++ )
      {
        int bc=cg[activeGrid].boundaryCondition(side,axis);
	if( bc>0 )
	{
	  addPrefix(validBCName,sideLabel[axis][side],bcLabel,numberOfBCNames);
          addPrefix(bcLabel,prefix,cmd,maxCommands);
  	  dialog.addOptionMenu(sideName[axis][side], cmd, validBCName, bc );
	}
        else if( bc<0 )
  	  dialog.addOptionMenu(sideName[axis][side], periodicName, periodicName,0);
        else 
  	  dialog.addOptionMenu(sideName[axis][side], interpolationName, interpolationName,0);
      }
    }
  ---- */
    
//      aString outflowOptions[]={"default outflow",
//                                "check for inflow at outflow",
//                                "expect inflow at outflow"};

//      dialog.addOptionMenu("outflow options", outflowOptions,outflowOptions,parameters.dbase.get<int >("checkForInflowAtOutFlow"));
    


//      aString editBC[] = {"edit left","edit right","edit bottom","edit top","edit back","edit front",""};
//      editBC[cg.numberOfDimensions()*2]="";   // chop off those that are not needed.
//      addPrefix(editBC,prefix,cmd,maxCommands);
//      int numRows=1;
//      dialog.setPushButtons( cmd, editBC, numRows ); 

    aString pushButtonCommands[]={"apply bc command","help","plot grid",""}; //
    addPrefix(pushButtonCommands,prefix,cmd,maxCommands);
    int numRows=5;
    dialog.setPushButtons( cmd, pushButtonCommands, numRows ); 

    // ----- Text strings ------
    const int numberOfTextStrings=20;
    aString textCommands[numberOfTextStrings];
    aString textLabels[numberOfTextStrings];
    aString textStrings[numberOfTextStrings];

    int nt=0;
    
    
    textCommands[nt] = "bc command";  textLabels[nt]=textCommands[nt];
    textStrings[nt] = getCommand();
    nt++; 

    textCommands[nt] = "default state";  textLabels[nt]=textCommands[nt];
    textStrings[nt] = "";
    for( int n=0; n<numberOfComponents; n++ )
    {
      textStrings[nt]+=sPrintF(buff,"%s=%g",(const char *)componentName[n],defaultState(n));
      if( n<numberOfComponents-1 ) textStrings[nt]+=",";
    }
    nt++; 

    textCommands[nt] = "order of extrap for interp neighbours";  textLabels[nt]=textCommands[nt];
    sPrintF(textStrings[nt], "%i", orderOfExtrapolationForInterpolationNeighbours); 
    nt++; 
    textCommands[nt] = "order of extrap for 2nd ghost line";  textLabels[nt]=textCommands[nt];
    sPrintF(textStrings[nt], "%i", orderOfExtrapolationForSecondGhostLine);  
    nt++; 

    // null strings terminal list
    textCommands[nt]="";   textLabels[nt]="";   textStrings[nt]="";  assert( nt<numberOfTextStrings );
    addPrefix(textCommands,prefix,cmd,maxCommands);
    dialog.setTextBoxes(cmd, textLabels, textStrings);


    delete [] cmd;
    delete [] gridName;
    delete [] gridLabel;
      
    // gui.buildPopup(pdeParametersMenu);
    // if( gi.graphicsIsOn() )
    //   dialog.openDialog(0);   // open the dialog here so we can reset the parameter values below

//     updatePDEparameters();
//     if( pde==Parameters::compressibleNavierStokes ||
// 	pde==Parameters::allSpeedNavierStokes )
//     {
//       fillCompressibleDialogValues(dialog,*this );
//     }
    
    if( executeCommand ) return 0;
  }
  
  const bool squaresWerePlotted=gi.getPlotTheColouredSquares();
  
  if( !executeCommand  )
  {
    gi.pushGUI(gui);
    gi.appendToTheDefaultPrompt("bc>");  

    // colour boundaries by bc number
    psp.getBoundaryColourOption()=GraphicsParameters::colourByBoundaryCondition;
    psp.set(GI_PLOT_THE_OBJECT_AND_EXIT,true);
    gi.setPlotTheColouredSquares(true);
  }
  int len=0;
  for(int it=0; ; it++)
  {
    if( pde==Parameters::compressibleNavierStokes ||
        pde==Parameters::compressibleMultiphase ||
	pde==Parameters::allSpeedNavierStokes )
    {
//       updatePDEparameters();
//       fillCompressibleDialogValues(dialog,*this );
    }


    if( !executeCommand )
    {
      gi.erase();
      PlotIt::plot(gi,cg,psp);  // plot the grid

      gi.getAnswer(answer,"");
    }
    else
    {
      if( it==0 ) 
        answer=command;
      else
        break;
    }
  
    if( answer(0,prefix.length()-1)==prefix )
      answer=answer(prefix.length(),answer.length()-1);   // strip off the prefix

    // gi.getMenuItem(pdeParametersMenu,answer,"change a pde parameter");


    if( answer=="done" )
      break;
    else if( answer=="plot grid" )
    {
      psp.set(GI_PLOT_THE_OBJECT_AND_EXIT,false);
      gi.erase();
      PlotIt::plot(gi,cg,psp);  // plot the grid
      psp.set(GI_PLOT_THE_OBJECT_AND_EXIT,true);
    }
    else if( len=answer.matches("boundary:") )
    {
      line=answer(len+1,answer.length());
      if( line!="all" )
      { // check that we have chosen a valid grid name
        
      }

      // reset the options
      bcOption1=""; 
      dialog.getOptionMenu("option:").setCurrentChoice("no option");
      bcOption2=""; 
      dialog.getOptionMenu("option2:").setCurrentChoice("no option2");

      bcBoundaryChoice=line;
      dialog.setTextLabel("bc command",(const char*)getCommand());
    }
    else if( len=answer.matches("bc number:") )
    {
      int bcNumber=-1;
      sScanF(answer(len+1,answer.length()),"%i",&bcNumber);
      if( bcNumber>0 )
      { // check that we have chosen a valid bcNumber
        
	// reset the options
	bcOption1=""; 
	dialog.getOptionMenu("option:").setCurrentChoice("no option");
	bcOption2=""; 
	dialog.getOptionMenu("option2:").setCurrentChoice("no option2");

	bcBoundaryChoice=sPrintF(buff,"bcNumber%i",bcNumber);
	dialog.setTextLabel("bc command",(const char*)getCommand());

        dialog.getOptionMenu("bc number:").setCurrentChoice(answer);
      }
    }
    else if( len=answer.matches("set face") )
    {
      faceChoice = ( answer=="set face left"   ? "(0,0)" :
		     answer=="set face right"  ? "(1,0)" :
		     answer=="set face bottom" ? "(0,1)" :
		     answer=="set face top"    ? "(1,1)" :
		     answer=="set face back"   ? "(0,2)" : 
		     answer=="set face front"  ? "(1,2)" : "" );

      dialog.setTextLabel("bc command",(const char*)getCommand());
      dialog.getOptionMenu("set face:").setCurrentChoice(answer(len+1,answer.length()-1));
    }
    else if( len=answer.matches("bc:") )
    {
      line=answer(len+1,answer.length());
      // check that we have chosen a valid bc
      // **** finish this *****
      int bcOption=0;
      for( i=0; i<numberOfValidBoundaryConditions; i++ )
      {
	if( line==bcNameList[i] )
	{
	  bcOption=i;
	  break;
	}
      }
      if( bcOption>=0 )
      {
	// reset the options
	bcOption1=""; 
	dialog.getOptionMenu("option:").setCurrentChoice("no option");
	bcOption2=""; 
	dialog.getOptionMenu("option2:").setCurrentChoice("no option2");

	bcTypeName=line;
	dialog.setTextLabel("bc command",(const char*)getCommand());
	dialog.getOptionMenu("bc:").setCurrentChoice((int)bcOption);
      }
      else
      {
	printf("ERROR: invalid `bc:' command=[%s]\n",(const char *)answer);
	gi.stopReadingCommandFile();
      }
    }
    else if( answer=="no option" ||
             answer=="uniform" ||
	     answer=="parabolic" ||
	     answer=="blasius" ||
             answer=="pressure" ||
	     answer=="jet" ||
	     answer=="user defined" )
    {
      
      int option=0;
      if( answer=="no option" )
      {
	bcOption1="";
      }
      else if( answer=="uniform" )
      {
	// bcOption1=", uniform(p=0.,u=0.,v=0.)";

	bcOption1=", uniform(";

        // we should popup a sibling dialog for parameters

      }
      else if( answer=="parabolic" )
      {
	bcOption1=", parabolic(d=.1,";
        // popup sibling dialog for parameters
      }
      else if( answer=="blasius" )
      {
	bcOption1=", blasius(R=1000.,";
      }
      else if( answer=="jet" )
      {
	bcOption1=", jet(d=.1,r=1.,x=0.,y=0,z=0.,";
      }
      else if( answer=="pressure" )
      {
	bcOption1=", pressure(.1*p+1.*p.n=0.)";
      }
      else if( answer=="user defined" )
      {
	bcOption1=", userDefinedBoundaryData";
      }
      else
      {
        Overture::abort("error");
      }
      if( answer=="uniform" || answer=="parabolic" || answer=="blasius" || answer=="jet" )
      {  // complete the option string
	for( int n=0; n<numberOfComponents; n++ )
	{
	  bcOption1+=sPrintF(buff,"%s=%g",(const char *)componentName[n],defaultState(n));
          if( n<numberOfComponents-1 ) bcOption1+=",";
	}
        bcOption1+=")";
   
      }
      // reset the option2
      bcOption2=""; 
      dialog.getOptionMenu("option2:").setCurrentChoice("no option2");
      
      dialog.setTextLabel("bc command",(const char*)getCommand());
      dialog.getOptionMenu("option:").setCurrentChoice(answer);

    }
    else if( answer=="no option2" ||
	     answer=="oscillate" ||
             answer=="ramp" )
    {
      if( answer=="no option2" )
      {
	bcOption2="";
      }
      else if( answer=="oscillate" )
      {
        bcOption2=", oscillate(t0=.5,omega=1.,a0=0.,a1=1.,u0=0,v0=0.,w0=0.)";
      }
      else if( answer=="ramp" )
      {
        bcOption2=", ramp(ta=0.,tb=1.,ua=0.,ub=1.,va=0.,vb=1.)";
      }
      else
      {
	Overture::abort("error");
      }
      dialog.setTextLabel("bc command",(const char*)getCommand());
      dialog.getOptionMenu("option2:").setCurrentChoice(answer);
    }
    else if( answer=="polynomial extrapolation" || 
             answer=="limited extrapolation" )
    {
      extrapolationOption = answer=="polynomial extrapolation" ? BoundaryConditionParameters::polynomialExtrapolation :
	BoundaryConditionParameters::extrapolateWithLimiter;
      dialog.getOptionMenu("extrap option:").setCurrentChoice((int)extrapolationOption);
      printF(" *** setting extrapolationOption to [%s]\n",(const char*)answer);
    }
    else if( answer=="apply bc command"  )
    {
      aString bcCommand = getCommand();
      gi.outputToCommandFile("* "+bcCommand+"\n");  // output current command as a comment
      setBoundaryConditionValues(bcCommand,originalBoundaryCondition,cg,*this);
 
    }
    else if( len=answer.matches("bc command") )
    {
      aString bcCommand=answer(len+1,answer.length()-1);
      setBoundaryConditionValues(bcCommand,originalBoundaryCondition,cg,*this);
 
    }
    else if( len=answer.matches("default state") )
    {
      answer=answer(len,answer.length()-1);
      inputParameterValues(answer,"default state",defaultState);

      aString state="";
      for( int n=0; n<numberOfComponents; n++ )
      {
	state+=sPrintF(buff,"%s=%g",(const char *)componentName[n],defaultState(n));
	if( n<numberOfComponents-1 ) state+=",";
      }
      dialog.setTextLabel("default state",(const char*)state);
    }
    
    else if( answer=="help" )
    {
      printF("\n"
             " ================================================================================================\n"
             " The objective of this menu is to build one or more `bc command' text strings\n"
             " Each `bc command' text string will assign boundary conditions to one or more boundaries\n"
             " Steps:\n" 
             "   1. Choose a `boundary:' or choose a `bc number' to assign.\n"
             "   2. Choose a 'bc:' such a noSlipWall, inflow etc.\n"
             "   3. If desired choose a primary option from `option:' (see below for more info on options) \n"
             "   4. If desired choose a modifier option from `option2'\n"
             "   5. Choose `apply bc command' to apply the bc or optionally edit \n"
             "      the `bc command string' to change values. After changing the string \n"
             "      hit return to apply the boundary condition.\n"
             );

      printF("The `bc command' text string is of the form  \n"
	     "                                                                             \n"
	     "       <grid name>(side,axis)=<boundary condition name> [,option] [,option] ...\n"
	     "                                                                             \n"
	     " to change the boundary condition on a given side of a given grid, or        \n"
	     "                                                                             \n"
	     "       <grid name>=<boundary condition name>  [,option] [,option] ...          \n"
	     "                                                                             \n"
	     " to change all boundaries on a given grid, or                                \n"
	     "                                                                             \n"
	     "       bcNumber<num>=<boundary condition name>  [,option] [,option] ... \n"
	     "                                                                             \n"
	     " to change all boundaries that *originally* had a boundary condition value of `num'\n"
	     " Here \n"
	     "   <grid name> is the name of the grid, side=0,1 and axis=0,1,2.  <grid name> can also be `all'.\n"
	     "                                                                             \n"
	     " The optional arguments specify data for the boundary conditions:            \n"
	     "    option = `uniform(p=1.,u=1.,...)'       : to specify a uniform inflow profile   \n"
	     "    option = `parabolic(d=2,p=1.,...)'      : to specify a parabolic inflow profile \n"
	     "    option = `blasius(p=1.,u=1.,R=1000)'  : to specify a laminar boundary layer profile \n"
	     "    option = `pressure(.1*p+1.*p.n=0.)'     : pressure boundary condition at outflow\n"
	     "    option = `oscillate(t0=.5,omega=1.,a0=0.,a1=1.,u0=0,v0=0.,w0=0.)'  : oscillating values    \n"
	     "    option = `ramp(ta=0.,tb=1.,ua=0.,ub=1.,va=0.,..)  : ramped inflow parameters    \n" 
	     "    option = `jet(d=.1,r=1.,x=0.,y=0,z=0.,p=1.,...)': to specify a parabolic jet inflow profile \n" 
	     "    option = `userDefinedBoundaryData : choose user defined data  \n"
	     " The pressure outflow boundary condition is a*p + b*p.n = c \n"
	     "  where p=pressure, p.n=normal derivative of p.\n"
	     "  Note that a and b should have the same sign or else the condition is unstable\n"
	     "                                                                               \n"
	     " The oscillate option is used with `uniform' or `parabolic' and takes the form:\n" 
	     "         [u0,v0,w0] + { a0+a1*cos((t-t0)*(2*pi*omega)) }*[ uniform or parabolic profile ]  \n"
	     "                                                                             \n"
	     "The userDefinedBoundaryData option calls the function userDefinedBoundaryValues that\n"
	     "  may be changed to suit your purposes. \n"
             " \n"
	     " Examples: \n"
	     "    square(0,0)=inflowWithVelocityGiven , uniform(p=1.,u=1.)     \n"
	     "    annulus=noSlipWall                         \n"
	     "    all=slipWall                               \n"
	     "    bcNumber2=slipWall                         \n"
	     "    square(0,1)=outflow , pressure(.1*p+1.*p.n=0.) \n"
             "    bcNumber2=inflowWithVelocityGiven, parabolic(d=1.,p=1.,u=1.) , ramp(ta=0.,tb=10.,ua=0.0,ub=1.)\n"
             "    square(0,0)=inflowWithVelocityGiven , parabolic(d=.2,p=1.,u=1.), oscillate(t0=.3,omega=2.5)\n"
             " ================================================================================================\n"
	);
      
    }
    else if( len=answer.matches("order of extrap for interp neighbours") )
    {
      sScanF(answer(len,answer.length()-1),"%i",&orderOfExtrapolationForInterpolationNeighbours);
      orderOfExtrapolationForInterpolationNeighbours=max(1,min(10,orderOfExtrapolationForInterpolationNeighbours));
      cout << " orderOfExtrapolationForInterpolationNeighbours=" << 
                orderOfExtrapolationForInterpolationNeighbours << endl;
      printF(" order=1 is constant extrapolation, order=2 is linear, order=3 is quadratic etc. \n");
      
      dialog.setTextLabel(0,sPrintF(line,"%i", orderOfExtrapolationForInterpolationNeighbours)); 
    }
    else if( len=answer.matches("order of extrap for 2nd ghost line") )
    {
      sScanF(answer(len,answer.length()-1),"%i",&orderOfExtrapolationForSecondGhostLine);
      orderOfExtrapolationForSecondGhostLine=max(1,min(10,orderOfExtrapolationForSecondGhostLine));
      cout << " orderOfExtrapolationForSecondGhostLine=" << 
                orderOfExtrapolationForSecondGhostLine << endl;
      printF(" order=1 is constant extrapolation, order=2 is linear, order=3 is quadratic etc. \n");
      dialog.setTextLabel(1,sPrintF(line,"%i", orderOfExtrapolationForSecondGhostLine)); 
    }
    else
    {
      if( executeCommand )
      {
	returnValue= 1;  // when executing a single command, return 1 if the command was not recognised.
        break;
      }
      else
      {
	cout << "Unknown response=[" << answer << "]\n";
	gi.stopReadingCommandFile();
      }
       
    }

  }
//  delete [] pdeParametersMenu;

//  updatePDEparameters();  // update parameters such as ReynoldsNumber, MachNumber, ... to be consistent.


  if( !executeCommand  )
  {
    if( !squaresWerePlotted ) gi.setPlotTheColouredSquares(false);
    gi.popGUI();
    gi.unAppendTheDefaultPrompt();
  }

 return returnValue;
}


