#include "DomainSolver.h"
#include "GenericGraphicsInterface.h"

// ===================================================================================================================
/// \brief Build the dialog that shows the various options for time stepping.
/// \param dialog (input) : graphics dialog to use.
///
/// <b> Dialog Menu Options: </b>\n
///  - method
///    - the list time stepping methods depends on the PDE being solved. 
///  - accuracy
///    - second order accurate : solve the problem to 2nd order accuracy in space.
///    - fourth order accurate : solve the problem to 4th order accuracy in space.
///  - time accuracy 
///    - solve for steady state :
///    - second order accurate in time : 
///    - fourth order accurate in time :
///  - choose grids for implicit : choose a list of grids that should be treated implicitly.
///  - use local time stepping (toggle) : 
// ==================================================================================================================
int DomainSolver::
buildTimeSteppingDialog(DialogData & dialog )
{
  
  dialog.setWindowTitle("Time Stepping Parameters");
  dialog.setExitCommand("close time stepping", "close");
  dialog.setOptionMenuColumns(1);

  const int maxNumberOfTimeSteppingMethods=Parameters::numberOfTimeSteppingMethods+10;
  aString *cmd = new aString [maxNumberOfTimeSteppingMethods];

  int nt=0;

  assert( nt<maxNumberOfTimeSteppingMethods );
  cmd[nt]="";

  if( nt>0 )
    dialog.addOptionMenu("method", cmd, cmd, (int)parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod"));

  delete [] cmd;
    
  aString accuracyCommands[] =  { "second order accurate",
				  "fourth order accurate",
				  ""     };
      
  dialog.addOptionMenu("accuracy", accuracyCommands, accuracyCommands,
				      parameters.dbase.get<int >("orderOfAccuracy")==2 ? 0 : 1);

  aString timeAccuracyCommands[] =  { "solve for steady state",
				      "second order accurate in time",
				      "fourth order accurate in time",
				      ""     };
      
  dialog.addOptionMenu("time accuracy", timeAccuracyCommands, timeAccuracyCommands,
				      parameters.dbase.get<int >("orderOfTimeAccuracy")==2 ? 1 : 
				      parameters.dbase.get<int >("orderOfTimeAccuracy")==4 ? 2 : 0);

  aString predictorOrderCommands[] =  { "default order predictor"
                                        "first order predictor",
					"second order predictor",
					"third order predictor",
					"fourth order predictor",
					 ""     };
      
  dialog.addOptionMenu("predictor order", predictorOrderCommands, predictorOrderCommands,
		       parameters.dbase.get<int>("predictorOrder"));

  // push buttons
  aString pbCommands[] = {"choose grids for implicit",
			  ""};

  const int numRows=1;
  dialog.setPushButtons( pbCommands, pbCommands, numRows ); 


// old way:
//   if( parameters.dbase.get<Parameters::PDE >("pde")==Parameters::incompressibleNavierStokes )
//   {
//     // build a menu to indicate which grids are solved implicitly

//     aString *cmd = new aString[cg.numberOfComponentGrids()+3];
//     aString *label = new aString[cg.numberOfComponentGrids()+3];
//     int *initialState = new int [cg.numberOfComponentGrids()+3];
//     int i;
//     for( i=0; i<cg.numberOfComponentGrids(); i++ )
//     {
//       cmd[i]=cg[i].getName();
//       label[i]="implicit grid: "+cmd[i];
//       initialState[i]=0;       // **** fix ******
//     }
//     i=cg.numberOfComponentGrids();
//     cmd[i]="all"; label[i]="implicit grid: "+cmd[i]; initialState[i]=0; i++;
//     cmd[i]="none"; label[i]="implicit grid: "+cmd[i]; initialState[i]=0; i++;
//     cmd[i]=""; label[i]="";
//     dialog.addPulldownMenu("implicit grids", label, cmd, GI_TOGGLEBUTTON,initialState);
//     delete [] cmd;
//     delete [] label;
//     delete [] initialState;
//   }

  const int maxNumberOfToggleButtons=3+1;
  aString tbCommands[maxNumberOfToggleButtons]; 
  int tbState[maxNumberOfToggleButtons];
  int ntb=0;
  if( parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")==Parameters::steadyStateRungeKutta )
  {
    tbCommands[ntb]="use local time stepping";
    tbState[ntb] = parameters.dbase.get<int >("useLocalTimeStepping");
    ntb++;
  }
    
  tbCommands[ntb]="adjust dt for moving bodies";
  tbState[ntb]=parameters.dbase.get<bool >("adjustTimeStepForMovingBodies");
  ntb++;

  tbCommands[ntb]="use full implicit system";
  tbState[ntb]=parameters.dbase.get<bool >("useFullSystemForImplicitTimeStepping");
  ntb++;

  tbCommands[ntb]="apply explicit BCs to implicit grids";
  tbState[ntb]=parameters.dbase.get<bool >("applyExplicitBCsToImplicitGrids");
  ntb++;

  tbCommands[ntb]="";  // null termination string
  assert( ntb<maxNumberOfToggleButtons );

  const int numColumns=1;
  dialog.setToggleButtons(tbCommands, tbCommands, tbState, numColumns);

  // ----- Text strings ------
  const int numberOfTextStrings=20;
  aString textCommands[numberOfTextStrings];
  aString textLabels[numberOfTextStrings];
  aString textStrings[numberOfTextStrings];

  nt=0;
    
  // we may not know whether this is a steady state solver at this point
//   if( !parameters.isSteadyStateSolver() )
//   {
    textCommands[nt] = "final time";  textLabels[nt]=textCommands[nt];
    sPrintF(textStrings[nt], "%g", parameters.dbase.get<real >("tFinal"));  nt++; 
//   }
//   else
//   {
    textCommands[nt] = "max iterations";  textLabels[nt]=textCommands[nt];
    sPrintF(textStrings[nt], "%i", parameters.dbase.get<int >("maxIterations"));  nt++; 
//  }
    
  textCommands[nt] = "cfl"; textLabels[nt]=textCommands[nt];
  sPrintF(textStrings[nt], "%5.3f", parameters.dbase.get<real >("cfl")); nt++; 

  textCommands[nt] = "dtMax"; textLabels[nt]=textCommands[nt];
  sPrintF(textStrings[nt], "%9.3e", parameters.dbase.get<real >("dtMax")); nt++; 

  textCommands[nt] = "implicit factor"; textLabels[nt]=textCommands[nt];
  sPrintF(textStrings[nt], "%9.3e (1=BE,0=FE)", parameters.dbase.get<real >("implicitFactor")); nt++; 

  textCommands[nt] = "recompute dt every"; textLabels[nt]=textCommands[nt];
  sPrintF(textStrings[nt], "%i steps", parameters.dbase.get<int >("maximumStepsBetweenComputingDt")); nt++; 
    
  textCommands[nt] = "slow start"; textLabels[nt]=textCommands[nt];
  sPrintF(textStrings[nt], "%g (seconds)", parameters.dbase.get<real >("slowStartTime")); nt++; 

  textCommands[nt] = "slow start cfl"; textLabels[nt]=textCommands[nt];
  sPrintF(textStrings[nt], "%g", parameters.dbase.get<real >("slowStartCFL")); nt++; 

  textCommands[nt] = "fixup unused frequency"; textLabels[nt]=textCommands[nt];
  sPrintF(textStrings[nt], "%i", parameters.dbase.get<int >("fixupFrequency")); nt++; 

  textCommands[nt] = "cflMin, cflMax"; textLabels[nt]=textCommands[nt];
  sPrintF(textStrings[nt], "%g, %g", parameters.dbase.get<real >("cflMin"),parameters.dbase.get<real >("cflMax")); nt++; 

  textCommands[nt] = "preconditioner frequency"; textLabels[nt] =textCommands[nt];
  sPrintF(textStrings[nt],"%i",parameters.dbase.get<int>("preconditionerFrequency")); nt++;

  textCommands[nt] = "number of PC corrections"; textLabels[nt] =textCommands[nt];
  sPrintF(textStrings[nt],"%i",parameters.dbase.get<int>("numberOfPCcorrections")); nt++;


  // null strings terminal list
  textCommands[nt]="";   textLabels[nt]="";   textStrings[nt]="";  assert( nt<numberOfTextStrings );
  dialog.setTextBoxes(textCommands, textLabels, textStrings);

  return 0;

}


//\begin{>>DomainSolverInclude.tex}{\subsection{getTimeSteppingOption}}   
int DomainSolver::
getTimeSteppingOption(const aString & answer,
		      DialogData & dialog )  
//================================================================================
// /Description:
//    Look for time stepping options in the string "answer"
//
// /answer (input) : check this command for a change to the time stepping options
//
// /Return value: return 1 if the command was found, 0 otherwise.
//\end{DomainSolverInclude.tex}  
//====================================================================
{
  assert( parameters.dbase.get<GenericGraphicsInterface* >("ps") !=NULL );
  GenericGraphicsInterface & gi = *parameters.dbase.get<GenericGraphicsInterface* >("ps");
  GraphicsParameters & psp = parameters.dbase.get<GraphicsParameters >("psp");

  int found=true; 
  char buff[180];
  aString answer2;
  int len=0;

  if( answer=="cfl" || answer=="cfl (cfl=)" ) // old way 
  {
    gi.inputString(answer2,sPrintF(buff,"Enter cfl (default value=%e)",parameters.dbase.get<real >("cfl")));
    if( answer2!="" )
      sScanF(answer2,"%e",&parameters.dbase.get<real >("cfl"));
    printF(" cfl=%9.3e\n",parameters.dbase.get<real >("cfl"));
    parameters.dbase.get<real >("cflMin")=parameters.dbase.get<real >("cfl")*.9;
    parameters.dbase.get<real >("cflMax")=min(max(parameters.dbase.get<real >("cfl"),.95),parameters.dbase.get<real >("cfl")*1.1);
  }
  else if( len=answer.matches("cflMin, cflMax") )
  {
    printF("INFO: cflMin and cflMax are used to determine when the time step for implicit\n"
	   "      time stepping should be changed. Decrease cflMin and increase cflMax to\n"
	   "      reduce the frequency of refactors -- but don't make cflMax too large.\n");
      
    sScanF(answer(len,answer.length()),"%e %e",&parameters.dbase.get<real >("cflMin"),&parameters.dbase.get<real >("cflMax"));
    printF(" cflMin=%g cflMax=%g \n",parameters.dbase.get<real >("cflMin"),parameters.dbase.get<real >("cflMax"));
    dialog.setTextLabel("cflMin, cflMax",
				       sPrintF(answer2,"%g, %g",parameters.dbase.get<real >("cflMin"),parameters.dbase.get<real >("cflMax")));
  }
  else if( answer.matches("cfl") )
  {
    sScanF(answer,"cfl %e",&parameters.dbase.get<real >("cfl"));
    printF(" cfl=%9.3e\n",parameters.dbase.get<real >("cfl"));
    parameters.dbase.get<real >("cflMin")=parameters.dbase.get<real >("cfl")*.9;
    parameters.dbase.get<real >("cflMax")=min(max(parameters.dbase.get<real >("cfl"),.95),parameters.dbase.get<real >("cfl")*1.1);
    dialog.setTextLabel("cfl",sPrintF(answer2,"%g", parameters.dbase.get<real >("cfl")));
    dialog.setTextLabel("cflMin, cflMax",
				       sPrintF(answer2,"%g, %g",parameters.dbase.get<real >("cflMin"),parameters.dbase.get<real >("cflMax")));
  }
  else if( answer=="forwardEuler" || answer=="forward Euler" )
  {
    parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")=Parameters::forwardEuler;
    dialog.getOptionMenu("method").setCurrentChoice("forward Euler");
  }
  else if( answer=="adamsBashforth2" || answer=="adams order 2" )
  {
    parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")=Parameters::adamsBashforth2;
    parameters.dbase.get<Parameters::ImplicitMethod >("implicitMethod")=Parameters::notImplicit;
    dialog.getOptionMenu("method").setCurrentChoice("adams order 2");
  }
  else if( answer=="adamsPredictorCorrector2" || answer=="adams PC")
  {
    parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")=Parameters::adamsPredictorCorrector2;
    parameters.dbase.get<int >("orderOfPredictorCorrector")=2;
    parameters.dbase.get<Parameters::ImplicitMethod >("implicitMethod")=Parameters::notImplicit;
    dialog.getOptionMenu("method").setCurrentChoice("adams PC");
  }
  else if( answer=="adamsPredictorCorrector4" || answer=="adams PC order 4")
  {
    parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")=Parameters::adamsPredictorCorrector4;
    parameters.dbase.get<int >("orderOfPredictorCorrector")=4;
    parameters.dbase.get<Parameters::ImplicitMethod >("implicitMethod")=Parameters::notImplicit;
    dialog.getOptionMenu("method").setCurrentChoice("adams PC order 4");
  }
  else if( answer=="variable time step PC" )
  {
    parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")=Parameters::variableTimeStepAdamsPredictorCorrector;
    parameters.dbase.get<int >("orderOfPredictorCorrector")=2;
    parameters.dbase.get<Parameters::ImplicitMethod >("implicitMethod")=Parameters::notImplicit;
    dialog.getOptionMenu("method").setCurrentChoice(answer);
  }
  else if( answer=="midpoint" )
  {
    parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")=Parameters::midPoint;
    parameters.dbase.get<Parameters::ImplicitMethod >("implicitMethod")=Parameters::notImplicit;
    dialog.getOptionMenu("method").setCurrentChoice(answer);
  }
  else if( answer.matches("steady state RK") )
  {
    parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")=Parameters::steadyStateRungeKutta;
    parameters.dbase.get<Parameters::ImplicitMethod >("implicitMethod")=Parameters::notImplicit;
    parameters.dbase.get<int >("useLocalTimeStepping")=true;
    if( answer=="steady state RK-line" )
    {
      parameters.dbase.get<Parameters::ImplicitMethod >("implicitMethod")=Parameters::lineImplicit;
      parameters.setGridIsImplicit();
    }
      
    dialog.getOptionMenu("method").setCurrentChoice(answer);
    dialog.setToggleState("use local time stepping",parameters.dbase.get<int >("useLocalTimeStepping"));  
  }
  else if( answer.matches("steady state Newton") )
  {
    parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")=Parameters::steadyStateNewton;
    parameters.dbase.get<Parameters::ImplicitMethod >("implicitMethod")=Parameters::backwardEuler;
    parameters.dbase.get<int >("refactorFrequency")=1;
    parameters.dbase.get<int >("useLocalTimeStepping")=true;
      
    dialog.getOptionMenu("method").setCurrentChoice(answer);
    dialog.setToggleState("use local time stepping",parameters.dbase.get<int >("useLocalTimeStepping"));  
  }
  else if( dialog.getToggleValue(answer,"use full implicit system",
                                 parameters.dbase.get<bool >("useFullSystemForImplicitTimeStepping")) ){ }
  else if( dialog.getToggleValue(answer,"apply explicit BCs to implicit grids",
                                 parameters.dbase.get<bool >("applyExplicitBCsToImplicitGrids")) ){ }
  // *** scLC
  else if( answer=="Runge-Kutta" )
  {
    parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")=Parameters::rKutta;
    parameters.dbase.get<Parameters::ImplicitMethod >("implicitMethod")=Parameters::notImplicit;
    parameters.dbase.get<IntegerArray >("timeStepType") = 0; // Explicit by default.
    dialog.getOptionMenu("method").setCurrentChoice(answer);
  }
  // *** ecLC
  else if( answer=="second order accurate" )
  {
    parameters.dbase.get<int >("orderOfAccuracy")=2;
    dialog.getOptionMenu("accuracy").setCurrentChoice(parameters.dbase.get<int >("orderOfAccuracy")==2 ? 0 : 1);
  }
  else if( answer=="fourth order accurate" )
  {
    MappedGrid & mg=cg[0];
    if( mg.discretizationWidth(0)<5 )
    {
      printf("ERROR: This grid does not have a discretization width of at least 5. Current width=%i\n",
	     "   The order of accuracy will not be changed\n", mg.discretizationWidth(0));
    }
    else
    {
      parameters.dbase.get<int >("orderOfAccuracy")=4;
      dialog.getOptionMenu("accuracy").setCurrentChoice(parameters.dbase.get<int >("orderOfAccuracy")==2 ? 0 : 1);
    }
  }
  else if( answer=="second order accurate in time" || 
	   answer=="fourth order accurate in time" ||
	   answer=="solve for steady state" )
  {
    parameters.dbase.get<int >("orderOfTimeAccuracy")= answer=="second order accurate in time" ? 2 : 
      answer== "fourth order accurate in time" ? 4 : 0;
    dialog.getOptionMenu("time accuracy").setCurrentChoice(parameters.dbase.get<int >("orderOfAccuracy")==0 ? 0 :
									  parameters.dbase.get<int >("orderOfAccuracy")==2 ? 1 : 2);
  }
  else if( answer=="implicit" )
  {
    parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")=Parameters::implicit;
    parameters.dbase.get<int >("orderOfPredictorCorrector")=2;

    parameters.dbase.get<Parameters::ImplicitMethod >("implicitMethod")=Parameters::crankNicolson;
    parameters.setGridIsImplicit();  // by default all grids are implicit for the implicit time stepping method
  }
  else if( answer=="all speed implicit" )
    parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")=Parameters::implicitAllSpeed;
  else if( answer=="linearized all speed implicit" )
  {
    parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")=Parameters::implicitAllSpeed;
    parameters.dbase.get<int >("linearizeImplicitMethod")=TRUE;
  }
  else if( answer=="implicit factor" )
  {
    gi.inputString(answer2,sPrintF(buff,"Enter the implicit factor .5=CN, 1=BE (default value=%e)",
				   parameters.dbase.get<real >("implicitFactor")));
    if( answer2!="" )
      sScanF(answer2,"%e",&parameters.dbase.get<real >("implicitFactor"));
    printF(" implicitFactor=%9.3e\n",parameters.dbase.get<real >("implicitFactor"));
  }
  else if( answer.matches("implicit factor") )
  {
    sScanF(answer,"implicit factor %e",&parameters.dbase.get<real >("implicitFactor")); 
    printF(" implicit factor=%9.3e\n",parameters.dbase.get<real >("implicitFactor"));
    dialog.setTextLabel("implicit factor",
				       sPrintF(answer2,"%9.3e (1=BE,0=FE)", parameters.dbase.get<real >("implicitFactor")));
  }
  else if( answer=="slow start time interval" )
  {
    gi.inputString(answer2,sPrintF(buff,"Enter the slow start time interval (default value=%e)",
				   parameters.dbase.get<real >("slowStartTime")));
    if( answer2!="" )
      sScanF(answer2,"%e",&parameters.dbase.get<real >("slowStartTime"));
    printF(" slowStartTime=%9.3e\n",parameters.dbase.get<real >("slowStartTime"));
  }
  else if( answer=="slow start cfl" )
  {
    gi.inputString(answer2,sPrintF(buff,"Enter the slow start initial cfl value (default value=%e)",
				   parameters.dbase.get<real >("slowStartCFL")));
    if( answer2!="" )
      sScanF(answer2,"%e",&parameters.dbase.get<real >("slowStartCFL"));
    printF(" slowStartCFL=%9.3e\n",parameters.dbase.get<real >("slowStartCFL"));
  }
  else if( answer.matches("slow start cfl") )
  {
    sScanF(answer,"slow start cfl %e",&parameters.dbase.get<real >("slowStartCFL"));
    printF(" The slow start initial cfl value=%9.3e\n",parameters.dbase.get<real >("slowStartCFL"));
    dialog.setTextLabel("slow start cfl",sPrintF(answer2,"%g", parameters.dbase.get<real >("slowStartCFL")));
  }
  else if( answer.matches("slow start") )
  {
    sScanF(answer,"slow start %e",&parameters.dbase.get<real >("slowStartTime"));
    printF(" The slow start time interval=%9.3e\n",parameters.dbase.get<real >("slowStartTime"));
    dialog.setTextLabel("slow start",sPrintF(answer2,"%g (seconds)", parameters.dbase.get<real >("slowStartTime")));
  }
  else if( dialog.getToggleValue(answer,"use local time stepping",parameters.dbase.get<int >("useLocalTimeStepping")) ){ }
  else if( dialog.getToggleValue(answer,"adjust dt for moving bodies",
						parameters.dbase.get<bool >("adjustTimeStepForMovingBodies")) ){}//
  else if( answer.matches("recompute dt every") )
  {
    printF("The time step, dt,  is recomputed every time the solution is plotted/saved \n");
    printF("In addition you may specify the maximum number of steps that will be taken \n");
    printF("before dt is recomputed. Use this if the solution is not plotted very often. \n");
    sScanF(answer,"recompute dt every %i",&parameters.dbase.get<int >("maximumStepsBetweenComputingDt"));
    printF(" recompute dt=%i\n",parameters.dbase.get<int >("maximumStepsBetweenComputingDt"));
    dialog.setTextLabel("recompute dt every",
				       sPrintF(answer2,"%i steps", parameters.dbase.get<int >("maximumStepsBetweenComputingDt"))); 
  }
  else if( answer=="recompute dt interval" ) // old way
  {
    printF("The time step, dt,  is recomputed every time the solution is plotted/saved \n");
    printF("In addition you may specify the maximum number of steps that will be taken \n");
    printF("before dt is recomputed. Use this if the solution is not plotted very often. \n");
    gi.inputString(answer2,sPrintF(buff,"Enter the max number of steps between recomputing dt"
				   "(current=%i)",parameters.dbase.get<int >("maximumStepsBetweenComputingDt")));
    if( answer2!="" )
      sScanF(answer2,"%i",&parameters.dbase.get<int >("maximumStepsBetweenComputingDt"));
    if( parameters.dbase.get<int >("maximumStepsBetweenComputingDt")>0 )
      printF(" maximumStepsBetweenComputingDt=%i\n",parameters.dbase.get<int >("maximumStepsBetweenComputingDt"));
    else
    {
      printF(" the answer must be a positive integer\n");
      parameters.dbase.get<int >("maximumStepsBetweenComputingDt")=100;
    }
  }
  else if( len=answer.matches("fixup unused frequency") )
  {
    sScanF(answer(len,answer.length()-1),"%i",&parameters.dbase.get<int >("fixupFrequency"));
    printF("Unused points on the grid can sometimes get large values in them.\n"
	   "The `fixup unused frequency' specifies how often to reset the unused points\n");
    printF("fixupFrequency=%i\n",parameters.dbase.get<int >("fixupFrequency"));
    dialog.setTextLabel("fixup unused frequency",sPrintF("%i",parameters.dbase.get<int >("fixupFrequency")));
  }
  else if( len=answer.matches("max iterations") )
  {
    sScanF(answer(len,answer.length()-1),"%i",&parameters.dbase.get<int >("maxIterations"));
    dialog.setTextLabel("max iterations",sPrintF("%i",parameters.dbase.get<int >("maxIterations")));      
  }
  else if( len=answer.matches("plot iterations") )
  {
    sScanF(answer(len,answer.length()-1),"%i",&parameters.dbase.get<int >("plotIterations"));
    dialog.setToggleState("plot iterations",parameters.dbase.get<int >("plotIterations"));      
  }
  else if( answer=="dtMax" )
  {
    gi.inputString(answer2,sPrintF(buff,"Enter dtMax (default value=%e)",parameters.dbase.get<real >("dtMax")));
    if( answer2!="" )
      sScanF(answer2,"%e",&parameters.dbase.get<real >("dtMax"));
    printF(" dtMax=%9.3e\n",parameters.dbase.get<real >("dtMax"));
  }
  else if( answer(0,4)=="dtMax" )
  {
    sScanF(answer,"dtMax %e",&parameters.dbase.get<real >("dtMax")); 
    printF(" dtMax=%9.3e\n",parameters.dbase.get<real >("dtMax"));
    dialog.setTextLabel("dtMax",sPrintF(answer2,"%9.3e", parameters.dbase.get<real >("dtMax")));
  }
  else if( answer=="final time (tf=)" ) // for backward compatibility
  {
    gi.inputString(answer2,sPrintF(buff,"Enter the final time (default value=%e)",parameters.dbase.get<real >("tFinal")));
    if( answer2!="" )
      sScanF(answer2,"%e",&parameters.dbase.get<real >("tFinal"));
    printF(" parameters.tFinal=%9.3e\n",parameters.dbase.get<real >("tFinal"));
  }
  else if( answer.matches("final time") )
  {
    sScanF(answer,"final time %e",&parameters.dbase.get<real >("tFinal"));
    printF(" tFinal=%9.3e\n",parameters.dbase.get<real >("tFinal"));
    dialog.setTextLabel("final time",sPrintF(answer2,"%9.3e", parameters.dbase.get<real >("tFinal")));   // ***** 0 may not be the correct position
  }
  else if ( dialog.getTextValue(answer,"preconditioner frequency", "%i", 
                                        parameters.dbase.get<int>("preconditionerFrequency"))){}
  else if ( dialog.getTextValue(answer,"number of PC corrections", "%i", 
                                parameters.dbase.get<int>("numberOfPCcorrections"))){ }
  else if( answer=="default order predictor" ||
           answer=="first order predictor" || 
           answer=="second order predictor" ||
           answer=="third order predictor" ||
           answer=="fourth order predictor" )
  {
    parameters.dbase.get<int>("predictorOrder") = 
      answer=="default order predictor" ? 0 : 
      answer=="first order predictor" ? 1 : 
      answer=="second order predictor" ? 2 :
      answer=="third order predictor" ? 3 : 4;
  }
  else if( answer=="choose grids for implicit" )
  {
//\begin{>>setParametersInclude.tex}{\subsubsection{Choosing grids for implicit time stepping.}\label{sec:implicitMenu}}
//\no function header:
//
//  When the option {\tt `choose grids for implicit'} is chosen from the main parameter menu one can
//  specify which grids should be treated implicitly or explicitly with the {\bf implicit} time stepping
//  option. Type a line of the form
//  \begin{verbatim}
//           <grid name>=[explicit][implicit] 
//  \end{verbatim}
//  where {\tt <grid name>} is the name of a grid or {\tt `all'}. Type {\tt `help'} to see the names.
//  Examples:
//  \begin{verbatim}
//       square=explicit
//       all=implicit
//       cylinder=implicit
//  \end{verbatim}
//  Type {\tt `done'} when finished.
//\end{setParametersInclude.tex}
    // *** scLC
    if( (parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")!=Parameters::implicit) &&  
	(parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")!=Parameters::rKutta) )
    {
      printf("WARNING: This option only currently only works for timeSteppingMethod==implicit or timeSteppingMethod==rKutta! \n");
      parameters.setGridIsImplicit(-1,0);
      parameters.dbase.get<IntegerArray >("timeStepType")=0;
    }
    printF("Set grids to be implicit, semi-implicit or explicit. Examples: \n"
	   "  square=explicit \n"
	   "  annulus=semi \n"
	   "  all=implicit \n" );

    gi.appendToTheDefaultPrompt("implicit>");
    aString gridName;
    int implicit;// implicit used to be a bool.
    for( ;; )
    {
      gi.inputString(answer2,"Specify grids to be implicit, semi-implicit or explicit. (or type `help' or `done')");
      if( answer2=="done" || answer2=="exit" )
	break;
      else if( answer2=="help" )
      {
	printF("Specify grids to be implicit, semi-implicit or explicit. Type a string of the form     \n"
	       "                                                                             \n"
	       "       <grid name>=[explicit][semi][implicit]\n"
	       "                                                                             \n"
	       " By default all grids are implicit.                                          \n"
	       " Examples: \n"
	       "     square=explicit                            \n"
	       "     annulus=semi                            \n"
	       "     all=implicit                               \n"
	  );
	// *** ecLC
	printF("Here are the names of the grids: \n");
	for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
	  printF(" grid %i : name=%s \n",grid,(const char*)cg[grid].mapping().getName(Mapping::mappingName));
      }
      else
      {
	int length=answer2.length();
	int i,mark=-1;
	for( i=0; i<length; i++ )
	{
	  if( answer2[i]=='=' )
	  {
	    mark=i-1;
	    break;
	  }
	}
	if( mark<0 )
	{
	  printF("unknown form of answer=[%s]. Try again or type `help' for examples.\n",(const char *)answer2);
	  gi.stopReadingCommandFile();
	  continue;
	}
	else
	{
	  gridName=answer2(0,mark);  // this is the name of the grid or `all'
	  Range G(-1,-1);
	  if( gridName=="all" )
	    G=Range(0,cg.numberOfComponentGrids()-1);
	  else
	  {
	    // search for the name of the grid
	    for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
	    {
	      if( gridName==cg[grid].mapping().getName(Mapping::mappingName) )
	      {
		G=Range(grid,grid);
		break;
	      }
	    }
	  }
	  if( G.getBase()==-1  )
	  {
	    printF("Unknown grid name = <%s> \n",(const char *)gridName);
	    gi.stopReadingCommandFile();
	    continue;
	  }

// *** scLC
	  //This could probably be done in a better way! 
	  if( answer2(mark+2,mark+3)=="im" ){
	    implicit=1;
	  }
	  else if( answer2(mark+2,mark+3)=="se" ){
	    implicit=2;
	  }
	  else {
	    implicit=0;
	  }
		   
	  for( int grid=G.getBase(); grid<=G.getBound(); grid++ )
	  {
	    parameters.setGridIsImplicit(grid,implicit);
	    printF("Setting time stepping to be %s for grid %s\n",
		   (parameters.getGridIsImplicit(grid)==1 ? "implicit" : 
		    (parameters.getGridIsImplicit(grid)==2 ? "semi-implicit" : 
		     "explicit")),
		   (const char*)cg[grid].mapping().getName(Mapping::mappingName));
	  }
	  // *** ecLC
	}
      }
    }
    gi.unAppendTheDefaultPrompt();
  }
//   else if( answer(0,13)=="implicit grid:" )
//   {
//     // answer is of the form "implicit grid: gridName [0,1]"
//     aString number=answer(answer.length()-2,answer.length()-1);
//     int toggleState=0;
//     sScanF(number,"%i",&toggleState);
      
//     answer2=answer(0,answer.length()-3);  // strip off last number
//     aString gridName=answer2(15,answer2.length()-1);
//     printF("implicit grid: gridName=[%s]\n",(const char*)gridName);
//     PullDownMenu& pulldown = dialog.getPulldownMenu(0);


//     if( gridName=="all" || gridName=="none" )
//     {
//       pulldown.setToggleState(cg.numberOfComponentGrids(),0); // turn off "all" state
//       pulldown.setToggleState(cg.numberOfComponentGrids()+1,0); // turn off "none" state

//       toggleState = gridName=="all" ? 1 : 0;
//       for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
//       {
// 	parameters.setGridIsImplicit(grid,toggleState);
// 	pulldown.setToggleState(grid,toggleState);
//       }
//     }
//     else
//     {

//       for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
//       {
// 	if( gridName==cg[grid].getName() )
// 	{
// 	  parameters.setGridIsImplicit(grid,toggleState);
// 	  printF("Setting time stepping to be %s for grid %s\n",
// 		 (parameters.getGridIsImplicit(grid)==1 ? "implicit" : 
// 		  (parameters.getGridIsImplicit(grid)==2 ? "semi-implicit" : 
// 		   "explicit")),(const char*)cg[grid].getName());

// 	  pulldown.setToggleState(grid,toggleState);
// 	}
//       }
//     }
//   }
  else 
  {
    found =false;
  }


  return found;
}

