#include "DomainSolver.h"
#include "ShowFileReader.h"
#include "DataPointMapping.h"
#include "ShowFileReader.h"
#include "ParallelUtility.h"

namespace // make the following local to this file
{
// Here are the possible options for user defined forcings. Add new options to this enum.
enum UserDefinedForcingOptions
{
  noForcing,
  constantForcing,
  gaussianForcing
};
}



//\begin{>>DomainSolverInclude.tex}{\subsection{userDefinedForcing}}  
int DomainSolver:: 
userDefinedForcing( realMappedGridFunction& f, const realMappedGridFunction & u, int iparam[], real rparam[] )
//==============================================================================================
// /Description:
//   User defined forcing. This function is called to actually evaluate the user defined forcing
//   The function setupUserDefinedForcing is first 
//   called to assign the option and parameters. Rewrite or add new options to 
//   this function and to setupUserDefinedForcing to supply your own forcing option.
//
// /u (input) : current solution
// /f (input/output) : add to this forcing function
// /iparam[] (input) : holds some integer parameters
// /rparam[] (input) : holds some real parameters
// 
// /Notes:
//  \begin{itemize}
//    \item You must fill in the realMappedGridFunction f.
//    \item The `parameters' object holds many useful parameters.
//    \item The current time "t" is extracted below from rparam.  
//  \end{itemize}
//
// /Return values: 0=success, non-zero=failure.
//\end{DomainSolverInclude.tex}  
//==============================================================================================
{
  // Look for the userDefinedForcing sub-directory in the data-base
  if( !parameters.dbase.get<DataBase >("modelData").has_key("userDefinedForcingData") )
  {
    // if the directory is not there then assume that there is no user defined forcing
    return 0;
  }
  DataBase & db = parameters.dbase.get<DataBase >("modelData").get<DataBase>("userDefinedForcingData");

  UserDefinedForcingOptions & option= db.get<UserDefinedForcingOptions>("option");

  if( option==noForcing ) return 0;

  const real t =rparam[1];       // ** here is the current time ***  
  const int & grid = iparam[0];

  const int & numberOfComponents=parameters.dbase.get<int >("numberOfComponents");
  const int & numberOfDimensions=parameters.dbase.get<int >("numberOfDimensions");
  const int & rc = parameters.dbase.get<int >("rc");   //  density = u(all,all,all,rc)  (if appropriate for this PDE)
  const int & uc = parameters.dbase.get<int >("uc");   //  u velocity component =u(all,all,all,uc)
  const int & vc = parameters.dbase.get<int >("vc");  
  const int & wc = parameters.dbase.get<int >("wc");
  const int & tc = parameters.dbase.get<int >("tc");   //  temperature
  const int & sc = parameters.dbase.get<int >("sc");   //  mass fraction lambda
  const int & pc = parameters.dbase.get<int >("pc");
  
  MappedGrid & mg = *u.getMappedGrid();
  
  mg.update(MappedGrid::THEvertex | MappedGrid::THEcenter );  // make sure the vertex array has been created
  #ifndef USE_PPP
    realArray & vertex = mg.vertex();  // grid points
    const realArray & ug = u;
    realArray & fg = f;
  #else
    // In parallel, we operate on the arrays local to each processor
    realSerialArray vertex; getLocalArrayWithGhostBoundaries(mg.vertex(),vertex);
    realSerialArray ug;     getLocalArrayWithGhostBoundaries(u,ug);
    realSerialArray fg;     getLocalArrayWithGhostBoundaries(f,fg);
  #endif

  Index I1,I2,I3;
  getIndex( mg.dimension(),I1,I2,I3 );          // all points including ghost points.
  // getIndex( mg.gridIndexRange(),I1,I2,I3 );  // boundary plus interior points.
  #ifdef USE_PPP
    // restrict bounds to local processor, include ghost
    bool ok = ParallelUtility::getLocalArrayBounds(u,ug,I1,I2,I3,1);   
    if( !ok ) return 0;  // no points on this processor
  #endif

  if( option==constantForcing )
  {
    const RealArray & constantForcingParameters = db.get<RealArray>("constantForcingParameters");
    for( int n=0; n<numberOfComponents; n++ )
    {
      if( constantForcingParameters(n)!=0. )
      {
	fg(I1,I2,I3,n)+=constantForcingParameters(n);
      }
    }
  }
  else if( option==gaussianForcing )
  {
    int & numberOfSources = db.get<int>("numberOfSources");
    RealArray & gaussianParameters = db.get<RealArray>("gaussianParameters");

    // Add a Gaussian source term
    int n;
    for( int m=0; m<numberOfSources; m++ )
    {
      for( int n=0; n<numberOfComponents; n++ )
      {
	real an,bn,pn,xn,yn,zn;

	an=gaussianParameters(0,n,m);
	bn=gaussianParameters(1,n,m);
	pn=gaussianParameters(2,n,m);
	xn=gaussianParameters(3,n,m);
	yn=gaussianParameters(4,n,m);
	zn=gaussianParameters(5,n,m);
	if( an!=0. )
	{
	  RealArray rad2;
	  if( numberOfDimensions==2 )
	    rad2 = ( SQR(vertex(I1,I2,I3,axis1)-xn)+
		     SQR(vertex(I1,I2,I3,axis2)-yn) );
	  else
	    rad2 = ( SQR(vertex(I1,I2,I3,axis1)-xn)+
		     SQR(vertex(I1,I2,I3,axis2)-yn)+
		     SQR(vertex(I1,I2,I3,axis3)-zn) );
	  fg(I1,I2,I3,n)+=an*exp( -bn*pow(rad2,pn*.5) );
	}
      }
    }
  }
  else
  {
    cout << "userDefinedForcing: Unknown option =" << option << endl;
  }

  return 0;
}




//\begin{>>DomainSolverInclude.tex}{\subsection{setupUserDefinedForcing}}  
int DomainSolver::
setupUserDefinedForcing()
//==============================================================================================
// /Description:
//    Setup User defined forcing. This function is used to setup and define the forcing to use.
// The function userDefinedForcing (above) is called to actually assign the forcing.
//  Rewrite or add new options to  this routine to supply your own forcing.
// Choose the "user defined forcing" option from the initial conditions options to have this routine
// called.
//
// /Return values: 0=success, non-zero=failure.
//\end{DomainSolverInclude.tex}  
//==============================================================================================
{
  GenericGraphicsInterface & gi = *parameters.dbase.get<GenericGraphicsInterface* >("ps");

  const int & numberOfComponents=parameters.dbase.get<int >("numberOfComponents");
  const int & numberOfDimensions=parameters.dbase.get<int >("numberOfDimensions");
  const int & rc = parameters.dbase.get<int >("rc");   //  density = u(all,all,all,rc)  (if appropriate for this PDE)
  const int & uc = parameters.dbase.get<int >("uc");   //  u velocity component =u(all,all,all,uc)
  const int & vc = parameters.dbase.get<int >("vc");  
  // const int & wc = parameters.dbase.get<int >("wc");
  const int & tc = parameters.dbase.get<int >("tc");   //  temperature
  const int & sc = parameters.dbase.get<int >("sc");   //  mass fraction lambda
  //  const int & pc = parameters.dbase.get<int >("pc");
  
  // here is a menu of possible forcing options
  aString menu[]=  
  {
    "no forcing",
    "constant forcing",
    "gaussian forcing",
    "exit",
    ""
  };
  aString answer,answer2;
  char buff[100];
  gi.appendToTheDefaultPrompt(">user defined");

  // Make a sub-directory in the data-base to store variables used here and in userDefinedForcing
  if( !parameters.dbase.get<DataBase >("modelData").has_key("userDefinedForcingData") )
    parameters.dbase.get<DataBase >("modelData").put<DataBase>("userDefinedForcingData");

  DataBase & db = parameters.dbase.get<DataBase >("modelData").get<DataBase>("userDefinedForcingData");
  // first time through allocate variables 
  if( !db.has_key("option") )
  {
    UserDefinedForcingOptions & option= db.put<UserDefinedForcingOptions>("option",noForcing);

    db.put<int>("numberOfSources");
    db.put<RealArray>("gaussianParameters");

  }

  UserDefinedForcingOptions & option= db.get<UserDefinedForcingOptions>("option");

  int & numberOfSources = db.get<int>("numberOfSources");

 
  for( ;; )
  {
    gi.getMenuItem(menu,answer,"enter an option");
    
    if( answer=="exit" || answer=="done" )
    {
      break;
    }
    else if( answer=="no forcing" )
    {
      option=noForcing;
    }
    else if( answer=="constant forcing" )
    {
      // define a constant forcing
      option=constantForcing;

      if( !db.has_key("constantForcingParameters") )
	db.put<RealArray>("constantForcingParameters");
      RealArray & constantForcingParameters = db.get<RealArray>("constantForcingParameters");

      constantForcingParameters.redim(numberOfComponents);
      constantForcingParameters=0.;
      
      printF("The constant forcing is of the form\n"
             "   f(x,t; n) = g(n)\n"
             " n = component number, n=0,1,..,%i \n",numberOfComponents-1 );

      // query for changes:
      for( int nn=0; ; nn++ )
      {
        gi.inputString(answer2,sPrintF("Enter n, g(n) (Enter `done' to finish)"));
        if( answer2=="done" ) break;
	
        int n=-1;
        real gn=0.;
        sScanF(answer2,"%i %e",&n,&gn);  
        if( n>=0 && n<numberOfComponents )
	{
          constantForcingParameters(n)=gn;
	  printF(" Setting constant forcing g(%i)=%g\n",n,gn);
	}
	else
	{
          printF("DomainSolver::setupUserDefinedForcing:ERROR: invalid input: n=%i \n",n);
	  gi.stopReadingCommandFile();
	  break;
	}
      }
    }
    else if( answer=="gaussian forcing" )
    {
      // define a Gaussian forcing
      option=gaussianForcing;

      if( !db.has_key("gaussianParameters") )
	db.put<RealArray>("gaussianParameters");
      RealArray & gaussianParameters = db.get<RealArray>("gaussianParameters");


      gi.inputString(answer2,sPrintF("Enter the number of Gaussian sources (default = 1 per component)"));
      sScanF(answer2,"%i",&numberOfSources);
      printF(" numberOfSources = %i\n",numberOfSources);
      
      printF("The Gaussian forcing is of the form\n"
             "   f(x,t; n) = Sum_m { a(n,m)*exp( -b(n,m)*| x - x0(n,m) |^p(n,m) ) }\n"
             " m = source number, m=0,1,..,%i \n"
             " n = component number, n=0,1,..,%i \n"
             " a(n,m) = amplitude \n"
             " b(n,m) = exponent \n" 
             " x0(n,m) = centre \n" 
             " p(n,m) = power \n",numberOfSources-1,numberOfComponents-1 );
      real an=0.,bn=10.,pn=2.,xn=0.,yn=0.,zn=0.;
      gaussianParameters.redim(6,numberOfComponents,numberOfSources);
      // assign defaults
      for( int m=0; m<numberOfSources; m++ )
      {
	for( int n=0; n<numberOfComponents; n++ )
	{
	  gaussianParameters(0,n,m)=an;
	  gaussianParameters(1,n,m)=bn;
	  gaussianParameters(2,n,m)=pn;
	  gaussianParameters(3,n,m)=xn;
	  gaussianParameters(4,n,m)=yn;
	  gaussianParameters(5,n,m)=zn;
	}
      }
      for( int nn=0; ; nn++ )
      {
        gi.inputString(answer2,sPrintF("Enter m, n, a, b,p, x,y,z (Enter `done' to finish)"));
        if( answer2=="done" ) break;
	
        int m=-1, n=-1;
        sScanF(answer2,"%i %i %e %e %e %e %e %e ",&m,&n,&an,&bn,&pn,&xn,&yn,&zn);  
        if( m>=0 && m<numberOfSources && n>=0 && n<numberOfComponents )
	{
          gaussianParameters(0,n,m)=an;
          gaussianParameters(1,n,m)=bn;
          gaussianParameters(2,n,m)=pn;
          gaussianParameters(3,n,m)=xn;
          gaussianParameters(4,n,m)=yn;
          gaussianParameters(5,n,m)=zn;
	  printF(" Setting Gaussian parameters for m=%i n=%i: a=%g, b=%g, p=%g, x=%g,y=%g,z=%g\n",
		 m,n,an,bn,pn,xn,yn,zn);
	}
	else
	{
          printF("DomainSolver::setupUserDefinedForcing:ERROR: invalid input: m=%i and n=%i \n",m,n);
	  gi.stopReadingCommandFile();
	  break;
	}
      }
    }
    else 
    {
      cout << "Unknown option =" << answer << endl;
      gi.stopReadingCommandFile();
    }
    
  }
  
  gi.unAppendTheDefaultPrompt();
  return 0;
}



//! This routine is called when DomainSolver is finished and can be used to clean up memory.
void DomainSolver::
userDefinedForcingCleanup()
{
  if( parameters.dbase.get<int >("myid")==0 ) 
    printf("***userDefinedForcingCleanup: delete arrays\n");

  if( parameters.dbase.get<DataBase >("modelData").has_key("userDefinedForcingData") )
  {
    DataBase & db = parameters.dbase.get<DataBase >("modelData").get<DataBase>("userDefinedForcingData");
  }
  
}

