// ========================================================================================================
/// \class Cgins
/// \brief Solve the incompressible Navier-Stokes equations.
/// \details Cgins can be used to solve the incompressible Navier-Stokes with heat transfer.
// ========================================================================================================

#include "Cgins.h"
#include "Oges.h"
#include "Ogshow.h"
#include "ParallelUtility.h"
#include "viscoPlasticMacrosCpp.h"

#include "ProjectVelocity.h"

// CG equation-domain solver for the Incompressible Navier-Stokes Equations

namespace {
  ProjectVelocity projector;  // ******************* fix this *******************
}

// ===================================================================================================================
/// \brief Constructor for the Cgcns class.
///
/// \param cg_ (input) : use this CompositeGrid.
/// \param ps (input) : pointer to a graphics object to use.
/// \param show (input) : pointer to a show file to use.
/// \param plotOption_ (input) : plot option 
/// 
///  \note InsParameters (passed to the DomainSolver constructor above) replaces the base class Parameters
// ==================================================================================================================
Cgins::
Cgins(CompositeGrid & cg_, 
      GenericGraphicsInterface *ps /* =NULL */, 
      Ogshow *show /* =NULL */ , 
      const int & plotOption_ /* =1 */) 
    : DomainSolver(*(new InsParameters),cg_,ps,show,plotOption_)
{
  className="Cgins";
  name="ins";
}



// ===================================================================================================================
/// \brief Destructor for the Cgins class.
// ==================================================================================================================
Cgins::
~Cgins()
{
  delete &parameters;
  
  for( int i=0; i<tzForcingVector.size(); i++)
    delete [] tzForcingVector[i];
}


//===============================================================================================
/// \brief Update the solver to match a new grid.
/// \param cg (input): composite grid.
//===============================================================================================
int Cgins::
updateToMatchGrid(CompositeGrid & cg)
{
  printF("\n $$$$$$$$$$$$$$$ Cgins: updateToMatchGrid(CompositeGrid & cg) $$$$$$$$$$$$\n\n");
  
  if( pp==NULL )
    pp=new realCompositeGridFunction;
  p().updateToMatchGrid(cg);
  // create the pressure equation
  updatePressureEquation(cg,gf[current]);

  return DomainSolver::updateToMatchGrid(cg);

}


//\begin{>>CompositeGridSolverInclude.tex}{\subsection{updateForMovingGrids}} 
int Cgins::
updateForMovingGrids(GridFunction & cgf)
//=========================================================================================
// /Description:
//    Update the CompositeGridSolver after grids have moved or have been adapted.
//
//\end{CompositeGridSolverInclude.tex}  
//=========================================================================================
{
  real cpu0,timeForUpdatePressure=0.;
  if( movingGridProblem() )
  {
    cpu0=getCPU();
    cgf.u.updateToMatchGrid( cgf.cg );    
    updateGeometryArrays( cgf );
    parameters.dbase.get<RealArray>("timing")(Parameters::timeForMovingUpdate)+=getCPU()-cpu0;

    cpu0=getCPU();
    // update the pressure equation
    updatePressureEquation(cgf.cg, cgf);
    timeForUpdatePressure=getCPU()-cpu0;

  }

  if( parameters.isAdaptiveGridProblem() )
  {
    real cpu0=getCPU();
    // update the pressure equation
    updatePressureEquation(cgf.cg, cgf);
    parameters.dbase.get<RealArray>("timing")(Parameters::timeForUpdatePressureEquation)+=getCPU()-cpu0;
    timeForUpdatePressure=getCPU()-cpu0;
  }
  if( parameters.isMovingGridProblem()  && parameters.isAdaptiveGridProblem() )
  {
    // update cgf.gridVelocity arrays
    cgf.updateGridVelocityArrays();
    // don't do this here if we have just moved the grids and not adapted!
    // *no* cgf.gridVelocityTime=cgf.t -1.e10;  // this will force a recomputation of the grid velocity the next time...
  }
  

  parameters.dbase.get<RealArray>("timing")(Parameters::timeForUpdatePressureEquation)+=timeForUpdatePressure;
  parameters.dbase.get<RealArray>("timing")(Parameters::timeForTimeIndependentVariables)+=timeForUpdatePressure;
  return 0;
}


int Cgins::
updateForNewTimeStep(GridFunction & cgf, const real & dt)
//=========================================================================================
// /Description:
//    Update geometry arrays after the time step has changed.
//
//\end{CompositeGridSolverInclude.tex}  
//=========================================================================================
{
  const int geometryHasChanged=false;
  updateDivergenceDamping( cgf.cg,geometryHasChanged ); 
  return 0;
}

int Cgins::
updateGeometryArrays(GridFunction & cgf)
{
  if( debug() & 4 ) printF(" --- Cgins::updateGeometryArrays ---\n");

  real cpu0=getCPU();

  int grid;
  for( grid=0; grid<cgf.cg.numberOfComponentGrids(); grid++ )
  {
    if( !cgf.cg[grid].isRectangular() || twilightZoneFlow() || parameters.isAxisymmetric() ||
	parameters.gridIsMoving(grid) )
      cgf.cg[grid].update(MappedGrid::THEcenter | MappedGrid::THEvertex );  
  }
  parameters.dbase.get<RealArray>("timing")(Parameters::timeForUpdatePressureEquation)+=getCPU()-cpu0;

  return DomainSolver::updateGeometryArrays(cgf);
}

void Cgins::
updateTimeIndependentVariables( CompositeGrid & cg0, GridFunction & cgf )
// ==========================================================================================
// This function is called after grids have been added or removed
// ==========================================================================================
{
  updatePressureEquation(cg0,cgf);
}

// ======================================================================================================
/// \brief:  Set the titles and labels that go on the show file output 
// ======================================================================================================
void Cgins::
saveShowFileComments( Ogshow &show )
{
  char buffer[200]; 

  // save comments that go at the top of each plot

  aString timeLine="";
  if(  parameters.dbase.has_key("timeLine") )
    timeLine=parameters.dbase.get<aString>("timeLine");

  aString showFileTitle[5];
  if( parameters.dbase.get<InsParameters::PDEModel >("pdeModel")==InsParameters::BoussinesqModel )
    showFileTitle[0]=sPrintF(buffer,"Incompressible NS (Boussinesq), nu=%8.2e, k=%8.2e",
			     parameters.dbase.get<real >("nu"),parameters.dbase.get<real >("kThermal"));
  else if( parameters.dbase.get<InsParameters::PDEModel >("pdeModel")==InsParameters::viscoPlasticModel )
  {
    // declare and lookup visco-plastic parameters (macro)
    declareViscoPlasticParameters;
    showFileTitle[0]=sPrintF(buffer,"INS-VP (eta,yield,exp)=(%.4g,%.4g,%.4g), k=%8.2e",
			     etaViscoPlastic,yieldStressViscoPlastic,exponentViscoPlastic,
                             parameters.dbase.get<real >("kThermal"));
  }
  else 
    showFileTitle[0]=sPrintF(buffer,"Incompressible NS, nu=%8.2e",parameters.dbase.get<real >("nu"));

  if( parameters.isAxisymmetric() )
    showFileTitle[1]=sPrintF(buffer,"axisymmetric, %s",(const char*)timeLine);
  else
    showFileTitle[1]=timeLine;

  if( parameters.dbase.get<bool >("useSecondOrderArtificialDiffusion") )
    showFileTitle[1]+=sPrintF(buffer,", ad21=%3.1f ad22=%3.1f",parameters.dbase.get<real >("ad21"),parameters.dbase.get<real >("ad22"));
  if( parameters.dbase.get<bool >("useFourthOrderArtificialDiffusion") )
    showFileTitle[1]+=sPrintF(buffer,", ad41=%3.1f ad42=%3.1f",parameters.dbase.get<real >("ad41"),parameters.dbase.get<real >("ad42"));

  showFileTitle[2]="";  // marks end of titles

  for( int i=0; showFileTitle[i]!=""; i++ )
    show.saveComment(i,showFileTitle[i]);

}  

// ===================================================================================================================
/// \brief Output run-time parameters for the header.
/// \param file (input) : write values to this file.
///
// ===================================================================================================================
void
Cgins::
writeParameterSummary( FILE * file )
{
  DomainSolver::writeParameterSummary( file );

  const InsParameters::PDEModel & pdeModel = parameters.dbase.get<InsParameters::PDEModel>("pdeModel");

  if ( file==parameters.dbase.get<FILE* >("checkFile") )
  {
    fprintf(parameters.dbase.get<FILE* >("checkFile"),"\\caption{Incompressible Navier Stokes, gridName, $\\nu=%3.2f$, $t=%2.1f$, ",
	    parameters.dbase.get<real >("nu"),parameters.dbase.get<real >("tFinal"));

    return;
    
  }

  Parameters::TimeSteppingMethod &timeSteppingMethod = 
             parameters.dbase.get<Parameters::TimeSteppingMethod>("timeSteppingMethod");

  fprintf(file," nu=%e, cdv=%g, cDt=%g, pressure solver=%s \n",parameters.dbase.get<real >("nu"),
	  parameters.dbase.get<real >("cdv"),parameters.dbase.get<real >("cDt"),
          (const char*)pressureSolverParameters.getSolverTypeName());
      
  if( pdeModel==InsParameters::BoussinesqModel )
  {
    fprintf(file," Boussinesq: kThermal=%g, thermalConductivity=%g\n",parameters.dbase.get<real >("kThermal"),
            parameters.dbase.get<real >("thermalConductivity"));
    ArraySimpleFixed<real,3,1,1,1> & gravity = parameters.dbase.get<ArraySimpleFixed<real,3,1,1,1> >("gravity");
    if( gravity[0]!=0. || gravity[1]!=0. || gravity[2]!=0. )
    {
      real thermalExpansivity=1.;
      parameters.dbase.get<ListOfShowFileParameters >("pdeParameters").getParameter("thermalExpansivity",thermalExpansivity);
      fprintf(file," Coefficient of thermal expansivity =%8.2e \n",thermalExpansivity);
    }
  }
  else if( pdeModel==InsParameters::twoPhaseFlowModel )
  {
    fprintf(file," twoPhaseFlowModel: kThermal=%g\n",parameters.dbase.get<real >("kThermal"));
  }

  if( pdeModel==InsParameters::BoussinesqModel || pdeModel==InsParameters::twoPhaseFlowModel )
  {
    ArraySimpleFixed<real,3,1,1,1> & gravity = parameters.dbase.get<ArraySimpleFixed<real,3,1,1,1> >("gravity");
    if( gravity[0]!=0. || gravity[1]!=0. || gravity[2]!=0. )
    {
      fprintf(file," gravity is on, acceleration due to gravity = (%8.2e,%8.2e,%8.2e) \n",
                     gravity[0],gravity[1],gravity[2]);
    }
    else
    {
      fprintf(file," gravity is off.\n");
    }
    
  }
  if( timeSteppingMethod==Parameters::implicit )
  {
    InsParameters::ImplicitVariation & implicitVariation = 
             parameters.dbase.get<InsParameters::ImplicitVariation>("implicitVariation");
    
    fprintf(file," implicit variation = %s\n",(implicitVariation==InsParameters::implicitViscous ? "implicitViscous" :
            implicitVariation==InsParameters::implicitAdvectionAndViscous ? "implicitAdvectionAndViscous" :
	    implicitVariation==InsParameters::implicitFullLinearized ? "implicitFullLinearized" : "unknown"));

    fprintf(file," refactor frequency=%i\n",parameters.dbase.get<int>("refactorFrequency"));

  }
  

}

//\begin{>>CompositeGridSolverInclude.tex}{\subsection{project}} 
int Cgins::
project(GridFunction & cgf)
// ===================================================================================
// /Description:
//   Project the solution to be divergence free (approximately)
// The projection uses the Oges object "poisson" from this class -- it thus will
// use the same parameter values set in OverBlown. The projection Poisson equation uses
// Neumann BC's all around so that it may have different BC's from the NS equations.
//\end{CompositeGridSolverInclude.tex}  
// ===================================================================================
{
  if( parameters.dbase.get<bool>("simulateGridMotionOnly") ) return 0;
  
  real time0=getCPU();
  const int myid=Communication_Manager::My_Process_Number;

  if( parameters.dbase.get<bool >("projectInitialConditions") )
  {
    printF(">>>>>Cgins::project: project the initial conditions <<<<\n");

    if( Parameters::checkForFloatingPointErrors )
      checkSolution(cgf.u,"Cgins::project: cgf.u at start");

    // *new* way, avoids generating a separate matrix for the projection operator
    CompositeGrid & cg = cgf.cg;
    realCompositeGridFunction & u = cgf.u;

    int numberOfSmoothsPerProjection=5;
    projector.setNumberOfSmoothsPerProjectionIteration(numberOfSmoothsPerProjection);
    projector.setVelocityComponent(parameters.dbase.get<int >("uc"));
    projector.setPoissonSolver(poisson);
    projector.setCompare3Dto2D(parameters.dbase.get<int >("compare3Dto2D"));
    projector.setIsAxisymmetric(parameters.isAxisymmetric());

    const int uc = parameters.dbase.get<int >("uc");
    const int vc = parameters.dbase.get<int >("vc");
    const int wc = parameters.dbase.get<int >("wc");

    Range V(uc,uc+parameters.dbase.get<int >("numberOfDimensions")-1);

    int numberOfSmoothsPerProjectionIteration=5;
    int minimumNumberOfProjectionIterations=1;
    real convergenceTolerance=.95;
    // int numberOfSmoothingSubIterations=1;  // only do 1 because of BC's and interp pts      

    int maximumNumberOfProjectionIterations=9;
    int nitsm=10;  // number of smoothing steps

    // ********************
#ifdef USE_PPP
    //   maximumNumberOfProjectionIterations=1;
    //   nitsm=1;
#endif      

    real divMax;
    real divMaxOld=1.e6;

    const real cdv=parameters.dbase.get<real >("cdv");
    const real advectionCoefficient=parameters.dbase.get<real >("advectionCoefficient");
    const real nu = parameters.dbase.get<real >("nu");

    // ------------------------------------------------------------------------------
    // --- Set parameters so that the RHS to the pressure equation is just div(u) ---
    // ------------------------------------------------------------------------------

    parameters.dbase.get<int>("initialConditionsAreBeingProjected")=1; // new way, this is used in assignPressureRHS

    parameters.dbase.get<real >("cdv")=1.e6;
    parameters.dbase.get<real >("advectionCoefficient")=0.;
    parameters.dbase.get<real >("nu")=0.;
      
    dt=parameters.dbase.get<real >("cDt");  // this will make the divergenceDampingWeight==1.
      

    BoundaryConditionParameters bcParams;
    bcParams.ghostLineToAssign= parameters.dbase.get<int >("orderOfAccuracy")/2;  // *note*
      
    // this will create the pressure equation.
    updateToMatchGrid(gf[0].cg); 


    int grid;
    int extra[3] = { 0,0,0 };  
    Range all;
    Index I1,I2,I3;
      
    // for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    // displayMask(cg[grid].mask(),"mask in project",parameters.dbase.get<FILE* >("debugFile"));

    // realCompositeGridFunction divergence(cg,all,all,all);  // holds divergence  ****** no need ***************
    assert( numberOfExtraFunctionsToUse>0 );
    realCompositeGridFunction divergence;  // use fn[0] as a workspace
#ifdef USE_PPP
    divergence.updateToMatchGrid(cgf.cg); 
#else
    divergence.link(fn[0],Range(0,0));
#endif
    aString buff;
      
    for( int it=0; it<maximumNumberOfProjectionIterations; it++ )
    {
      // ---smoothing iterations, smooth more the first few times
      int numberOfSmoothingSubIterations = (it<5) ? max(nitsm,numberOfSmoothsPerProjectionIteration) : nitsm; 
      smoothVelocity( cgf,numberOfSmoothingSubIterations ); 

      if( debug() & 4 || debug() & 32 )
	cgf.u.display(sPrintF(buff,"projectVelocity>>> Solution after smooth velocity: u, it=%i",it),parameters.dbase.get<FILE* >("debugFile"),"%5.2f ");
	
      if( Parameters::checkForFloatingPointErrors )
        checkSolution(cgf.u,"Cgins::project: cgf.u after smoothVelocity");

      real divMaxAfterSmooth = projector.computeDivergence( u,divergence );   // compute divergence for rhs 
      if( it==0 ) divMaxOld=divMaxAfterSmooth;
      if( debug() & 4 )
	printF(" projectVelocity>>> iteration=%i, divergence after smooth    =%f\n",it,divMaxAfterSmooth);

      // *** undo the div(u)=0 BC so the projection will work better on the boundary.
      if( false || parameters.dbase.get<int >("orderOfAccuracy")==2 ) // turn off for 4th order for testing
      {
	u.applyBoundaryCondition(V,BCTypes::extrapolate,BCTypes::allBoundaries);
	if( parameters.dbase.get<int >("orderOfAccuracy")==4 )
	  u.applyBoundaryCondition(V,BCTypes::extrapolate,BCTypes::allBoundaries,0.,0.,bcParams); 
      }
      if( parameters.dbase.get<int >("orderOfAccuracy")==4 ) // trouble at outflow with standard BC's
      {
	u.applyBoundaryCondition(V,BCTypes::extrapolate,BCTypes::allBoundaries);
	u.applyBoundaryCondition(V,BCTypes::extrapolate,BCTypes::allBoundaries,0.,0.,bcParams); // **** 990106
	u.finishBoundaryConditions();
      }


      solveForTimeIndependentVariables( cgf ); 

      if( debug() & 4 || debug() & 32 )
	p().display(sPrintF(buff,"projectVelocity>>> Pressure after solve: p() it=%i",it),parameters.dbase.get<FILE* >("debugFile"),"%9.6f ");
        
      for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
      {
	// update interior points (and periodic boundaries)
	for( int axis=0; axis<cg.numberOfDimensions(); axis++ )
	  extra[axis] = cg[grid].isPeriodic()(axis) ? 0 : -1;
	getIndex(extendedGridIndexRange(cg[grid]),I1,I2,I3,extra[0],extra[1],extra[2]); 

	// **** used optimized operators here ****
	MappedGridOperators & op = *(u[grid].getOperators());
	realMappedGridFunction & px = fn[0][grid];   // use fn[0] as a work space
	realMappedGridFunction & pg = p()[grid];
	realArray & uu = u[grid];
	  
	Range V(uc,uc+cg.numberOfDimensions()-1);
	  
#ifdef USE_PPP
	realSerialArray pgLocal;  getLocalArrayWithGhostBoundaries(pg,pgLocal);
	realSerialArray pxLocal;  getLocalArrayWithGhostBoundaries(px,pxLocal);
	realSerialArray uuLocal;  getLocalArrayWithGhostBoundaries(uu,uuLocal);
#else
	realSerialArray & pgLocal = pg;
	realSerialArray & pxLocal = px;
	realSerialArray & uuLocal = uu;
#endif  

	bool ok = ParallelUtility::getLocalArrayBounds(uu,uuLocal,I1,I2,I3);
	if( ok )
	{
	  op.derivative(MappedGridOperators::xDerivative,pgLocal,pxLocal,I1,I2,I3);
	  uuLocal(I1,I2,I3,uc)-=pxLocal(I1,I2,I3,0);
	  op.derivative(MappedGridOperators::yDerivative,pgLocal,pxLocal,I1,I2,I3);
	  uuLocal(I1,I2,I3,vc)-=pxLocal(I1,I2,I3,0);
	  if( cg.numberOfDimensions()==3 )
	  {
	    op.derivative(MappedGridOperators::zDerivative,pgLocal,pxLocal,I1,I2,I3);
	    uuLocal(I1,I2,I3,wc)-=pxLocal(I1,I2,I3,0);
	  }
	}
	if( false ) // old way
	{
	  op.derivative(MappedGridOperators::xDerivative,pg,px,I1,I2,I3);

	  // px=0;
	  
	  uu(I1,I2,I3,uc)-=px(I1,I2,I3,0);

	  if( debug() & 4 )
	  {
	    ::display(pg,sPrintF(buff,"projectVelocity>>> p, grid=%i it=%i",grid,it),parameters.dbase.get<FILE* >("debugFile"),"%5.2f ");
	    ::display(px,sPrintF(buff,"projectVelocity>>> px, grid=%i it=%i",grid,it),parameters.dbase.get<FILE* >("debugFile"),"%5.2f ");
	  }
	  op.derivative(MappedGridOperators::yDerivative,pg,px,I1,I2,I3);
	  uu(I1,I2,I3,vc)-=px(I1,I2,I3,0);
	  if( cg.numberOfDimensions()==3 )
	  {
	    op.derivative(MappedGridOperators::zDerivative,pg,px,I1,I2,I3);
	    uu(I1,I2,I3,wc)-=px(I1,I2,I3,0);
	  }
	}
	  
	if( debug() & 4 )
	{
	  ::display(px,sPrintF(buff,"projectVelocity>>> py, grid=%i it=%i",grid,it),parameters.dbase.get<FILE* >("debugFile"),"%5.2f ");
	}

	// 	  u[grid](I1,I2,I3,uc  )-=p()[grid].x(I1,I2,I3)(I1,I2,I3);
	// 	  u[grid](I1,I2,I3,uc+1)-=p()[grid].y(I1,I2,I3)(I1,I2,I3);
	// 	  if( cg.numberOfDimensions()==3 )
	// 	    u[grid](I1,I2,I3,uc+2)-=p()[grid].z(I1,I2,I3)(I1,I2,I3);


      }

      if( debug() & 4 || debug() & 32 )
	cgf.u.display(sPrintF(buff,"projectVelocity>>> Solution after project: u it=%i",it),parameters.dbase.get<FILE* >("debugFile"),"%9.6f ");

      if( (it % 2) == 1)
	fixupUnusedPoints(cgf.u);
      // u.display("project velocity u, before interpolate");
    
      // interpolate first, needed for extrapolate-neighbours
      interpolate(cgf,V);

      if( debug() & 4 || debug() & 32 )
	cgf.u.display("projectVelocity>>> Solution after interpolate: u",parameters.dbase.get<FILE* >("debugFile"),"%9.6f ");

      // assign bc's to get ghost point values
      DomainSolver::applyBoundaryConditions(cgf); 

      if( false  || parameters.dbase.get<int >("orderOfAccuracy")==2 )// turn off for 4th order for testing
      {
	u.applyBoundaryCondition(V,BCTypes::extrapolate,BCTypes::allBoundaries);
	if( parameters.dbase.get<int >("orderOfAccuracy")==4 )
	  u.applyBoundaryCondition(V,BCTypes::extrapolate,BCTypes::allBoundaries,0.,0.,bcParams); // **** 990106
      }
      if( parameters.dbase.get<int >("orderOfAccuracy")==4 )
      {
	u.applyBoundaryCondition(V,BCTypes::extrapolate,BCTypes::allBoundaries);
	u.applyBoundaryCondition(V,BCTypes::extrapolate,BCTypes::allBoundaries,0.,0.,bcParams); // **** 990106
	u.finishBoundaryConditions();
      }
	
	
      if( debug() & 4 || debug() & 32 )
	cgf.u.display(sPrintF(buff,"projectVelocity>>> Solution after applyBC: u, it=%i",it),parameters.dbase.get<FILE* >("debugFile"),"%9.6f ");
    
      divMax = projector.computeDivergence( u,divergence );  // compute new divergence to test for convergence
      if( debug() & 4 )
	printF(" projectVelocity>>> iteration=%i, divergence after project =%f\n",it,divMax);

      if( (TRUE || debug() & 2) )
      {
	printF(" projectVelocity>>> iteration=%i, (new div)/(old div)=%6.3f, divergence after projection=%9.2e \n", 
	       it,min(1.e6,divMax/max(REAL_MIN,divMaxOld)),divMax);
	fPrintF(parameters.dbase.get<FILE* >("debugFile"),"******* projectVelocity>>> iteration=%i, (new div)/(old div)=%6.3f, divergence after projection=%9.2e \n", 
		it,min(1.e6,divMax/max(REAL_MIN,divMaxOld)),divMax);
      }
      //   psp.set(GI_TOP_LABEL,"After project");
      //   parameters.dbase.get<GenericGraphicsInterface* >("ps")->contour(u,psp);
	
      //  --- stop when the divergence is nolonger decreasing
      if(divMax==0. || (it>minimumNumberOfProjectionIterations && divMax/divMaxOld > convergenceTolerance))
	break;
      divMaxOld=divMax;


    }

    // reset: 
    parameters.dbase.get<int>("initialConditionsAreBeingProjected")=0; // new way 

    parameters.dbase.get<real >("cdv")=cdv;  // reset
    parameters.dbase.get<real >("advectionCoefficient")=advectionCoefficient;
    parameters.dbase.get<real >("nu")=nu;

    dt=0.;
  }
  
  else
  {
    updateToMatchGrid(cgf.cg); 
  }
  

  if( debug() & 16 ) 
  {
    fPrintF(parameters.dbase.get<FILE* >("debugFile")," \n ****Solution after projectVelocity and BC's**** \n");
    outputSolution( cgf.u,0. );
  }
  
  real time=getCPU()-time0;
  printF(">>>>>Time to project = %8.2e s <<<<\n",time);

  return 0;
}
  
