! **************************************************************************************************
!   Define the full implicit matrix for the incompressible Navier-Stokes
! **************************************************************************************************


! Macro's for forming the general implicit matrix
#Include "insImp.h"


! =============================================================
! macro to declare temporary variables:
! =============================================================
#beginMacro declareInsImpTemporaryVariables()
 declareTemporaryVariables(2,2)
 declareParametricDerivativeVariables(uu,3)   ! declare temp variables uu, uur, uus, ...
 declareParametricDerivativeVariables(vv,3) 
 declareParametricDerivativeVariables(ww,3) 
 declareParametricDerivativeVariables(pp,3) 
 declareParametricDerivativeVariables(qq,3) 
 declareParametricDerivativeVariables(uul,3)   ! declare temp variables uu, uur, uus, ...
 declareParametricDerivativeVariables(vvl,3) 
 declareParametricDerivativeVariables(wwl,3) 
 declareParametricDerivativeVariables(qql,3) 
 declareJacobianDerivativeVariables(aj,3)     ! declareJacobianDerivativeVariables(aj,DIM)
 real radi 
 real yyCoeff(0:maxWidthDim-1)
 integer bc0
#endMacro

! =====================================================================================
! This macro is used in insImp.h and is used to look up parameters etc. for this PDE
! =====================================================================================
#beginMacro initializePdeParameters()
 if( pdeModel.ne.standardModel .and. pdeModel.ne.BoussinesqModel )then
   write(*,'("insImpINS:ERROR: unexpected pdeModel=",i3)') pdeModel
   stop 2734
 end if
 ! make sure these are initialized (used in evaluation of adCoeffl even if cd22=0. )
 ulx=0.
 uly=0.
 ulz=0.
 vlx=0.
 vly=0.
 vlz=0.
 wlx=0.
 wly=0.
 wlz=0.

#endMacro

! ==============================================================================================================
!   Fill in the coefficients for the Incompressible Navier-Stokes Equations
! ==============================================================================================================
#beginMacro fillCoeffPDE()
if( fillCoefficients.eq.1 )then


 if( fillCoefficientsScalarSystem.ge.fillCoeffU .and. fillCoefficientsScalarSystem.le.fillCoeffW )then
   ! --- fill coefficients for a scalar system for a velocity component: I - nu*Delta - A.D. 

   ! write(*,'("@@@@ insImpINS: (u,v,w) fillCoefficientsScalarSystem=",i4)') fillCoefficientsScalarSystem
   ! '

 beginLoops()

  ! Evaluate the jacobian derivatives used by the coefficient and forward derivatives:
  ! opEvalJacobianDerivatives(MAXDER) : MAXDER = max number of derivatives to precompute.
  opEvalJacobianDerivatives(aj,1)

  ! evaluate the coeff operators 
  getCoeff(identity, iCoeff,aj)
  getCoeff(laplacian, lapCoeff,aj)

  ! dissCoeff = dr^2*D_rr + ds^2*D_ss [ + dt^2*D_tt ] 
  getCoeff(r2Dissipation, dissCoeff,aj )

  ! First evaluate the parametric derivatives of u, used later by the spatial derivatives. 
  ! MAXDER = max number of parametric derivatives to precompute.
  opEvalParametricDerivative(u,uc,uu,1)    ! computes uur, uus 
  ! Evaluate the spatial derivatives of u (uses uur, uus):
  getOp(x, u,uc,uu,aj,u0x)       ! u.x
  getOp(y, u,uc,uu,aj,u0y)       ! u.y
  #If $DIM == 3
   getOp(z, u,uc,uu,aj,u0z)       ! u.z
  #End

  ! parametric derivatives of v: 
  opEvalParametricDerivative(u,vc,vv,1)        ! computes vvr, vvs 
  getOp(x, u,vc,vv,aj,v0x)       ! v.x
  getOp(y, u,vc,vv,aj,v0y)       ! v.y 
  #If $DIM == 3
   getOp(z, u,vc,vv,aj,v0z)      ! v.z

   opEvalParametricDerivative(u,wc,ww,1)        ! computes vvr, vvs 
   getOp(x, u,wc,ww,aj,w0x)      ! w.x
   getOp(y, u,wc,ww,aj,w0y)      ! w.y 
   getOp(z, u,wc,ww,aj,w0z)      ! w.z
  #End

 getArtificialDissipationCoeff(adCoeff, u0x,u0y,u0z, v0x,v0y,v0z, w0x,w0y,w0z )

 ! Form : I - nuDt*Delta -adCoeff*dt*( urr + uss + .. )
 setCoeff3(cmpu,eqnu,coeff,iCoeff,-nuDt*lapCoeff,-adCoeff*dt*dissCoeff)

 #If $DIM == 2 
  if( isAxisymmetric.eq.1 )then
    ! add axisymmetric corrections 
    radi = radiusInverse(i1,i2,i3)
    if( radi.ne.0. )then
     getCoeff(y, yCoeff,aj)
     if( fillCoefficientsScalarSystem.eq.fillCoeffU )then
      addCoeff1(cmpu,eqnu,coeff, -nuDt*radi*yCoeff)                        ! -nu*dt*( U_r/r )
     else
      addCoeff2(cmpu,eqnu,coeff, -nuDt*radi*yCoeff, nuDt*radi**2*iCoeff)   ! -nu*dt*( V_r/r - V/r^2 )
     end if
    else
     ! corrections on the axis
     getCoeff(yy, yyCoeff,aj)
     if( fillCoefficientsScalarSystem.eq.fillCoeffU )then
      addCoeff1(cmpu,eqnu,coeff, -nuDt*yyCoeff)                            ! -nu*dt*( U_rr )
     else
      addCoeff1(cmpu,eqnu,coeff, -nuDt*.5*yyCoeff)                         ! -nu*dt*( .5*V_rr )
     end if
    end if
    
  end if

 #End

 endLoops()


 else if( fillCoefficientsScalarSystem.eq.fillCoeffT )then
   ! --- fill coefficients for a scalar system for the Temperature: I - kappa*Delta - A.D. 

   ! write(*,'("@@@@ insImpINS: T: fillCoefficientsScalarSystem=",i4)') fillCoefficientsScalarSystem

 beginLoops()

  ! Evaluate the jacobian derivatives used by the coefficient and forward derivatives:
  ! opEvalJacobianDerivatives(MAXDER) : MAXDER = max number of derivatives to precompute.
  opEvalJacobianDerivatives(aj,1)

  ! evaluate the coeff operators 
  getCoeff(identity, iCoeff,aj)
  getCoeff(laplacian, lapCoeff,aj)

  ! dissCoeff = dr^2*D_rr + ds^2*D_ss [ + dt^2*D_tt ] 
  getCoeff(r2Dissipation, dissCoeff,aj )

  setCoeff3(cmpu,eqnu,coeff,iCoeff,-kDt*lapCoeff,-adcBoussinesq*dt*dissCoeff)

 #If $DIM == 2 

  if( isAxisymmetric.eq.1 )then
    ! add axisymmetric corrections 
    radi = radiusInverse(i1,i2,i3)
    if( radi.ne.0. )then
     getCoeff(y, yCoeff,aj)
     addCoeff1(cmpu,eqnu,coeff, -kDt*radi*yCoeff)                        ! -nu*dt*( U_r/r )
    else
     ! corrections on the axis
     getCoeff(yy, yyCoeff,aj)
     addCoeff1(cmpu,eqnu,coeff, -kDt*yyCoeff)                            ! -nu*dt*( U_rr )
    end if
  end if

 #End

 endLoops()


 else
   ! --- fill coefficients for the full system ---
 beginLoops()

  ! Evaluate the jacobian derivatives used by the coefficient and forward derivatives:
  ! opEvalJacobianDerivatives(MAXDER) : MAXDER = max number of derivatives to precompute.
  opEvalJacobianDerivatives(aj,1)

  ! evaluate the coeff operators 
  getCoeff(identity, iCoeff,aj)
  getCoeff(laplacian, lapCoeff,aj)
  getCoeff(x, xCoeff,aj)
  getCoeff(y, yCoeff,aj)
  #If $DIM == 3
   getCoeff(z, zCoeff,aj)
  #End

  ! dissCoeff = dr^2*D_rr + ds^2*D_ss [ + dt^2*D_tt ] 
  getCoeff(r2Dissipation, dissCoeff,aj )

  ! for testing, get coeff for div( s grad )
  ! getOpCoeffDivScalarGrad(s(i1,i2,i3,0))

  ! **** get a fourth order dissipation u.rrrr + u.ssss 

  ! evaluate forward derivatives of the current solution: 

  ! First evaluate the parametric derivatives of u, used later by the spatial derivatives. 
  ! MAXDER = max number of parametric derivatives to precompute.
  opEvalParametricDerivative(u,uc,uu,1)    ! computes uur, uus 
  ! Evaluate the spatial derivatives of u (uses uur, uus):
  getOp(x, u,uc,uu,aj,u0x)       ! u.x
  getOp(y, u,uc,uu,aj,u0y)       ! u.y
  #If $DIM == 3
   getOp(z, u,uc,uu,aj,u0z)       ! u.z
  #End

  ! parametric derivatives of v: 
  opEvalParametricDerivative(u,vc,vv,1)        ! computes vvr, vvs 
  getOp(x, u,vc,vv,aj,v0x)       ! v.x
  getOp(y, u,vc,vv,aj,v0y)       ! v.y 
  #If $DIM == 3
   getOp(z, u,vc,vv,aj,v0z)      ! v.z

   opEvalParametricDerivative(u,wc,ww,1)        ! computes vvr, vvs 
   getOp(x, u,wc,ww,aj,w0x)       ! w.x
   getOp(y, u,wc,ww,aj,w0y)       ! w.y 
   getOp(z, u,wc,ww,aj,w0z)       ! w.z
  #End

 getArtificialDissipationCoeff(adCoeff, u0x,u0y,u0z, v0x,v0y,v0z, w0x,w0y,w0z )

 if( gridIsMoving.ne.0 )then
   ugv = uu - gv(i1,i2,i3,0)
   vgv = vv - gv(i1,i2,i3,1)
  #If $DIM == 3
   wgv = ww - gv(i1,i3,i3,2)
  #End
 else
   ugv = uu
   vgv = vv
  #If $DIM == 3
   wgv = ww
  #End
 end if

 #If $DIM == 2 
  ! Form : I - nuDt*Delta + aDt*u*Dx + aDt*v*Dy + ...
  !    u0*ux + v0*uy + u*u0x + v*u0y 
  ! moving: (u-gv)*ux =  u*ux - gv*ux -> u0*ux+u*u0x-gv*ux = (u0-gv)*ux + u0x*u (linearized form)
  setCoeff6(cmpu,eqnu,coeff,iCoeff,-nuDt*lapCoeff,adt*ugv*xCoeff,adt*vgv*yCoeff,bDt*u0x*iCoeff,-adCoeff*dt*dissCoeff)
  addCoeff1(cmpv,eqnu,coeff, bDt*u0y*iCoeff)  ! v*u0y in u eqn
  ! addCoeff1(cmpv,eqnu,coeff, bDt*iCoeff)  ! for testing -- add v to u-eqn
  !    u0*vx + v0*vy + u*v0x + v*v0y 
  setCoeff6(cmpv,eqnv,coeff,iCoeff,-nuDt*lapCoeff,adt*ugv*xCoeff,adt*vgv*yCoeff,bDt*v0y*iCoeff,-adCoeff*dt*dissCoeff)
  addCoeff1(cmpu,eqnv,coeff, bDt*v0x*iCoeff)  ! u*v0x in v eqn

  if( isAxisymmetric.eq.1 )then
    ! add axisymmetric corrections 
    radi = radiusInverse(i1,i2,i3)
    if( radi.ne.0. )then
     addCoeff1(cmpu,eqnu,coeff, -nuDt*radi*yCoeff)                        ! -nu*dt*( U_r/r )
     addCoeff2(cmpv,eqnv,coeff, -nuDt*radi*yCoeff, nuDt*radi**2*iCoeff)   ! -nu*dt*( V_r/r - V/r^2 )
    else
     ! corrections on the axis
     getCoeff(yy, yyCoeff,aj)
     addCoeff1(cmpu,eqnu,coeff, -nuDt*yyCoeff)                            ! -nu*dt*( U_rr )
     addCoeff1(cmpv,eqnv,coeff, -nuDt*.5*yyCoeff)                         ! -nu*dt*( .5*V_rr )
    end if
    
  end if

 #Else
  ! Form : I - nuDt*Delta + aDt*u*Dx + aDt*v*Dy + aDt*w*dz
  !    u0*ux + v0*uy + w0*uz + u*u0x + v*u0y + w*u0z 
  setCoeff7(cmpu,eqnu,coeff,iCoeff,-nuDt*lapCoeff,adt*ugv*xCoeff,adt*vgv*yCoeff,adt*wgv*zCoeff,bDt*u0x*iCoeff,-adCoeff*dt*dissCoeff)
  addCoeff1(cmpv,eqnu,coeff, bDt*u0y*iCoeff)  ! v*u0y in u eqn
  addCoeff1(cmpw,eqnu,coeff, bDt*u0z*iCoeff)  ! w*u0z in u eqn

  !    u0*vx + v0*vy + w0*vz + u*v0x + v*v0y + w*v0z 
  setCoeff7(cmpv,eqnv,coeff,iCoeff,-nuDt*lapCoeff,adt*ugv*xCoeff,adt*vgv*yCoeff,adt*wgv*zCoeff,bDt*v0y*iCoeff,-adCoeff*dt*dissCoeff)
  addCoeff1(cmpu,eqnv,coeff, bDt*v0x*iCoeff)  ! u*v0x in u eqn
  addCoeff1(cmpw,eqnv,coeff, bDt*v0z*iCoeff)  ! w*v0z in u eqn

  !    u0*wx + v0*wy + w0*wz + u*w0x + v*w0y + w*w0z 
  setCoeff7(cmpw,eqnw,coeff,iCoeff,-nuDt*lapCoeff,adt*ugv*xCoeff,adt*vgv*yCoeff,adt*wgv*zCoeff,bDt*w0z*iCoeff,-adCoeff*dt*dissCoeff)
  addCoeff1(cmpu,eqnw,coeff, bDt*w0x*iCoeff)  ! u*w0x in u eqn
  addCoeff1(cmpv,eqnw,coeff, bDt*w0y*iCoeff)  ! v*w0y in u eqn

 #End

 if( pdeModel.eq.BoussinesqModel )then

  ! ---------------------------------------------------------
  ! ---------- add the temperature equation -----------------
  ! ---------------------------------------------------------

  opEvalParametricDerivative(u,qc,qq,1)        ! computes parametric derivatives of T: qqr, qqs 
  getOp(x, u,qc,qq,aj,q0x)       ! q.x
  getOp(y, u,qc,qq,aj,q0y)       ! q.y 
  #If $DIM == 3
   getOp(z, u,qc,qq,aj,q0z)      ! q.z
  #End

 #If $DIM == 2 

  setCoeff5(cmpq,eqnq,coeff,iCoeff,-kDt*lapCoeff,adt*ugv*xCoeff,adt*vgv*yCoeff,-adcBoussinesq*dt*dissCoeff)
  addCoeff1(cmpu,eqnq,coeff, bDt*q0x*iCoeff)  ! u*q0x  
  addCoeff1(cmpv,eqnq,coeff, bDt*q0y*iCoeff)  ! v*q0y 

  addCoeff1(cmpq,eqnu,coeff, teDt*gravity(0)*iCoeff)  ! add buoyancy to u eqn 
  addCoeff1(cmpq,eqnv,coeff, teDt*gravity(1)*iCoeff)  ! add buoyancy to v eqn 

  if( isAxisymmetric.eq.1 )then
    ! add axisymmetric corrections 
    radi = radiusInverse(i1,i2,i3)
    if( radi.ne.0. )then
     addCoeff1(cmpq,eqnq,coeff, -kDt*radi*yCoeff)                        ! -nu*dt*( U_r/r )
    else
     ! corrections on the axis
     getCoeff(yy, yyCoeff,aj)
     addCoeff1(cmpq,eqnq,coeff, -kDt*yyCoeff)                            ! -nu*dt*( U_rr )
    end if
  end if

 #Else

  setCoeff6(cmpq,eqnq,coeff,iCoeff,-kDt*lapCoeff,adt*ugv*xCoeff,adt*vgv*yCoeff,adt*wgv*zCoeff,-adcBoussinesq*dt*dissCoeff)
  addCoeff1(cmpu,eqnq,coeff, bDt*q0x*iCoeff)  ! u*q0x 
  addCoeff1(cmpv,eqnq,coeff, bDt*q0y*iCoeff)  ! v*q0y 
  addCoeff1(cmpw,eqnq,coeff, bDt*q0z*iCoeff)  ! w*q0z 

  addCoeff1(cmpq,eqnu,coeff, teDt*gravity(0)*iCoeff)  ! add buoyancy to u eqn 
  addCoeff1(cmpq,eqnv,coeff, teDt*gravity(1)*iCoeff)  ! add buoyancy to v eqn 
  addCoeff1(cmpq,eqnw,coeff, teDt*gravity(2)*iCoeff)  ! add buoyancy to w eqn 

 #End

 endif  ! pdeModel.eq.BoussinesqModel

 endLoops()
end if 

end if
#endMacro

! ===================================================================================
! Macro to evaluate the RHS for the INS equations
! ===================================================================================
#beginMacro assignRHSPDE()
if( evalRightHandSide.eq.1 .or. evalResidual.eq.1 )then

! **** to do : optimize this for backward-Euler : fe=0, fi=0 !!

! NOTE: For moving grid problems we must eval the RHS as some mask==0 (exposed) points
beginLoopsNoMask()
  ! Evaluate the jacobian derivatives used by the coefficient and forward derivatives:
  ! opEvalJacobianDerivatives(MAXDER) : MAXDER = max number of derivatives to precompute.
  opEvalJacobianDerivatives(aj,1)

  ! evaluate forward derivatives of the current solution: 

  ! First evaluate the parametric derivatives of u, used later by the spatial derivatives. 
  ! MAXDER = max number of parametric derivatives to precompute.
  opEvalParametricDerivative(u,uc,uu,2)
  ! Evaluate the spatial derivatives of u:
  getOp(x, u,uc,uu,aj,u0x)       ! u.x
  getOp(y, u,uc,uu,aj,u0y)       ! u.y

  getOp(xx, u,uc,uu,aj,u0xx)       ! u.xx
  getOp(yy, u,uc,uu,aj,u0yy)       ! u.yy

  ! Evaluate the spatial derivatives of v:
  opEvalParametricDerivative(u,vc,vv,2)
  getOp(x, u,vc,vv,aj,v0x)       ! v.x
  getOp(y, u,vc,vv,aj,v0y)       ! v.y 

  getOp(xx, u,vc,vv,aj,v0xx)       ! v.xx
  getOp(yy, u,vc,vv,aj,v0yy)       ! v.yy

  ! Evaluate the spatial derivatives of p:
  opEvalParametricDerivative(u,pc,pp,1)
  getOp(x, u,pc,pp,aj,p0x)       ! p.x
  getOp(y, u,pc,pp,aj,p0y)       ! p.y 

 #If $DIM == 3

  getOp(z, u,uc,uu,aj,u0z)       ! u.z
  getOp(z, u,vc,vv,aj,v0z)       ! v.z
  getOp(z, u,pc,pp,aj,p0z)       ! p.z

  getOp(zz, u,uc,uu,aj,u0zz)       ! u.zz
  getOp(zz, u,vc,vv,aj,v0zz)       ! v.zz

  opEvalParametricDerivative(u,wc,ww,2)
  getOp(x, u,wc,ww,aj,w0x)       ! w.x
  getOp(y, u,wc,ww,aj,w0y)       ! w.y
  getOp(z, u,wc,ww,aj,w0z)       ! w.z

  getOp(xx, u,wc,ww,aj,w0xx)       ! w.xx
  getOp(yy, u,wc,ww,aj,w0yy)       ! w.yy
  getOp(zz, u,wc,ww,aj,w0zz)       ! w.zz

 #End

 if( evalLinearizedDerivatives.eq.1 )then

  #If $DIM == 2
   opEvalParametricDerivative(ul,uc,uul,1)
   getOp(x, ul,uc,uul,aj,ulx)       ! ul.x
   getOp(y, ul,uc,uul,aj,uly)       ! ul.y
 
   opEvalParametricDerivative(ul,vc,vvl,1)
   getOp(x, ul,vc,vvl,aj,vlx)       ! vl.x
   getOp(y, ul,vc,vvl,aj,vly)       ! vl.y
  #Else
   opEvalParametricDerivative(ul,uc,uul,1)
   getOp(x, ul,uc,uul,aj,ulx)       ! ul.x
   getOp(y, ul,uc,uul,aj,uly)       ! ul.y
   getOp(z, ul,uc,uul,aj,ulz)       ! ul.y
   opEvalParametricDerivative(ul,vc,vvl,1)
   getOp(x, ul,vc,vvl,aj,vlx)       ! vl.x
   getOp(y, ul,vc,vvl,aj,vly)       ! vl.y
   getOp(z, ul,vc,vvl,aj,vlz)       ! vl.y
   opEvalParametricDerivative(ul,wc,wwl,1)
   getOp(x, ul,wc,wwl,aj,wlx)       ! wl.x
   getOp(y, ul,wc,wwl,aj,wly)       ! wl.y
   getOp(z, ul,wc,wwl,aj,wlz)       ! wl.y
  #End
 end if

 if( pdeModel.eq.BoussinesqModel )then
  ! Evaluate the spatial derivatives of q:
  opEvalParametricDerivative(u,qc,qq,2)
  getOp(x, u,qc,qq,aj,q0x)       ! q.x
  getOp(y, u,qc,qq,aj,q0y)       ! q.y 
  getOp(xx,u,qc,qq,aj,q0xx)       ! q.xx
  getOp(yy,u,qc,qq,aj,q0yy)       ! q.yy
  #If $DIM == 3
   getOp(z, u,qc,qq,aj,q0z)       ! q.z  *wdh* 080720
   getOp(zz,u,qc,qq,aj,q0zz)      ! q.zz
  #End
 end if

 if( gridIsMoving.ne.0 )then
   ugv = uu - gv(i1,i2,i3,0)
   vgv = vv - gv(i1,i2,i3,1)
  #If $DIM == 3
   wgv = ww - gv(i1,i3,i3,2)
  #End
 else
   ugv = uu
   vgv = vv
  #If $DIM == 3
   wgv = ww
  #End
 end if

 ! eval the nonlinear coeff. of artificial dissipation: 
 getArtificialDissipationCoeff(adCoeff, u0x,u0y,u0z, v0x,v0y,v0z, w0x,w0y,w0z )

 #If $DIM == 2
 if( evalRightHandSide.eq.1 )then


  if( gridIsImplicit.eq.0 )then

    ! ********** explicit *********
    fe(i1,i2,i3,uc) = -ugv*u0x -vgv*u0y - p0x + nu*(u0xx+u0yy) + adCoeff*uDiss22(u,uc)
    fe(i1,i2,i3,vc) = -ugv*v0x -vgv*v0y - p0y + nu*(v0xx+v0yy) + adCoeff*uDiss22(u,vc)

    if( isAxisymmetric.eq.1 )then
      radi=radiusInverse(i1,i2,i3)
      if( radi.ne.0. )then
        fe(i1,i2,i3,uc) = fe(i1,i2,i3,uc) + nu*( u0y*radi )            ! add u_r/r 
        fe(i1,i2,i3,vc) = fe(i1,i2,i3,vc) + nu*( (v0y -vv*radi)*radi ) ! add v_r/r - v/r^2 
      else
        fe(i1,i2,i3,uc) = fe(i1,i2,i3,uc) + nu*( u0yy )    ! add u_rr 
        fe(i1,i2,i3,vc) = fe(i1,i2,i3,vc) + nu*( .5*v0yy ) ! add .5*vrr
      end if
    end if

    if( pdeModel.eq.BoussinesqModel )then

     fe(i1,i2,i3,uc) = fe(i1,i2,i3,uc) -thermalExpansivity*gravity(0)*qq 
     fe(i1,i2,i3,vc) = fe(i1,i2,i3,vc) -thermalExpansivity*gravity(1)*qq 
     fe(i1,i2,i3,qc) = -ugv*q0x -vgv*q0y + kThermal*(q0xx+q0yy) + adcBoussinesq*uDiss22(u,qc)
      if( isAxisymmetric.eq.1 )then
        radi=radiusInverse(i1,i2,i3)
        if( radi.ne.0. )then
          fe(i1,i2,i3,qc) = fe(i1,i2,i3,qc) + kThermal*( q0y*radi )            ! add u_r/r 
        else
          fe(i1,i2,i3,qc) = fe(i1,i2,i3,qc) + kThermal*( q0yy )    ! add u_rr 
        end if
      end if

    end if

  else 
    ! ********** implicit *********

    ! eval the nonlinear coeff. of artificial dissipation for the linearized solution:
    getArtificialDissipationCoeff(adCoeffl, ulx,uly,ulz, vlx,vly,vlz, wlx,wly,wlz )

    ! implicit method -- compute explicit part
    fe(i1,i2,i3,uc) = -ugv*u0x -vgv*u0y - p0x + (adCoeff-adCoeffl)*uDiss22(u,uc)
    fe(i1,i2,i3,vc) = -ugv*v0x -vgv*v0y - p0y + (adCoeff-adCoeffl)*uDiss22(u,vc)
    if( nonlinearTermsAreImplicit.eq.1 )then
      ! include linearized terms u0*ulx + ul*u0x 

      if( gridIsMoving.ne.0 )then
       ugvl = uul - gvl(i1,i2,i3,0)
       vgvl = vvl - gvl(i1,i2,i3,1)
      else
       ugvl = uul
       vgvl = vvl
      end if

      ! ulterm = uul*u0x + vvl*u0y + bImp*(uu*ulx +  vv*uly)
      ! vlterm = uul*v0x + vvl*v0y + bImp*(uu*vlx +  vv*vly)

      ! moving: (u-gv)*ux ->  u*ux - gv*ux -> ul*ux +u*ulx +gvl*ux = (ul-gvl)*ux + ulx*u 

      ulterm = ugvl*u0x + vgvl*u0y + bImp*(uu*ulx +  vv*uly)
      vlterm = ugvl*v0x + vgvl*v0y + bImp*(uu*vlx +  vv*vly)

      ! ++ ulterm = uul*u0x + vvl*u0y + bImp*(uu*ulx )
      ! ++ vlterm = uul*v0x + vvl*v0y + bImp*(vv*vly)
      ! ++ ulterm = uul*u0x + vvl*u0y + bImp*( vv )
      ! ++ ulterm = uul*u0x + vvl*u0y 
      ! ++ vlterm = uul*v0x + vvl*v0y 

      fe(i1,i2,i3,uc) = fe(i1,i2,i3,uc) + ulterm
      fe(i1,i2,i3,vc) = fe(i1,i2,i3,vc) + vlterm
    end if

    if( implicitOption.eq.computeImplicitTermsSeparately )then
      ! implicit method -- compute implicit part 
      fi(i1,i2,i3,uc) = nuImp*(u0xx+u0yy)+ adCoeffl*uDiss22(u,uc)   !  I think nuImp==nu now
      fi(i1,i2,i3,vc) = nuImp*(v0xx+v0yy)+ adCoeffl*uDiss22(u,vc) 
      if( nonlinearTermsAreImplicit.eq.1 )then
        ! include linearized terms u0*ulx + ul*u0x 
        fi(i1,i2,i3,uc) = fi(i1,i2,i3,uc) - aImp*( ulterm )
        fi(i1,i2,i3,vc) = fi(i1,i2,i3,vc) - aImp*( vlterm )
      end if
      if( isAxisymmetric.eq.1 )then
       radi=radiusInverse(i1,i2,i3)
       if( radi.ne.0. )then
         fi(i1,i2,i3,uc) = fi(i1,i2,i3,uc) + nuImp*( u0y*radi )            ! add u_r/r 
         fi(i1,i2,i3,vc) = fi(i1,i2,i3,vc) + nuImp*( (v0y -vv*radi)*radi ) ! add v_r/r - v/r^2 
       else
         fi(i1,i2,i3,uc) = fi(i1,i2,i3,uc) + nuImp*( u0yy )    ! add u_rr 
         fi(i1,i2,i3,vc) = fi(i1,i2,i3,vc) + nuImp*( .5*v0yy ) ! add .5*vrr
       end if
      end if

    end if

    if( pdeModel.eq.BoussinesqModel )then

     ! tImp=1 if the buoyancy term is implicit, 0 if explicit
     fe(i1,i2,i3,uc) = fe(i1,i2,i3,uc) -(1.-tImp)*thermalExpansivity*gravity(0)*qq 
     fe(i1,i2,i3,vc) = fe(i1,i2,i3,vc) -(1.-tImp)*thermalExpansivity*gravity(1)*qq 
     fe(i1,i2,i3,qc) = -ugv*q0x -vgv*q0y 
 
     if( nonlinearTermsAreImplicit.eq.1 )then
       ! include linearized terms u0*ulx + ul*u0x 
 
       opEvalParametricDerivative(ul,qc,qql,1)
       getOp(x, ul,qc,qql,aj,qlx)       ! ql.x
       getOp(y, ul,qc,qql,aj,qly)       ! ql.y
 
       qlterm = ugvl*q0x + vgvl*q0y + bImp*(uu*qlx +  vv*qly)
       fe(i1,i2,i3,qc) = fe(i1,i2,i3,qc) + qlterm
     end if
 
     if( implicitOption.eq.computeImplicitTermsSeparately )then
       ! implicit method -- compute implicit part 
       fi(i1,i2,i3,uc) = fi(i1,i2,i3,uc) -tImp*thermalExpansivity*gravity(0)*qq 
       fi(i1,i2,i3,vc) = fi(i1,i2,i3,vc) -tImp*thermalExpansivity*gravity(1)*qq 

       fi(i1,i2,i3,qc) = kThermal*(q0xx+q0yy) + adcBoussinesq*uDiss22(u,qc)  
       if( nonlinearTermsAreImplicit.eq.1 )then
         ! include linearized terms u0*ulx + ul*u0x 
         fi(i1,i2,i3,qc) = fi(i1,i2,i3,qc) - aImp*( qlterm )
       end if
       if( isAxisymmetric.eq.1 )then
        radi=radiusInverse(i1,i2,i3)
        if( radi.ne.0. )then
          fi(i1,i2,i3,qc) = fi(i1,i2,i3,qc) + kThermal*( q0y*radi ) 
        else
          fi(i1,i2,i3,qc) = fi(i1,i2,i3,qc) + kThermal*( q0yy )    
        end if
       end if
     end if

    end if

  end if

 end if

 if( evalResidual.eq.1 )then
   ! residual in 2D: (NOTE: currently ul is not available when evaluating the residual)
   fe(i1,i2,i3,uc) = fi(i1,i2,i3,uc) - ugv*u0x -vgv*u0y - p0x + nu*(u0xx+u0yy)+ adCoeff*uDiss22(u,uc)
   fe(i1,i2,i3,vc) = fi(i1,i2,i3,vc) - ugv*v0x -vgv*v0y - p0y + nu*(v0xx+v0yy)+ adCoeff*uDiss22(u,vc)
   if( isAxisymmetric.eq.1 )then
     radi=radiusInverse(i1,i2,i3)
     if( radi.ne.0. )then
       fe(i1,i2,i3,uc) = fe(i1,i2,i3,uc) + nu*( u0y*radi )            ! add u_r/r 
       fe(i1,i2,i3,vc) = fe(i1,i2,i3,vc) + nu*( (v0y -vv*radi)*radi ) ! add v_r/r - v/r^2 
     else
       fe(i1,i2,i3,uc) = fe(i1,i2,i3,uc) + nu*( u0yy )    ! add u_rr 
       fe(i1,i2,i3,vc) = fe(i1,i2,i3,vc) + nu*( .5*v0yy ) ! add .5*vrr
     end if
   end if
   if( pdeModel.eq.BoussinesqModel )then

    fe(i1,i2,i3,uc) = fe(i1,i2,i3,uc) -thermalExpansivity*gravity(0)*qq 
    fe(i1,i2,i3,vc) = fe(i1,i2,i3,vc) -thermalExpansivity*gravity(1)*qq 
    fe(i1,i2,i3,qc) = fi(i1,i2,i3,qc) - ugv*q0x -vgv*q0y  +kThermal*(q0xx+q0yy) + adcBoussinesq*uDiss22(u,qc)
    if( isAxisymmetric.eq.1 )then
     radi=radiusInverse(i1,i2,i3)
     if( radi.ne.0. )then
       fe(i1,i2,i3,qc) = fe(i1,i2,i3,qc) + kThermal*( q0y*radi )  ! add u_r/r 
     else
       fe(i1,i2,i3,qc) = fe(i1,i2,i3,qc) + kThermal*( q0yy )      ! add u_rr 
     end if
    end if
   end if
 end if

 #Else

  ! ---- 3D -----

 if( evalRightHandSide.eq.1 )then


  if( gridIsImplicit.eq.0 )then
    ! explicit
    fe(i1,i2,i3,uc) = -ugv*u0x -vgv*u0y -wgv*u0z - p0x+ nu*(u0xx+u0yy+u0zz) + adCoeff*uDiss23(u,uc)
    fe(i1,i2,i3,vc) = -ugv*v0x -vgv*v0y -wgv*v0z - p0y+ nu*(v0xx+v0yy+v0zz) + adCoeff*uDiss23(u,vc)
    fe(i1,i2,i3,wc) = -ugv*w0x -vgv*w0y -wgv*w0z - p0z+ nu*(w0xx+w0yy+w0zz) + adCoeff*uDiss23(u,wc) 

    if( pdeModel.eq.BoussinesqModel )then
     fe(i1,i2,i3,uc) = fe(i1,i2,i3,uc) -thermalExpansivity*gravity(0)*qq
     fe(i1,i2,i3,vc) = fe(i1,i2,i3,vc) -thermalExpansivity*gravity(1)*qq
     fe(i1,i2,i3,wc) = fe(i1,i2,i3,wc) -thermalExpansivity*gravity(2)*qq
     fe(i1,i2,i3,qc) = -ugv*q0x -vgv*q0y -wgv*q0z + kThermal*(q0xx+q0yy+q0zz) + adcBoussinesq*uDiss23(u,qc)
    end if
  else 

    ! eval the nonlinear coeff. of artificial dissipation for the linearized solution:
    getArtificialDissipationCoeff(adCoeffl, ulx,uly,ulz, vlx,vly,vlz, wlx,wly,wlz )

    ! implicit method -- compute explicit part
    fe(i1,i2,i3,uc) = -ugv*u0x -vgv*u0y -wgv*u0z - p0x + (adCoeff-adCoeffl)*uDiss23(u,uc)
    fe(i1,i2,i3,vc) = -ugv*v0x -vgv*v0y -wgv*v0z - p0y + (adCoeff-adCoeffl)*uDiss23(u,vc)
    fe(i1,i2,i3,wc) = -ugv*w0x -vgv*w0y -wgv*w0z - p0z + (adCoeff-adCoeffl)*uDiss23(u,wc)

    if( nonlinearTermsAreImplicit.eq.1 )then
      ! include linearized terms u0*ulx + ul*u0x + ...
      if( gridIsMoving.ne.0 )then
       ugvl = uul - gvl(i1,i2,i3,0)
       vgvl = vvl - gvl(i1,i2,i3,1)
       wgvl = wwl - gvl(i1,i3,i3,2)
      else
       ugvl = uul
       vgvl = vvl
       wgvl = wwl
      end if
      ! ulterm = uul*u0x + vvl*u0y + bImp*(uu*ulx +  vv*uly)
      ulterm = ugvl*u0x + vgvl*u0y + wgvl*u0z + bImp*(uu*ulx + vv*uly + ww*ulz)
      vlterm = ugvl*v0x + vgvl*v0y + wgvl*v0z + bImp*(uu*vlx + vv*vly + ww*vlz)
      wlterm = ugvl*w0x + vgvl*w0y + wgvl*w0z + bImp*(uu*wlx + vv*wly + ww*wlz)

      fe(i1,i2,i3,uc) = fe(i1,i2,i3,uc) + ulterm
      fe(i1,i2,i3,vc) = fe(i1,i2,i3,vc) + vlterm
      fe(i1,i2,i3,wc) = fe(i1,i2,i3,wc) + wlterm
    end if

    if( implicitOption.eq.computeImplicitTermsSeparately )then
      ! implicit method -- compute implicit part 
      fi(i1,i2,i3,uc) = nuImp*(u0xx+u0yy+u0zz) + adCoeffl*uDiss23(u,uc)
      fi(i1,i2,i3,vc) = nuImp*(v0xx+v0yy+v0zz) + adCoeffl*uDiss23(u,vc) 
      fi(i1,i2,i3,wc) = nuImp*(w0xx+w0yy+w0zz) + adCoeffl*uDiss23(u,wc) 
      if( nonlinearTermsAreImplicit.eq.1 )then
        ! include linearized terms u0*ulx + ul*u0x 
        fi(i1,i2,i3,uc) = fi(i1,i2,i3,uc) - aImp*( ulterm )
        fi(i1,i2,i3,vc) = fi(i1,i2,i3,vc) - aImp*( vlterm )
        fi(i1,i2,i3,wc) = fi(i1,i2,i3,wc) - aImp*( wlterm )
      end if
    end if

    ! add Boussinesq terms 
    if( pdeModel.eq.BoussinesqModel )then   
     fe(i1,i2,i3,uc) = fe(i1,i2,i3,uc) -(1.-tImp)*thermalExpansivity*gravity(0)*qq 
     fe(i1,i2,i3,vc) = fe(i1,i2,i3,vc) -(1.-tImp)*thermalExpansivity*gravity(1)*qq 
     fe(i1,i2,i3,wc) = fe(i1,i2,i3,wc) -(1.-tImp)*thermalExpansivity*gravity(2)*qq 
     fe(i1,i2,i3,qc) = -ugv*q0x -vgv*q0y -wgv*q0z 

     if( nonlinearTermsAreImplicit.eq.1 )then
       opEvalParametricDerivative(ul,qc,qql,1)
       getOp(x, ul,qc,qql,aj,qlx)       ! ql.x
       getOp(y, ul,qc,qql,aj,qly)       ! ql.y
       getOp(z, ul,qc,qql,aj,qlz)       ! ql.z
       qlterm = ugvl*q0x + vgvl*q0y + wgvl*q0z + bImp*(uu*qlx + vv*qly+ ww*qlz)
       fe(i1,i2,i3,qc) = fe(i1,i2,i3,qc) + qlterm
     end if

     if( implicitOption.eq.computeImplicitTermsSeparately )then
      ! implicit method -- compute implicit part 
      fi(i1,i2,i3,uc) = fi(i1,i2,i3,uc) -tImp*thermalExpansivity*gravity(0)*qq 
      fi(i1,i2,i3,vc) = fi(i1,i2,i3,vc) -tImp*thermalExpansivity*gravity(1)*qq 
      fi(i1,i2,i3,wc) = fi(i1,i2,i3,wc) -tImp*thermalExpansivity*gravity(2)*qq 
      fi(i1,i2,i3,qc) = kThermal*(q0xx+q0yy+q0zz) + adcBoussinesq*uDiss23(u,qc)
      if( nonlinearTermsAreImplicit.eq.1 )then
        fi(i1,i2,i3,qc) = fi(i1,i2,i3,qc) - aImp*( qlterm )
      end if
     end if

    end if


  end if
 end if
 if( evalResidual.eq.1 )then
   ! residual in 3D: (NOTE: currently ul is not available when evaluating the residual)
   fe(i1,i2,i3,uc) = fi(i1,i2,i3,uc) -ugv*u0x -vgv*u0y -wgv*u0z - p0x + nu*(u0xx+u0yy+u0zz) + adCoeff*uDiss23(u,uc) 
   fe(i1,i2,i3,vc) = fi(i1,i2,i3,vc) -ugv*v0x -vgv*v0y -wgv*v0z - p0y + nu*(v0xx+v0yy+v0zz) + adCoeff*uDiss23(u,vc)
   fe(i1,i2,i3,wc) = fi(i1,i2,i3,wc) -ugv*w0x -vgv*w0y -wgv*w0z - p0z + nu*(w0xx+w0yy+w0zz) + adCoeff*uDiss23(u,wc)  
   if( pdeModel.eq.BoussinesqModel )then  
    fe(i1,i2,i3,uc) = fe(i1,i2,i3,uc) -thermalExpansivity*gravity(0)*qq
    fe(i1,i2,i3,vc) = fe(i1,i2,i3,vc) -thermalExpansivity*gravity(1)*qq 
    fe(i1,i2,i3,wc) = fe(i1,i2,i3,wc) -thermalExpansivity*gravity(2)*qq 
    fe(i1,i2,i3,qc) = fi(i1,i2,i3,qc) -ugv*q0x -vgv*q0y -wgv*q0z +kThermal*(q0xx+q0yy+q0zz) + adcBoussinesq*uDiss23(u,qc)
   end if
 end if

 #End
endLoopsNoMask()
end if
#endMacro


! ===============================================================================================
!  Fill in the matrix BCs on a face for the INS equations 
! ===============================================================================================
#beginMacro fillMatrixBoundaryConditionsPDE()

if( fillCoefficients.eq.1 )then

if( fillCoefficientsScalarSystem.ne.fillCoeffT )then

  ! ---- fill BC coeffs for (u,v,w) ---

  bc0 = bc(side,axis)
  if( bc0.eq.slipWall .and. fillCoefficientsScalarSystem.gt.0 )then
    if( fillCoefficientsScalarSystem.eq.(axis+1) )then
      bc0=dirichletBoundaryCondition
    else
      bc0=neumannBoundaryCondition
    end if
  end if

if( bc0.eq.dirichletBoundaryCondition .or.\
    bc0.eq.noSlipWall .or.\
    bc0.eq.inflowWithVelocityGiven .or.\
    bc0.eq.interfaceBoundaryCondition )then
  
 ! Dirichlet BC
 beginLoopsMixedBoundary()

  ! zero out equations for u,v, [w]    ** but not the T eqn since these may be a Neumann BC ! **
  zeroMatrixCoefficients( coeff,eqnu,eqnu+ndu-1, i1,i2,i3 )

  ! Evaluate the jacobian derivatives used by the coefficient and forward derivatives:
  ! opEvalJacobianDerivatives(MAXDER) : MAXDER = max number of derivatives to precompute.
!   opEvalJacobianDerivatives(aj,1)

  ! evaluate the coeff operators 
  getCoeff(identity, iCoeff,aj)
!   getCoeff(x, xCoeff,aj)
!   getCoeff(y, yCoeff,aj)
  #If $DIM == 3
!    getCoeff(z, zCoeff,aj)
  #End

  do n=0,ndu-1
   setCoeff5(cmpu+n,eqnu+n,coeff,iCoeff,,,,)
  end do 

 endLoops()

else if( bc0.eq.outflow )then

 if( outflowOption.eq.1 )then
  ! Neumann BC at outflow if outflowOption==1
  beginLoopsMixedBoundary()
   getNormalForCurvilinearGrid(side,axis,i1,i2,i3)
   do n=0,ndu-1
     fillMatrixNeumann(coeff, cmpu+n,eqnu+n, i1,i2,i3, an,0.,1. )
   end do
  endLoops()
 end if

else if( bc0.eq.neumannBoundaryCondition )then

  ! Neumann BC (used by slipWall and scalar systems)
  beginLoopsMixedBoundary()
   getNormalForCurvilinearGrid(side,axis,i1,i2,i3)
   do n=0,ndu-1
     fillMatrixNeumann(coeff, cmpu+n,eqnu+n, i1,i2,i3, an,0.,1. )
   end do
  endLoops()

else if( bc0.eq.slipWall )then

  ! SLIP-WALL

  ! NOTE: Here we assume the matrix already includes the interior equations on the boundary 

  ! NOTE: what about corners ???

  ! boundary values use:
  !    n.u = f
  !    tau.(Lu) = tau.g   (tangential component of the equations on the boundary
  ! To avoid a zero pivot we combine the above equations as
  !     (n.u) n + ( tau.(Lu) ) tau = n f + tau g 
  !
  ! OR:  ( tau.(Lu) ) tau = Lu - (n.(Lu)) n 
  !    (n.u) n + Lu - (n.(Lu)) n = n f +  g - (n.g) n 
  !       
 if( fillCoefficientsScalarSystem.ge.fillCoeffU .and. fillCoefficientsScalarSystem.le.fillCoeffW )then
   write(*,'(" insImpINS: slipWall BC not finished for scalar systems")')
   stop 8130
 end if

 getCoeff(identity, iCoeff,aj)

 beginLoopsMixedBoundary()

  getNormalForCurvilinearGrid(side,axis,i1,i2,i3)

  beginMatrixLoops(m1,m2,m3)

  ! Form the matrix for "n.u"  -->  nDot(mc3(m1,m2,m3,c)) = iCoeff(m1,m2,m3)*an(c-cmpu)
  !nDot=0
  !nDot(mc3(0,0,0,cmpu))=an(0)
  !nDot(mc3(0,0,0,cmpv))=an(1)
  !nDot(mc3(0,0,0,cmpw))=an(2)

  ! Form the matrix for "n.Lu"
#If $DIM == 2 
  nDotL(0)=an(0)*coeff(mce2(m1,m2,m3,cmpu,eqnu),i1,i2,i3)+\
           an(1)*coeff(mce2(m1,m2,m3,cmpu,eqnv),i1,i2,i3)
  nDotL(1)=an(0)*coeff(mce2(m1,m2,m3,cmpv,eqnu),i1,i2,i3)+\
           an(1)*coeff(mce2(m1,m2,m3,cmpv,eqnv),i1,i2,i3)
#Else
  nDotL(0)=an(0)*coeff(mce3(m1,m2,m3,cmpu,eqnu),i1,i2,i3)+\
           an(1)*coeff(mce3(m1,m2,m3,cmpu,eqnv),i1,i2,i3)+\
           an(2)*coeff(mce3(m1,m2,m3,cmpu,eqnw),i1,i2,i3)
  nDotL(1)=an(0)*coeff(mce3(m1,m2,m3,cmpv,eqnu),i1,i2,i3)+\
           an(1)*coeff(mce3(m1,m2,m3,cmpv,eqnv),i1,i2,i3)+\
           an(2)*coeff(mce3(m1,m2,m3,cmpv,eqnw),i1,i2,i3)
  nDotL(2)=an(0)*coeff(mce3(m1,m2,m3,cmpw,eqnu),i1,i2,i3)+\
           an(1)*coeff(mce3(m1,m2,m3,cmpw,eqnv),i1,i2,i3)+\
           an(2)*coeff(mce3(m1,m2,m3,cmpw,eqnw),i1,i2,i3)
#End

  ! form the matrix for  Lu + [ (n.u) - (n.(Lu)) ] n 

  !  eqnu:  (n1*u1+n2*u2+n3*u3)*n1 + L1(u) - nDotL*n1    
  !  eqnv:  (n1*u1+n2*u2+n3*u3)*n2 + L2(u) - nDotL*n2
  !  eqnw:  (n1*u1+n2*u2+n3*u3)*n3 + L3(u) - nDotL*n3
  !  nDotL = L1(u)*n1 + L2(u)*n2 + L3(u)*n3 
  do e=eqnu,eqnu+nd-1
  do c=cmpu,cmpu+nd-1
   coeff(MCE(m1,m2,m3,c,e),i1,i2,i3)= coeff(MCE(m1,m2,m3,c,e),i1,i2,i3) + \
       an(e-eqnu)*(iCoeff(MA(m1,m2,m3))*an(c-cmpu)-nDotL(c-cmpu))
  end do
  end do

 endMatrixLoops()

 ! fill ghost pt eqn's with a vector symmetry condition:
 fillMatrixVectorSymmetry(coeff, cmpu,eqnu, i1,i2,i3, an )

 endLoops()

else if( bc0.eq.axisymmetric )then

 if( fillCoefficientsScalarSystem.eq.fillCoeffU )then

  ! BC on an axisymmetric side : scalar matrix for U 
  beginLoopsMixedBoundary()
 
   ! Evaluate the jacobian derivatives used by the coefficient and forward derivatives:
   ! opEvalJacobianDerivatives(MAXDER) : MAXDER = max number of derivatives to precompute.
   opEvalJacobianDerivatives(aj,1)

   ! getCoeff(identity, iCoeff,aj)
   !zeroMatrixCoefficients( coeff,eqnv,eqnv, i1,i2,i3 )  ! set v eqn coeffs to zero
   !setCoeff1(cmpv,eqnv,coeff,iCoeff)                ! dirichlet: V= 
   !fillMatrixExtrapolation(coeff,cmpv,eqnv,i1,i2,i3,orderOfExtrapolation,1)  ! extrap ghost for V 
 
   getNormalForCurvilinearGrid(side,axis,i1,i2,i3)
   fillMatrixNeumann(coeff, cmpu,eqnu, i1,i2,i3, an,0.,1. )  ! Neuman: U_r = 
 
  endLoops()

 else if( fillCoefficientsScalarSystem.eq.fillCoeffV )then

  ! BC on an axisymmetric side : scalar matrix for V 
  beginLoopsMixedBoundary()
 
   ! Evaluate the jacobian derivatives used by the coefficient and forward derivatives:
   ! opEvalJacobianDerivatives(MAXDER) : MAXDER = max number of derivatives to precompute.
   opEvalJacobianDerivatives(aj,1)
   getCoeff(identity, iCoeff,aj)
 
   zeroMatrixCoefficients( coeff,eqnv,eqnv, i1,i2,i3 )  ! set v eqn coeffs to zero
   setCoeff1(cmpv,eqnv,coeff,iCoeff)                ! dirichlet: V= 
   fillMatrixExtrapolation(coeff,cmpv,eqnv,i1,i2,i3,orderOfExtrapolation,1)  ! extrap ghost for V 
 
   !getNormalForCurvilinearGrid(side,axis,i1,i2,i3)
   !fillMatrixNeumann(coeff, cmpu,eqnu, i1,i2,i3, an,0.,1. )  ! Neuman: U_r = 
 
  endLoops()

 else

  ! BC on an axisymmetric side  
  beginLoopsMixedBoundary()
 
   ! Evaluate the jacobian derivatives used by the coefficient and forward derivatives:
   ! opEvalJacobianDerivatives(MAXDER) : MAXDER = max number of derivatives to precompute.
   opEvalJacobianDerivatives(aj,1)
   getCoeff(identity, iCoeff,aj)
 
   zeroMatrixCoefficients( coeff,eqnv,eqnv, i1,i2,i3 )  ! set v eqn coeffs to zero
   setCoeff1(cmpv,eqnv,coeff,iCoeff)                ! dirichlet: V= 
   fillMatrixExtrapolation(coeff,cmpv,eqnv,i1,i2,i3,orderOfExtrapolation,1)  ! extrap ghost for V 
 
   getNormalForCurvilinearGrid(side,axis,i1,i2,i3)
   fillMatrixNeumann(coeff, cmpu,eqnu, i1,i2,i3, an,0.,1. )  ! Neuman: U_r = 
 
  endLoops()
 end if
else if( bc0.gt.0 )then
  write(*,'("insimp:BC: ERROR unknown bc=",i4)') bc0
  stop 5501
end if

if( bc0.eq.dirichletBoundaryCondition .or.\
    bc0.eq.noSlipWall.or.\
    bc0.eq.inflowWithVelocityGiven .or.\
    (bc0.eq.outflow .and. outflowOption.eq.0) .or.\
    bc0.eq.interfaceBoundaryCondition )then

 ! === extrapolation ===

 orderOfExtrap=orderOfExtrapolation
 ! write(*,'("insimp:BC: orderOfExtrapolationForOutflow=",i4)') orderOfExtrapolationForOutflow
 if( bc0.eq.outflow .and. orderOfExtrapolationForOutflow.gt.0 )then
   orderOfExtrap=orderOfExtrapolationForOutflow
 end if
 if( orderOfExtrap.lt.1 .or. orderOfExtrap.gt.maxOrderOfExtrapolation )then
  write(*,'("insimp:BC:ERROR: requesting orderOfExtrap=",i6)') orderOfExtrap
  stop 5502
 end if
 beginLoopsMixedBoundary()

  ! i1m=i1-is1  ! ghost point
  ! i2m=i2-is2
  ! i3m=i3-is3
  ! do m=0,ndc-1
  !  coeff(m,i1m,i2m,i3m)=0.  ! init all elements to zero
  ! end do

  do n=0,ndu-1
   c=cmpu+n
   e=eqnu+n
   fillMatrixExtrapolation(coeff,c,e,i1,i2,i3,orderOfExtrap,1)
  end do

 endLoops()
end if  
end if


if( pdeModel.eq.BoussinesqModel .and. (fillCoefficientsScalarSystem.eq.0 .or. fillCoefficientsScalarSystem.eq.fillCoeffT) )then  
 ! ----------------------------
 ! --- Assign the BCs for T --- (this is duplicated from insImpVP.bf : we could share 
 ! ----------------------------
 bc0 = bc(side,axis)

 a0 = mixedCoeff(tc,side,axis,grid)
 a1 = mixedNormalCoeff(tc,side,axis,grid)
 !write(*,'(" insImpINS: T BC: bc=",i3," (a0,a1)=(",f5.1,",",f5.1,") for side,axis,grid=",3i3)') \
 !   bc0,a0,a1,side,axis,grid
 ! '
 if( bc0.eq.dirichletBoundaryCondition .or.\
     bc0.eq.noSlipWall .or.\
     bc0.eq.slipWall .or.\
     bc0.eq.inflowWithVelocityGiven .or.\
     bc0.eq.outflow .or.\
     bc0.eq.axisymmetric .or.\
     bc0.eq.interfaceBoundaryCondition )then

   if( bc0.eq.outflow .or. bc0.eq.axisymmetric )then
    ! outflow is Neumann
    a0=0.
    a1=1.
   end if

  if( a1.ne.0. )then
   ! Mixed BC 
   beginLoopsMixedBoundary()
    getNormalForCurvilinearGrid(side,axis,i1,i2,i3)
    fillMatrixNeumann(coeff, cmpq,eqnq, i1,i2,i3, an,a0,a1 )
   endLoops()
  else
   ! Dirichlet + extrap ghost line values


   getCoeff(identity, iCoeff,aj)
   beginLoopsMixedBoundary()
    zeroMatrixCoefficients( coeff,eqnq,eqnq,i1,i2,i3 )
    setCoeff5(cmpq,eqnq,coeff,a0*iCoeff,,,,)
    fillMatrixExtrapolation(coeff,cmpq,eqnq,i1,i2,i3,orderOfExtrap,1)
   endLoops()
  end if

!  else if( bc0.eq.outflow )then
!    ! === extrapolation ===
!   beginLoopsMixedBoundary()
!    fillMatrixExtrapolation(coeff,cmpq,eqnq,i1,i2,i3,orderOfExtrap,1)
!   endLoops()

 else if( bc0.gt.0 )then
  write(*,'("insImpINS:T: ERROR unknown bc=",i4)') bc0 
  stop 9167
 end if 
endif

end if ! end if( fillCoefficients.eq.1 )
#endMacro


! ===============================================================================================
!  Compute the residual for BCs
! ===============================================================================================
#beginMacro getBoundaryResidualPDE()
if( evalResidualForBoundaryConditions.eq.1 )then

if( bc(side,axis).eq.dirichletBoundaryCondition .or.\
    bc(side,axis).eq.noSlipWall.or.\
    bc(side,axis).eq.inflowWithVelocityGiven.or.\
    bc(side,axis).eq.interfaceBoundaryCondition )then
  
 ! Dirichlet BC
 beginLoops()

  ! do n=0,ntdc-1             ! ntdc = number of time depenedent components
  do n=0,nd-1  ! ntdc = number of time depenedent components
   ! fe(i1,i2,i3,uc+n)=0.  
   fe(i1,i2,i3,uc+n)=fi(i1,i2,i3,uc+n)-u(i1,i2,i3,uc+n)
  end do 

 endLoops()

else if( bc(side,axis).eq.outflow )then

else if( bc(side,axis).eq.axisymmetric )then

! do nothing for now


else if( bc(side,axis).eq.slipWall )then
  ! SLIP-WALL
 
  ! boundary values use:
  !    n.u = f
  !    tau.(Lu) = tau.g   (tangential component of the equations on the boundary
  ! To avoid a zero pivot we combine the above equations as
  !     (n.u) n + ( tau.(Lu) ) tau = n f + tau g 
  !
  ! OR:  ( tau.(Lu) ) tau = Lu - (n.(Lu)) n 
  !    (n.u) n + Lu - (n.(Lu)) n = n f +  g - (n.g) n 
  !       

 beginLoops()

  do n=0,nd-1
   fe(i1,i2,i3,uc+n)=0.  ! *** do this for now 
  end do 

 endLoops()

else if( bc(side,axis).gt.0 )then

  write(*,'("insimp:residual:BC: ERROR unknown bc=",i4)') bc(side,axis)

end if

end if ! end if( evalResidualForBoundaryConditions.eq.1 )
#endMacro


! ********* Here we now define the suboutine insImpINS ******************
INSIMP(insImpINS,INS)
