// ==================================================================================
//   cgmp : main program for the multi-phsyics solver
// ==================================================================================

#include "Cgmp.h"
#include "PlotStuff.h"
#include "Ogshow.h"
#include "ParallelUtility.h"
#include "display.h"

#include "Cgad.h"
#include "Cgasf.h"
#include "Cgcns.h"
#include "Cgins.h"
#include "Cgsm.h"

#include "CgSolverUtil.h"

//\begin{>>CgmpInclude.tex}{\subsection{getModelInfo}} 
int Cgmp::
getModelInfo( std::vector<aString> & modelName )
// ========================================================================================================
// /Description:
//    This function is used to provide Cgmp with information about the 'models' (i.e. PDE solver classes that
//   have been derived from the class DomainSolver).
//
//  /modelName (output) : an array of names of the models. 
// 
//\end{CgmpInclude.tex} 
// ========================================================================================================
{
  modelName.push_back("Cgins");
  modelName.push_back("Cgcns");
  modelName.push_back("Cgad"); 
  modelName.push_back("Cgasf"); 
  modelName.push_back("Cgsm"); 

  return 0;
}

//\begin{>>CgmpInclude.tex}{\subsection{buildModel}}
DomainSolver* Cgmp::
buildModel( const aString & modelName, 
            CompositeGrid & cg, GenericGraphicsInterface *ps /* =NULL */, Ogshow *show /* =NULL */, const int & plotOption /* =1 */ )
// ========================================================================================================
// /Description:
//    This function acts as a virtual constructor to 'new' a copy of a DomainSolver model.
//
//  /modelName (input) : build an model with this name.
//  /cg,ps,show,plotOption (input) : parameters that are passed to the DomainSolver constructor. cg will be the CompositeGrid
//      that corresponds to the domain over-which the model is valid. 
//
//  /Return value: a pointer to the model. 
// 
//\end{CgmpInclude.tex} 
// ========================================================================================================
{
  DomainSolver *solver=NULL;
  if( modelName=="Cgasf")
  {
    solver = new Cgasf(cg,ps,show,plotOption);
  }
  else if( modelName=="Cgins")
  {
    solver = new Cgins(cg,ps,show,plotOption);
  }
  else if( modelName=="Cgcns")
  {
    solver = new Cgcns(cg,ps,show,plotOption);
  }
  else if( modelName=="Cgad" )
  {
    solver = new Cgad(cg,ps,show,plotOption);
  }
  else if( modelName=="Cgsm" )
  {
    solver = new Cgsm(cg,ps,show,plotOption);
  }
  else
  {
    printF("Cgmp:buildModel:ERROR: unknown modelName=[%s]\n",(const char*)modelName);
    Overture::abort("Cgmp:buildModel:ERROR: unknown modelName");
  }

  // Tell the domain solver that this is a multi-domain problem.
  solver->parameters.dbase.get<int>("multiDomainProblem")=1;
  
  return solver;
}



int
main(int argc, char *argv[])
{
  Overture::start(argc,argv);  // initialize Overture and A++/P++
  // Optimization_Manager::setForceVSG_Update(Off);
  const int myid=Communication_Manager::My_Process_Number;

  int plotOption=true;
  bool smartRelease=false;
  bool reportMemory=false;
  bool loadBalance=false;
  
  aString commandFileName="";
  if( argc > 1 )
  { // look at arguments for "noplot" or some other name
    aString line;
    for( int i=1; i<argc; i++ )
    {
      line=argv[i];
      if( line=="-noplot" || line=="noplot" )
        plotOption=false;
      else if( line=="-nopause" || line=="-abortOnEnd" || line=="-nodirect" ||
               line=="-readCollective" || line=="-writeCollective" ||
               line=="nopause" || line=="abortOnEnd" || line=="nodirect" )
        continue; // these commands are processed by getGraphicsInterface below 
      else if( line=="memory" )
      {
	reportMemory=true;
        Diagnostic_Manager::setTrackArrayData(TRUE);
      }
      else if( line=="loadBalance" || line=="-loadBalance" )
      {
	loadBalance=true;
      }
      else if( line=="release" )
      {
        smartRelease=true;
        printF("*** turn on smart release of memory ***\n");
        Diagnostic_Manager::setSmartReleaseOfInternalMemory( ON );
      }
      else if( commandFileName=="" )
      {
        commandFileName=line;    
        printF("cgmp: reading commands from file [%s]\n",(const char*)commandFileName);
      }
      
    }
  }
  else
  {
    printF("Usage: `cgmp [options][file.cmd]' \n"
	   "     options:                            \n" 
	   "          noplot:   run without graphics \n" 
	   "          nopause: do not pause \n" 
	   "          abortOnEnd: abort if command file ends \n" 
	   "          memory:   run with A++ memory tracking\n" 
	   "          release:  run with A++ smart release of memeory\n"
	   "     file.cmd: read this command file \n");
    
  }


  GenericGraphicsInterface & ps = *Overture::getGraphicsInterface("cgmp",false,argc,argv);

  // By default start saving the command file called "cgmp.cmd"
  aString logFile="cgmp.cmd";
  ps.saveCommandFile(logFile);
  printF("User commands are being saved in the file `%s'\n",(const char *)logFile);

  ps.appendToTheDefaultPrompt("cgmp>");

  // read from a command file if given
  if( commandFileName!="" )
  {
    printF("read command file =%s\n",(const char*)commandFileName);
    ps.readCommandFile(commandFileName);
  }

  aString nameOfShowFile;

#if 0
  aString nameOfOGFile, nameOfShowFile;
  ps.inputString(nameOfOGFile,"Enter the name of the (old) overlapping grid file:");

  printF("Try to read the overlapping grid file : %s\n",(const char*)nameOfOGFile);
  
  // create and read in a CompositeGrid
  #ifdef USE_PPP
    // On Parallel machines always add at least this many ghost lines on local arrays
    const int numGhost=2;
    MappedGrid::setMinimumNumberOfDistributedGhostLines(numGhost);


  #endif
  // Reduce the width of extrapolation for interp neighbours
  // and increase parallel ghost boundary width to 2 -- this will mean that we should
  // get the same result independent of np
  GenericMappedGridOperators::setDefaultMaximumWidthForExtrapolateInterpolationNeighbours(3);

  CompositeGrid m;
  getFromADataBase(m,nameOfOGFile);
  CompositeGrid & cg = m.numberOfMultigridLevels()==1 ? m : m.multigridLevel[0];
  if( m.numberOfMultigridLevels() > 1 )
    m.update(CompositeGrid::THEmultigridLevel);
#else

  CompositeGrid cg;
  readOrBuildTheGrid(ps, cg, loadBalance);

#endif

  cg.update(GridCollection::THEdomain);

  const int numberOfDomains=cg.numberOfDomains();
  printF(" >>> numberOfDomains=%i\n",numberOfDomains);

//  Interpolant *interpolant = new Interpolant[numberOfDomains];

//   Interpolant interpolant(cg);
//   interpolant.setImplicitInterpolationMethod(Interpolant::iterateToInterpolate);

//   #ifdef USE_PPP
//   if( !interpolant.interpolationIsExplicit() )
//   {
//     printf("cgmp:ERROR: The parallel composite grid interpolator needs explicit interpolation ****\n");
//     Overture::abort();
//   }
//   #endif

  
  Ogshow *show=NULL;

  Cgmp & mpSolver = *new Cgmp(cg,&ps,show,plotOption);
  
  mpSolver.setParametersInteractively();

  mpSolver.solve();

  mpSolver.printStatistics();

  ps.unAppendTheDefaultPrompt();
//  delete show;

  if( reportMemory )
    Diagnostic_Manager::report();

  if(  myid==0 && false ) 
  {
    printf("\n +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
    for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
      cg[grid].displayComputedGeometry();
    printf(" +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
  }


//  delete pDomainSolver;
//   for( int d=0; d<numberOfDomains; d++ )
//   {
//     delete solver[d];  // do this so we shutdown PETSc (if used).
//   }
//   delete [] solver;
//   delete [] interpolant;
  

  delete &mpSolver;  // do this so we shutdown PETSc before P++  (if PETSc is being used).

  Overture::finish();          
  if( smartRelease )
  {
    int totalNumberOfArrays=GET_NUMBER_OF_ARRAYS;
    if( totalNumberOfArrays>0 )
    {
      printf("\n**** WARNING: Number of A++ is %i >0 \n",totalNumberOfArrays);
    }
  }
  
  return 0;
}


