#include "Maxwell.h"
#include "SquareMapping.h"
#include "BoxMapping.h"
#include "AnnulusMapping.h"
#include "MatrixTransform.h"
#include "DataPointMapping.h"
#include "CompositeGridOperators.h"
#include "display.h"
#include "UnstructuredMapping.h"
#include "ParallelUtility.h"

#include "ULink.h"

extern bool verifyUnstructuredConnectivity( UnstructuredMapping &umap, bool verbose );

#define FOR_3D(i1,i2,i3,I1,I2,I3) \
int I1Base =I1.getBase(),   I2Base =I2.getBase(),  I3Base =I3.getBase();  \
int I1Bound=I1.getBound(),  I2Bound=I2.getBound(), I3Bound=I3.getBound(); \
for(i3=I3Base; i3<=I3Bound; i3++) \
for(i2=I2Base; i2<=I2Bound; i2++) \
for(i1=I1Base; i1<=I1Bound; i1++)

#define FOR_3(i1,i2,i3,I1,I2,I3) \
I1Base =I1.getBase(),   I2Base =I2.getBase(),  I3Base =I3.getBase();  \
I1Bound=I1.getBound(),  I2Bound=I2.getBound(), I3Bound=I3.getBound(); \
for(i3=I3Base; i3<=I3Bound; i3++) \
for(i2=I2Base; i2<=I2Bound; i2++) \
for(i1=I1Base; i1<=I1Bound; i1++)

// ======================================================================================
//  Return the max value of a scalar over all processors
//  /processor: return the result to this processor (-1 equals all processors)
// ======================================================================================
real Maxwell::
getMaxValue(real value, int processor /* = -1 */)
{
  real maxValue=value;
  #ifdef USE_PPP 
  if( processor==-1 )
    MPI_Allreduce(&value, &maxValue, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
  else
    MPI_Reduce        (&value, &maxValue, 1, MPI_DOUBLE, MPI_MAX, processor, MPI_COMM_WORLD);
  #endif
  return maxValue;
}

int Maxwell::
getMaxValue(int value, int processor /* = -1 */)
{
  int maxValue=value;
  #ifdef USE_PPP 
  if( processor==-1 )
    MPI_Allreduce(&value, &maxValue, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
  else
    MPI_Reduce        (&value, &maxValue, 1, MPI_INT, MPI_MAX, processor, MPI_COMM_WORLD);
  #endif
  return maxValue;
}

real Maxwell::
getMinValue(real value, int processor /* = -1 */ )
{
  real minValue=value;
  #ifdef USE_PPP 
  if( processor==-1 )
    MPI_Allreduce(&value, &minValue, 1, MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD);
  else
    MPI_Reduce        (&value, &minValue, 1, MPI_DOUBLE, MPI_MIN, processor, MPI_COMM_WORLD);
  #endif
  return minValue;
}

int Maxwell::
getMinValue(int value, int processor /* = -1 */)
{
  int minValue=value;
  #ifdef USE_PPP 
  if( processor==-1 )
    MPI_Allreduce(&value, &minValue, 1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
  else
    MPI_Reduce        (&value, &minValue, 1, MPI_INT, MPI_MIN, processor, MPI_COMM_WORLD);
  #endif
  return minValue;
}

//! Determine the time step
int Maxwell::
computeTimeStep()
// =============================================================================================
// =============================================================================================
{
  real time0=getCPU();

  assert( cgp!=NULL );
  CompositeGrid & cg= *cgp;
  const int numberOfComponentGrids = cg.numberOfComponentGrids();
  const int numberOfDimensions = cg.numberOfDimensions();
  
  deltaT=REAL_MAX*.01;
  
//    RealArray dtGrid(numberOfComponentGrids);  // time step for each grid by itself
//    dtGrid=REAL_MAX;
  
  const real cMax=max(cGrid);
  for( int grid=0; grid<numberOfComponentGrids; grid++ )
  {
    MappedGrid & mg = cg[grid];

    const int numberOfDimensions=mg.numberOfDimensions();
    c = cGrid(grid);
    eps = epsGrid(grid);
    mu = muGrid(grid);

    real dtg=REAL_MAX*.01;
    if( mg.getGridType()==MappedGrid::structuredGrid )
    {
      real dx[3];
      if( mg.isRectangular() )
      {
	mg.getDeltaX(dx);

        if( numberOfDimensions==2 )
	  dtg=cfl*1./( c*sqrt( 1./(dx[0]*dx[0])+1./(dx[1]*dx[1]) ) );  
        else
  	  dtg=cfl*1./( c*sqrt( 1./(dx[0]*dx[0])+1./(dx[1]*dx[1])+1./(dx[2]*dx[2]) ) ); 
	
	dxMinMax(grid,0)=numberOfDimensions==2 ? min(dx[0],dx[1]) : min(dx[0],dx[1],dx[2]);
	dxMinMax(grid,1)=numberOfDimensions==2 ? max(dx[0],dx[1]) : max(dx[0],dx[1],dx[2]);
	
	// printF(" computeTimeStep: dx=%8.2e dy=%8.2e c=%8.2e, dtg=%8.2e\n",dx[0],dx[1],c,dtg);
      }
      else
      {
	mg.update(MappedGrid::THEinverseVertexDerivative);
	const realArray & rx = mg.inverseVertexDerivative();
	const intArray & mask = mg.mask();
	
      
	Index I1,I2,I3;
	getIndex( mg.indexRange(),I1,I2,I3);

	// Grid spacings on unit square:
	real dr1 = mg.gridSpacing(axis1);
	real dr2 = mg.gridSpacing(axis2);
	real dr3 = mg.gridSpacing(axis3);

	// parallel version here --- also broadcast max error in forcing.bC *************************
        #ifdef USE_PPP
          realSerialArray rxLocal; getLocalArrayWithGhostBoundaries(rx,rxLocal);
          intSerialArray maskLocal; getLocalArrayWithGhostBoundaries(mask,maskLocal);
        #else
          const realSerialArray & rxLocal = rx;
          const intSerialArray & maskLocal = mask;
        #endif

	real *rxp = rxLocal.Array_Descriptor.Array_View_Pointer3;
	const int rxDim0=rxLocal.getRawDataSize(0);
	const int rxDim1=rxLocal.getRawDataSize(1);
	const int rxDim2=rxLocal.getRawDataSize(2);
	const int rxDim3=mg.numberOfDimensions();   // note
#undef RX
#define RX(i0,i1,i2,i3,i4) rxp[i0+rxDim0*(i1+rxDim1*(i2+rxDim2*(i3+rxDim3*(i4))))]

	const int *maskp = maskLocal.Array_Descriptor.Array_View_Pointer2;
	const int maskDim0=maskLocal.getRawDataSize(0);
	const int maskDim1=maskLocal.getRawDataSize(1);
	const int md1=maskDim0, md2=md1*maskDim1; 
#define MASK(i0,i1,i2) maskp[(i0)+(i1)*md1+(i2)*md2]


	int includeGhost=0;
	bool ok = ParallelUtility::getLocalArrayBounds(rx,rxLocal,I1,I2,I3,includeGhost);

	int i1,i2,i3;
	real a11Min=REAL_MAX*.001;
	real a11Max=-a11Min;
	//  **** this is a guess **** check this.
	const real alpha0=1.;

	dxMinMax(grid,0)=REAL_MAX*.01; 
	dxMinMax(grid,1)=0.;
	dtg = REAL_MAX*.01;

	real a11,a12,a22;
	if( ok )
	{
	  if( numberOfDimensions==2 )
	  {

	    FOR_3D(i1,i2,i3,I1,I2,I3)
	    {
	      
	      if( MASK(i1,i2,i3)>0 )
	      {
		a11 = ( RX(i1,i2,i3,0,0)*RX(i1,i2,i3,0,0) + RX(i1,i2,i3,0,1)*RX(i1,i2,i3,0,1) );
		a12 = ( RX(i1,i2,i3,0,0)*RX(i1,i2,i3,1,0) + RX(i1,i2,i3,0,1)*RX(i1,i2,i3,1,1) )*2.;
		a22 = ( RX(i1,i2,i3,1,0)*RX(i1,i2,i3,1,0) + RX(i1,i2,i3,1,1)*RX(i1,i2,i3,1,1) );


		a11=1./sqrt( a11 *(1./(alpha0*dr1*dr1)) 
			     +abs(a12)*(.25/(alpha0*dr1*dr2))
			     +a22 *(1./(alpha0*dr2*dr2)) 
		  );

		a11Min=min(a11Min,a11);
		a11Max=max(a11Max,a11);
          
	      }

	    }
	    
	  }
	  else
	  { // ***** 3D ********

#define rxDotRx(axis,dir) (RX(i1,i2,i3,axis,0)*RX(i1,i2,i3,dir,0) \
			 + RX(i1,i2,i3,axis,1)*RX(i1,i2,i3,dir,1) \
			 + RX(i1,i2,i3,axis,2)*RX(i1,i2,i3,dir,2))
      
	    // There wold be a factor of 4 for the worst case plus/minus wave but we also
	    // divide by a factor of 4 for the 2nd-order time stepping.
	    FOR_3D(i1,i2,i3,I1,I2,I3)
	    {
	      if( MASK(i1,i2,i3)>0 )
	      {
		a11=1./sqrt(  rxDotRx(0,0) *(1./(dr1*dr1)) 
			      +rxDotRx(1,1) *(1./(dr2*dr2))
			      +rxDotRx(2,2) *(1./(dr3*dr3))
			      +abs(rxDotRx(1,0))*(.5/(dr2*dr1))  
			      +abs(rxDotRx(2,0))*(.5/(dr3*dr1)) 
			      +abs(rxDotRx(2,1))*(.5/(dr3*dr2)) );

		// ** a11 =  pow(a11,-.5);
		
		a11Min=min(a11Min,a11);
		a11Max=max(a11Max,a11);

	      }
	    }
#undef rxDotRx
	  }

	  dxMinMax(grid,0)=a11Min;  // ********** need to broadcast ***********
	  dxMinMax(grid,1)=a11Max;
	  dtg = (cfl/c) * dxMinMax(grid,0); // *wdh* 071029 : divide by c !
	}
	
      }
    
      dtg=getMinValue(dtg);  // compute min over all processors
      dxMinMax(grid,0)=getMinValue(dxMinMax(grid,0));
      dxMinMax(grid,1)=getMaxValue(dxMinMax(grid,1));
      
      if( artificialDissipation>0. )
      {
	// Here is the correction for artificial dissipation
        //
        // The equation for dt looks like
        //   dt*dt *c*c*(  1/dx^2 + 1/dy^2 ) = 1 - beta*dt
        //
        real gamma = dtg*dtg;
        real beta;
        const real adc = artificialDissipation*SQR(cMax); // scale dissipation by c^2 *wdh* 041103

        beta = .5*adc*( numberOfDimensions*pow(2.,real(orderOfArtificialDissipation)) );
        real factor=2.;  // safety factor
	beta *=factor;

        dtg = sqrt( gamma + pow(beta*gamma*.5,2.) ) - beta*gamma*.5;

        if( debug & 4 )
	  fprintf(pDebugFile," getTimeStep: Correct for art. dissipation: new dt=%9.3e (old = %9.3e, new/old=%4.2f) myid=%i\n",
		 dtg,sqrt(gamma),dtg/sqrt(gamma),myid);
	
      }
      
      if( timeSteppingMethod==modifiedEquationTimeStepping )
      {
	if( true || orderOfAccuracyInTime==2 || orderOfAccuracyInTime==4 )
	{
	  dtg*=1.; // Check this for 3D
	}
	else
	{
          printF("getTimeStep:ERROR: modifiedEquationTimeStepping -- orderOfAccuracyInTime=%i ??\n",
          orderOfAccuracyInTime);
	  
          Overture::abort("getTimeStep:ERROR: modifiedEquationTimeStepping -- orderOfAccuracyInTime?? ");
	}
	
      }
      else
      {

	if( orderOfAccuracyInSpace==2 )
	{
	}
	else if( orderOfAccuracyInSpace==4 )
	  dtg*=sqrt(3./4.);
	else if( orderOfAccuracyInSpace==6 )
	  dtg*=sqrt(.6618);
	else if( orderOfAccuracyInSpace==8 )
	  dtg*=sqrt(.6152);
	else
	{
          Overture::abort("getTimeStep:ERROR: modifiedEquationTimeStepping -- orderOfAccuracyInSpace?? ");
	}

	if( orderOfAccuracyInTime==4 )
	{
	  dtg*=1.41/2.;
	}
	else if( orderOfAccuracyInTime==6 )
	{
	  dtg*=.84/2.;
	}
	else if( orderOfAccuracyInTime==8 )
	{
	  dtg*=.46/2.;
	}
	else if( orderOfAccuracyInTime==3 && method==dsi )
	{
	  dtg*=(12./7.)/2.;   // ABS3
	}
	else if( orderOfAccuracyInTime!=2 )
	{
	  Overture::abort();
	}
      }
      
      
      printF(" computeTimeStep: grid=%i c=%8.2e, dtg=%8.2e min-dx=%8.2e max-dx=%8.2e\n",grid,c,dtg,dxMinMax(grid,0),dxMinMax(grid,1));
      printf(" computeTimeStep: grid=%i c=%8.2e, dtg=%8.2e min-dx=%8.2e max-dx=%8.2e myid=%i\n",
                grid,c,dtg,dxMinMax(grid,0),dxMinMax(grid,1),myid);

#undef RX
    
    }
    else
    {
      // unstructured grid.

//        UnstructuredMapping & map = (UnstructuredMapping &) mg.mapping().getMapping();
//        const int numberOfElements = map.getNumberOfElements();
//        const intArray & faces = map.getFaces();
//        const intArray & elementFace = map.getElementFaces();
//        const realArray & nodes = map.getNodes();

//        real dsMax=0.;
//        int e;
//        for( e=0; e<numberOfElements; e++ )
//        {
//  	const int numFacesThisElement=map.getNumberOfFacesThisElement(e);      

//  	// select two faces from the element:
//  	int f1=elementFace(e,0);
//  	int f2=elementFace(e,numFacesThisElement-1);
      
//  	int n0=faces(f1,0), n1=faces(f1,1);
//  	real ds1=SQR( nodes(n1,0)-nodes(n0,0) )+SQR( nodes(n1,1)-nodes(n0,1) );
//  	n0=faces(f2,0), n1=faces(f2,1);
//  	real ds2=SQR( nodes(n1,0)-nodes(n0,0) )+SQR( nodes(n1,1)-nodes(n0,1) );

//  	dsMax = max( dsMax, 1./ds1+1./ds2 );
//        }
    
//        if( map.getNumberOfFacesThisElement(0)==3 )
//        {
//  	dsMax*=4.;  // **** fudge for triangular grids -- fix this ---
//        }
    
#if 0
      mg.update(MappedGrid::THEminMaxEdgeLength);
      real dsMax =8./(mg.minimumEdgeLength(0)*mg.minimumEdgeLength(0));

      printF(" computeTimeStep: dsMax=%e c=%e\n",dsMax,c);

      dtg=cfl*1./( c*sqrt(dsMax) );
#else

      UnstructuredMapping::EntityTypeEnum cellType = UnstructuredMapping::EntityTypeEnum(cg.numberOfDimensions());
      UnstructuredMapping::EntityTypeEnum cellBdyType = UnstructuredMapping::EntityTypeEnum(cg.numberOfDimensions()-1);
      UnstructuredMapping::EntityTypeEnum faceType = UnstructuredMapping::Face;
      UnstructuredMapping::EntityTypeEnum edgeType = UnstructuredMapping::Edge;

      UnstructuredMapping &umap = (UnstructuredMapping &)mg.mapping().getMapping();

      int rDim = umap.getRangeDimension();
      int dDim = umap.getDomainDimension();

      bool vCent = mg.isAllVertexCentered();
      realArray &cFArea = vCent ? mg.centerArea() : mg.faceArea();
      const realArray &cFNorm = vCent ? mg.centerNormal() : mg.faceNormal();
      const realArray &cEArea = vCent ? mg.faceArea() : mg.centerArea();
      const realArray &cENorm = vCent ? mg.faceNormal() : mg.centerNormal();

      realArray cellBdyCenters, cellCenters;
      getCenters( mg, cellBdyType, cellBdyCenters);
      getCenters( mg, cellType, cellCenters);

      const realArray &nodes  = umap.getNodes();
      const intArray &edges = umap.getEntities(edgeType);

      UnstructuredMappingIterator iter,iter_end;
      UnstructuredMappingAdjacencyIterator aiter, aiter_end;

      iter_end = umap.end(UnstructuredMapping::Face);
      real minLen = REAL_MAX;
      real minFEL = REAL_MAX;
      for ( iter = umap.begin(UnstructuredMapping::Face); iter!=iter_end; iter++ )
      {
	int f=*iter;
	real area = cFArea(f,0,0);
	real lsum = 0.;
	aiter_end = umap.adjacency_end(iter, UnstructuredMapping::Edge);
	for ( aiter = umap.adjacency_begin(iter, UnstructuredMapping::Edge); aiter!=aiter_end; aiter++ )
	{
	  int e = *aiter;
	  real edgeL = 0;
	  int v1 = edges(e,0);
	  int v2 = edges(e,1);
	  for ( int a=0; a<rDim; a++ )
	    edgeL += ( nodes(v2,a)-nodes(v1,a) )*( nodes(v2,a)-nodes(v1,a) );
	  //	      lsum += sqrt(edgeL);
	  lsum = max(lsum,edgeL);
	}

	lsum = sqrt(lsum);

	minLen = min(minLen, area/lsum);
	minFEL = min(lsum,minFEL);
      }

      iter_end = umap.end(cellBdyType);
      real minCCL = REAL_MAX;
      for ( iter=umap.begin(cellBdyType); iter!=iter_end; iter++ )
      {
	int e = *iter;
	aiter_end = umap.adjacency_end(iter,cellType);
	real edgeL = 0;
	for ( aiter=umap.adjacency_begin(iter,cellType); aiter!=aiter_end; aiter++ )
	{
	  real L = 0;
	  int c = *aiter;
	  for ( int a=0; a<rDim; a++ )
	    L += (cellCenters(c,a)-cellBdyCenters(e,a))*(cellCenters(c,a)-cellBdyCenters(e,a));
	  edgeL += sqrt(L);
	}

	minCCL = min(minCCL, edgeL);
	if ( rDim==2 )
	  minLen = min(edgeL,minLen);
	else
	{
	  aiter_end = umap.adjacency_end(iter,edgeType);
	  for ( aiter=umap.adjacency_begin(iter,edgeType); aiter!=aiter_end; aiter++ )
	  {
	    real area=0;
	    for ( int a=0; a<rDim; a++ )
	      area+=edgeAreaNormals(*aiter,0,0,a)*edgeAreaNormals(*aiter,0,0,a);
	    minLen = min(minLen, sqrt(area)/(2*edgeL));
	  }
	}
	  
      }

      minLen = min(minLen,minCCL);
      mg.update(MappedGrid::THEminMaxEdgeLength);
      cout<<"minFEL = "<<minFEL<<", minCCL = "<<minCCL<<", minLen = "<<minLen<<", minEdge = "<<mg.minimumEdgeLength(0)<<endl;
      minLen = min(minLen,mg.minimumEdgeLength(0));
      dtg=cfl*minLen/c;

#endif
      
    }
    
    
    deltaT=min(deltaT,dtg);
    
  } // end for grid

  deltaT=getMinValue(deltaT);  // min value over all processors
  printF("==== computeTimeStep: deltaT=%8.2e\n",deltaT);
  if( debug & 4 )
    fprintf(pDebugFile,"==== computeTimeStep: deltaT=%8.2e, myid=%i\n",deltaT,myid);

  timing(timeForComputingDeltaT)+=getCPU()-time0;
  return 0;
}




//! Setup and initialization. Build the grid and solution fields. 
int Maxwell::
setupGrids()
// ===================================================================================
// Build a grid
// ===================================================================================
{
  real time0=getCPU();

  // real xa=0, xb=1., ya=0., yb=1.;
  real xa=xab[0][0], xb=xab[1][0], ya=xab[0][1], yb=xab[1][1], za=xab[0][2], zb=xab[1][2];
  
  real dx[3];
  dx[0]=(xb-xa)/(nx[0]-1);
  dx[1]=(yb-ya)/(nx[1]-1);
  dx[2]=(zb-za)/(nx[2]-1);

//   xa=0, xb=1., ya=0., yb=1.25;
//   SquareMapping rectangle(xa,xb,ya,yb);
//square.setMappingCoordinateSystem( Mapping::general ); 

  Mapping *mapPointer;

  if( gridType==square && elementType==structuredElements )
  {
    mapPointer= new SquareMapping(xa,xb,ya,yb);  mapPointer->incrementReferenceCount();
  }
  else if ( gridType==box )
  {
    mapPointer = new BoxMapping(xa,xb,ya,yb,za,zb); mapPointer->incrementReferenceCount();

    for( int axis=0; axis<mapPointer->getDomainDimension(); axis++ )
    {
      mapPointer->setGridDimensions(axis,nx[axis]);
      if( bcOption==useAllPeriodicBoundaryConditions )
	mapPointer->setIsPeriodic(axis,Mapping::derivativePeriodic);
    }

    if ( elementType!=structuredElements )
    {
      UnstructuredMapping & uns = * new UnstructuredMapping;
      BoxMapping &bmap = (BoxMapping &)*mapPointer;
      mapPointer->decrementReferenceCount();
      mapPointer= &uns;  mapPointer->incrementReferenceCount();
      if ( bcOption==useAllPeriodicBoundaryConditions )
	uns.addGhostElements(true);
      else
	uns.addGhostElements(false);

      printF(" **** elementType=%i \n",elementType);
	  
      //	  UnstructuredMapping etype;
      uns.addGhostElements(true);
      uns.buildFromAMapping(bmap);
      verifyUnstructuredConnectivity(uns,true);
      for( int axis=0; axis<bmap.getDomainDimension() && bcOption==useAllPeriodicBoundaryConditions; axis++ )
      {
	uns.setIsPeriodic(axis,Mapping::derivativePeriodic);
	uns.setBoundaryCondition(Start,axis,-1);
	uns.setBoundaryCondition(End  ,axis,-1);
      }
      //	  if ( !bcOption==useAllPeriodicBoundaryConditions )
      //	    uns.expandGhostBoundary();

    }
  }
  else if( gridType==annulus )
  {
    mapPointer= new AnnulusMapping;  mapPointer->incrementReferenceCount();
  }
  else if( gridType==rotatedSquare )
  {
    Mapping *sq=new SquareMapping(xa,xb,ya,yb); 
    sq->incrementReferenceCount();
    MatrixTransform & mat = * new MatrixTransform(*sq);
    mapPointer= &mat;   mapPointer->incrementReferenceCount();
    if( sq->decrementReferenceCount()==0 ) delete sq;
    
    // mat.rotate(axis3,90.*twoPi/360.);
    // mat.rotate(axis3,45.*twoPi/360.);
    mat.rotate(axis3,30.*twoPi/360.);
    for( int axis=0; axis<mapPointer->getDomainDimension(); axis++ )
    {
      mapPointer->setGridDimensions(axis,nx[axis]);
    }

    UnstructuredMapping & uns = * new UnstructuredMapping;
    mapPointer->decrementReferenceCount();
    mapPointer= &uns;  mapPointer->incrementReferenceCount();
    uns.addGhostElements(true);

    printF(" **** elementType=%i \n",elementType);
      
    int domainDimension = 2;
    if ( elementType==defaultUnstructured && domainDimension==2)
      uns.buildFromARegularMapping(mat);
    else if ( domainDimension==2 )
    {
      uns.buildFromARegularMapping(mat,elementType==triangles ? UnstructuredMapping::triangle : elementType==quadrilaterals ? UnstructuredMapping::quadrilateral : UnstructuredMapping::hexahedron);
    }
    else
    {
      uns.buildFromAMapping(mat);
      for( int axis=0; axis<uns.getDomainDimension() && 
	     bcOption==useAllPeriodicBoundaryConditions; axis++ )
      {
	uns.setIsPeriodic(axis,Mapping::derivativePeriodic);
	uns.setBoundaryCondition(Start,axis,-1);
	uns.setBoundaryCondition(End  ,axis,-1);
      }
    }
    

    //            verifyUnstructuredConnectivity(uns,true);
    
    if( mat.decrementReferenceCount() == 0 )
      delete &mat;

  }
  else if( gridType==skewedSquare || gridType==sineSquare ||
           gridType==chevron || gridType==chevbox || gridType==sineByTriangles )
  {
    // build a DataPointMapping
    DataPointMapping & dpm = *new DataPointMapping;
    mapPointer= &dpm;   mapPointer->incrementReferenceCount();
    Index I1,I2,I3;

    // *wdh* 050818 -- use 2nd-order interpolation so the metrics are periodic when the grid is periodic
//    dpm.setOrderOfInterpolation(4);
    dpm.setOrderOfInterpolation(2);

//    int numberOfGhostLines=0; // add more for 4th-order and periodic derivatives of mapping at ghost pts (?)
    // int numberOfGhostLines=4; // add more for 4th-order and periodic derivatives of mapping at ghost pts (?)
    int numberOfGhostLines=max(4,(orderOfAccuracyInSpace+2)/2);
//   int numberOfGhostLines=8; // add more for 4th-order and periodic derivatives of mapping at ghost pts (?)

    I1=Range(-numberOfGhostLines,nx[0]+numberOfGhostLines-1);
    I2=Range(-numberOfGhostLines,nx[1]+numberOfGhostLines-1);
    I3= gridType!=chevbox ? Range(0,0) : Range(-numberOfGhostLines,nx[2]+numberOfGhostLines-1);
    
    int rDim = gridType==chevbox ? 3 : 2;
    realArray x(I1,I2,I3,rDim);
    realArray r1(I1),r2(I2),r3(I3);

    r1.seqAdd(-numberOfGhostLines*dx[0],dx[0]);
    r2.seqAdd(-numberOfGhostLines*dx[1],dx[1]);
    if ( gridType==chevbox )
      r3.seqAdd(-numberOfGhostLines*dx[2],dx[2]);

    if( gridType==sineSquare || gridType==sineByTriangles )
    {
      realArray bottom(I1);
      real amplitude=.1;    // amplitude of the sine wave
      bottom=(cos(twoPi*r1)-1.)*amplitude;
    
      int i2=0, i3=0;
      for( i2=I2.getBase(); i2<=I2.getBound(); i2++ )
      {
	x(I1,i2,i3,0)=r1(I1);
	x(I1,i2,i3,1)=bottom(I1)+r2(i2);
      }
    }
    else if( gridType==chevron || gridType==chevbox)
    {
      real amplitudeY=dx[1]*chevronAmplitude;    // amplitude of the chevron oscillation
      realArray yShift(I1);
      // .5 seemed stable, .75 was not stable
      real freqX=(Pi/dx[0])*chevronFrequency; 
      // real freq=(Pi/dx[0])*.75;  // this is a lower frequency perturbation

      bool useMoreFrequencies=true;
      if( !useMoreFrequencies )
      {
        yShift=cos(freqX*r1)*amplitudeY;
      }
      else
      {
        yShift=cos(freqX*r1)*amplitudeY + cos(.5*freqX*r1)*amplitudeY*1.5;
      }
      
      // display(yShift,"yShift for the chevron grid");
      

      // remove perturbation from the ends
      bool removePerturbationAtEdges=false; // false; // true;
      int extra=numberOfGhostLines+5;
      if( removePerturbationAtEdges )
      {
	yShift(Range(I1.getBase(),I1.getBase()+extra))=0.;
	yShift(Range(I1.getBound()-extra,I1.getBound()))=0.;
      }
      
      // here is a perturbation in the x-direction
      real amplitudeX=dx[0]*chevronAmplitude;  
      realArray xShift(I2);
      real freqY=(Pi/dx[1])*chevronFrequency; 
      xShift=cos(freqY*r2)*amplitudeX; 
      if( useMoreFrequencies )
      {
        xShift=cos(freqY*r2)*amplitudeX+ cos(.5*freqY*r2)*amplitudeX*1.5;
      }
      
      if( removePerturbationAtEdges )
      {
	xShift(Range(I2.getBase(),I2.getBase()+extra))=0.;
	xShift(Range(I2.getBound()-extra,I2.getBound()))=0.;
      }

      int i2=0, i3=0;
      if ( gridType==chevron )
      {
        if( true )
	{
	  // box with random perturbations
          int seed=184273654;
          srand(seed);
	  int i1;
	  for( i2=I2.getBase(); i2<=I2.getBound(); i2++ )
	    for( i1=I1.getBase(); i1<=I1.getBound(); i1++ )
	    {
	      real d1=(-1.+2.*rand()/RAND_MAX)*.25*dx[0];  // random number between [-.25,.25]*dx[0]
	      real d2=(-1.+2.*rand()/RAND_MAX)*.25*dx[1];
		
	      x(i1,i2,i3,0)= i1*dx[0]+d1;
	      x(i1,i2,i3,1)= i2*dx[1]+d2;
	    }

          // make periodic
          real xba=1., yba=1., zba=1.;
          Range Rx=2;
          for( i1=I1.getBase(); i1<0; i1++ )
	  {
            x(i1,I2,I3,Rx)=x(i1+nx[0]-1,I2,I3,Rx);
            x(i1,I2,I3,0)-=xba; 
	  }
          for( i1=nx[0]-1; i1<=I1.getBound(); i1++ )
	  {
            x(i1,I2,I3,Rx)=x(i1-nx[0]+1,I2,I3,Rx); 
            x(i1,I2,I3,0)+=xba; 
	  }
	  
          for( i2=I2.getBase(); i2<0; i2++ )
	  {
            x(I1,i2,I3,Rx)=x(I1,i2+nx[1]-1,I3,Rx);
            x(I1,i2,I3,1)-=yba; 
	  }
          for( i2=nx[1]-1; i2<=I2.getBound(); i2++ )
	  {
            x(I1,i2,I3,Rx)=x(I1,i2-nx[1]+1,I3,Rx); 
            x(I1,i2,I3,1)+=yba; 
	  }
	  
	}
	else
	{
	  for( i2=I2.getBase(); i2<=I2.getBound(); i2++ )
	  {
	    x(I1,i2,i3,0)=r1(I1);
	    if( !removePerturbationAtEdges ||
		(i2>=I2.getBase() + extra &&
		 i2<=I2.getBound()- extra) )
	    {
	      x(I1,i2,i3,1)=yShift(I1)+r2(i2);
	    }
	    else
	    {
	      x(I1,i2,i3,1)=r2(i2);
	    }
	  
	  }
	  if( amplitudeX != 0. )
	  {
	    for( i2=I2.getBase(); i2<=I2.getBound(); i2++ )
	      x(I1,i2,i3,0)+=xShift(i2);
	  }
	}
	
      }
      else 
      {
        // chevBox
        if( true )
	{
	  // box with random perturbations
          int seed=184273654;
          srand(seed);
	  int i1;
	  for( i3=I3.getBase(); i3<=I3.getBound(); i3++ )
	    for( i2=I2.getBase(); i2<=I2.getBound(); i2++ )
	      for( i1=I1.getBase(); i1<=I1.getBound(); i1++ )
	      {
                real d1=(-1.+2.*rand()/RAND_MAX)*.25*dx[0];  // random number between [-.25,.25]*dx[0]
                real d2=(-1.+2.*rand()/RAND_MAX)*.25*dx[1];
                real d3=(-1.+2.*rand()/RAND_MAX)*.25*dx[2];
		
                // if( i1<=0 || i1>=nx[0]-1 ) d1=0.;
                // if( i2<=0 || i2>=nx[1]-1 ) d2=0.;
                // if( i3<=0 || i3>=nx[2]-1 ) d3=0.;

		x(i1,i2,i3,0)= i1*dx[0]+d1;
		x(i1,i2,i3,1)= i2*dx[1]+d2;
		x(i1,i2,i3,2)= i3*dx[2]+d3;
	      }

          // make periodic
          real xba=1., yba=1., zba=1.;
          Range Rx=3;
          for( int mm=0; mm<=1; mm++ )
	  {
	    for( i1=I1.getBase(); i1<0; i1++ )
	    {
	      x(i1,I2,I3,Rx)=x(i1+nx[0]-1,I2,I3,Rx);
	      x(i1,I2,I3,0)-=xba; 
	    }
	    for( i1=nx[0]-1; i1<=I1.getBound(); i1++ )
	    {
	      x(i1,I2,I3,Rx)=x(i1-nx[0]+1,I2,I3,Rx); 
	      x(i1,I2,I3,0)+=xba; 
	    }
	  
	    for( i2=I2.getBase(); i2<0; i2++ )
	    {
	      x(I1,i2,I3,Rx)=x(I1,i2+nx[1]-1,I3,Rx);
	      x(I1,i2,I3,1)-=yba; 
	    }
	    for( i2=nx[1]-1; i2<=I2.getBound(); i2++ )
	    {
	      x(I1,i2,I3,Rx)=x(I1,i2-nx[1]+1,I3,Rx); 
	      x(I1,i2,I3,1)+=yba; 
	    }
	  
	    for( i3=I3.getBase(); i3<0; i3++ )
	    {
	      x(I1,I2,i3,Rx)=x(I1,I2,i3+nx[2]-1,Rx);
	      x(I1,I2,i3,2)-=zba; 
	    }
	    for( i3=nx[2]-3; i3<=I3.getBound(); i3++ )
	    {
	      x(I1,I2,i3,Rx)=x(I1,I2,i3-nx[2]+1,Rx);
	      x(I1,I2,i3,2)+=zba; 
	    }
	  }
	  

	}
	else
	{
	  for( i3=I3.getBase(); i3<=I3.getBound(); i3++ )
	    for( i2=I2.getBase(); i2<=I2.getBound(); i2++ )
	    {
	      x(I1,i2,i3,0)=r1(I1);
	      x(I1,i2,i3,1)=yShift(I1)+r2(i2);
	    }

	  for( i3=I3.getBase(); i3<=I3.getBound(); i3++ )
	    x(I1,I2,i3,2)=r3(i3);

	  if( amplitudeX != 0. )
	  {
	    for( i2=I2.getBase(); i2<=I2.getBound(); i2++ )
	      x(I1,i2,I3,0)+=xShift(i2);
	  }
	  real amplitudeZ=dx[2]*chevronAmplitude;  
	  if( amplitudeZ != 0. )
	  {
	
	    realArray zShift(I3);
	    real freqY=(Pi/dx[1])*chevronFrequency; 
	    zShift=cos(freqY*r3)*amplitudeZ; 
	    if( useMoreFrequencies )
	    {
	      zShift=cos(freqY*r3)*amplitudeZ+ cos(.5*freqY*r3)*amplitudeZ*1.5;
	    }
	    for( i3=I3.getBase(); i3<=I3.getBound(); i3++ )
	    {
	      x(I1,I2,i3,2)+=zShift(i3);
	    }
	  }
	}
	
      }

    }
    else
    {
      throw "error";
    }
    // display(x,"Data points for DPM");

    const int domainDimension= gridType==chevbox ? 3 : 2;
    int ng=max(2,numberOfGhostLines);
    // int numberOfGhostLinesNew[2][3]={ng,ng,ng,ng,ng,ng}; //
    Mapping::IndexRangeType numberOfGhostLinesNew;
    for( int axis=0; axis<3; axis++ )for( int side=0; side<=1; side++ ) numberOfGhostLinesNew(side,axis)=ng;
    
    // tell the dpm to use the extra ghost lines (*wdh* 050818)
    dpm.setDomainDimension(domainDimension); 
    dpm.setRangeDimension(domainDimension); 
    dpm.setNumberOfGhostLines(numberOfGhostLinesNew);
    
    // Set periodic BC's before supplying data points so the ghost points are computed properly
    // *************** wdh* 050818 **
    if( bcOption==useAllPeriodicBoundaryConditions )
    {
      for( int axis=0; axis<domainDimension; axis++ )
      {
 	dpm.setIsPeriodic(axis,Mapping::derivativePeriodic);
 	dpm.setBoundaryCondition(Start,axis,-1);
 	dpm.setBoundaryCondition(End  ,axis,-1);
      }
    }

    dpm.setDataPoints(x,3,domainDimension,numberOfGhostLines);

    if( gridType==sineByTriangles || elementType!=structuredElements )
    {
      int axis;
      for( axis=0; axis<domainDimension; axis++ )
      {
	dpm.setIsPeriodic(axis,Mapping::derivativePeriodic);
	dpm.setBoundaryCondition(Start,axis,-1);
	dpm.setBoundaryCondition(End  ,axis,-1);
      }

      UnstructuredMapping & uns = * new UnstructuredMapping;
      mapPointer->decrementReferenceCount();
      mapPointer= &uns;  mapPointer->incrementReferenceCount();
      if ( bcOption==useAllPeriodicBoundaryConditions )
	uns.addGhostElements(true);
      else
	uns.addGhostElements(false);

      printF(" **** elementType=%i \n",elementType);
      

      if ( elementType==defaultUnstructured && domainDimension==2)
	uns.buildFromARegularMapping(dpm);
      else if ( domainDimension==2 )
      {
	uns.buildFromARegularMapping(dpm,elementType==triangles ? UnstructuredMapping::triangle : elementType==quadrilaterals ? UnstructuredMapping::quadrilateral : UnstructuredMapping::hexahedron);
      }
      else
      {
// 	  uns.buildFromAMapping(dpm);
// 	  verifyUnstructuredConnectivity(uns,true);

	uns.addGhostElements(true);
	uns.buildFromAMapping(dpm);
	for( int axis=0; axis<uns.getDomainDimension() && 
	       bcOption==useAllPeriodicBoundaryConditions; axis++ )
	{
	  uns.setIsPeriodic(axis,Mapping::derivativePeriodic);
	  uns.setBoundaryCondition(Start,axis,-1);
	  uns.setBoundaryCondition(End  ,axis,-1);
	}
      }

      //            verifyUnstructuredConnectivity(uns,true);

      if( dpm.decrementReferenceCount() == 0 )
        delete &dpm;
      
    }
  }
  else if( gridType==perturbedSquare || gridType==perturbedBox )
  {
    // ************ perturbed square or box **********

    const int rangeDimension = gridType==perturbedSquare ? 2 : 3; 
    const int domainDimension= rangeDimension; 

    // build a DataPointMapping
    DataPointMapping & dpm = *new DataPointMapping;
    mapPointer= &dpm;   mapPointer->incrementReferenceCount();
    Index I1,I2,I3;

    dpm.setOrderOfInterpolation(2);

    int numberOfGhostLines=max(4,(orderOfAccuracyInSpace+2)/2);

    I1=Range(-numberOfGhostLines,nx[0]+numberOfGhostLines-1);
    I2=Range(-numberOfGhostLines,nx[1]+numberOfGhostLines-1);
    I3= domainDimension==2 ? Range(0,0) : Range(-numberOfGhostLines,nx[2]+numberOfGhostLines-1);
    
    realArray x(I1,I2,I3,rangeDimension);

    int seed=184273654;
    srand(seed);          // supply seed to the random number generator

    int i1=0, i2=0, i3=0;
    if ( gridType==perturbedSquare )
    {
      //  === square with random perturbations ===

      for( i2=I2.getBase(); i2<=I2.getBound(); i2++ )
	for( i1=I1.getBase(); i1<=I1.getBound(); i1++ )
	{
	  real d1=(-1.+2.*rand()/RAND_MAX)*.25*dx[0];  // random number between [-.25,.25]*dx[0]
	  real d2=(-1.+2.*rand()/RAND_MAX)*.25*dx[1];
		
	  x(i1,i2,i3,0)= i1*dx[0]+d1;
	  x(i1,i2,i3,1)= i2*dx[1]+d2;
	}

      // make periodic
      real xba=1., yba=1., zba=1.;
      Range Rx=2;
      for( i1=I1.getBase(); i1<0; i1++ )
      {
	x(i1,I2,I3,Rx)=x(i1+nx[0]-1,I2,I3,Rx);
	x(i1,I2,I3,0)-=xba; 
      }
      for( i1=nx[0]-1; i1<=I1.getBound(); i1++ )
      {
	x(i1,I2,I3,Rx)=x(i1-nx[0]+1,I2,I3,Rx); 
	x(i1,I2,I3,0)+=xba; 
      }
	  
      for( i2=I2.getBase(); i2<0; i2++ )
      {
	x(I1,i2,I3,Rx)=x(I1,i2+nx[1]-1,I3,Rx);
	x(I1,i2,I3,1)-=yba; 
      }
      for( i2=nx[1]-1; i2<=I2.getBound(); i2++ )
      {
	x(I1,i2,I3,Rx)=x(I1,i2-nx[1]+1,I3,Rx); 
	x(I1,i2,I3,1)+=yba; 
      }
	
    }
    else 
    {
      //  === Box with random perturbations ===
      // box with random perturbations
      for( i3=I3.getBase(); i3<=I3.getBound(); i3++ )
	for( i2=I2.getBase(); i2<=I2.getBound(); i2++ )
	  for( i1=I1.getBase(); i1<=I1.getBound(); i1++ )
	  {
	    real d1=(-1.+2.*rand()/RAND_MAX)*.25*dx[0];  // random number between [-.25,.25]*dx[0]
	    real d2=(-1.+2.*rand()/RAND_MAX)*.25*dx[1];
	    real d3=(-1.+2.*rand()/RAND_MAX)*.25*dx[2];
		
	    // if( i1<=0 || i1>=nx[0]-1 ) d1=0.;
	    // if( i2<=0 || i2>=nx[1]-1 ) d2=0.;
	    // if( i3<=0 || i3>=nx[2]-1 ) d3=0.;

	    x(i1,i2,i3,0)= i1*dx[0]+d1;
	    x(i1,i2,i3,1)= i2*dx[1]+d2;
	    x(i1,i2,i3,2)= i3*dx[2]+d3;
	  }

      // make periodic
      real xba=1., yba=1., zba=1.;
      Range Rx=3;
      for( int mm=0; mm<=1; mm++ )  // loop twice to make sure all corners points are correct
      {
	for( i1=I1.getBase(); i1<0; i1++ )
	{
	  x(i1,I2,I3,Rx)=x(i1+nx[0]-1,I2,I3,Rx);
	  x(i1,I2,I3,0)-=xba; 
	}
	for( i1=nx[0]-1; i1<=I1.getBound(); i1++ )
	{
	  x(i1,I2,I3,Rx)=x(i1-nx[0]+1,I2,I3,Rx); 
	  x(i1,I2,I3,0)+=xba; 
	}
	  
	for( i2=I2.getBase(); i2<0; i2++ )
	{
	  x(I1,i2,I3,Rx)=x(I1,i2+nx[1]-1,I3,Rx);
	  x(I1,i2,I3,1)-=yba; 
	}
	for( i2=nx[1]-1; i2<=I2.getBound(); i2++ )
	{
	  x(I1,i2,I3,Rx)=x(I1,i2-nx[1]+1,I3,Rx); 
	  x(I1,i2,I3,1)+=yba; 
	}
	  
	for( i3=I3.getBase(); i3<0; i3++ )
	{
	  x(I1,I2,i3,Rx)=x(I1,I2,i3+nx[2]-1,Rx);
	  x(I1,I2,i3,2)-=zba; 
	}
	for( i3=nx[2]-3; i3<=I3.getBound(); i3++ )
	{
	  x(I1,I2,i3,Rx)=x(I1,I2,i3-nx[2]+1,Rx);
	  x(I1,I2,i3,2)+=zba; 
	}
      }
      
    }
    
    int ng=max(2,numberOfGhostLines);
    // int numberOfGhostLinesNew[2][3]={ng,ng,ng,ng,ng,ng}; //
    Mapping::IndexRangeType numberOfGhostLinesNew;
    for( int axis=0; axis<3; axis++ )for( int side=0; side<=1; side++ ) numberOfGhostLinesNew(side,axis)=ng;
    
    dpm.setDomainDimension(domainDimension); 
    dpm.setRangeDimension(rangeDimension); 
    dpm.setNumberOfGhostLines(numberOfGhostLinesNew);
    
    // Set periodic BC's before supplying data points so the ghost points are computed properly
    if( bcOption==useAllPeriodicBoundaryConditions )
    {
      for( int axis=0; axis<domainDimension; axis++ )
      {
 	dpm.setIsPeriodic(axis,Mapping::derivativePeriodic);
 	dpm.setBoundaryCondition(Start,axis,-1);
 	dpm.setBoundaryCondition(End  ,axis,-1);
      }
    }

    dpm.setDataPoints(x,3,domainDimension,numberOfGhostLines);

    
  }
  else if( gridType==squareByTriangles || gridType==squareByQuads ||
	   (gridType==square && elementType==Maxwell::defaultUnstructured) )
  {
    Mapping & sq= *new SquareMapping(xa,xb,ya,yb); 
    sq.incrementReferenceCount();

    int side,axis;
    for( axis=0; axis<sq.getDomainDimension(); axis++ )
    {
      sq.setGridDimensions(axis,nx[axis]);
      sq.setIsPeriodic(axis,Mapping::derivativePeriodic);
      sq.setBoundaryCondition(Start,axis,-1);
      sq.setBoundaryCondition(End  ,axis,-1);
    }

    UnstructuredMapping & uns = * new UnstructuredMapping;
    mapPointer= &uns;   mapPointer->incrementReferenceCount();
    uns.addGhostElements(true);
    if( gridType==squareByTriangles )
    {
      uns.buildFromARegularMapping(sq,UnstructuredMapping::triangle);
    }
    else
    {

      uns.buildFromARegularMapping(sq,UnstructuredMapping::quadrilateral);

//        // For debugging:
//        UnstructuredMappingIterator iter;
//        for( iter=uns.begin(UnstructuredMapping::Region); iter!=uns.end(UnstructuredMapping::Region); iter++ )
//        {
//  	printf(" Element %i is valid \n",*iter);
//        }

    }

    //    uns.getEntities(UnstructuredMapping::Face).display("Faces");
    //    uns.getEntities(UnstructuredMapping::Edge).display("Edges");
    //        verifyUnstructuredConnectivity(uns,true);

//      GenericGraphicsInterface & ps = *gip;
//      GraphicsParameters params;
//      params.set(GI_PLOT_UNS_EDGES,true);
//      params.set(GI_PLOT_UNS_FACES,true);

//      PlotIt::plot(ps,uns,params);
    
    if( sq.decrementReferenceCount()==0 ) delete &sq;
  }
  else if( gridType==compositeGrid )
  {
    // In this case we grab the Mapping from the first component grid.
     
    // create and read in a CompositeGrid
    delete cgp;
    
    cgp=new CompositeGrid;
    CompositeGrid & cg = *cgp;
    getFromADataBase(cg,nameOfGridFile);

    mapPointer=&(cg[0].mapping().getMapping());
  }
  else
  {
    throw "error";
  }
  
  if( cgp==NULL )
  {
    // ****************************************
    // *********** single grid  ***************
    // ****************************************

    Mapping & map = *mapPointer;
  
    bool unstructured=!(elementType==structuredElements);//map.getClassName()=="UnstructuredMapping";

    int side,axis;
    if( !unstructured )
    {
      for( axis=0; axis<map.getDomainDimension(); axis++ )
      {
	map.setGridDimensions(axis,nx[axis]);
	if( map.getIsPeriodic(axis)==Mapping::notPeriodic )
	{
	  if( bcOption==useAllPeriodicBoundaryConditions )
	  {
	    map.setIsPeriodic(axis,Mapping::derivativePeriodic);
	    map.setBoundaryCondition(Start,axis,-1);
	    map.setBoundaryCondition(End  ,axis,-1);
	  }
	  else if( bcOption==useAllDirichletBoundaryConditions )
	  {
	    map.setBoundaryCondition(Start,axis,dirichlet);
	    map.setBoundaryCondition(End  ,axis,dirichlet);
	  }
	  else if( bcOption==useAllPerfectElectricalConductorBoundaryConditions )
	  {
	    map.setBoundaryCondition(Start,axis,perfectElectricalConductor);
	    map.setBoundaryCondition(End  ,axis,perfectElectricalConductor);
	  }
	}
      
      }
    }
  
    // Build a MappedGrid
    mgp = new MappedGrid(map);
    MappedGrid & mg = *mgp;
  
    if( !unstructured )
    {
      const int numberOfGhostPoints=max(2,orderOfAccuracyInSpace/2); 
      printF(">>>>Create the MappedGrid with %i ghost points, orderOfAccuracyInSpace=%i\n",numberOfGhostPoints,
	     orderOfAccuracyInSpace  );
      
      for( axis=0; axis<map.getDomainDimension(); axis++ )
	for( side=Start; side<=End; side++ )
	  mg.setNumberOfGhostPoints(side,axis,numberOfGhostPoints);
  
    }

    cgp= new CompositeGrid(mg.numberOfDimensions(),1);
    CompositeGrid & cg = *cgp;
    cg[0].reference(mg);
    cg.updateReferences();
    
//    mg.update(MappedGrid::THEcenter | MappedGrid::THEvertex );  

//    cg.update(MappedGrid::THEmask | MappedGrid::THEcenter | MappedGrid::THEvertex );  
    cg.update( MappedGrid::THEcenter | MappedGrid::THEvertex | MappedGrid::THEmask );  

    if ( unstructured )
    {
      cg.update(MappedGrid::THEcorner| MappedGrid::THEfaceArea | MappedGrid::THEfaceNormal | MappedGrid::THEcellVolume | MappedGrid::THEcenterNormal | MappedGrid::THEcenterArea );

      for ( int g=0; false && bcOption==useAllPeriodicBoundaryConditions && g<cg.numberOfGrids(); g++ )
      {
	// we do this trick for convenience when working with periodic boundaries and dsi schemes
	bool vCent = mg.isAllVertexCentered();
	realArray &cFArea = vCent ? mg.centerArea() : mg.faceArea();
	realArray &cFNorm = vCent ? mg.centerNormal() : mg.faceNormal();
	realArray &cEArea = vCent ? mg.faceArea() : mg.centerArea();
	realArray &cENorm = vCent ? mg.faceNormal() : mg.centerNormal();

	MappedGrid &mg = cg[g];
	const IntegerArray &perEImages = *mg.getUnstructuredPeriodicBC(UnstructuredMapping::Edge);
	    
	for ( int e=perEImages.getBase(0); e<=perEImages.getBound(0); e++ )
	{
	  cEArea(perEImages(e,0),0,0) = cEArea(perEImages(e,1),0,0);
	  for ( int r=0; r<mg.numberOfDimensions(); r++ )
	    cENorm(perEImages(e,0),0,0,r) = cENorm(perEImages(e,1),0,0,r);
	}

	const IntegerArray &perHImages = *mg.getUnstructuredPeriodicBC(UnstructuredMapping::Face);
	for ( int h=perHImages.getBase(0); mg.numberOfDimensions()>2 &&h<=perHImages.getBound(0); h++ )
	{
	  cFArea(perHImages(h,0),0,0) = cFArea(perHImages(h,1),0,0);
	  for ( int r=0; r<mg.numberOfDimensions(); r++ )
	    cFNorm(perHImages(h,0),0,0,r) = cFNorm(perHImages(h,1),0,0,r);
	}

      }
    }

//  display(mg.vertex(),"mg.vertex()");
    // display(mg.center(),"mg.center()");

//   display(mg.dimension(),"mg.dimension()");
//   display(mg.gridIndexRange(),"mg.gridIndexRange()");
  
//    GenericGraphicsInterface & ps = *gip;
//    PlotIt::plot(ps,mg);
  
//  mg.updateReferences();

//   mg.update(MappedGrid::THEcenterJacobian | MappedGrid::THEinverseCenterDerivative );
//   display(mg.center(),"mg.center()");
//   display(mg.centerJacobian(),"mg.centerJacobian()");
//   display(mg.inverseCenterDerivative(),"inverseCenterDerivative");
  

    if( method==defaultMethod )
    {
      if( mg.isRectangular() )
	method=yee;
      else
	method=dsi;
    }
  
    if( mapPointer->decrementReferenceCount()==0 ) delete mapPointer;

  }
  else
  {
    // *********************************************
    // *********** CompositeGrid *******************
    // *********************************************

    CompositeGrid & cg = *cgp;

    //    cg.update( MappedGrid::THEcenter | MappedGrid::THEvertex | MappedGrid::THEmask );  

    if ( cg[0].getGridType()==MappedGrid::unstructuredGrid )
    {
      UnstructuredMapping &umap = (UnstructuredMapping &) cg[0].mapping().getMapping();
      umap.expandGhostBoundary();
      verifyUnstructuredConnectivity(umap,true);
      //	umap.expandGhostBoundary();
      //	verifyUnstructuredConnectivity(umap,true);

      cg.destroy( MappedGrid::THEcenter | MappedGrid::THEvertex | MappedGrid::THEmask |
		  MappedGrid::THEcorner | MappedGrid::THEcellVolume | MappedGrid::THEcenterNormal |
		  MappedGrid::THEfaceArea | MappedGrid::THEfaceNormal | 
		  MappedGrid::THEcellVolume  | MappedGrid::THEcenterArea );	

      cg.update( MappedGrid::THEcenter | MappedGrid::THEvertex | MappedGrid::THEmask |
		 MappedGrid::THEcorner | MappedGrid::THEcellVolume | MappedGrid::THEcenterNormal |
		 MappedGrid::THEfaceArea | MappedGrid::THEfaceNormal | 
		 MappedGrid::THEcellVolume  | MappedGrid::THEcenterArea );	

    }
    else
    {
      // cg.update(MappedGrid::THEmask );
      
      // *wdh* 031202 cg.update(MappedGrid::THEcenter | MappedGrid::THEvertex );  
    }

    pinterpolant = new Interpolant(cg);
#ifdef USE_PPP
    if( !pinterpolant->interpolationIsExplicit() )
    {
      printF("*** ERROR: The parallel composite grid interpolator needs explicit interpolation ****\n");
      Overture::abort();
    }
#endif

    //kkc can now use dsi, this would override the command line spec    method=nfdtd;
    
    // Set the default order of accuracy from the grid parameters
    int minDiscretizationWidth=INT_MAX;
    int minInterpolationWidth=INT_MAX;
    Range R=cg.numberOfDimensions();
    const IntegerArray & iw = cg.interpolationWidth;
    // iw.display("iw");
    for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    {
      MappedGrid & mg = cg[grid];
      const IntegerArray & dw = mg.discretizationWidth();
      
      // dw.display("dw");
      
      minDiscretizationWidth=min(minDiscretizationWidth,min(dw(R)));
      
      for( int grid2=0; grid2<cg.numberOfComponentGrids(); grid2++ )
      {
        if( grid!=grid2 )
  	  minInterpolationWidth=min( minInterpolationWidth,min(iw(R,grid,grid2)));
      }
    }
    if( minInterpolationWidth==INT_MAX ) minInterpolationWidth=minDiscretizationWidth;
    printF(" *** minDiscretizationWidth=%i, minInterpolationWidth=%i ****\n",minDiscretizationWidth,
	   minInterpolationWidth);

    const int maxOrderOfAccuracy=8;  // *************
    
    orderOfAccuracyInSpace=min(maxOrderOfAccuracy,minDiscretizationWidth-1,minInterpolationWidth-1);
    if( orderOfAccuracyInSpace%2 ==1 )
      orderOfAccuracyInSpace--;   // must be even
    
    orderOfAccuracyInTime =orderOfAccuracyInSpace;
    orderOfArtificialDissipation=orderOfAccuracyInSpace;
    
    printF("***Setting orderOfAccuracyInSpace=%i, orderOfAccuracyInTime=%i, orderOfArtificialDissipation=%i\n",
	   orderOfAccuracyInSpace,orderOfAccuracyInTime,orderOfArtificialDissipation);

    if( orderOfAccuracyInSpace>4 )
    {
      printF("***Setting useConservative=false by default for order of accuracy >4.\n");
      useConservative=false;
    }
    
  }

  timing(timeForInitialize)+=getCPU()-time0;

  // ********** By default do not solve for H in 3D ************
  if( cgp->numberOfDimensions()==3 )
  {
    solveForMagneticField=false;


  }
  else
  {
    kz=0; // *wdh* 040626 
  }
  
  // These next arrays hold (eps,mu,c) in the case when they are constant on each grid but
  // may vary from grid to grid
  const int numberOfComponentGrids = mgp!=NULL ? 1 : cgp->numberOfComponentGrids();
  epsGrid.redim(numberOfComponentGrids); epsGrid=eps;
  muGrid.redim(numberOfComponentGrids);  muGrid=mu;
  cGrid.redim(numberOfComponentGrids);   cGrid=c;

 // subtract out the incident field before apply NRBC's
  adjustFarFieldBoundariesForIncidentField.redim(numberOfComponentGrids);
  adjustFarFieldBoundariesForIncidentField=0;

  return 0;
}

//! Setup and initialization. Build the solution fields. 
int Maxwell::
setupGridFunctions()
// ===================================================================================
//    Build grid functions
// ===================================================================================
{
  real time0=getCPU();

  assert( cgp!=NULL );
  CompositeGrid & cg= *cgp;
  const int numberOfDimensions = cg.numberOfDimensions();

  int grid,side,axis;
  for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
  {
    MappedGrid & mg = cg[grid];

    if( mg.getGridType()==MappedGrid::structuredGrid )
    {
      for( axis=0; axis<mg.numberOfDimensions(); axis++ )
	for( side=0; side<=1; side++ )
	{
	  int bc0 = mg.boundaryCondition(side,axis);
	  if( bc0>0 )
	  {
	    if( bcOption==useAllDirichletBoundaryConditions )
	    {
	      mg.setBoundaryCondition(side,axis,dirichlet);
	    }
	    else if( bcOption==useAllPerfectElectricalConductorBoundaryConditions )
	    {
	      mg.setBoundaryCondition(side,axis,perfectElectricalConductor);
	    }
	  }
      
	}
    }
    else
    {
      UnstructuredMapping & umap = (UnstructuredMapping &) mg.mapping().getMapping();
      int dDim = umap.getDomainDimension();
      int rDim = umap.getRangeDimension();

	
      UnstructuredMapping::EntityTypeEnum etype,etypeBdy;
      etype = UnstructuredMapping::EntityTypeEnum(dDim);
      etypeBdy = UnstructuredMapping::Edge;//UnstructuredMapping::EntityTypeEnum(dDim-1);

      string ghostCellTag = string("Ghost ")+UnstructuredMapping::EntityTypeStrings[etype].c_str();
      string ghostBdyTag = string("Ghost ")+UnstructuredMapping::EntityTypeStrings[etypeBdy].c_str();
      string bctag = string("__bcnum ")+UnstructuredMapping::EntityTypeStrings[etypeBdy].c_str();
      string perTag = string("periodic ")+UnstructuredMapping::EntityTypeStrings[etype].c_str();
      // set a tag on all the bounding entities to specify the pec bc
      std::string pecTag("PEC BC");
      if ( bcOption )
      {
	int bcv = dirichlet;
	    
	if ( bcOption==useAllPerfectElectricalConductorBoundaryConditions )
	  bcv = perfectElectricalConductor;

	UnstructuredMapping::tag_entity_iterator git = umap.tag_entity_begin(ghostCellTag);
	UnstructuredMapping::tag_entity_iterator git_end = umap.tag_entity_end(ghostCellTag);

	UnstructuredMappingAdjacencyIterator ebdy,ebdy_end;

	int nPec=0;
	for ( ; git!=git_end; git++ )
	{
	  string bctagc = string("__bcnum ")+UnstructuredMapping::EntityTypeStrings[etype].c_str();
	  //		umap.addTag(git->et,git->e,bctagc,(void *)bcv);
	  umap.setBC(git->et,git->e,bcv);

	  etypeBdy = UnstructuredMapping::Edge;//UnstructuredMapping::EntityTypeEnum(dDim-1);
	  string ghostBdyTag = string("Ghost ")+UnstructuredMapping::EntityTypeStrings[etypeBdy].c_str();
	  string bctag = string("__bcnum ")+UnstructuredMapping::EntityTypeStrings[etypeBdy].c_str();
	  ebdy_end = umap.adjacency_end(*git,etypeBdy);
	  for ( ebdy=umap.adjacency_begin(*git, etypeBdy); ebdy!=ebdy_end; ebdy++ )
	  {
	    if ( !umap.hasTag(etypeBdy,*ebdy,bctag) )
	    {
			
	      if ( bcOption==useAllPerfectElectricalConductorBoundaryConditions )
		umap.addTag(etypeBdy,*ebdy,pecTag,0);

	      //			umap.addTag(etypeBdy,*ebdy,bctag,(void *)bcv);
	      umap.setBC(etypeBdy,*ebdy,bcv);
	      nPec++;
	    }

	    if ( umap.hasTag(git->et, git->e, perTag ) )
	      umap.deleteTag(git->et, git->e, perTag );
	  }

	  if ( rDim==3 )
	  {
	    etypeBdy = UnstructuredMapping::Face;//UnstructuredMapping::EntityTypeEnum(dDim-1);
	    ghostBdyTag = string("Ghost ")+UnstructuredMapping::EntityTypeStrings[etypeBdy].c_str();
	    bctag = string("__bcnum ")+UnstructuredMapping::EntityTypeStrings[etypeBdy].c_str();
	    ebdy_end = umap.adjacency_end(*git,etypeBdy);
	    for ( ebdy=umap.adjacency_begin(*git, etypeBdy); ebdy!=ebdy_end; ebdy++ )
	    {
	      if (  !umap.hasTag(etypeBdy,*ebdy,bctag) )
	      {
			    
		if ( bcOption==useAllPerfectElectricalConductorBoundaryConditions )
		  umap.addTag(etypeBdy,*ebdy,pecTag,0);
			    
		//			    umap.addTag(etypeBdy,*ebdy,bctag,(void *)bcv);
		umap.setBC(etypeBdy,*ebdy,bcv);
		nPec++;
	      }
			
	      if ( umap.hasTag(git->et, git->e, perTag ) )
		umap.deleteTag(git->et, git->e, perTag );
	    }
	  }
	}

	if ( !nPec )
	{
	  string bdyEdgeTag = string("boundary ")+UnstructuredMapping::EntityTypeStrings[int(UnstructuredMapping::Edge)].c_str();
	  string bdyFaceTag = string("boundary ")+UnstructuredMapping::EntityTypeStrings[int(UnstructuredMapping::Face)].c_str();
	  string bdyVertTag = string("boundary ")+UnstructuredMapping::EntityTypeStrings[int(UnstructuredMapping::Vertex)].c_str();
		
	  UnstructuredMapping::tag_entity_iterator bit = umap.tag_entity_begin(bdyEdgeTag);
	  UnstructuredMapping::tag_entity_iterator bit_end = umap.tag_entity_end(bdyEdgeTag);
		
	  UnstructuredMappingAdjacencyIterator ebdy,ebdy_end;

	  string bctagc = string("__bcnum ")+UnstructuredMapping::EntityTypeStrings[int(UnstructuredMapping::Edge)].c_str();
	  for ( ; bit!=bit_end; bit++ )
	  {
	    if ( !umap.hasTag(bit->et,bit->e,bctagc) )
	    {
	      //			umap.addTag(bit->et,bit->e,bctagc,(void *)bcv);
	      umap.setBC(bit->et,bit->e,bcv);
	      if ( bcOption==useAllPerfectElectricalConductorBoundaryConditions )
		umap.addTag(bit->et,bit->e,pecTag,0);
	      nPec++;
	    }
		    
	  }

	  bit = umap.tag_entity_begin(bdyVertTag);
	  bit_end = umap.tag_entity_end(bdyVertTag);
	  bctagc = string("__bcnum ")+UnstructuredMapping::EntityTypeStrings[int(UnstructuredMapping::Vertex)].c_str();
	  for ( ; bit!=bit_end; bit++ )
	  {
	    if ( !umap.hasTag(bit->et,bit->e,bctagc) )
	    {
	      //			umap.addTag(bit->et,bit->e,bctagc,(void *)bcv);
	      umap.setBC(bit->et,bit->e,bcv);
	      if ( bcOption==useAllPerfectElectricalConductorBoundaryConditions )
		umap.addTag(bit->et,bit->e,pecTag,0);
	      nPec++;
	    }
		    
	  }

	  if ( rDim==3 )
	  {
	    bit = umap.tag_entity_begin(bdyFaceTag);
	    bit_end = umap.tag_entity_end(bdyFaceTag);
	    bctagc = string("__bcnum ")+UnstructuredMapping::EntityTypeStrings[int(UnstructuredMapping::Face)].c_str();
	    for ( ; bit!=bit_end; bit++ )
	    {
	      if ( !umap.hasTag(bit->et,bit->e,bctagc) )
	      {
		//			    umap.addTag(bit->et,bit->e,bctagc,(void *)bcv);
		umap.setBC(bit->et,bit->e,bcv);
		if ( bcOption==useAllPerfectElectricalConductorBoundaryConditions )
		  umap.addTag(bit->et,bit->e,pecTag,0);
		nPec++;
	      }
	    }
	  }
		
	}

	cout<<"unstructured grid "<<grid<<" has "<<nPec<<" non-periodic boundary entities"<<endl;
      }
      cout<<"unstructured grid "<<grid<<" has "<<std::distance(umap.tag_entity_begin(perTag),umap.tag_entity_end(perTag))<<" periodic boundary entities"<<endl;
	
    }
  
  }

  numberOfTimeLevels=2;  // keep this many levels of u
  if( true && method==nfdtd )
  {
    if( usingPMLBoundaryConditions() ||
        true )  // for combined dissipation with advance
      numberOfTimeLevels=3;  // needed for PML 
  }

  int numberOfComponents;

  if ( method!=nfdtd )
  {// both fields are needed by the staggered schemes
    solveForElectricField = solveForMagneticField = true;
  }

  if( gridType!=compositeGrid )
  {
    // ****************************************
    // *********** single grid  ***************
    // ****************************************

    MappedGrid & mg = cg[0];
    Mapping & map = mg.mapping().getMapping();
  
    bool unstructured=map.getClassName()=="UnstructuredMapping";

    if( method==nfdtd && !mg.isRectangular() )
    {
      // The energy estimate may not be correct unless we ensure that the
      // jacobian derivatives are periodic
      mg.update( MappedGrid::THEinverseVertexDerivative );
      realMappedGridFunction & rx = mg.inverseVertexDerivative();
      rx.periodicUpdate();
    }

    if( method==defaultMethod )
    {
      if( mg.isRectangular() )
	method=yee;
      else
	method=dsi;
    }
  

    Range all;
    if( numberOfDimensions==2 )
    {
      numberOfComponents= method==yee || method==nfdtd ? 
	(int)numberOfComponentsRectangularGrid : 
	(int)numberOfComponentsCurvilinearGrid; // number of components
    }
    else
    {
      numberOfComponents=(int(solveForElectricField)+int(solveForMagneticField))*3;
    }
      
    //kkc 040304 ops apparently needed now for all schemes
    MappedGridOperators *ops = new MappedGridOperators(mg);
    this->op = ops; //kkc 040304 I guess we need this too?

    if ( method==nfdtd /* kkc 040310 method!=dsi && method!=yee*/ )
    {      
      fields = new realMappedGridFunction [numberOfTimeLevels];
      for( int n=0; n<numberOfTimeLevels; n++ )
      {
	fields[n].updateToMatchGrid(mg,all,all,all,numberOfComponents);
	fields[n]=0.;
	      
	if( numberOfDimensions==2 )
	{
	  if( numberOfComponents==3 )
	  {
	    fields[n].setName("Ex",ex);
	    fields[n].setName("Ey",ey);
	    fields[n].setName("Hz",hz);
	  }
	  else
	  {
	    fields[n].setName("Ex10",ex10);
	    fields[n].setName("Ey10",ey10);
	    fields[n].setName("Ex01",ex01);
	    fields[n].setName("Ey01",ey01);
	    fields[n].setName("Hz",hz11);
	  }
	  fields[n].setOperators(*ops);
	}
	else
	{
	  //kkc 040304 XXX  BILL, I assumed the operators were set for these guys since you
	  //            use them elsewhere 
	  int c=0;
	  if( solveForElectricField )
	  {
	    ex=c; fields[n].setName("Ex",ex); c++;
	    ey=c; fields[n].setName("Ey",ey); c++;
	    ez=c; fields[n].setName("Ez",ez); c++;
	  }
	  else
	  {
	    ex=c;  //  ex always points to first component
	  }
	  if( solveForMagneticField )
	  {
	    hx=c; fields[n].setName("Hx",hx); c++;
	    hy=c; fields[n].setName("Hy",hy); c++;
	    hz=c; fields[n].setName("Hz",hz); c++;
	  }
	  else
	  {
	    hz=c-1;   // hz always points to the last component
	  }
	}
      }

    }
    else
    {
      // yee and dsi gridfunction setup
      //	  fields = new realMappedGridFunction [3*numberOfTimeLevels];
      //kkc 040304 dsi and yee schemes now use faceCenteredAll gridfunctions
      fields = new realMappedGridFunction [2*numberOfTimeLevels];
      if ( numberOfDimensions==2 )
      {
	ex = 0;
	ey = ez = 1;
	hz = 0;
      }
      else
      {
	ex=hx=0;
	ey=hy=1;
	ez=hz=2;

      }

      MappedGridOperators *ops = new MappedGridOperators(mg);
      for( int n=0; n<numberOfTimeLevels; n++ )
      {
	if ( numberOfDimensions==2 )
	{
	  fields[n].updateToMatchGrid(mg,GridFunctionParameters::cellCentered);
	  if ( method==dsiMatVec )
	  {
	    //		      fields[n+numberOfTimeLevels].updateToMatchGrid(mg,GridFunctionParameters::faceCenteredAll);
	    //		      fields[n+numberOfTimeLevels].updateToMatchGrid(mg,GridFunctionParameters::edgeCentered,2);
	    fields[n+numberOfTimeLevels].updateToMatchGrid(mg,GridFunctionParameters::edgeCentered);
	    fields[n+numberOfTimeLevels].setName("E.n",ex);
	  }
	  else
	  {
	    fields[n+numberOfTimeLevels].updateToMatchGrid(mg,GridFunctionParameters::faceCenteredAll,mg.numberOfDimensions()); //XXX the last argument should go away when the dsi schemes are fully implemented
	    fields[n+numberOfTimeLevels].setName("Ex",ex);
	    fields[n+numberOfTimeLevels].setName("Ey",ey);
	  }
	  fields[n].setName("Hz",hz);
	}
	else
	{
	  if ( method!=dsiMatVec )
	  {
	    fields[n].updateToMatchGrid(mg,GridFunctionParameters::faceCenteredAll,numberOfDimensions);
	    fields[n+numberOfTimeLevels].updateToMatchGrid(mg,GridFunctionParameters::edgeCentered,numberOfDimensions); 		   
	    fields[n].setName("Hx",hx);
	    fields[n].setName("Hy",hy);
	    fields[n].setName("Hz",hz);
	    fields[n+numberOfTimeLevels].setName("Ex",ex);
	    fields[n+numberOfTimeLevels].setName("Ey",ey);
	    fields[n+numberOfTimeLevels].setName("Ez",ez);

	  }
	  else
	  {
	    fields[n].updateToMatchGrid(mg,GridFunctionParameters::faceCenteredAll);
	    fields[n+numberOfTimeLevels].updateToMatchGrid(mg,GridFunctionParameters::edgeCentered); 
	    fields[n].setName("H.n",0);
	    fields[n+numberOfTimeLevels].setName("E.n",0);
	  }

	}
	//	      fields[n+numberOfTimeLevels].updateToMatchGrid(mg,GridFunctionParameters::faceCenteredAxis1,all,all,all,2);
	//	      fields[n+2*numberOfTimeLevels].updateToMatchGrid(mg,GridFunctionParameters::faceCenteredAxis2,all,all,all,2);

	//	      fields[n].setName("Hz",hz);
	//	      fields[n+numberOfTimeLevels].setName("E{.n,x}",ex);
	//	      fields[n+numberOfTimeLevels].setName("Ey",ey);
// 	      fields[n+numberOfTimeLevels].setName("Ex{10}",ey);
// 	      fields[n+2*numberOfTimeLevels].setName("Ex01",ex);
// 	      fields[n+2*numberOfTimeLevels].setName("Ex01",ey);
	fields[n].setOperators(*ops);
	fields[n+numberOfTimeLevels].setOperators(*ops);
	//	      fields[n+2*numberOfTimeLevels].setOperators(*ops);
      }

    }
  
    if( artificialDissipation>0. )
    {
      dissipation=new realMappedGridFunction;
      if ( method==nfdtd || numberOfDimensions==2 )
	dissipation->updateToMatchGrid(mg,GridFunctionParameters::cellCentered);
      else
      {
	if ( method==dsiMatVec )
	  dissipation->updateToMatchGrid(mg,GridFunctionParameters::faceCenteredAll); 
	else
	  dissipation->updateToMatchGrid(mg,GridFunctionParameters::faceCenteredAll,numberOfDimensions); 
      }
	
      *dissipation = 0;
    }
  
    if( method==nfdtd )
    {
      op = new MappedGridOperators(mg);
      op->useConservativeApproximations(useConservative);
      op->setOrderOfAccuracy(orderOfAccuracyInSpace);
    }

  }
  else
  {
    // *********************************************
    // *********** CompositeGrid *******************
    // *********************************************

    if ( /*kkc 040310 method!=dsi*/ method==nfdtd ) //kkc dsi needs to work with unstructured grids...
    {
      Range all;
      if( numberOfDimensions==2 )
      {
	numberOfComponents= method==yee || method==nfdtd ? 
	  (int)numberOfComponentsRectangularGrid : 
	  (int)numberOfComponentsCurvilinearGrid; 
	    
      }
      else
      {
	numberOfComponents=(int(solveForElectricField)+int(solveForMagneticField))*3;
      }
	
      numberOfFields=2;
	
      cgfields = new realCompositeGridFunction [numberOfTimeLevels];
	
      for( int n=0; n<numberOfTimeLevels; n++ )
      {
	cgfields[n].updateToMatchGrid(cg,all,all,all,numberOfComponents);
	cgfields[n]=0.;
	    
	if( numberOfDimensions==2 )
	{
	  if( numberOfComponents==3 )
	  {
	    cgfields[n].setName("Ex",ex);
	    cgfields[n].setName("Ey",ey);
	    cgfields[n].setName("Hz",hz);
	  }
	  else
	  {
	    cgfields[n].setName("Ex10",ex10);
	    cgfields[n].setName("Ey10",ey10);
	    cgfields[n].setName("Ex01",ex01);
	    cgfields[n].setName("Ey01",ey01);
	    cgfields[n].setName("Hz",hz11);
	  }
	}
	else
	{
	  int c=0;
	  if( solveForElectricField )
	  {
	    ex=c; cgfields[n].setName("Ex",ex); c++;
	    ey=c; cgfields[n].setName("Ey",ey); c++;
	    ez=c; cgfields[n].setName("Ez",ez); c++;
	  }
	  if( solveForMagneticField )
	  {
	    hx=c; cgfields[n].setName("Hx",hx); c++;
	    hy=c; cgfields[n].setName("Hy",hy); c++;
	    hz=c; cgfields[n].setName("Hz",hz); c++;
	  }
		
	}
	
      }
	
      // *wdh* Do this as needed in computeDissipation
//  	if( artificialDissipation>0. )
//  	{
//  	  cgdissipation=new realCompositeGridFunction;
//  	  cgdissipation->updateToMatchGrid(cg,all,all,all,numberOfComponents);
//  	  if( numberOfComponents==3 )
//  	  {
//  	    cgdissipation->setName("Ex dissipation",ex);
//  	    cgdissipation->setName("Ey dissipation",ey);
//  	    cgdissipation->setName("Hz dissipation",hz);
//  	  }
//  	}
	
      if( method==nfdtd )
      {
	cgop = new CompositeGridOperators(cg);
	cgop->useConservativeApproximations(useConservative);
	cgop->setOrderOfAccuracy(orderOfAccuracyInSpace);
      }
    }
    else // dsi scheme
    {
      Range all;
      if ( numberOfDimensions==2 )
      {
	ex = 0;
	ey = ez = 1;
	hz = 0;
      }
      else
      {
	ex=hx=0;
	ey=hy=1;
	ez=hz=2;

      }

      CompositeGridOperators *ops = new CompositeGridOperators(cg);
      dsi_cgfieldsE0 = new realCompositeGridFunction[numberOfTimeLevels];
      //		  dsi_cgfieldsE1 = new realCompositeGridFunction[numberOfTimeLevels];
      dsi_cgfieldsH  = new realCompositeGridFunction[numberOfTimeLevels];
      for( int n=0; n<numberOfTimeLevels; n++ )
      {
	if ( numberOfDimensions==2 )
	{
	  dsi_cgfieldsH[n].updateToMatchGrid(cg,GridFunctionParameters::cellCentered);
	  if ( method==dsiMatVec )
	  {
	    //		      dsi_cgfieldsE0[n].updateToMatchGrid(cg,GridFunctionParameters::edgeCentered,2);
	    dsi_cgfieldsE0[n].updateToMatchGrid(cg,GridFunctionParameters::edgeCentered);
	    dsi_cgfieldsE0[n].setName("E.n",ex);
	    if( artificialDissipation>0. )
	    {
	      cgdissipation=new realCompositeGridFunction;
	      cgdissipation->updateToMatchGrid(cg,GridFunctionParameters::cellCentered);
	      cgdissipation->setName("H dissp",0);
	      e_cgdissipation=new realCompositeGridFunction;
	      e_cgdissipation->updateToMatchGrid(cg,GridFunctionParameters::edgeCentered);
	      e_cgdissipation->setName("E.n dissp",0);
	    }
	  }
	  else
	  {
	    dsi_cgfieldsE0[n].updateToMatchGrid(cg,GridFunctionParameters::faceCenteredAll,cg.numberOfDimensions()); //XXX the last argument should go away when the dsi schemes are fully implemented
	    dsi_cgfieldsE0[n].setName("Ex",ex);
	    dsi_cgfieldsE0[n].setName("Ey",ey);
	  }
	  dsi_cgfieldsH[n].setName("Hz",hz);
	}
	else
	{

	  if ( method!=dsiMatVec )
	  {
	    dsi_cgfieldsH[n].updateToMatchGrid(cg,GridFunctionParameters::faceCenteredAll,cg.numberOfDimensions());
	    dsi_cgfieldsE0[n].updateToMatchGrid(cg,GridFunctionParameters::edgeCentered,cg.numberOfDimensions()); 		   
	    dsi_cgfieldsH[n].setName("Hx",hx);
	    dsi_cgfieldsH[n].setName("Hy",hy);
	    dsi_cgfieldsH[n].setName("Hz",hz);
	    dsi_cgfieldsE0[n].setName("Ex",ex);
	    dsi_cgfieldsE0[n].setName("Ey",ey);
	    dsi_cgfieldsE0[n].setName("Ez",ez);

	  }
	  else
	  {
	    dsi_cgfieldsH[n].updateToMatchGrid(cg,GridFunctionParameters::faceCenteredAll);
	    dsi_cgfieldsE0[n].updateToMatchGrid(cg,GridFunctionParameters::edgeCentered); 
	    dsi_cgfieldsH[n].setName("H.n",0);
	    dsi_cgfieldsE0[n].setName("E.n",0);
	    if( artificialDissipation>0. && !cgdissipation )
	    {
	      cgdissipation=new realCompositeGridFunction;
	      cgdissipation->updateToMatchGrid(cg,GridFunctionParameters::faceCenteredAll);
	      cgdissipation->setName("H dissp",0);

	      e_cgdissipation=new realCompositeGridFunction;
	      e_cgdissipation->updateToMatchGrid(cg,GridFunctionParameters::edgeCentered);
	      e_cgdissipation->setName("E dissp",0);
	    }

	  }

	}
	//	      fields[n+numberOfTimeLevels].updateToMatchGrid(mg,GridFunctionParameters::faceCenteredAxis1,all,all,all,2);
	//	      fields[n+2*numberOfTimeLevels].updateToMatchGrid(mg,GridFunctionParameters::faceCenteredAxis2,all,all,all,2);

	//	      fields[n].setName("Hz",hz);
	//	      fields[n+numberOfTimeLevels].setName("E{.n,x}",ex);
	//	      fields[n+numberOfTimeLevels].setName("Ey",ey);
// 	      fields[n+numberOfTimeLevels].setName("Ex{10}",ey);
// 	      fields[n+2*numberOfTimeLevels].setName("Ex01",ex);
// 	      fields[n+2*numberOfTimeLevels].setName("Ex01",ey);
	dsi_cgfieldsH[n].setOperators(*ops);
	dsi_cgfieldsE0[n].setOperators(*ops);
	//	      fields[n+2*numberOfTimeLevels].setOperators(*ops);
      }

#if 0
      ex = 0;
      ey = 1;
      hz = 0;

      for( int n=0; n<numberOfTimeLevels; n++ )
      {
	dsi_cgfieldsE0[n].updateToMatchGrid(cg,GridFunctionParameters::faceCenteredAll,all,all,all,Range(2));
	dsi_cgfieldsE1[n].updateToMatchGrid(cg,GridFunctionParameters::faceCenteredAll,all,all,all,Range(2));
	dsi_cgfieldsH[n].updateToMatchGrid(cg,GridFunctionParameters::cellCentered,all,all,all);

	dsi_cgfieldsE0[n].setName("Ex{10}",ex);
	dsi_cgfieldsE0[n].setName("Ey{10}",ey);
	dsi_cgfieldsE1[n].setName("Ex01",ex);
	dsi_cgfieldsE1[n].setName("Ey01",ey);
	dsi_cgfieldsH[n].setName("Hz",hz);
      }
#endif
    }
  }
  
  if( method==yee )
  {
    methodName="Yee";
  }
  else if( method==dsi )
  {
    methodName="DSI";
  }
  else if ( method==dsiMatVec)
  {
    methodName="DSI-MatVec";
  }
  else if( method==nfdtd )
  {
    methodName="NFDTD";
  }
  else if( method==dsiNew )
  {
    methodName="DSI-new";
  }
  else
  {
    printF("Maxwell:setup:ERROR: unknown method=%i\n",(int)method);
    throw "error";
  }

  dxMinMax.redim(cgp->numberOfComponentGrids(),2);
  dxMinMax=0.;
  
  if( useVariableDissipation && method==nfdtd )
    buildVariableDissipation();

  if ( method==dsiMatVec )
  {
    real tm0 = getCPU();
    setupDSICoefficients();
    timing(timeForDSIMatrix) += getCPU()-tm0;
  }

  if( initialConditionOption==annulusEigenfunctionInitialCondition )
  {
    // try to guess the cylinder radius and length

    if( cg[0].isRectangular() )
    {
      // assume grid 0 is the inner core 
      MappedGrid & mg = cg[0];
      real dx[3]={1.,1.,1.}, xab[2][3]={0.,0.,0.,0.,0.,0.};
      mg.getRectangularGridParameters( dx, xab );

      cylinderAxisStart=min(xab[0][2],xab[1][2]);
      cylinderAxisEnd  =max(xab[0][2],xab[1][2]);
 
      printF(" setupGridFunctions: I guess that the cylinder extends from [%8.2e,%8.2e] in the axial direction\n",
             cylinderAxisStart,cylinderAxisEnd    );
	
    }
      
  }

  rc=numberOfComponents;  // position of density in TZ functions

  if( useChargeDensity && pRho==NULL )
  {
    // allocate the grid function that holds the charge density
    pRho=new realCompositeGridFunction(cg);
  }

  numberOfSequences=numberOfComponents;
  if( computeEnergy )
    numberOfSequences+=2; // save the energy and delta(energy)

  computeTimeStep();
  timing(timeForInitialize)+=getCPU()-time0;

  
  return 0;
}

