#include "Maxwell.h"
#include "PlotStuff.h"
#include "CompositeGridOperators.h"
#include "Oges.h"
#include "ParallelUtility.h"
#include "App.h"
#include "Ogshow.h"

void Maxwell::
displayBoundaryConditions(FILE *file /* = stdout */)
{
//===================================================================================
// /Description:
//   Print names for boundary conditions
//
// /file (input) : write to this file.
//===================================================================================
  assert( file!=NULL );
  assert( cgp!=NULL );
  CompositeGrid & cg= *cgp;

  int maxNameLength=3;
  int grid;
  for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    maxNameLength=max( maxNameLength,cg[grid].getName().length());

  char buff[80];
  sPrintF(buff," %%4i: %%%is     %%i    %%i    %%3i :",maxNameLength); // build a format string


  aString blanks="                                                                           ";
  fPrintF(file," grid   name%s side axis    boundary condition and name\n",
           (const char *)blanks(0,min(maxNameLength-3,blanks.length()-1)));
  fPrintF(file," ----   ----%s ---- ----    ---------------------------\n",
           (const char *)blanks(0,min(maxNameLength-3,blanks.length()-1)));
  for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
  {
    for( int axis=axis1; axis<cg.numberOfDimensions(); axis++ )
    for( int side=Start; side<=End; side++ )
    {
      int bc=cg[grid].boundaryCondition()(side,axis);
      fPrintF(file,buff,grid,(const char *)cg[grid].getName(),side,axis,bc);

      if( bc > 0 && bc<numberOfBCNames)
        fPrintF(file," %s \n",(const char*)bcName[bc]);
      else if( bc==0 )
        fPrintF(file," %s \n","none");
      else if( bc<0 )
        fPrintF(file," %s \n","periodic");
      else
        fPrintF(file," %s \n","unknown");
    }
  }
}

void Maxwell::
computeNumberOfStepsAndAdjustTheTimeStep(const real & t,
                                         const real & tFinal,
					 const real & nextTimeToPlot, 
					 int & numberOfSubSteps, 
					 real & dtNew,
                                         const bool & adjustTimeStep /* = TRUE */ )
// =====================================================================================
// /Description:
// determine how many steps we should take to reach the next output time, but do
// not take more than `maximumStepsBetweenComputingDt'
//
// /nextTimeToPlot (input):  
// /numberOfSteps (output): Take this many sub-steps
// /dtNew (input/output) : On input this is the time step. On output this value may be changed
//     if adjustTimeStep=TRUE.
// /adjustTimeStep (input) : if TRUE alter the time step to exactly reach the nexTimeToPrint.
//\end{CompositeGridSolverInclude.tex}  
// =====================================================================================
{
  if ( dtNew<=REAL_MIN && tFinal<=REAL_MIN )
    { // kkc I use dt=0 to test the interpolation in the dsi scheme
      if ( adjustTimeStep ) dtNew=0.;
       return;
    }
  else if( dtNew<=0. )
  {
    printF("\n\ncomputeNumberOfStepsAndAdjustTheTimeStep:ERROR: dtNew<=0., dtNew=%e\n",dtNew);
    printF(" t=%e, tFinal=%e, nextTimeToPlot=%e, tPlot=%e \n",t,tFinal,nextTimeToPlot,tPlot);
    Overture::abort("error");
  }
  
  real timeInterval=min(tFinal,nextTimeToPlot)-t;
  if(  timeInterval<=0. )
    timeInterval+=tPlot;
  assert( timeInterval>0. );

  if( timeInterval/dtNew > INT_MAX ) 
  {
    printF("computeNumberOfStepsAndAdjustTheTimeStep:ERROR: time step too small? dtNew=%e, timeInterval=%e \n",dtNew,
           timeInterval);
    printF(" t=%e, tFinal=%e, nextTimeToPlot=%e, tPrint=%e \n",t,tFinal,nextTimeToPlot,tPlot);
    Overture::abort("error");
  }
  
  numberOfSubSteps=max(1,int(timeInterval/dtNew+.9999));   // used to be +.5 or +1.

  const int maximumStepsBetweenComputingDt=INT_MAX;
    
  if( numberOfSubSteps > maximumStepsBetweenComputingDt )
  {
    // no need to adjust dt in this case since we will recompute dt anyway
    numberOfSubSteps=maximumStepsBetweenComputingDt;
  }
  else if( adjustTimeStep )
  {
    dtNew=timeInterval/numberOfSubSteps;
    if( true || debug & 2 )
      printF(" ---- adjust time step: timeInterval=%e, numberOfSubSteps=%i, dtNew=%e\n",
	     timeInterval,numberOfSubSteps,dtNew);
  }

  if( dtNew<0. )
  {
    printF("computeNumberOfStepsAndAdjustTheTimeStep:ERROR: dtNew<=0., dtNew=%e, numberOfSubSteps=%i\n",dtNew,
           numberOfSubSteps);
    printF(" t=%e, tFinal=%e, nextTimeToPlot=%e, tPrint=%e timeInterval=%e\n",t,tFinal,nextTimeToPlot,
            tPlot,timeInterval);
    Overture::abort("error");
  }

}

int Maxwell::
outputResultsAfterEachTimeStep( int current, real t, real dt, int stepNumber, real nextTimeToPlot )
// ===================================================================================
// /Description:
//     Save any results that need to be output at every time step:
//          (1) values at probes.
// ===================================================================================
{
  const bool okToOutput = myid<=0;
  
  if( probes.size(1)>0 && (stepNumber % frequencyToSaveProbes)==0 )
  {
    assert( cgp!=NULL );
    CompositeGrid & cg= *cgp;

    // output probe values
    if( probeFile==NULL && okToOutput )
    {
      // remove leading blanks in name
      int i=0, len=probeFileName.length();
      while( i<len && probeFileName[i]==' ' ) i++;
      probeFileName=probeFileName(i,len-1);
      if( okToOutput ) printF("Opening the probe file named [%s]\n",(const char*)probeFileName);
      probeFile=fopen((const char*)probeFileName,"w" ); 
    }
    
    // Write probe data in the format
    //     t_1 ex_1 ey_1 hz_1
    //     t_2 ex_2 ey_2 hz_2
    //     t_3 ex_3 ey_3 hz_3
    //       ...
    const int numberOfProbes=probes.size(1);
    if( okToOutput ) fPrintF(probeFile,"%e ",t);
    IntegerArray uprobeLoc(1);
    RealArray uprobeVal(1,3);

    for( int i=0; i<numberOfProbes; i++ )
    {
      int grid=probeGridLocation(3,i);
      assert( grid>=0 && grid<cg.numberOfComponentGrids() );
      realMappedGridFunction & u =mgp!=NULL ? fields[current] : getCGField(HField,current)[grid];//cgfields[current][grid];

      const int i1=probeGridLocation(0,i);
      const int i2=probeGridLocation(1,i);
      const int i3=probeGridLocation(2,i);
      
      if ( cg[grid].getGridType()==MappedGrid::structuredGrid )
      {
	if( cg.numberOfDimensions()==2 )
	{
	  real uex=u(i1,i2,i3,ex), uey=u(i1,i2,i3,ey), uhz=u(i1,i2,i3,hz);
	  if( okToOutput ) fPrintF(probeFile,"%e %e %e ",uex,uey,uhz);
	}
	else
	{
	  real uex=u(i1,i2,i3,ex), uey=u(i1,i2,i3,ey), uez=u(i1,i2,i3,ez);
	  if( okToOutput ) fPrintF(probeFile,"%e %e %e ",uex,uey,uez);
	}
      }
      else
      {
	uprobeLoc(0) = i1;
	reconstructDSIAtEntities(t, HField, uprobeLoc, u, uprobeVal);
	if ( cg.numberOfDimensions()==2 && okToOutput)
	  fPrintF(probeFile,"%e ",uprobeVal(0));
	else if ( okToOutput )
	  fPrintF(probeFile,"%e %e %e ",uprobeVal(0),uprobeVal(1),uprobeVal(2));
	      
      }
      
      
    }
    if( okToOutput ) fPrintF(probeFile,"\n");
    
  }


  if( plotIntensity )
  {
    computeIntensity(current,t,dt,stepNumber,nextTimeToPlot);
  }


  return 0;
}

namespace
{
int numInSum=-1;
real integralOne=0.; // for testing, integrate a constant function, f(t)=1, in time 
}

  

// =======================================================================================
/// \brief Compute the time averaged intensity.
/// \details Compute the time averaged intensity to be output at t=nextTimeToPlot
///         
/// ======================================================================================
int Maxwell::
computeIntensity(int current, real t, real dt, int stepNumber, real nextTimeToPlot)
{
  if( !plotIntensity || nextTimeToPlot<=0. ) 
    return 1;

  assert( cgp!=NULL );
  CompositeGrid & cg= *cgp;

  
// exTrue(x,y,t) sin(twoPi*(kx*(x)+ky*(y)-cc*(t)))*(-ky/(eps*cc))

  int gridp=0;  // ** fix this *** we need to know which grid to get c from in order to compute the period.

  const real c = cGrid(gridp);
  const real cc= c*sqrt( kx*kx+ky*ky +kz*kz);
  real period = 1./cc;

  if( period > tPlot )
  {
    printF("computeIntensity:ERROR: period=%8.2e > tPlot=%8.2e : unable to compute the time average energy density\n",
           period,tPlot);
    return 1;
  }
  
  if( pIntensity==NULL )
  {
    pIntensity = new realCompositeGridFunction(cg);
    numInSum=-1;
  }

  const real ta = nextTimeToPlot-period;
  const real tb=nextTimeToPlot;
  if( t+dt <= ta )
  {
    numInSum=-1;
    return 0;
  }
  
  //           <------- period ----->
  // +------+--|--+-----+-----+-----+
  //           ta       <-dt->      tb 
  //        t0    t1    t2 ...      tm

  //  Integral = .5*( fa+f1)*(t1-ta) + .5*( f1+f2 )*dt + .5*( f2+f3 ) + ... 
  //             .5*( (1-alpha)*f0 + alpha*f1 + f1)*(t1-ta) + .5*( f1+f2 )*dt + .5*( f2+f3 ) + ...
  //      fa = (1-alpha)*f0 + alpha*f1
  //      alpha = (ta-t0)/dt 

  real weight=dt;
  if( numInSum<0 )
  {
    // This is the first time through
    numInSum=0;
    integralOne=0.;

    assign(*pIntensity,0.);

    real t0=t, t1=t+dt;
    real alpha=(ta-t0)/dt;
    weight = .5*(1.-alpha)*(t1-ta);
    printF("computeIntensity:(1) t=%8.2e: ta=%8.2e, tb=%8.2e, alpha=%8.2e\n",t,ta,tb,alpha);
  }
  else if( numInSum==1 )
  {
    // second time through 
    real t0=t-dt, t1=t;
    real alpha=(ta-t0)/dt;
    weight= .5*(1.+alpha)*(t1-ta) +.5*dt;
    printF("computeTimeAveragedIntensity:(2) t=%8.2e: ta=%8.2e, tb=%8.2e, alpha=%8.2e\n",t,ta,tb,alpha);
  }
  if( fabs(t-nextTimeToPlot)<dt*.01 )
  {
    printF("computeTimeAveragedIntensity:(3) t=%8.2e: nextTimeToPlot=%8.2e, dt=%8.2e\n",t,nextTimeToPlot,dt);
    weight=.5*dt;  // last entry in sum
  } 

  numInSum++;
  
  realCompositeGridFunction & intensity = *pIntensity;

  for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
  {
    const real eps = epsGrid(grid);
    const real mu = muGrid(grid);
    // const real cc= c*sqrt( kx*kx+ky*ky +kz*kz);

    MappedGrid & mg = cg[grid];
    realMappedGridFunction & u = cgfields[current][grid];
    realMappedGridFunction & v = intensity[grid];

#ifdef USE_PPP
    realSerialArray uLocal; getLocalArrayWithGhostBoundaries(u,uLocal);
    realSerialArray vLocal; getLocalArrayWithGhostBoundaries(v,vLocal);
#else
    const realSerialArray & uLocal = u;
    const realSerialArray & vLocal = v;
#endif


    Index Iv[3], &I1=Iv[0], &I2=Iv[1], &I3=Iv[2];
    getIndex(mg.dimension(),I1,I2,I3);
    const int includeGhost=1;
    bool ok = ParallelUtility::getLocalArrayBounds(u,uLocal,I1,I2,I3,includeGhost); 

    real factor = weight*cGrid(grid)*eps/period;

    integralOne += weight/period;

    if( ok )
    {
	
      if( mg.numberOfDimensions()==2 )
      {
	// vLocal(I1,I2,I3)+= factor*( SQR(uLocal(I1,I2,I3,ex))+SQR(uLocal(I1,I2,I3,ey)) );

        // compute the intensity as (eps/2)*E^2 + (mu/2)*H^2 
	vLocal(I1,I2,I3)+= (.5*factor)*( SQR(uLocal(I1,I2,I3,ex))+SQR(uLocal(I1,I2,I3,ey)) ) +
                           (.5*factor*mu/eps)*( SQR(uLocal(I1,I2,I3,hz) ) );
      }
      else
      {
	Overture::abort("finish me -- we need H here");
	vLocal(I1,I2,I3)+= factor*( SQR(uLocal(I1,I2,I3,ex))+SQR(uLocal(I1,I2,I3,ey))+
				    SQR(uLocal(I1,I2,I3,ez)) );
      }
    } // end if ok 
  } // end for grid
  
  if( fabs(t-nextTimeToPlot)<dt*.01 )
  {
    printF("computeIntensity: t=%8.2e: period=%8.2e, integralOne=%8.2e, err=%8.2e\n",
	   t,period,integralOne,fabs(1.-integralOne));
  }
  return 0;
}



int Maxwell::
outputResults( int current, real t, real dt )
// ===================================================================================
// /Description:
//     Save any results after time intervals of tPlot
// 
// ===================================================================================
{
  // *************************************************************
  // write to the check file for regression and convergence tests
  // *************************************************************

  // kkc fix range to work with staggered schemes
  int base = ex;
  int bound = method==nfdtd ? hz : max(ey,ez) + hz + 1;
  Range C(base,bound);  // fix this 

  int numberOfComponents= C.length(); 
  int numberToOutput= numberOfComponents+1; // include div(E)
//    printF("\n ***** outputResults: maximumError.getLength(0)=%i numberOfComponents=%i\n",
//  	 maximumError.getLength(0),numberOfComponents);
  
  if( maximumError.getLength(0)>= (numberToOutput-1) && myid==0 )
  {
    fPrintF(checkFile,"%9.2e %i  ",t,numberToOutput);
    int c;
    real err,uc;
    for( c=0; c<numberToOutput-1; c++ )
    {
      err = maximumError(c); // error(c) > checkFileCutoff(c) ? error(c) : 0.;
      uc = solutionNorm(c); 
      // if( uc<checkFileCutoff(c) ) uc=0.;
      fPrintF(checkFile,"%i %9.2e %10.3e  ",c,err,uc);
    }

    c=numberToOutput-1;
    err=divEMax/max(REAL_MIN*100.,gradEMax);
    uc=gradEMax;
    if ( method==nfdtd )
      fPrintF(checkFile,"%i %9.2e %10.3e  ",c,err,uc);
    else
      fPrintF(checkFile,"%i %9.2e %10.3e  ",c,divEMax,gradEMax); // kkc this is actually divEMax and divHMax (divH stored in gradE)
    fPrintF(checkFile,"\n");

    // *************************************************************
    if( computeEnergy )
    {
      maximumError(numberOfComponents )=totalEnergy;
      maximumError(numberOfComponents+1)=totalEnergy-initialTotalEnergy;
    }
    
    saveSequenceInfo( t, maximumError );

  }
  else if( solutionNorm.getLength(0)>= (numberToOutput-1) && myid<=0 )
  {
    for( int fileio=0; fileio<2; fileio++ )
    {
      FILE *output = fileio==0 ? logFile : stdout;

      if ( method==nfdtd )
	fPrintF(output,"-->t=%10.4e dt=%8.2e |div(E)|=%8.2e |div(E)|/|grad(E)|=%8.2e, |grad(E)|=%8.2e, max(u):[",
                t,dt,divEMax,divEMax/max(REAL_MIN*100.,gradEMax),gradEMax);
      else
	fPrintF(output,"-->t=%10.4e dt=%8.2e |div(E)|=%8.2e |div(H)=%8.2e, max(u):[",
                t,dt,divEMax,gradEMax);
      for( int c=C.getBase(); c<=C.getBound(); c++ )
	fPrintF(output,"%8.2e,",solutionNorm(c));

      fPrintF(output,"]\n");
	  
    }
  }

  // compute the current energy (if computeEnergy==true)
  getEnergy( current, t, dt );

  return 0;
}


//! Solve the equations
int Maxwell::
solve(GL_GraphicsInterface &gi )
{
  real time0=getCPU();
  
  Range all;
  assert( cgp!=NULL );
  CompositeGrid & cg= *cgp;

  // ************** move this to setup ****************************
#if 0
  checkErrors= checkErrors &&
             (
               (initialConditionOption==planeWaveInitialCondition&&bcOption==useAllDirichletBoundaryConditions) || 
               (initialConditionOption==planeWaveInitialCondition&&bcOption==useAllPeriodicBoundaryConditions) ||
	       
	       forcingOption==twilightZoneForcing ||
	       forcingOption==planeWaveBoundaryForcing ||
	       initialConditionOption==gaussianPlaneWave ||
               initialConditionOption==squareEigenfunctionInitialCondition ||
               initialConditionOption==annulusEigenfunctionInitialCondition ||
               compareToReferenceShowFile );
#endif
  // printF("\n ****** solve: checkErrors=%i *****\n",checkErrors);
  
  if( mgp!=NULL )
  {

    MappedGrid & mg = *mgp;

    if( errp==NULL && checkErrors ) 
    {
      // Create a grid function to hold the errors for plotting
      if ( method==nfdtd )
      {
	int numberOfComponents = fields[0].getLength(3);
	errp = new realMappedGridFunction[1];
	errp->updateToMatchGrid(mg,all,all,all,numberOfComponents);
	errp->setName("Ex error",ex);
	errp->setName("Ey error",ey);
	errp->setName("Hz error",hz);
      }
      else
      {
	errp = new realMappedGridFunction[2];
	if ( cg.numberOfDimensions()==2 )
	{
	  errp[0].updateToMatchGrid(mg,GridFunctionParameters::edgeCentered,2);
	  errp[1].updateToMatchGrid(mg,GridFunctionParameters::cellCentered);
	  errp[0].setName("Ex error",ex);
	  errp[0].setName("Ey error",ey);
	  errp[1].setName("Hz error",hz);
	}
	else
	{
	  errp[0].updateToMatchGrid(mg,GridFunctionParameters::edgeCentered,3);
	  errp[0].setName("Ex error",ex);
	  errp[0].setName("Ey error",ey);
	  errp[0].setName("Ez error",ez);
	  errp[1].updateToMatchGrid(mg,GridFunctionParameters::faceCenteredAll,3);
	  errp[1].setName("Hx error",hx);
	  errp[1].setName("Hy error",hy);
	  errp[1].setName("Hz error",hz);
	}
      }
    }
  }
  else
  {
    if( checkErrors )//&& cg[0].getGridType()==MappedGrid::structuredGrid )
    {
      if ( method==nfdtd )
      {
	cgerrp = new realCompositeGridFunction [1];
	int numberOfComponents = cgfields[0][0].getLength(3);
	cgerrp->updateToMatchGrid(*cgp,all,all,all,numberOfComponents);
	
	if( cg.numberOfDimensions()==2 )
	{
	  assert( numberOfComponents==3 );
	  cgerrp->setName("Ex error",ex);
	  cgerrp->setName("Ey error",ey);
	  cgerrp->setName("Hz error",hz);
	}
	else
	{
	  if( solveForElectricField )
	  {
	    cgerrp->setName("Ex error",ex);
	    cgerrp->setName("Ey error",ey);
	    cgerrp->setName("Ez error",ez);
	  }
	  if( solveForMagneticField )
	  {
	    cgerrp->setName("Hx error",hx);
	    cgerrp->setName("Hy error",hy);
	    cgerrp->setName("Hz error",hz);
	  }
	      
	}
      }
      else
      {
	cgerrp = new realCompositeGridFunction [2];
	if ( cg.numberOfDimensions()==2 )
	{
	  cgerrp[0].updateToMatchGrid(*cgp,GridFunctionParameters::edgeCentered,2);
	  cgerrp[1].updateToMatchGrid(*cgp,GridFunctionParameters::cellCentered);
	  cgerrp[0].setName("Ex error",ex);
	  cgerrp[0].setName("Ey error",ey);
	  cgerrp[1].setName("Hz error",hz);	  
	}
	else
	{
	  cgerrp[0].updateToMatchGrid(*cgp,GridFunctionParameters::edgeCentered,3);
	  cgerrp[1].updateToMatchGrid(*cgp,GridFunctionParameters::faceCentered,3);

	  cgerrp[0].setName("Ex error",ex);
	  cgerrp[0].setName("Ey error",ey);
	  cgerrp[0].setName("Ez error",ez);
	      
	  cgerrp[1].setName("Hx error",hx);
	  cgerrp[1].setName("Hy error",hy);
	  cgerrp[1].setName("Hz error",hz);

	}
      }
    }
    
  }
    
  
  // realMappedGridFunction & err = *errp;

  real t = 0.;
  real dtb2,dtNew;
  real & dt= deltaT;
  int maxNumberOfTimeSteps=INT_MAX;
  int numberOfSubSteps;
  
  int current=0;
  int next=1;

  // ******************** fix this **************
  real dx[3];
  cg[0].getDeltaX(dx);

  // adjust the time step so we reach tPlot exactly
  //   dt=deltaT is computed in computeTimeStep in setup

  if ( tPlot<=0 ) //kkc 040304
    tPlot = tFinal;


  if( true )
  {
    // sanity check for dt 
    real dtmx = ParallelUtility::getMaxValue(dt);
    real dtmn = ParallelUtility::getMinValue(dt);
    if( dtmx!=dtmn )
    {
      printf(" ERROR: dt=%10.4e, dtmx=%10.4e dtmn=%10.4e, myid=%i \n",dt,dtmx,dtmn,myid);
      Overture::abort("Error");
    }
  }
  
  // We want dt*numStepsToPlot = tPlot
  const real dt0=dt;
  int numStepsToPlot = int(std::ceil( tPlot/dt ));// kkc is this what you mean?int(tPlot/dt+.999999);
  // int numSteps = int(tFinal/dt+1.);
  // numSteps = (numSteps/numStepsToPlot)*numStepsToPlot;

  // *wdh int numSteps= int(std::ceil( tFinal/dt ));//kkc is this what you meant?int(tFinal/tPlot+.999999)*numStepsToPlot;
  // we choose the number of steps so that we exactly reach times that are multiples of tPlot
  int numSteps= int(ceil(tFinal/tPlot))*numStepsToPlot; 

  dt= numSteps ? tFinal/numSteps : 0 ;
  
  printF("--- dt0=%9.3e dt=%9.3e tPlot=%f, tPlot/dt=%f numStepsToPlot=%i numSteps=%i \n",dt0,dt,tPlot,tPlot/dt,numStepsToPlot,numSteps);
  if( debug & 1 )
  {
    fprintf(pDebugFile,"--- dt0=%9.3e dt=%9.3e tPlot=%f, tPlot/dt=%f numStepsToPlot=%i numSteps=%i \n",
	    dt0,dt,tPlot,tPlot/dt,numStepsToPlot,numSteps);
    fflush(0);
  }
  assert( dt<= dt0 );
  

  // Initial conditions
  assignInitialConditions( current,t,dt ); 
  dtNew=dt;

  real nextTimeToPlot=0.;
//    nextTimeToPlot=min(int(t/tPlot+.5)*tPlot+tPlot,tFinal);   //  ...new time to print:
//    computeNumberOfStepsAndAdjustTheTimeStep(t,tFinal,nextTimeToPlot,numberOfSubSteps,dtNew );
//    dt=dtNew;

   
  for( int fileio=0; fileio<2; fileio++ )
  {
    FILE *file = fileio==0 ? logFile : stdout; 
    fPrintF(file,"\n"
	    "******************************************************************\n"
	    "           Cgmx : Maxwell Solver Version 0.0b                       \n"
	    "           ----------------------------------                       \n");

    fPrintF(file," Using method %s\n",(const char *)methodName);

    if( timeSteppingMethod==modifiedEquationTimeStepping )
      fPrintF(file," Time stepping method is modifiedEquation\n");
    

    fPrintF(file," order of accuracy: space=%i, time=%i\n",orderOfAccuracyInSpace,orderOfAccuracyInTime);
    fPrintF(file," artificial diffusion: coefficient=%8.2e, order=%i\n",artificialDissipation,
	    orderOfArtificialDissipation);
    fPrintF(file," divergence damping coefficient=%8.2e\n",divergenceDamping);
            
    if( forcingOption==twilightZoneForcing )
      fPrintF(file," Twilightzone flow is on.");
    if( forcingOption==twilightZoneForcing && twilightZoneOption==polynomialTwilightZone )
      fPrintF(file," Polynomial solution, degreeSpace=%i, degreeTime=%i\n",degreeSpace,degreeTime);
    else
      fPrintF(file,"\n");

    if( method==nfdtd )
    {
      MappedGridOperators & mgop = mgp!=NULL ? *op : (*cgop)[0];
      fPrintF(file," curvilinear grid operators use %s difference approximation\n",
	      mgop.usingConservativeApproximations() ? "conservative" : "non-conservative" );
      
    }
    if( projectFields && poisson!=NULL  )
    {
      OgesParameters & ogesParameters=poisson->parameters;
      if( !poisson->isSolverIterative() )
      {
	fPrintF(file," project fields using solver: solver=%s, \n",
		(const char*)ogesParameters.getSolverName()); 
      }
      else
      {
	real tolerance;
	int maximumNumberOfIterations;
      
	ogesParameters.get(OgesParameters::THErelativeTolerance,tolerance);
	ogesParameters.get(OgesParameters::THEmaximumNumberOfIterations,maximumNumberOfIterations);
	fPrintF(file," project fields using solver: solver=%s, \n"
		"                         : tolerance=%8.2e, max number of iterations=%i (0=choose default)\n",
		(const char*)ogesParameters.getSolverName(),tolerance,maximumNumberOfIterations);
      }
	
    }
    if( useVariableDissipation )
    {
      fPrintF(file," use variable dissipation (only add dissipation near interpolation points\n");
    }
    if(  max(adjustFarFieldBoundariesForIncidentField)>0 )
    {
      fPrintF(file," adjust far field boundaries for any incident fields.\n");
    }
    
    if( true )
    {
      fPrintF(file," PML parameters: width=%i, strength=%g, power=%i\n",numberLinesForPML,pmlLayerStrength,pmlPower);
    }
    

    fPrintF(file,"  tFinal=%f, dt=%9.3e, tPlot=%9.3e cfl=%3.2f ad=%3.2f  \n",
	    tFinal,dt,tPlot,cfl,artificialDissipation );

    int maxNameLength=3;
    int grid;
    for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
      maxNameLength=max( maxNameLength,cg[grid].getName().length());

    aString blanks="                                                                           ";
    fPrintF(file,"               Grid Data\n"
	    "               ---------\n"
	    "grid     name%s  gridIndexRange(0:1,0:2)           gridPoints   hmx      hmn  \n",
	    (const char *)blanks(0,min(maxNameLength-3,blanks.length()-1)));
    char buff[180];
    sPrintF(buff,"%%4i: %%%is   ([%%2i:%%5i],[%%2i:%%5i],[%%2i:%%5i])  %%8i    %%8.2e %%8.2e \n",maxNameLength);
    real maxMax=0.,maxMin=0.,minMin=REAL_MAX;
    int numberOfGridPoints=0;
    for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    {
      MappedGrid & c = cg[grid];
      real & hMin = dxMinMax(grid,0);
      real & hMax = dxMinMax(grid,1);
      maxMax=max(maxMax,hMax);
      maxMin=max(maxMin,hMin);
      minMin=min(minMin,hMin);
    
      int numGridPoints = c.mask().elementCount();
    
      // fPrintF(file,"%4i: %20s ([%2i:%5i],[%2i:%5i],[%2i:%5i])  %8i   %8.2e %8.2e \n",
      fPrintF(file,buff,grid, (const char *)cg[grid].getName(),
	      c.gridIndexRange(Start,axis1),c.gridIndexRange(End,axis1),
	      c.gridIndexRange(Start,axis2),c.gridIndexRange(End,axis2),
	      c.gridIndexRange(Start,axis3),c.gridIndexRange(End,axis3),
	      numGridPoints,hMax,hMin);
      numberOfGridPoints+=numGridPoints;
    }
    fPrintF(file," total number of grid points =%i, min(hmn)=%6.2e, max(hmn)=%6.2e, max(hmx)=%6.2e,  \n\n",
	    numberOfGridPoints,minMin,maxMin,maxMax);

    displayBoundaryConditions(file);

    fPrintF(file,"******************************************************************\n\n");
    
  }

  plotOptions=1;
  if( !gi.graphicsIsOn() && !gi.readingFromCommandFile() )
    plotOptions=0;
  
  char buff[80];           // buffer for sprintf
  numberOfStepsTaken=0;
  int iPrint=0;
  
  outputResultsAfterEachTimeStep( current,t,dt,0,nextTimeToPlot );
    
  for( int stepNumber=0; dt>REAL_MIN && stepNumber<maxNumberOfTimeSteps; stepNumber++ )                // take some time steps
  {

    if( t+.5*dt>nextTimeToPlot )  // plot solution 
    {
      // compute errors
      getErrors( current,t,dt );

      if( frequencyToSaveInShowFile>0 && (iPrint % frequencyToSaveInShowFile == 0) )
        saveShow( current,t,dt );  // save the current solution in the show file
      
      int finished = plot(current, t, dt );
      if( finished ) break;
      if( t >tFinal-.5*dt )
      {
	// we are done (unless tFinal is increased in the next call to plot). plot solution at final time
	//	if( plotOptions & 1 )
	{
	  plotOptions=1;  // plot and wait 
	  plot(current, t, dt );
	}
        if( t >tFinal-.5*dt ) // tFinal may have been increased, so check again
	{ 
	  finished=true;
	  break;
	}
      }
      iPrint++;
    }

    nextTimeToPlot=min(int(t/tPlot+.5)*tPlot+tPlot,tFinal);   //  ...new time to print:
    // ** we cannot change the time step now: 
    bool adjustTimeStep=false;

    computeNumberOfStepsAndAdjustTheTimeStep(t,tFinal,nextTimeToPlot,numberOfSubSteps,dtNew,adjustTimeStep );

    // no: dt=dtNew;  

    real time1=getCPU();

    for( int subStep=0; subStep<numberOfSubSteps; subStep++)
    {
      if( method==yee )
      {
	// advance on a rectangular grid
	advanceR( current,t,dt,dx, fields );
      }
      else if( method==dsi || method==dsiMatVec )      
      {
	//  	if( cg[0].getGridType()==MappedGrid::structuredGrid )
//  	{
//  	  advanceC( current,t,dt, fields );
//  	}
//  	else
//  	{
//  	  advanceUnstructuredDSI(current,t,dt, fields );
//  	}

        advanceDSI(current,t,dt);

      }
      else if( method==nfdtd )
      {
	advanceNFDTD( numberOfStepsTaken,current,t,dt );
      }
      else if( method==dsiNew )
      {
	advanceNew( current,t,dt, fields );
      }
      else
      {
	throw "error";
      }
      t+=dt;
      current = next;
      next= (next+1) % numberOfTimeLevels;

      outputResultsAfterEachTimeStep( current,t,dt,stepNumber,nextTimeToPlot );

      numberOfStepsTaken++;
    }
    
    timing(timeForAdvance)+=getCPU()-time1;
    
  }

  if ( !numberOfStepsTaken && dt<REAL_MIN )
    {
      getErrors(current,t,dt );
      outputResults(current,t,dt);
    }
    

  timing(totalTime)+=getCPU()-time0;
  
  printStatistics();
  
  return 0;
}
