#include "Cgsm.h"
#include "SmParameters.h"
#include "CompositeGridOperators.h"
#include "display.h"
#include "ParallelUtility.h"


#define FOR_3D(i1,i2,i3,I1,I2,I3) \
int I1Base =I1.getBase(),   I2Base =I2.getBase(),  I3Base =I3.getBase();  \
int I1Bound=I1.getBound(),  I2Bound=I2.getBound(), I3Bound=I3.getBound(); \
for(i3=I3Base; i3<=I3Bound; i3++) \
for(i2=I2Base; i2<=I2Bound; i2++) \
for(i1=I1Base; i1<=I1Bound; i1++)

#define FOR_3(i1,i2,i3,I1,I2,I3) \
I1Base =I1.getBase(),   I2Base =I2.getBase(),  I3Base =I3.getBase();  \
I1Bound=I1.getBound(),  I2Bound=I2.getBound(), I3Bound=I3.getBound(); \
for(i3=I3Base; i3<=I3Bound; i3++) \
for(i2=I2Base; i2<=I2Bound; i2++) \
for(i1=I1Base; i1<=I1Bound; i1++)


// ===================================================================================================================
// \brief Determine the time step
/// \details 
// ===================================================================================================================
real Cgsm::
getTimeStep( GridFunction & gf )
{
  real time0=getCPU();

  SmParameters::PDEModel & pdeModel = parameters.dbase.get<SmParameters::PDEModel>("pdeModel");
  SmParameters::PDEVariation & pdeVariation = parameters.dbase.get<SmParameters::PDEVariation>("pdeVariation");

  // There is no need to recompute the time step for linear-elasticity 
  bool & recomputeDt = parameters.dbase.get<bool>("recomputeDt");
  if( !recomputeDt && 
      pdeModel==SmParameters::linearElasticity &&
      ( pdeVariation!=SmParameters::hemp ) )
    return deltaT;

  SmParameters::TimeSteppingMethodSm & timeSteppingMethodSm = 
    parameters.dbase.get<SmParameters::TimeSteppingMethodSm>("timeSteppingMethodSm");

  real & cfl = parameters.dbase.get<real>("cfl");
  
  recomputeDt=false;

  if( pdeVariation==SmParameters::godunov ||
      pdeVariation==SmParameters::hemp )
  {
    // =============================================================================
    // ============= Time-stepping eigenvalues are provided else-where =============
    // =============================================================================

    bool & computeTimeSteppingEigenValues = parameters.dbase.get<bool>("computeTimeSteppingEigenValues");
    real & reLambda = parameters.dbase.get<real>("realPartOfTimeSteppingEigenValue");
    real & imLambda = parameters.dbase.get<real>("imaginaryPartOfTimeSteppingEigenValue");
    
    computeTimeSteppingEigenValues=true;  // this is used in advanceFOS

    real dt =1.e-9;  // choose a small non-zero value 
    // Call the next routine to compute the time stepping eigenvalue: 
    advanceFOS( current, gf.t, dt );

    computeTimeSteppingEigenValues=false;  // reset 

    assert( reLambda>=0. && imLambda>=0. );

    if( timeSteppingMethodSm==SmParameters::improvedEuler )
    { // the stability region for the midpoint rule extends to -2 on the Re axis
      // and about sqrt(3) along the imaginary (NOT REALLY -- it does not touch the Im axis!!)
      deltaT =cfl/sqrt( SQR(reLambda/2.)+SQR(imLambda/sqrt(3.)) );
    }
    else
    {
      deltaT= cfl/sqrt( SQR(reLambda/2.)+SQR(imLambda/1.) );  // ***********is this ok ****
    }
    


    deltaT=ParallelUtility::getMinValue(deltaT);  // min value over all processors
    printF("==== getTimeStep: FOS: reLamba=%8.2e imLambda=%8.2e, cfl=%8.2e deltaT=%8.2e\n",
               reLambda,imLambda,cfl,deltaT);

    timing(timeForComputingDeltaT)+=getCPU()-time0;

    return deltaT;

  }
  


  const int numberOfComponentGrids = cg.numberOfComponentGrids();
  const int numberOfDimensions = cg.numberOfDimensions();
  const int & orderOfAccuracyInSpace = parameters.dbase.get<int>("orderOfAccuracy");
  const int & orderOfAccuracyInTime  = parameters.dbase.get<int>("orderOfTimeAccuracy");

  deltaT=REAL_MAX;
  
  real & rho=parameters.dbase.get<real>("rho");
  real & mu = parameters.dbase.get<real>("mu");
  real & lambda = parameters.dbase.get<real>("lambda");
  RealArray & muGrid = parameters.dbase.get<RealArray>("muGrid");
  RealArray & lambdaGrid = parameters.dbase.get<RealArray>("lambdaGrid");

  // NOTE: For now we only approximate the time step -- we need to work out
  //       the correct formula for elasticity


  // cMax : fastest wave speed on all grids 
  // real cMax=max(2.*muGrid+lambdaGrid)/rho;
  real cMax= sqrt( max(lambdaGrid + 2.*muGrid)/rho ); // *wdh* 081120 -- sqrt added 

  for( int grid=0; grid<numberOfComponentGrids; grid++ )
  {
    MappedGrid & mg = cg[grid];

    const int numberOfDimensions=mg.numberOfDimensions();
    lambda=lambdaGrid(grid);
    mu=muGrid(grid);
    // c = fastest wave speed on this grid 
    const real c = sqrt( (lambda+2.*mu)/rho );   // *wdh* 081120 -- sqrt added

    real dtg;

    real dx[3];
    if( mg.isRectangular() )
    {
      mg.getDeltaX(dx);

      if( numberOfDimensions==2 )
	dtg=cfl*1./( c*sqrt( 1./(dx[0]*dx[0])+1./(dx[1]*dx[1]) ) );  
      else
	dtg=cfl*1./( c*sqrt( 1./(dx[0]*dx[0])+1./(dx[1]*dx[1])+1./(dx[2]*dx[2]) ) ); 
	
      dxMinMax(grid,0)=numberOfDimensions==2 ? min(dx[0],dx[1]) : min(dx[0],dx[1],dx[2]);
      dxMinMax(grid,1)=numberOfDimensions==2 ? max(dx[0],dx[1]) : max(dx[0],dx[1],dx[2]);
	
      // printF(" getTimeStep: dx=%8.2e dy=%8.2e c=%8.2e, dtg=%8.2e\n",dx[0],dx[1],c,dtg);
    }
    else
    {
      mg.update(MappedGrid::THEinverseVertexDerivative);
      const realArray & rx = mg.inverseVertexDerivative();
      const intArray & mask = mg.mask();
	
      
      Index I1,I2,I3;
      getIndex( mg.indexRange(),I1,I2,I3);

      // Grid spacings on unit square:
      real dr1 = mg.gridSpacing(axis1);
      real dr2 = mg.gridSpacing(axis2);
      real dr3 = mg.gridSpacing(axis3);

      if( true )
      {
	// parallel version here --- also broadcast max error in forcing.bC *************************
#ifdef USE_PPP
	realSerialArray rxLocal;  getLocalArrayWithGhostBoundaries(rx,rxLocal);
	intSerialArray maskLocal;  getLocalArrayWithGhostBoundaries(mask,maskLocal);
#else
	const realSerialArray & rxLocal = rx; 
	const intSerialArray & maskLocal  =  mask; 
#endif

	real *rxp = rxLocal.Array_Descriptor.Array_View_Pointer3;
	const int rxDim0=rxLocal.getRawDataSize(0);
	const int rxDim1=rxLocal.getRawDataSize(1);
	const int rxDim2=rxLocal.getRawDataSize(2);
	const int rxDim3=mg.numberOfDimensions();   // note
#undef RX
#define RX(i0,i1,i2,i3,i4) rxp[i0+rxDim0*(i1+rxDim1*(i2+rxDim2*(i3+rxDim3*(i4))))]

	const int *maskp = maskLocal.Array_Descriptor.Array_View_Pointer2;
	const int maskDim0=maskLocal.getRawDataSize(0);
	const int maskDim1=maskLocal.getRawDataSize(1);
	const int md1=maskDim0, md2=md1*maskDim1; 
#define MASK(i0,i1,i2) maskp[(i0)+(i1)*md1+(i2)*md2]


	I1 = Range(max(I1.getBase(),rxLocal.getBase(0)),min(I1.getBound(),rxLocal.getBound(0)));
	I2 = Range(max(I2.getBase(),rxLocal.getBase(1)),min(I2.getBound(),rxLocal.getBound(1)));
	I3 = Range(max(I3.getBase(),rxLocal.getBase(2)),min(I3.getBound(),rxLocal.getBound(2)));

	int i1,i2,i3;
	real a11Min=REAL_MAX*.1;
	real a11Max=-a11Min;
	//  **** this is a guess **** check this.
	const real alpha0=1.;

	real a11,a12,a22;
	if( numberOfDimensions==2 )
	{

	  FOR_3D(i1,i2,i3,I1,I2,I3)
	  {
	      
	    if( MASK(i1,i2,i3)>0 )
	    {
	      a11 = ( RX(i1,i2,i3,0,0)*RX(i1,i2,i3,0,0) + RX(i1,i2,i3,0,1)*RX(i1,i2,i3,0,1) );
	      a12 = ( RX(i1,i2,i3,0,0)*RX(i1,i2,i3,1,0) + RX(i1,i2,i3,0,1)*RX(i1,i2,i3,1,1) )*2.;
	      a22 = ( RX(i1,i2,i3,1,0)*RX(i1,i2,i3,1,0) + RX(i1,i2,i3,1,1)*RX(i1,i2,i3,1,1) );

 
	      a11=1./sqrt( a11 *(1./(alpha0*dr1*dr1)) 
			   +abs(a12)*(.25/(alpha0*dr1*dr2))
			   +a22 *(1./(alpha0*dr2*dr2)) 
		);

	      a11Min=min(a11Min,a11);
	      a11Max=max(a11Max,a11);
          
	    }

	  }
	    
	}
	else
	{ // ***** 3D ********

#define rxDotRx(axis,dir) (RX(i1,i2,i3,axis,0)*RX(i1,i2,i3,dir,0) \
			 + RX(i1,i2,i3,axis,1)*RX(i1,i2,i3,dir,1) \
			 + RX(i1,i2,i3,axis,2)*RX(i1,i2,i3,dir,2))
      
	  // There would be a factor of 4 for the worst case plus/minus wave but we also
	  // divide by a factor of 4 for the 2nd-order time stepping.
	  FOR_3D(i1,i2,i3,I1,I2,I3)
	  {
	    if( MASK(i1,i2,i3)>0 )
	    {
	      a11=1./sqrt(  rxDotRx(0,0) *(1./(dr1*dr1)) 
			    +rxDotRx(1,1) *(1./(dr2*dr2))
			    +rxDotRx(2,2) *(1./(dr3*dr3))
			    +abs(rxDotRx(1,0))*(.5/(dr2*dr1))  
			    +abs(rxDotRx(2,0))*(.5/(dr3*dr1)) 
			    +abs(rxDotRx(2,1))*(.5/(dr3*dr2)) );

	      // ** a11 =  pow(a11,-.5);
		
	      a11Min=min(a11Min,a11);
	      a11Max=max(a11Max,a11);

	    }
	  }
#undef rxDotRx
	}

	dxMinMax(grid,0)=a11Min;  // ********** need to broadcast ***********
	dxMinMax(grid,1)=a11Max;
	dtg = (cfl/c) * dxMinMax(grid,0);  // *wdh* 071029 : divide by c !
	

      }
      else // *** old way ***
      {
#undef RX
#define RX(i1,i2,i3,m,n) rx(i1,i2,i3,m+numberOfDimensions*(n))

	realArray a11;

	if( numberOfDimensions==2 )
	{
	  realArray a12,a22;
	  a11 = ( RX(I1,I2,I3,0,0)*RX(I1,I2,I3,0,0) + RX(I1,I2,I3,0,1)*RX(I1,I2,I3,0,1) );
	  a12 = ( RX(I1,I2,I3,0,0)*RX(I1,I2,I3,1,0) + RX(I1,I2,I3,0,1)*RX(I1,I2,I3,1,1) )*2.;
	  a22 = ( RX(I1,I2,I3,1,0)*RX(I1,I2,I3,1,0) + RX(I1,I2,I3,1,1)*RX(I1,I2,I3,1,1) );

	  //  **** this is a guess **** check this.
	  const real alpha0=1.;

	  a11=pow(a11 *(1./(alpha0*dr1*dr1)) 
		  +abs(a12)*(.25/(alpha0*dr1*dr2))
		  +a22 *(1./(alpha0*dr2*dr2)) 
		  ,-.5);
	}
	else
	{ // ***** 3D ********

#define rxDotRx(axis,dir) (RX(I1,I2,I3,axis,0)*RX(I1,I2,I3,dir,0) \
			 + RX(I1,I2,I3,axis,1)*RX(I1,I2,I3,dir,1) \
			 + RX(I1,I2,I3,axis,2)*RX(I1,I2,I3,dir,2))
      
	  // There wold be a factor of 4 for the worst case plus/minus wave but we also
	  // divide by a factor of 4 for the 2nd-order time stepping.
	  a11=(  rxDotRx(0,0) *(1./(dr1*dr1)) 
		 +rxDotRx(1,1) *(1./(dr2*dr2))
		 +rxDotRx(2,2) *(1./(dr3*dr3))
		 +abs(rxDotRx(1,0))*(.5/(dr2*dr1))  
		 +abs(rxDotRx(2,0))*(.5/(dr3*dr1)) 
		 +abs(rxDotRx(2,1))*(.5/(dr3*dr2)) );

	  a11 = pow(a11,-.5);
	  
#undef rxDotRx
	}
	
	where( mask(I1,I2,I3)>0 )
	{
	  dxMinMax(grid,0)=min(a11);
	  dxMinMax(grid,1)=max(a11);
          
	  dtg = (cfl/c) * dxMinMax(grid,0); // *wdh* 071029 : divide by c !
	}
	  
      }
	
//  	dtg = cfl * min( 
//  	  pow(
//  	    a11 *(1./(alpha0*dr1*dr1)) 
//  	    +abs(a12)*(.25/(alpha0*dr1*dr2))
//  	    +a22 *(1./(alpha0*dr2*dr2)) 
//  	    ,-.5)
//  	  );
      

    }
    
      
    if( artificialDissipation>0. )
    {
      // Here is the correction for artificial dissipation
      //
      // The equation for dt looks like
      //   dt*dt *c*c*(  1/dx^2 + 1/dy^2 ) = 1 - beta*dt
      //
      real gamma = dtg*dtg;
      real beta;
      // const real adc=artificialDissipation*SQR(cMax); // scale dissipation by c^2 *wdh* 041103
      const real adc=artificialDissipation; // do not scale *wdh* 090216

      beta = .5*adc*( numberOfDimensions*pow(2.,real(orderOfArtificialDissipation)) );
      real factor=2.;  // safety factor
      beta *=factor;

      dtg = sqrt( gamma + pow(beta*gamma*.5,2.) ) - beta*gamma*.5;

      printF(" getTimeStep: Correct for art. dissipation: new dt=%9.3e (old = %9.3e, new/old=%4.2f)\n",
	     dtg,sqrt(gamma),dtg/sqrt(gamma));
	
    }
      
    if( timeSteppingMethodSm==SmParameters::modifiedEquationTimeStepping )
    {
      if( true || orderOfAccuracyInTime==2 || orderOfAccuracyInTime==4 )
      {
	dtg*=1.; // Check this for 3D
      }
      else
      {
	printF("getTimeStep:ERROR: modifiedEquationTimeStepping -- orderOfAccuracyInTime=%i ??\n",
	       orderOfAccuracyInTime);
	  
	Overture::abort("getTimeStep:ERROR: modifiedEquationTimeStepping -- orderOfAccuracyInTime?? ");
      }
	
    }
    else if( timeSteppingMethodSm==SmParameters::forwardEuler  )
    {
      // what should this be ? 
    }
    else if( timeSteppingMethodSm==SmParameters::improvedEuler  )
    {
      // what should this be ? 
    }
    else if( timeSteppingMethodSm==SmParameters::adamsBashforth2  )
    {
      // what should this be ? 
    }
    else if( timeSteppingMethodSm==SmParameters::adamsPredictorCorrector2  )
    {
      // what should this be ? 
    }
    else if( timeSteppingMethodSm==SmParameters::adamsPredictorCorrector4  )
    {
      // what should this be ? 
    }
    else
    {

      if( orderOfAccuracyInSpace==2 )
      {
      }
      else if( orderOfAccuracyInSpace==4 )
	dtg*=sqrt(3./4.);
      else if( orderOfAccuracyInSpace==6 )
	dtg*=sqrt(.6618);
      else if( orderOfAccuracyInSpace==8 )
	dtg*=sqrt(.6152);
      else
      {
	Overture::abort("getTimeStep:ERROR: modifiedEquationTimeStepping -- orderOfAccuracyInSpace?? ");
      }

      if( orderOfAccuracyInTime==4 )
      {
	dtg*=1.41/2.;
      }
      else if( orderOfAccuracyInTime==6 )
      {
	dtg*=.84/2.;
      }
      else if( orderOfAccuracyInTime==8 )
      {
	dtg*=.46/2.;
      }
      else if( orderOfAccuracyInTime==3 ) // && method==dsi
      {
	dtg*=(12./7.)/2.;   // ABS3
      }
      else if( orderOfAccuracyInTime!=2 )
      {
	Overture::abort();
      }
    }
      
    dtg=ParallelUtility::getMinValue(dtg);  // compute min over all processors
      
    printF(" getTimeStep: grid=%i c=%8.2e, dtg=%8.2e\n",grid,c,dtg);

#undef RX
    
    deltaT=min(deltaT,dtg);
    
  } // end for grid

  deltaT=ParallelUtility::getMinValue(deltaT);  // min value over all processors
  printF("==== getTimeStep: deltaT=%8.2e\n",deltaT);

  timing(timeForComputingDeltaT)+=getCPU()-time0;

  return deltaT;
}

