#include "Cgsm.h"
#include "SmParameters.h"
#include "GenericGraphicsInterface.h"
#include "ParallelUtility.h"

namespace
{
enum UserDefinedBoundaryConditions
{
  ellipseDeform,
  testDeform,
  tractionForcing,
  pressureForce,    
  piston
};
 
}

// =========================================================================================
/// \brief Interactively define user specific values for boundary conditions. 
/// \details This function will be called when interactively choosing boundary conditions and the
/// option userDefinedBoundaryData is used.
///
/// You may add a time dependence to an existing boundary condition or you may define a 
/// new boundary condition. In this function you should prompt for the boundary condition
/// to be used as well as any parameters that will be needed. Parameters can be saved using
/// the setUserBoundaryConditionParameters function. 
///
/// \note:  The actual boundary values are NOT assigned in this routine. This is done in the
///    userDefinedBoundaryValues function.
/// 
/// \param side,axis,grid (input): assign boundary conditions for this face.
///
// =========================================================================================
int SmParameters::
chooseUserDefinedBoundaryValues(int side, int axis, int grid, CompositeGrid & cg)
{
  Parameters & parameters = *this;
  assert( parameters.dbase.get<GenericGraphicsInterface* >("ps") !=NULL );
  GenericGraphicsInterface & gi = *parameters.dbase.get<GenericGraphicsInterface* >("ps");


  printf("Choose boundary values for (side,axis,grid)=(%i,%i,%i) gridName=%s\n",side,axis,grid,
	 (const char*)cg[grid].getName());

  aString menu[]=
  {
    "!user boundary values",
    "ellipse deform",
    "test deform",
    "traction forcing",
    "pressure force",
    "time dependent inflow",
    "piston",
    "done",
    ""
  };
  aString answer,answer2;
  Index Ib1,Ib2,Ib3;
//   const int pc=parameters.dbase.get<int >("pc");
//   const int uc=parameters.dbase.get<int >("uc");
//   const int vc=parameters.dbase.get<int >("vc");
//   const int wc=parameters.dbase.get<int >("wc");
  
  for(;;)
  {
    gi.getMenuItem(menu,answer,"choose a menu item");
    if( answer=="done" || answer=="exit" )
    {
      break;
    }
    else if( answer=="ellipse deform" )
    {
      parameters.setUserBcType(side,axis,grid,ellipseDeform);    // set the bcType to be a unique value.
      parameters.setBcIsTimeDependent(side,axis,grid,true);  // this condition IS time dependent


      RealArray values(2);
      real deformationAmplitude=.2;
      real deformationFrequency=1.;
      gi.inputString(answer2,"Enter the amplitude and frequency");
      if( answer2!="" )
      {
	sScanF(answer2,"%e %e",&deformationAmplitude,&deformationFrequency);
      }
      printf("***userDefinedBoundaryValues: amplitude=%9.3e frequency=%9.3e for ellipse deform\n",
	     deformationAmplitude,deformationFrequency);

      values(0)=deformationAmplitude;
      values(1)=deformationFrequency;

      // save the parameters to be used when evaluating the time dependent BC's:
      parameters.setUserBoundaryConditionParameters(side,axis,grid,values);

    }
    else if( answer=="test deform" )
    {
      parameters.setUserBcType(side,axis,grid,testDeform);    // set the bcType to be a unique value.
      parameters.setBcIsTimeDependent(side,axis,grid,true);  // this condition IS time dependent

    }
    else if( answer=="traction forcing" )
    {
      parameters.setUserBcType(side,axis,grid,tractionForcing);    // set the bcType to be a unique value.
      parameters.setBcIsTimeDependent(side,axis,grid,true);  // this condition IS time dependent

      RealArray values(3);
      values=0.;
      gi.inputString(answer2,"Enter the traction force");
      if( answer2!="" )
      {
	sScanF(answer2,"%e %e %e",&values(0),&values(1),&values(2));
      }
      printf("***traction force= (%9.3e,%9.3e,%9.3e)\n",values(0),values(1),values(2));

      // save the parameters to be used when evaluating the time dependent BC's:
      parameters.setUserBoundaryConditionParameters(side,axis,grid,values);
    }
    else if( answer=="pressure force" )
    {
      parameters.setUserBcType(side,axis,grid,pressureForce);    // set the bcType to be a unique value.
      parameters.setBcIsTimeDependent(side,axis,grid,false);  // this condition is NOT time dependent

      RealArray values(1);
      values=0.;
      gi.inputString(answer2,"Enter the pressure force");
      if( answer2!="" )
      {
	sScanF(answer2,"%e",&values(0));
      }
      printf("***pressure force= %9.3e for grid=%i (side,axis)=(%i,%i)\n",values(0),grid,side,axis);

      // save the parameters to be used when evaluating the time dependent BC's:
      parameters.setUserBoundaryConditionParameters(side,axis,grid,values);
    }
    else if( answer=="piston" )
    {
      parameters.setUserBcType(side,axis,grid,piston);    // set the bcType to be a unique value
      parameters.setBcIsTimeDependent(side,axis,grid,true);  // this condition IS time dependent

      RealArray values(2);
      real velx = 0.1;
      real vely = 0.0;
      gi.inputString(answer2,"Enter the x and y velocities");
      if( answer2!="" )
      {
	sScanF(answer2,"%e %e",&velx,&vely);
      }
      printf("***userDefinedBoundaryValues: velx=%9.3e vely=%9.3e for piston\n",velx,vely);

      values(0)=velx;
      values(1)=vely;

      // save the parameters to be used when evaluating the time dependent BC's:
      parameters.setUserBoundaryConditionParameters(side,axis,grid,values);
    }
    else
    {
      printF("Unknown answer =[%s]\n",(const char*)answer2);
      gi.stopReadingCommandFile();
    }
  }

  return 0;
}




// =========================================================================================
/// \brief Assign user specific values for boundary conditions. 
/// \details The user may fill in
/// the boundaryData array with right-hand-side values for boundary conditions.
/// The user is also required to provide the time derivative of the boundary values.
/// 
/// \note  This function shows examples of applying a user defined boundary condition as
/// well as adding a time dependence to some pre-defined boundary conditions.
///
/// \param t (input) : current time.
/// \param u (input) : the current solution.
/// \param gridVelocity (input) : if the grid is moving, this is the grid velocity.
/// \param grid (input): the component grid we are assigning.
/// \param forcingType (input) : if forcingType==computeForcing then return the rhs for the 
///  boundary condition; if forcingType==computeTimeDerivativeOfForcing then return the 
///   first time derivative of the forcing.
///
// =========================================================================================
int Cgsm::
userDefinedBoundaryValues(const real & t, 
			  realMappedGridFunction & u, 
			  realMappedGridFunction & gridVelocity,
			  const int & grid,
			  int side0 /* = -1 */,
			  int axis0 /* = -1 */,
			  ForcingTypeEnum forcingType /* =computeForcing */)
{
  // printf("***userDefinedBoundaryValues\n");

  MappedGrid & mg = *u.getMappedGrid();

  assert( side0>=-1 && side0<2 );
  assert( axis0>=-1 && axis0<parameters.dbase.get<int >("numberOfDimensions") );
  
  const int axisStart= axis0==-1 ? 0 : axis0;
  const int axisEnd  = axis0==-1 ? parameters.dbase.get<int >("numberOfDimensions")-1 : axis0;
  const int sideStart= side0==-1 ? 0 : side0;
  const int sideEnd  = side0==-1 ? 1 : side0;

  int numberOfSidesAssigned=0;

  Range C(0,parameters.dbase.get<int >("numberOfComponents")-1);
  const int pc=parameters.dbase.get<int >("pc");
  const int uc=parameters.dbase.get<int >("uc");
  const int vc=parameters.dbase.get<int >("vc");
  const int wc=parameters.dbase.get<int >("wc");
  const int tc=parameters.dbase.get<int >("tc");
  
  const int v1c=parameters.dbase.get<int >("v1c");
  const int v2c=parameters.dbase.get<int >("v2c");
  const int v3c=parameters.dbase.get<int >("v3c");

  const int s11c=parameters.dbase.get<int >("s11c");
  const int s12c=parameters.dbase.get<int >("s12c");
  const int s13c=parameters.dbase.get<int >("s13c");
  const int s21c=parameters.dbase.get<int >("s21c");
  const int s22c=parameters.dbase.get<int >("s22c");
  const int s23c=parameters.dbase.get<int >("s23c");
  const int s31c=parameters.dbase.get<int >("s31c");
  const int s32c=parameters.dbase.get<int >("s32c");
  const int s33c=parameters.dbase.get<int >("s33c");

  const bool gridIsMoving = parameters.gridIsMoving(grid); // true if the grid is moving

  mg.update(MappedGrid::THEvertex);
  const realArray & vertex = mg.vertex();  // here is the array of grid points
  
 int includeGhost=1;
  #ifdef USE_PPP
    realSerialArray uLocal; getLocalArrayWithGhostBoundaries(u,uLocal);
    realSerialArray xLocal; getLocalArrayWithGhostBoundaries(vertex,xLocal);
  #else
    const realSerialArray & uLocal = u;
    const realSerialArray & xLocal = vertex;
  #endif

  const int numberOfDimensions = mg.numberOfDimensions();
  
  mg.update(MappedGrid::THEvertexBoundaryNormal);

  int axis;
  Index Ib1,Ib2,Ib3;
  
  for( axis=axisStart; axis<=axisEnd; axis++ )
  {
    for( int side=sideStart; side<=sideEnd; side++ )
    {
       #ifdef USE_PPP
	 realSerialArray & normal = *(mg.rcData->pVertexBoundaryNormal[axis][side]); 
       #else
        realArray & normal = mg.vertexBoundaryNormal(side,axis);
       #endif


      if( parameters.userBcType(side,axis,grid)==ellipseDeform )
      {
        RealArray values(2);
	parameters.getUserBoundaryConditionParameters(side,axis,grid,values);
        // printf("***userDefinedBoundaryValues: assign (side,axis,grid)=(%i,%i,%i) variableInflow force=%i, "
        //        "values=%f,%f,%f\n",side,axis,grid,(int)forcingType,values(0),values(1),values(2));

        getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);
	
	numberOfSidesAssigned++;

	bool ok=ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3,includeGhost);
	if( !ok ) continue;  // no points on this processor

	RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);
	real deformationAmplitude=values(0);
	real deformationFrequency=values(1);

        real omega=pow(sin(deformationFrequency*Pi*t),2.);  // omega varies in the interval [0,1]

	real rad=1.;  // initial radius 
	real a =  deformationAmplitude*sin(omega);
	real b = -deformationAmplitude*sin(omega);
      
	printf("***userDefinedBoundaryValues:ellipseDeform: t=%9.3e, da=%9.3e\n",t,a);

        // Here are the displacements from a circle to form an ellipse
	bd(Ib1,Ib2,Ib3,uc)=a*xLocal(Ib1,Ib2,Ib3,0);  //  x=r*cos(theta) 
	bd(Ib1,Ib2,Ib3,vc)=b*xLocal(Ib1,Ib2,Ib3,1);  //  y=r*sin(theta)
	if( numberOfDimensions>2 )
	  bd(Ib1,Ib2,Ib3,wc)=0.;

	// ::display(bd(Ib1,Ib2,Ib3,Range(uc,vc)),"bd: boundary data");
	
        // NOTE: For the first order system we need to also specify the velocity and acceleration
        if( ((SmParameters&)parameters).isFirstOrderSystem() )
 	{
          real omegat=2.*deformationFrequency*Pi*sin(deformationFrequency*Pi*t)*cos(deformationFrequency*Pi*t);
          real at = deformationAmplitude*cos(omega)*omegat;	  
          real bt = -at;
 	  bd(Ib1,Ib2,Ib3,v1c)=at*xLocal(Ib1,Ib2,Ib3,0);  // velocity
 	  bd(Ib1,Ib2,Ib3,v2c)=bt*xLocal(Ib1,Ib2,Ib3,1);
        
          real omegatt = 2.*SQR(deformationFrequency*Pi)*(SQR(cos(deformationFrequency*Pi*t)) - 
                                                          SQR(sin(deformationFrequency*Pi*t)) );
          real att = -deformationAmplitude*cos(omega)*omegat*omegat + deformationAmplitude*cos(omega)*omegatt;
          real btt = -att;
 	  bd(Ib1,Ib2,Ib3,s11c)=att*xLocal(Ib1,Ib2,Ib3,0); //  acceleration  
 	  bd(Ib1,Ib2,Ib3,s12c)=btt*xLocal(Ib1,Ib2,Ib3,1);
 	}
	
      }
      else if( parameters.userBcType(side,axis,grid)==testDeform )
      {
        getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);
	
	numberOfSidesAssigned++;

	bool ok=ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3,includeGhost);
	if( !ok ) continue;  // no points on this processor

	RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);
      
	printf("***userDefinedBoundaryValues:testDeform: t=%9.3e\n",t);

	bd(Ib1,Ib2,Ib3,uc)=xLocal(Ib1,Ib2,Ib3,0)*(1.+t);  
	bd(Ib1,Ib2,Ib3,vc)=xLocal(Ib1,Ib2,Ib3,1)*(1.+t);  
	if( numberOfDimensions>2 )
	  bd(Ib1,Ib2,Ib3,wc)=xLocal(Ib1,Ib2,Ib3,2)*(1.+t); 

	// ::display(bd(Ib1,Ib2,Ib3,Range(uc,vc)),"bd: boundary data");
	
      }
      else if( parameters.userBcType(side,axis,grid)==tractionForcing )
      {
        RealArray values(3);
	parameters.getUserBoundaryConditionParameters(side,axis,grid,values);

        getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);
	
	numberOfSidesAssigned++;

	bool ok=ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3,includeGhost);
	if( !ok ) continue;  // no points on this processor

	RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);
      
	printf("***userDefinedBoundaryValues: set tractionForce=(%8.2e,%8.2e,%8.2e) at t=%9.3e for"
               "(side,axis,grid)=(%i,%i,%i)\n",values(0),values(1),values(2),t,side,axis,grid);

	bd(Ib1,Ib2,Ib3,uc)=values(0);
	bd(Ib1,Ib2,Ib3,vc)=values(1);
	if( numberOfDimensions>2 )
	  bd(Ib1,Ib2,Ib3,wc)=values(2);

        // NOTE: For the first order system we need to also specify: 
        if( ((SmParameters&)parameters).isFirstOrderSystem() )
 	{
          // time derivative of the traction: 
   	  bd(Ib1,Ib2,Ib3,v1c)=0.;  
   	  bd(Ib1,Ib2,Ib3,v2c)=0.;
        
          // The traction also appears here: *wdh* this is no longer needed
   	  // bd(Ib1,Ib2,Ib3,s11c)=bd(Ib1,Ib2,Ib3,uc);  
   	  // bd(Ib1,Ib2,Ib3,s12c)=bd(Ib1,Ib2,Ib3,vc);
	  // if( numberOfDimensions>2 )
	  //  bd(Ib1,Ib2,Ib3,s13c)=bd(Ib1,Ib2,Ib3,wc);
 	}

      }
      else if( parameters.userBcType(side,axis,grid)==pressureForce )
      {
        RealArray values(1);
	parameters.getUserBoundaryConditionParameters(side,axis,grid,values);
	const real p0=values(0);

        getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);
	
	numberOfSidesAssigned++;

	bool ok=ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3,includeGhost);
	if( !ok ) continue;  // no points on this processor

	RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);
      
	printf("***userDefinedBoundaryValues: set pressure force=%8.2e at t=%9.3e for"
               "(side,axis,grid)=(%i,%i,%i)\n",p0,t,side,axis,grid);

	bd(Ib1,Ib2,Ib3,uc)=-p0*normal(Ib1,Ib2,Ib3,0);
	bd(Ib1,Ib2,Ib3,vc)=-p0*normal(Ib1,Ib2,Ib3,1);
	if( numberOfDimensions>2 )
	  bd(Ib1,Ib2,Ib3,wc)=-p0*normal(Ib1,Ib2,Ib3,2);

        // NOTE: For the first order system we need to also specify: 
        if( ((SmParameters&)parameters).isFirstOrderSystem() )
 	{
          // time derivative of the traction: 
   	  bd(Ib1,Ib2,Ib3,v1c)=0.;  
   	  bd(Ib1,Ib2,Ib3,v2c)=0.;
        
          // The traction also appears here:  *wdh* this is no longer needed
   	  // bd(Ib1,Ib2,Ib3,s11c)=bd(Ib1,Ib2,Ib3,uc);  
   	  // bd(Ib1,Ib2,Ib3,s12c)=bd(Ib1,Ib2,Ib3,vc);
	  // if( numberOfDimensions>2 )
	  //   bd(Ib1,Ib2,Ib3,s13c)=bd(Ib1,Ib2,Ib3,wc);
 	}

      }
      else if( parameters.userBcType(side,axis,grid)==piston )
      {
        RealArray values(2);
	parameters.getUserBoundaryConditionParameters(side,axis,grid,values);

        getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);
	
	numberOfSidesAssigned++;

	bool ok=ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3,includeGhost);
	if( !ok ) continue;  // no points on this processor

	RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);
	real velx=values(0);
	real vely=values(1);

// 	bd(Ib1,Ib2,Ib3,uc)=bd(Ib1,Ib2,Ib3,uc)+velx*t;
// 	bd(Ib1,Ib2,Ib3,vc)=bd(Ib1,Ib2,Ib3,vc)+vely*t;

	bd(Ib1,Ib2,Ib3,uc)=velx*t;
	bd(Ib1,Ib2,Ib3,vc)=vely*t;

	//printf( "vx=%e, vy=%e %i %i %i %i\n",velx,vely,uc,vc,v1c,v2c );
	
        // NOTE: For the first order system we need to also specify the velocity and acceleration
        if( ((SmParameters&)parameters).isFirstOrderSystem() )
 	{
 	  bd(Ib1,Ib2,Ib3,v1c)=velx; // velocity
 	  bd(Ib1,Ib2,Ib3,v2c)=vely;
        
	  // 	  bd(Ib1,Ib2,Ib3,s11c)=0.0; //  acceleration  
	  // 	  bd(Ib1,Ib2,Ib3,s12c)=0.0;
 	}
	
        // bd(Ib1,Ib2,Ib3,uc)=p0*normal(Ib1,Ib2,Ib3,0);
        // bd(Ib1,Ib2,Ib3,vc)=p0*normal(Ib1,Ib2,Ib3,1);

      }
      else
      {
	// printf(" perturbed shear flow: mg.boundaryCondition(side,axis)=%i \n",mg.boundaryCondition(side,axis));
	
      }
    }
    
  }
  
  return numberOfSidesAssigned;
}
